// state.c

#include "shared.h"
//#include "m68kcpu.h"

#define READ(val, size) READ_STREAM(val,size,file)
#define WRITE(val, size) WRITE_STREAM(val,size, file)

#define SIG	'G+S1'

extern char RomPath[260];
extern char genZ_curVolume;

extern uint8 sat[0x400];               /* Internal copy of sprite attribute table */
extern uint8 vram[0x10000];            /* Video RAM (64Kx8) */
extern uint8 cram[0x80];               /* On-chip color RAM (64x9) */
extern uint8 vsram[0x80];              /* On-chip vertical scroll RAM (40x11) */
extern uint8 reg[0x20];                /* Internal VDP registers (23x8) */
extern uint16 addr;                    /* Address register */
extern uint16 addr_latch;              /* Latched A15, A14 of address */
extern uint8 code;                     /* Code register */
extern uint8 pending;                  /* Pending write flag */
extern uint16 buffer;                  /* Read buffer */
extern uint16 status;                  /* VDP status flags */
extern uint16 ntab;                    /* Name table A base address */
extern uint16 ntbb;                    /* Name table B base address */
extern uint16 ntwb;                    /* Name table W base address */
extern uint16 satb;                    /* Sprite attribute table base address */
extern uint16 hscb;                    /* Horizontal scroll table base address */
extern uint16 sat_base_mask;           /* Base bits of SAT */
extern uint16 sat_addr_mask;           /* Index bits of SAT */
extern uint8 is_color_dirty;           /* 1= One or more colors have changed */
extern uint8 color_dirty[0x40];        /* 1= This color is dirty */
extern uint8 border;                   /* Border color index */
extern uint8 is_border_dirty;          /* 1= The border color has changed */
extern uint8 bg_name_dirty[0x800];     /* 1= This pattern is dirty */
extern uint16 bg_name_list[0x800];     /* List of modified pattern indices */
extern uint16 bg_list_index;           /* # of modified patterns in list */
extern uint8 bg_pattern_cache[0x80000];/* Cached and flipped patterns */
extern uint8 playfield_shift;          /* Width of planes A, B (in bits) */
extern uint8 playfield_col_mask;       /* Vertical scroll mask */
extern uint16 playfield_row_mask;      /* Horizontal scroll mask */
extern uint32 y_mask;                  /* Name table Y-index bits mask */
extern int hint_pending;               /* 0= Line interrupt is pending */
extern int vint_pending;               /* 1= Frame interrupt is pending */
extern int counter;                    /* Raster counter */
extern int dma_fill;                   /* 1= DMA fill has been requested */
extern int im2_flag;                   /* 1= Interlace mode 2 is being used */
extern int frame_end;                  /* End-of-frame (IRQ line) */
extern int v_counter;                  /* VDP scan line counter */
extern int v_update;                  /* 1= VC was updated by a ctrl or HV read */


extern uint8 fm_status;            /* Read-only status flags */

void GetStateFileName(int num,char *str)
{
	char result[256];
	uint32 i=0,j=255,k=255,l=255;
	char chaine[256];  	
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"gs");
	sprintf(str,"%s%s%d",GENZ_SAVESDIR,result,num);
	str[0]=genZ_curVolume+'0';
}

int state_load(int num,int onlytest)
{
	uint32	tmp32;
	STREAM	file;
	uint32	sig;
	int i;
	int		height, width;
	char savename[255];
	
	GetStateFileName(num,savename);
	
	file = OPEN_STREAM(savename, "rb");
	if (file == NULL)
	{
		return -1;
	}
	
	READ(&sig, 4);
	if (sig != SIG)
	{
		CLOSE_STREAM(file);
		return -1;
	}
	
	if (onlytest)
	{
		CLOSE_STREAM(file);
		return 0;
	}
	
	menu_message("Loading");

	// gen.c stuff
	READ(work_ram, 0x10000);
	READ(zram, 0x2000);
	READ(&zbusreq, 1);
	READ(&zreset, 1);
	READ(&zbusack, 1);
	READ(&zirq, 1);
	READ(&zbank, 4);

	// io.c stuff
	READ(io_reg, 0x10);
	
	// vdp.c stuff
	READ(sat, 0x400);
	READ(vram, 0x10000);
	READ(cram, 0x80);
	READ(vsram, 0x80);
	READ(reg, 0x20);
	READ(&addr, 2);
	READ(&addr_latch, 2);
	READ(&code, 1);
	READ(&pending, 1);
	READ(&buffer, 2);
	READ(&status, 2);
	READ(&ntab, 2);
	READ(&ntbb, 2);
	READ(&ntwb, 2);
	READ(&satb, 2);
	READ(&hscb, 2);
	READ(&sat_base_mask, 2);
	READ(&sat_addr_mask, 2);
	READ(&border, 1);
	READ(&playfield_shift, 1);
	READ(&playfield_col_mask, 1);
	READ(&playfield_row_mask, 2);
	READ(&y_mask, 4);
	READ(&hint_pending, 4);
	READ(&vint_pending, 4);
	READ(&counter, 4);
	READ(&dma_fill, 4);
	READ(&im2_flag, 4);
	READ(&frame_end, 4);
	READ(&v_counter, 4);
	READ(&v_update, 4);
	
	// sound.c stuff
	READ(fm_reg[0], 0x100);
	READ(fm_reg[1], 0x100);
	READ(fm_latch, 2);
	READ(&fm_status, 2);
	READ(&timer[0].running, sizeof(int));
	READ(&timer[0].enable, sizeof(int));
	READ(&timer[0].count, sizeof(int));
	READ(&timer[0].base, sizeof(int));
	READ(&timer[0].index, sizeof(int));
	READ(&timer[1].running, sizeof(int));
	READ(&timer[1].enable, sizeof(int));
	READ(&timer[1].count, sizeof(int));
	READ(&timer[1].base, sizeof(int));
	READ(&timer[1].index, sizeof(int));
	
	// Window size
	READ(&height, sizeof(int));
	READ(&width, sizeof(int));
	if (height != bitmap.viewport.h)
	{
		bitmap.viewport.oh = bitmap.viewport.h;
		bitmap.viewport.h = height;
		bitmap.viewport.changed = 1;
	}
	if (width != bitmap.viewport.w)
	{
		bitmap.viewport.ow = bitmap.viewport.w;
		bitmap.viewport.w = width;
		bitmap.viewport.changed = 1;
	}

	// 68000 CPU
	if (genZ_cpu_cyclone)
	{
	READ(&tmp32, 4); CycloneCPU.d[0]=tmp32; //m68k_set_reg(M68K_REG_D0, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[1]=tmp32; //m68k_set_reg(M68K_REG_D1, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[2]=tmp32; //m68k_set_reg(M68K_REG_D2, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[3]=tmp32; //m68k_set_reg(M68K_REG_D3, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[4]=tmp32; //m68k_set_reg(M68K_REG_D4, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[5]=tmp32; //m68k_set_reg(M68K_REG_D5, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[6]=tmp32; //m68k_set_reg(M68K_REG_D6, tmp32);
	READ(&tmp32, 4); CycloneCPU.d[7]=tmp32; //m68k_set_reg(M68K_REG_D7, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[0]=tmp32; //m68k_set_reg(M68K_REG_A0, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[1]=tmp32; //m68k_set_reg(M68K_REG_A1, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[2]=tmp32; //m68k_set_reg(M68K_REG_A2, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[3]=tmp32; //m68k_set_reg(M68K_REG_A3, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[4]=tmp32; //m68k_set_reg(M68K_REG_A4, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[5]=tmp32; //m68k_set_reg(M68K_REG_A5, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[6]=tmp32; //m68k_set_reg(M68K_REG_A6, tmp32);
	READ(&tmp32, 4); CycloneCPU.a[7]=tmp32; //m68k_set_reg(M68K_REG_A7, tmp32);
	READ(&tmp32, 4); CycloneCPU.pc=tmp32; //m68k_set_reg(M68K_REG_PC, tmp32);	
	CycloneCPU.membase=0;
    CycloneCPU.pc=CycloneCPU.checkpc(CycloneCPU.pc);
   	READ(&tmp32, 1); //m68k_set_reg(M68K_REG_SR, tmp32);
	//low sr : flags
	CycloneCPU.xc = (tmp32&0x10)>>3;
	CycloneCPU.ccr = (tmp32&0xC)|((tmp32&0x1)<<1)|((tmp32&0x2)>>1);
	//high sr : srh
	READ(&tmp32, 1);
	CycloneCPU.srh = tmp32;
	//0xD2
	READ(&tmp32, 2);	//dummy
	}
	else
	{
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D0, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D1, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D2, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D3, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D4, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D5, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D6, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_D7, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A0, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A1, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A2, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A3, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A4, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A5, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A6, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_A7, tmp32);
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_PC, tmp32);	
	READ(&tmp32, 4); m68k_set_reg(M68K_REG_SR, tmp32);
	}
	
	
  	
	// Z80 CPU
	READ(&tmp32, 4); z80_set_reg(Z80_PC, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_SP, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_AF, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_BC, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DE, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_HL, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IX, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IY, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_R, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_I, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_AF2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_BC2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DE2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_HL2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IM, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IFF1, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IFF2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_HALT, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_NMI_STATE, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_IRQ_STATE, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DC0, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DC1, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DC2, tmp32);
	READ(&tmp32, 4); z80_set_reg(Z80_DC3, tmp32);
	
	// Close it
	CLOSE_STREAM(file);
	
	// Remake cache
	/*make_bg_pattern_cache();
	vdp_restore_colours();*/
	for (i=0;i<0x800;i++) 
    {
	    bg_name_list[i]=i;
	    bg_name_dirty[i]=0xFF;
    }
    bg_list_index=0x800;
    if(color_update)
    {
    	for(i = 0; i < 0x40; i += 1)
        {
        	color_update(i, *(uint16 *)&cram[i << 1]);                    
        }
        color_update(0x00, *(uint16 *)&cram[border << 1]);
        color_update(0x40, *(uint16 *)&cram[border << 1]);
        color_update(0x80, *(uint16 *)&cram[border << 1]);
    }    					
		
	// Flush fm
	//fm_flush();
	
	// Success
	return 0;
}

int state_save(int num)
{
	uint32	tmp32;
	STREAM	file;
	uint32	sig;	
	char savename[255];
	
	GetStateFileName(num,savename);
	
	file = OPEN_STREAM(savename, "w+b");
	if (file == NULL)
	{
		return -1;
	}
	
	menu_message("Saving");

	// Header
	sig = SIG;
	WRITE(&sig, 4);

	// gen.c stuff
	WRITE(work_ram, 0x10000);
	WRITE(zram, 0x2000);
	WRITE(&zbusreq, 1);
	WRITE(&zreset, 1);
	WRITE(&zbusack, 1);
	WRITE(&zirq, 1);
	WRITE(&zbank, 4);

	// io.c stuff
	WRITE(io_reg, 0x10);
	
	// vdp.c stuff
	WRITE(sat, 0x400);
	WRITE(vram, 0x10000);
	WRITE(cram, 0x80);
	WRITE(vsram, 0x80);
	WRITE(reg, 0x20);
	WRITE(&addr, 2);
	WRITE(&addr_latch, 2);
	WRITE(&code, 1);
	WRITE(&pending, 1);
	WRITE(&buffer, 2);
	WRITE(&status, 2);
	WRITE(&ntab, 2);
	WRITE(&ntbb, 2);
	WRITE(&ntwb, 2);
	WRITE(&satb, 2);
	WRITE(&hscb, 2);
	WRITE(&sat_base_mask, 2);
	WRITE(&sat_addr_mask, 2);
	WRITE(&border, 1);
	WRITE(&playfield_shift, 1);
	WRITE(&playfield_col_mask, 1);
	WRITE(&playfield_row_mask, 2);
	WRITE(&y_mask, 4);
	WRITE(&hint_pending, 4);
	WRITE(&vint_pending, 4);
	WRITE(&counter, 4);
	WRITE(&dma_fill, 4);
	WRITE(&im2_flag, 4);
	WRITE(&frame_end, 4);
	WRITE(&v_counter, 4);
	WRITE(&v_update, 4);
	
	// sound.c stuff
	WRITE(fm_reg[0], 0x100);
	WRITE(fm_reg[1], 0x100);
	WRITE(fm_latch, 2);
	WRITE(&fm_status, 2);
	WRITE(&timer[0].running, sizeof(int));
	WRITE(&timer[0].enable, sizeof(int));
	WRITE(&timer[0].count, sizeof(int));
	WRITE(&timer[0].base, sizeof(int));
	WRITE(&timer[0].index, sizeof(int));
	WRITE(&timer[1].running, sizeof(int));
	WRITE(&timer[1].enable, sizeof(int));
	WRITE(&timer[1].count, sizeof(int));
	WRITE(&timer[1].base, sizeof(int));
	WRITE(&timer[1].index, sizeof(int));
	
	// Window size
	WRITE(&bitmap.viewport.h, sizeof(int));
	WRITE(&bitmap.viewport.w, sizeof(int));
	
	// 68000 CPU
	if (genZ_cpu_cyclone)
	{
	tmp32 = CycloneCPU.d[0];/*m68k_get_reg(NULL, M68K_REG_D0);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[1];/*m68k_get_reg(NULL, M68K_REG_D1);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[2];/*m68k_get_reg(NULL, M68K_REG_D2);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[3];/*m68k_get_reg(NULL, M68K_REG_D3);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[4];/*m68k_get_reg(NULL, M68K_REG_D4);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[5];/*m68k_get_reg(NULL, M68K_REG_D5);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[6];/*m68k_get_reg(NULL, M68K_REG_D6);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.d[7];/*m68k_get_reg(NULL, M68K_REG_D7);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[0];/*m68k_get_reg(NULL, M68K_REG_A0);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[1];/*m68k_get_reg(NULL, M68K_REG_A1);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[2];/*m68k_get_reg(NULL, M68K_REG_A2);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[3];/*m68k_get_reg(NULL, M68K_REG_A3);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[4];/*m68k_get_reg(NULL, M68K_REG_A4);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[5];/*m68k_get_reg(NULL, M68K_REG_A5);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[6];/*m68k_get_reg(NULL, M68K_REG_A6);*/	WRITE(&tmp32, 4);
	tmp32 = CycloneCPU.a[7];/*m68k_get_reg(NULL, M68K_REG_A7);*/	WRITE(&tmp32, 4);
	
	
	
	tmp32=CycloneCPU.pc-CycloneCPU.membase;
	/*tmp32 = m68k_get_reg(NULL, M68K_REG_PC);*/	WRITE(&tmp32, 4);
	
	//low sr : flags
	tmp32=(CycloneCPU.ccr&0xC)|((CycloneCPU.ccr&0x1)<<1)|((CycloneCPU.ccr&0x2)>>1);
	tmp32|=(CycloneCPU.xc&0x2)<<3;	
	WRITE(&tmp32, 1); //m68k_set_reg(M68K_REG_SR, tmp32);
	//high sr : srh
	tmp32=CycloneCPU.srh;
	WRITE(&tmp32, 1);
	
	tmp32=0;
	WRITE(&tmp32, 2);	//dummy
	}
	else
	{
	tmp32 =m68k_get_reg(NULL, M68K_REG_D0);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D1); 	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D2);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D3);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D4);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D5);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D6);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_D7);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A0);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A1);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A2);	WRITE(&tmp32, 4);
	tmp32  =	m68k_get_reg(NULL, M68K_REG_A3);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A4);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A5);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A6);	WRITE(&tmp32, 4);
	tmp32 =	m68k_get_reg(NULL, M68K_REG_A7);	WRITE(&tmp32, 4);		
	tmp32 = m68k_get_reg(NULL, M68K_REG_PC);	WRITE(&tmp32, 4);
		
	tmp32 = m68k_get_reg(NULL, M68K_REG_SR);	WRITE(&tmp32, 4); 
	}
  	
	// Z80 CPU
	tmp32 = z80_get_reg(Z80_PC);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_SP);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_AF);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_BC);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DE);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_HL);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IX);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IY);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_R);				WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_I);				WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_AF2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_BC2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DE2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_HL2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IM);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IFF1);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IFF2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_HALT);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_NMI_STATE);		WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_IRQ_STATE);		WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DC0);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DC1);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DC2);			WRITE(&tmp32, 4);
	tmp32 = z80_get_reg(Z80_DC3);			WRITE(&tmp32, 4);
	
	// Close it
	CLOSE_STREAM(file);
	return 0;
}