#include "lj.h"

#include "datatypes.h"
#include "ljohnGP.h"
#include "tapwave.h"

#include "unes.h"
#include "unes_ppu.h"

extern int32 ljZ_TickCountGet(void);
extern VirtualNES Vnes;

// Assertion Support
static void
assert_check(int error, const char* message)
{
    if (error) {
        ErrDisplayFileLineMsg("PalmOS", (UInt16) error, message);
    }
}

#define xassert(error) {assert_check(error, #error);}

static void showError(const char* aWhat, Err aError) {
    ErrDisplayFileLineMsg("oink", (UInt16) aError, aWhat);
}


static int runcounter;

TwGfxType *lj_mGfxLib;
TwGfxSurfaceType *lj_curRenderingScreen,*lj_RenderingScreen[2];
uint16 *lj_curRenderingScreenPtr,*lj_RenderingScreenPtr[2];
uint32 lj_renderingmode;
uint32 lj_curRenderingScreenIndex;

int lj_padState;

extern int fskiprate;
extern int nes_main (void);
extern long lastTick;

void InitEmuVar()
{	
	unsigned long val;
	lj_padState=0;
}

int LaunchNesEmu(void)
{
 	InitEmuVar();
	return nes_main();
}

int lj_loadrom(void)
{
	ljz_std_showstat();  
	//Init vid
	TwGfxSurfaceInfoType surfInfo;
	surfInfo.size = sizeof(surfInfo);
	surfInfo.width = 256;
	surfInfo.height = 240;
	surfInfo.location = twGfxLocationAcceleratorMemoryNoBackingStore;
	surfInfo.pixelFormat = twGfxPixelFormatRGB565;
	xassert(TwGfxAllocSurface(lj_mGfxLib, &lj_RenderingScreen[0], &surfInfo));
	xassert(TwGfxAllocSurface(lj_mGfxLib, &lj_RenderingScreen[1], &surfInfo));

	if (TwGfxLockSurface(lj_RenderingScreen[0], (void**)&lj_RenderingScreenPtr[0]))
	{
	   return 1;
	}
	if (TwGfxLockSurface(lj_RenderingScreen[1], (void**)&lj_RenderingScreenPtr[1]))
	{
	   return 1;
	}
	
	memset(lj_RenderingScreenPtr[0],0,256*240*2);
	memset(lj_RenderingScreenPtr[1],0,256*240*2);
	
	lj_curRenderingScreenIndex=0;
	lj_curRenderingScreen=lj_RenderingScreen[lj_curRenderingScreenIndex];		
	lj_curRenderingScreenPtr=lj_RenderingScreenPtr[lj_curRenderingScreenIndex];
	
	unsigned long val1,val2;
	FtrGet(APPLI_ID,FTR_OPT_SCREENMODE_ID,(unsigned long*)&val1);
	FtrGet(APPLI_ID,FTR_OPT_SMOOTHING_ID,(unsigned long*)&val2);
	if (!val1) lj_renderingmode=0;
	else 
	if (val1==1) lj_renderingmode=(val2?2:1);
	else
	if (val1==2) lj_renderingmode=(val2?4:3);
	else
	lj_renderingmode=(val2?6:5);
	
	return LaunchNesEmu();
}
int lj_setupenv(struct TwGfxTypeTag *mGfxLib)
{
	lj_mGfxLib=mGfxLib;  
	return 0;
}

void lj_updatevpad(int pad)
{
	lj_padState=pad;
}

int lj_getvpad()
{
	return lj_padState;
}

void lj_changefskip(void)
{
	char tmp[64];
	int i;
	

	fskiprate++;
	if (fskiprate>SKEEZIX_FSKIP) fskiprate=0;
	
	
	if (fskiprate==AUTO_FSKIP) sprintf(tmp,"Frameskip : AUTO");
	else if (fskiprate==SKEEZIX_FSKIP) sprintf(tmp,"Frameskip : Skeezix");
	else sprintf(tmp,"Frameskip : %d",fskiprate);	
	menu_message(tmp);
	
	i=ljZ_TickCountGet();
	while ((ljZ_TickCountGet()-i)<300)
	{
	}
	lastTick=ljZ_TickCountGet()+Vnes.var.frame_time;
}

void lj_changerendering(void)
{
	lj_renderingmode++;
	if (lj_renderingmode==7) lj_renderingmode=0;
/*	if (lj_renderingmode>=5) StatHide();
	if (lj_renderingmode==0) StatShow();*/
	
	menu_message("Changing screen mode");
	lastTick=ljZ_TickCountGet();
	while ((ljZ_TickCountGet()-lastTick)<300)
	{
	}
	lastTick=ljZ_TickCountGet()+Vnes.var.frame_time;
}

void lj_end(void)
{	
	  int val;
 	if ((lj_renderingmode)&&(!(lj_renderingmode&1))) FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID,1);
	else FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID,0);	
	FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,Vnes.var.enablesound);
	switch (Vnes.var.sndfreq)
	{
		case 11025:val=2;break;
		case 22050:val=1;break;
		default:val=0;break;
	}
	FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,val);
	
	FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,((lj_renderingmode+1)>>1));
	
	FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,fskiprate);		

	CloseAll();
	
	ljz_std_showstat();  
}

void lj_pause(int pause)
{
	if (pause)
	{
	}
	else
	{
		lastTick=ljZ_TickCountGet()+Vnes.var.frame_time;
	}
	ljZ_PauseAudio(pause);	
}