//#include "defines.h"
//#include <stdlib.h>
//#include <ctype.h>
//#include "cleantyp.h"
#include "gpstdio.h"
#include "port_dummy.h"
#include "mystdio.h"


int fexists ( const char * filename )
  {
    GPFILEATTR attr;
    
    if (GpFileAttr(filename, &attr)!=SM_OK)
      return FALSE;
    
    return TRUE;
  }

//FILE * fopen ( const char * filename, const char * mode )
FILE * fopen_gp32 ( const char * filename, const char * mode )
  {
    FILE * fh;
    F_HANDLE gpfh;
    unsigned long m = 0;
    int append=FALSE;
    int change=FALSE;

    if (mode[0]=='\0')
      return NULL;
    else
    if ((mode[0]=='r') || (mode[0]=='R'))
      m |= OPEN_R;
    else
    if ((mode[0]=='w') || (mode[0]=='W'))
      m |= OPEN_W;
    else
    if ((mode[0]=='a') || (mode[0]=='A'))
      append=TRUE;
    else
      return NULL;

    if (mode[1]!='\0')
      {
        if (mode[1]=='+')
          change=TRUE;
      }

    if (m&OPEN_W) /* W */
      {
        /* Flush file */
        if (GpFileCreate(filename, ALWAYS_CREATE, &gpfh)!=SM_OK)
          return NULL;
      } else
    if (append)   /* A */
      {
        if (fexists(filename))
          {
            long s;

            if (GpFileOpen(filename, OPEN_R|(change?OPEN_W:0), &gpfh)!=SM_OK)
              return NULL;
            GpFileSeek ( gpfh, FROM_END, 0, &s );
          } else
          {
            if (GpFileCreate(filename, ALWAYS_CREATE, &gpfh)!=SM_OK)
              return NULL;
          }
      } else      /* R */
    if (m&OPEN_R)
      {
        ERR_CODE err =GpFileOpen(filename, m|(change?OPEN_W:0), &gpfh);
        if (err!=SM_OK)
          return NULL;
      } else
      return NULL;

    if ( (fh = malloc ( sizeof(FILE) ))==NULL )
      {
        GpFileClose ( gpfh );
        return NULL;
      }
    
    fh->fhandle = gpfh;

    return fh;
  }

//int fclose ( FILE * fh )
int fclose_gp32 ( FILE * fh )
  {
    if (fh==NULL)
      return EOF;
    
    if (GpFileClose(fh->fhandle)!=SM_OK)
      return EOF;

    free(fh);
    return 0;
  }

int getc ( FILE * fh )
  {
    unsigned char b;
    unsigned long r;

    if (fh==NULL)
      return EOF;
    
    if (GpFileRead(fh->fhandle, &b, 1, &r)!=SM_OK)
      return EOF;
    
    if (r!=1)
      return EOF;
    
    return b;
  }

int fgetc ( FILE * fh )
  {
    return getc(fh);
  }

int putc ( int c, FILE * fh )
  {
    unsigned char b = c;

    if (fh==NULL)
      return EOF;
    
    if (GpFileWrite(fh->fhandle, &b, 1)!=SM_OK)
      return EOF;
    
    return c;
  }

int fputc ( int c, FILE * fh )
  {
    return (putc(c,fh));
  }

long ftell ( FILE * fh )
  {
    long res;
    
    if (fh==NULL)
      return -1;
      
    if (GpFileSeek(fh->fhandle, FROM_CURRENT, 0, &res)!=SM_OK)
      return -1;
    
    return res;
  }

int fseek ( FILE * fh, long offset, int origin )
  {
    long old;
    long mode;

    if (fh==NULL)
      return -1;
    
    if (origin==SEEK_SET)
      mode = FROM_BEGIN;
    else
    if (origin==SEEK_CUR)
      mode = FROM_CURRENT;
    else
    if (origin==SEEK_END)
      {
        mode = FROM_END;
        offset = -offset;
      }
    else
      return -1;

    if (GpFileSeek(fh->fhandle, mode, offset, &old)!=SM_OK)
      return -1;
    
    return 0;
  }

size_t fread ( void * ptr, size_t size, size_t nobj, FILE * fh )
  {
    unsigned int obj;
    unsigned int ofs;
    
    if (nobj==0 || size==0)
      return 0;
      
    if (fh==NULL)
      return 0;
    
    for (obj=0,ofs=0;obj<nobj;obj++,ofs+=size)
      {
        unsigned long r;
        ERR_CODE err;
        err=GpFileRead(fh->fhandle, ((unsigned char*)ptr)+ofs, size, &r);
        if (err!=SM_OK || r!=size)
          return obj;
      }
    return obj;
  }

size_t fwrite ( const void * ptr, size_t size, size_t nobj, FILE * fh )
  {
    int obj,ofs;

    if (fh==NULL)
      return 0;
    
    if (nobj==0 || size==0)
      return 0;
    
    for (obj=0,ofs=0;obj<nobj;obj++,ofs+=size)
      {
        ERR_CODE err;
        err=GpFileWrite ( fh->fhandle, (void*)(((unsigned char*)ptr)+ofs), size );
        if (err!=SM_OK)
          return obj;
      }
    return obj;
  }

int remove ( const char * filename )
  {
    if (GpFileRemove(filename)!=SM_OK)
      return -1;
    else
      return 0;
  }

int rename ( const char * oldname, const char * newname )
  {
    if (GpFileRename(oldname, newname)!=SM_OK)
      return -1;
    else
      return 0;
  }

void rewind ( FILE * fh )
  {
    fseek ( fh, 0, SEEK_SET );
  }
  
int fgetpos( FILE * fh,fpos_t * fpos)
{
long r = 0;
ERR_CODE err;

	if (fh==NULL || fpos==0)
	  return 0;

	err=GpFileSeek(fh->fhandle, FROM_CURRENT, 0, &(r));
	*fpos=r;
    if (err!=SM_OK ) return -1;

    return 0;
  }
