
#include "shared.h"

#include "math.h"

#define PSG_FRQ		3579545


//volatile uint32 s_in_update;
// int32		s_nSamplesPerFrame;
// int32		s_MixBufPos;
// uint32		s_ClockCount;


t_psg s_Psg[8];



 uint32		s_Channel;				// 0 - 5;
 uint32		s_MainVolumeL;			// 0 - 15
 uint32		s_MainVolumeR;			// 0 - 15
 uint32		s_LfoFrq;
 BOOL			s_bLfoOn;
 uint32		s_LfoCtrl;

 uint8		s_Port[15];				// for debug purpose 



int32 s_NoiseTable[32768];


int32		s_VolumeTable[92]={
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
1,1,1,2,2,2,3,4,4,5,6,8,9,11,13,16,
19,23,27,32,38,46,55,65,77,92,109,130,155,184,219,260,
309,367,436,519,617,733,871,1036,1231,1463,1739,2067,2457,2920,3470,4125,
4902,5827,6925,8230,9782,11626,13818,16422,19518,23197,27570,32768};


void create_noise_table()
{
	int32	i;
	uint32	bit0;
	uint32	bit1;
	uint32	bit14;
	uint32	reg = 0x100;	// eLgE 

	for (i = 0; i < 32768; i++)
	{
		bit0 = reg & 1;
		bit1 = (reg & 2) >> 1;
		bit14 = (bit0 ^ bit1);
		reg >>= 1;
		reg |= (bit14 << 14);
		s_NoiseTable[i] = (bit0) ? 15 : -16;
	}
}



/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown routines                                           */
/*--------------------------------------------------------------------------*/

int psg_init(void)
{   
	//memset(s_Psg, 0, sizeof(s_Psg[0])*8);

	create_noise_table(); 
    
    return (0);
}

void psg_reset(void)
{
	memset(s_Psg, 0, sizeof(s_Psg[0])*8);

    s_Channel=0;
        
    s_MainVolumeL = 0;
	s_MainVolumeR = 0;
	s_LfoFrq = 0;
	s_bLfoOn = false;
	s_LfoCtrl = 0;
}

void psg_shutdown(void)
{
}

/*--------------------------------------------------------------------------*/
/* PSG emulation                                                            */
/*--------------------------------------------------------------------------*/

void psg_w(uint16 address, int8 data)
{
	int regNum=address&0xF;
	s_Port[regNum & 15] = data;
	switch (regNum & 15)
	{
		case 0:	// register select
			s_Channel = data & 7;
			break;

		case 1:	// main volume
			s_MainVolumeL = data >> 4;
			s_MainVolumeR = data & 0x0f;
			break;

		case 2:	// frequency low
			s_Psg[s_Channel].frq &= ~0xff;
			s_Psg[s_Channel].frq |= data;
			break;

		case 3:	// frequency high
			s_Psg[s_Channel].frq &= ~0xf00;
			s_Psg[s_Channel].frq |= (data & 0xf) << 8;
			break;

		case 4:	// ON, DDA, AL
//			printf("$0804 = %x\n", data);
			if (data & 0x80)
			{
				// ON rbg 0  1 ňʑJE^Zbg 
				if (!(s_Psg[s_Channel].bOn))
				{
					s_Psg[s_Channel].phase = 0;
				}
				s_Psg[s_Channel].bOn = true;
			}
			else
			{
				s_Psg[s_Channel].bOn = false;
			}

			if (data & 0x40)
			{
				s_Psg[s_Channel].waveIndex = 0;
				s_Psg[s_Channel].bDDA = true;
			}
			else
			{
				s_Psg[s_Channel].bDDA = false;
				s_Psg[s_Channel].daa_index=0;
				s_Psg[s_Channel].daa_phase=0;
			}

			s_Psg[s_Channel].volume = data & 0x1f;
			break;

		case 5:	// LAL, RAL
			s_Psg[s_Channel].volumeL = data >> 4;
			s_Psg[s_Channel].volumeR = data & 0xf;
			break;

		case 6:	// wave data
			if (s_Psg[s_Channel].bDDA)
			{
				s_Psg[s_Channel].daa_data[s_Psg[s_Channel].daa_index++] = data & 0x1f;
				s_Psg[s_Channel].daa_index&=(PSG_DIRECT_ACCESS_BUFSIZE-1);
			}
			else// if (!s_Psg[s_Channel].bOn)
			{
				s_Psg[s_Channel].wave[s_Psg[s_Channel].waveIndex++] = data & 0x1f;
				s_Psg[s_Channel].waveIndex &= 0x1f;
			}
			break;

		case 7:	// noise on, noise frq
//			printf("$0807 = %x\n", data);
			if (s_Channel >= 4)
			{
				s_Psg[s_Channel].bNoiseOn = (data & 0x80) != 0;
				s_Psg[s_Channel].noiseFrq = 0x20 - (data & 0x1f);
			}
			break;

		case 8:	// LFO frequency
			s_LfoFrq = data;
			break;

		case 9:	// LFO on, LFO control
			s_bLfoOn = (data & 0x80) != 0;
			s_LfoCtrl = data & 3;
			break;

		default:	// invalid write
			break;
	}

	return;


}

uint8 psg_r(uint16 address)
{
    int regNum=address&0xF;
    if (regNum == 0)
	{
		return s_Channel;
	}
	return s_Port[regNum & 15];

}




void PSG_Mix(int32*			pDst,int32			nSample)
{
	int32		i;
	int32		j;
	uint32		phase;					// must be 32-bit
	uint32		dp;						// delta phase
	int32		temp[32];
	int32		volumeL;
	int32		volumeR;
	int32		sampleL;
	int32		sampleR;
	int32*		p;

	for (i = 0; i < 6; i++) // 6 channels
	{
		if (!s_Psg[i].bOn)
			continue;

		p = pDst;

//   pow() ₽x̂Ńe[ugƂɂB 
//		volumeL = 91 - (s_Psg[i].volume + (s_MainVolumeL + s_Psg[i].volumeL) * 2);	// attenuation [dB]
//		volumeL = (Sint32)(32768.0 * pow(10.0, volumeL * (-1.5) / 20.0));

//		volumeR = 91 - (s_Psg[i].volume + (s_MainVolumeR + s_Psg[i].volumeR) * 2);	// attenuation [dB]
//		volumeR = (Sint32)(32768.0 * pow(10.0, volumeR * (-1.5) / 20.0));

		volumeL = s_VolumeTable[s_Psg[i].volume + (s_MainVolumeL + s_Psg[i].volumeL) * 2];
		volumeR = s_VolumeTable[s_Psg[i].volume + (s_MainVolumeR + s_Psg[i].volumeR) * 2];

		if (s_Psg[i].bDDA)
		{
/*			dp = (uint32)(65536.0 * 64.0 * (double)PSG_FRQ / (double)(0x1FF) / snd.sample_rate);
			phase = s_Psg[i].daa_phase;
			if ((phase>>22)>=s_Psg[i].daa_index) phase=0;
			
			for (j = 0; j < nSample; j++)	// mixing loop
			{
				sampleL  = (int32)s_Psg[i].daa_data[phase>>22] - 16;
				sampleR = sampleL;
								
				phase += dp;
				if ((phase>>22)>=s_Psg[i].daa_index) phase=0;//(s_Psg[i].daa_index)<<22;
				
				

				*p++ += sampleL * volumeL;
				*p++ += sampleR * volumeR;
			}
			s_Psg[i].daa_phase = phase;*/
		}
		else if (s_Psg[i].bNoiseOn)
		{
			// S{I[o[TvOŃ~bNX̂Œʏ dp SŊ 
			dp = (uint32)(131072.0 * (double)PSG_FRQ / 64.0 / (double)s_Psg[i].noiseFrq / snd.sample_rate / 4.0);

			phase = s_Psg[i].phase;
			for (j = 0; j < nSample; j++)	// mixing loop
			{
				// oversample
				sampleL  = s_NoiseTable[phase>>17];	phase += dp;
				sampleL += s_NoiseTable[phase>>17];	phase += dp;
				sampleL += s_NoiseTable[phase>>17];	phase += dp;
				sampleL += s_NoiseTable[phase>>17];	phase += dp;

				sampleL /= 4;
				sampleR = sampleL;

				sampleL = sampleL * volumeL;
				sampleR = sampleR * volumeR;
				*p++ += sampleL;
				*p++ += sampleR;
			}
			s_Psg[i].phase = phase;
		}
		else
		{
//			if (s_Psg[i].frq == 0)	continue;

			// tɕϊ 
			for (j = 0; j < 32; j++)
			{
				temp[j] = (int32)s_Psg[i].wave[j] - 16;
			}

			dp = (uint32)(65536.0 * 256.0 * 8.0 * (double)PSG_FRQ / (double)(s_Psg[i].frq+1) / snd.sample_rate / 4.0);

			phase = s_Psg[i].phase;
			for (j = 0; j < nSample; j++)	// mixing loop
			{
				sampleL  = temp[phase>>27];	phase += dp;
				sampleL += temp[phase>>27];	phase += dp;
				sampleL += temp[phase>>27];	phase += dp;
				sampleL += temp[phase>>27];	phase += dp;

				sampleL /= 4;
				sampleR = sampleL;

				*p++ += sampleL * volumeL;
				*p++ += sampleR * volumeR;
			}
			s_Psg[i].phase = phase;
		}
	}
}

void update_audio(Int16 *stream, int len)
{
//	memset(stream, 0, len*2*2);
	
	PSG_Mix(snd.buffer, len);
	
	int32 *pSrc=snd.buffer;
	for (;len;len--)
	{
		*stream++ = (int16)(((pSrc[0] * 227) >> 15));
		*stream++ = (int16)(((pSrc[1] * 227) >> 15));
		pSrc[0] = pSrc[1] = 0;
		pSrc += 2;
	}
}
