/*
	File:			Application.cpp	

	Description:	Contains the 68K portion of the Palm ARMlet.  It sets up a form so that
					the UI works with the Palm UI, and it calls the ARMlet.

	Author:		

	Copyright:		Copyright  2003 by Tapwave, Inc.

	Disclaimer:		IMPORTANT:  This Tapwave software is provided by Tapwave, Inc. ("Tapwave").  Your 
					use is subject to and governed by terms and conditions of the Software Development 
					Kit Agreement ("SDK Agreement") between you and Tapwave.  If you have not entered 
					into Tapwaves standard SDK Agreement with Tapwave, you have no right or license 
					to use, reproduce, modify, distribute or otherwise exploit this Tapwave software.  
					You may obtain a copy of Tapwaves standard SDK Agreement by calling 650-960-1817
					or visiting Tapwave at http://www.tapwave.com/developers/. 

	Change History (most recent first):
				
*/

#include <TapWave.h>
#include <TwRuntime.h>
#include <StringMgr.h>

#define __NES_ROMS_DIR__ "/PALM/Programs/LJZ/NES/Roms/"
#define __SNES_ROMS_DIR__ "/PALM/Programs/LJZ/SNES/Roms/"
#define __GB_ROMS_DIR__ "/PALM/Programs/LJZ/GB/Roms/"
#define __GENZ_ROMS_DIR__ "/PALM/Programs/LJZ/GENZ/Roms/"
#define __NEOPOPZ_ROMS_DIR__ "/PALM/Programs/LJZ/NGP/Roms/"
#define __OSWANZ_ROMS_DIR__ "/PALM/Programs/LJZ/WS/Roms/"

#define WELCOME_MESSAGE "Welcome to LJZ 0.5-RC3"


#include "ljZRsc.h"

/***********************************************************************
 *
 *	Internal Constants
 *
 ***********************************************************************/

#define FORM_LISTROM_ID 5000

//#define FORM_LISTINPUT_ID 5001

#define FTR_ROM_FILENAME_ID 6000
#define FTR_OPT_SMOOTHING_ID 6002
#define FTR_OPT_SCREENMODE_ID 6003
#define FTR_OPT_FSKIP_ID 6004
#define FTR_OPT_CPUCORE_ID 6005
#define FTR_OPT_EMU_ACCURACY_ID 6006
#define FTR_OPT_SOUND_ID 6010
#define FTR_OPT_SOUNDFREQ_ID 6011

#define FTR_OPT_SNESRENDER_ID 6020
#define FTR_OPT_SNESTRANSP_ID 6021
#define FTR_OPT_FAST_SPR_ID 6022

#define FTR_OPT_INPUTCONF_ID 6030

#define SYS_NES     1
#define SYS_SNES    2
#define SYS_GB      4
#define SYS_GENZ    8
#define SYS_NEOPOPZ 16
#define SYS_OSWANZ  32

#define APPLI_ID 'LJZZ'

#define VERSION 0x0053


#define BTN_Red		0x01
#define BTN_Green	0x02
#define BTN_Yellow	0x03
#define BTN_Blue	0x04
#define BTN_L		0x05
#define BTN_R		0x06
#define BTN_Func	0x07
#define BTN_Home	0x08
#define BTN_Up		0x09
#define BTN_Down	0x0A
#define BTN_Left	0x0B
#define BTN_Right	0x0C
#define BTN_TL		0x0D
#define BTN_BL		0x0E
#define BTN_TR		0x0F
#define BTN_BR		0x10
#define BTN_Anal	0x11


#define ASK_FOR_UPDATE \
	{EventType my_event; \
	my_event.eType=frmUpdateEvent;\
	EvtAddEventToQueue(&my_event);}
	
#define DRAW_SYSTEM_ICON(S,X,Y)\
	{\
		MemHandle hRsc=NULL;\
			switch (S)\
			{\
				case SYS_NES:hRsc = DmGetResource(bitmapRsc, NESBMP);break;\
				case SYS_SNES:hRsc = DmGetResource(bitmapRsc, SNESBMP);break;\
				case SYS_GB:hRsc = DmGetResource(bitmapRsc, GBBMP);break;\
				case SYS_GENZ:hRsc = DmGetResource(bitmapRsc, GENESISBMP);break;\
				case SYS_NEOPOPZ:hRsc = DmGetResource(bitmapRsc, NGPBMP);break;\
				case SYS_OSWANZ:hRsc = DmGetResource(bitmapRsc, WSBMP);break;\
			}\
			if (hRsc)\
			{	BitmapType* bitmapP = (BitmapType*) MemHandleLock(hRsc);\
				WinPaintBitmap(bitmapP, X, Y);\
				MemHandleUnlock(hRsc);\
				DmReleaseResource(hRsc);}\
   }

static char g_labelbutton[0x12][10]=
{"none","Red","Green","Yellow","Blue","L","R","Func","Home",
"Up","Down","Left","Right","TL","BL","TR","BR","Analog"};

//static unsigned char g_ljz_pref.g_inputconfinv[4][20];
static char g_inputfld[255];

static char g_lastQuickRomChar;

static char g_helpmsg[]="\
[LJZ] v0.5-RC3\n\
--------------\n\
Compiled on 2004/05/25\n\
\n\
* multiple screen mode (zoom/smoothing)\n\
* adjustable frameskipping\n\
* SRAM support\n\
* Load/Save state\n\
* Turbo mode\n\
* sound (except Wonderswan)\n\
* zipped rom\n\
\n\
Put your ROMS in \n\
/PALM/Programs/LJZ/xxx/Roms/\n\
Where xxx is NES,SNES,...\n\
\n\
[GB/GBC]\n\
* powered by fGB32(gnuboy)\n\
* from great Rlyeh & gnuboy team\n\
\n\
[GENESIS]\n\
* powered by GenesisPlus\n\
from Charles Mac Donald\n\
* powered by Cyclone 68000 ARM cpu core\n\
from Dave\n\
\n\
[NEOGEO POCKET]\n\
* ppowered by NeoGP32 from great THUNDERZ!\n\
\n\
[NES]\n\
* using/inspired by\n\
Shatbox,Nofrendo,Nester,NNNEsterJ,\n\
NesterDC,PocketNester,Nestopia\n\
* support mappers :\n\
0,1,2,3,4,5,6,7,8,9,10,11,13,15,\n\
16,17,18,19,21,22,23,24,25,26,\n\
32,33,34,40,41,42,43,44,45,64,\n\
66,68,69,70,71,72,73,74,75,76,\n\
77,78,79,80,82,83,85,86,87,88,\n\
89,90,91,92,93,94,95,96,97,99,\n\
100,101,105,112,113,114,115,117,\n\
118,119,122,187,255\n\
\n\
[SNES]\n\
* powered by OpenSnes9xGP(Snes9X 1.39a)\n\
from Laxer3a,ThunderZ,YoyoFR\n\
* spc playback through openspc++\n\
\n\
[WONDERSWAN]\n\
* powered by Oswan 0.7 \n\
* No audio, slow emulation\n\
* still beta emu\n\
----------\n\
Greetings to\n\
Skeezix\n\
Rlyeh\n\
Dave\n\
Sine909\n\
All the donators!\n\
....\n\
\n\
\n\
\tYoyoFR\n";

static int g_midi;
static unsigned char* g_midiData;
static MemHandle g_midiDataH;
static TwSmfType* g_smfHandle;

static int g_frmloaded;
static int g_config_mode=0;
static int g_rom_system;

typedef struct {
UInt16 g_SnesFastSprite;
UInt16 g_SnesTransparency;
UInt16 g_SnesRender;
UInt16 g_SnesSound;
UInt16 g_SnesSmoothing;
UInt16 g_SnesScreenMode;
UInt16 g_SnesFSkip;
UInt16 g_SnesSoundFreq;
UInt16 g_SnesASMCPU;

UInt16 g_NesSound;
UInt16 g_NesSmoothing;
UInt16 g_NesScreenMode;
UInt16 g_NesFSkip;
UInt16 g_NesSoundFreq;

UInt16 g_GBSound;
UInt16 g_GBSmoothing;
UInt16 g_GBScreenMode;
UInt16 g_GBFSkip;
UInt16 g_GBSoundFreq;

UInt16 g_GENZSound;
UInt16 g_GENZSmoothing;
UInt16 g_GENZScreenMode;
UInt16 g_GENZFSkip;
UInt16 g_GENZSoundFreq;
UInt16 g_GENZASMCPU;


UInt16 g_NEOPOPZSound;
UInt16 g_NEOPOPZSmoothing;
UInt16 g_NEOPOPZScreenMode;
UInt16 g_NEOPOPZFSkip;
UInt16 g_NEOPOPZSoundFreq;
UInt16 g_NEOPOPZEmuAccuracy;


UInt16 g_OSWANZSound;
UInt16 g_OSWANZSmoothing;
UInt16 g_OSWANZScreenMode;
UInt16 g_OSWANZFSkip;
UInt16 g_OSWANZSoundFreq;

UInt16 g_sysfilter;//|SYS_SNES|SYS_GB|SYS_GENZ;
UInt16 g_conf_system;
unsigned char g_inputconf[6][20];
char g_lastrom[256];
UInt16 g_firsttime;
} t_ljz_pref;

static t_ljz_pref g_ljz_pref;
/**********************************/

static UInt32 my_sin[512]={
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,
1,1,1,1,1,2,2,2,2,2,2,2,2,3,3,3,
3,3,3,3,4,4,4,4,4,4,5,5,5,5,5,5,
6,6,6,6,6,7,7,7,7,7,8,8,8,8,9,9,
9,9,10,10,10,10,11,11,11,11,12,12,12,12,13,13,
13,14,14,14,14,15,15,15,16,16,16,17,17,17,18,18,
18,19,19,19,20,20,20,21,21,21,22,22,22,23,23,23,
24,24,25,25,25,26,26,27,27,27,28,28,29,29,29,30,
30,31,31,31,32,32,33,33,34,34,35,35,35,36,36,37,
37,38,38,39,39,40,40,41,41,42,42,43,43,44,44,45,
45,46,46,47,47,48,48,49,49,50,50,51,51,52,52,53,
53,54,55,55,56,56,57,57,58,58,59,60,60,61,61,62,
62,63,64,64,65,65,66,67,67,68,68,69,70,70,71,72,
72,73,73,74,75,75,76,77,77,78,78,79,80,80,81,82,
82,83,84,84,85,86,86,87,88,88,89,90,90,91,92,93,
93,94,95,95,96,97,97,98,99,100,100,101,102,102,103,104,
105,105,106,107,108,108,109,110,110,111,112,113,113,114,115,116,
116,117,118,119,120,120,121,122,123,123,124,125,126,127,127,128,
129,130,130,131,132,133,134,134,135,136,137,138,138,139,140,141,
142,143,143,144,145,146,147,147,148,149,150,151,152,152,153,154,
155,156,157,158,158,159,160,161,162,163,163,164,165,166,167,168,
169,170,170,171,172,173,174,175,176,177,177,178,179,180,181,182,
183,184,184,185,186,187,188,189,190,191,192,193,193,194,195,196,
197,198,199,200,201,202,203,203,204,205,206,207,208,209,210,211,
212,213,214,214,215,216,217,218,219,220,221,222,223,224,225,226,
227,228,228,229,230,231,232,233,234,235,236,237,238,239,240,241,
242,243,244,245,246,247,247,248,249,250,251,252,253,254,255,256,
257,258,259,260,261,262,263,264,265,266,267,268,269,270,271,272,
273,274,274,275,276,277,278,279,280,281,282,283,284,285,286,287,
288,289,290,291,292,293,294,295,296,297,298,299,300,301,302,303,
304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,319,
};

static UInt32 my_cos[512]={
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,1,1,1,1,1,2,2,2,2,2,3,3,3,3,4,
4,4,4,5,5,5,6,6,6,7,7,7,8,8,8,9,
9,10,10,10,11,11,12,12,13,13,14,14,15,15,16,16,
17,17,18,18,19,19,20,21,21,22,22,23,24,24,25,25,
26,27,27,28,29,29,30,31,32,32,33,34,34,35,36,37,
37,38,39,40,41,41,42,43,44,45,45,46,47,48,49,50,
51,51,52,53,54,55,56,57,58,59,60,61,62,62,63,64,
65,66,67,68,69,70,71,72,73,74,75,76,78,79,80,81,
82,83,84,85,86,87,88,89,90,92,93,94,95,96,97,98,
99,101,102,103,104,105,106,108,109,110,111,112,114,115,116,117,
118,120,121,122,123,125,126,127,128,130,131,132,133,135,136,137,
138,140,141,142,144,145,146,147,149,150,151,153,154,155,157,158,
159,161,162,163,164,166,167,168,170,171,173,174,175,177,178,179,
181,182,183,185,186,187,189,190,191,193,194,196,197,198,200,201,
202,204,205,207,208,209,211,212,213,215,216,218,219,220,222,223,
224,226,227,229,230,231,233,234,236,237,238,240,241,242,244,245,
247,248,249,251,252,253,255,256,258,259,260,262,263,264,266,267,
268,270,271,272,274,275,276,278,279,281,282,283,285,286,287,288,
290,291,292,294,295,296,298,299,300,302,303,304,305,307,308,309,
311,312,313,314,316,317,318,319,321,322,323,324,326,327,328,329,
331,332,333,334,335,337,338,339,340,341,343,344,345,346,347,348,
350,351,352,353,354,355,356,357,359,360,361,362,363,364,365,366,
367,368,369,370,371,373,374,375,376,377,378,379,380,381,382,383,
384,385,386,387,387,388,389,390,391,392,393,394,395,396,397,398,
398,399,400,401,402,403,404,404,405,406,407,408,408,409,410,411,
412,412,413,414,415,415,416,417,417,418,419,420,420,421,422,422,
423,424,424,425,425,426,427,427,428,428,429,430,430,431,431,432,
432,433,433,434,434,435,435,436,436,437,437,438,438,439,439,439,
440,440,441,441,441,442,442,442,443,443,443,444,444,444,445,445,
445,445,446,446,446,446,447,447,447,447,447,448,448,448,448,448,
448,449,449,449,449,449,449,449,449,449,449,449,449,449,449,449,
};

static Int32 fix_cos[512]={
512,511,511,511,511,511,510,510,509,508,508,507,506,505,504,503,
502,500,499,498,496,495,493,491,489,488,486,484,482,479,477,475,
473,470,468,465,462,460,457,454,451,448,445,442,439,435,432,429,
425,422,418,414,411,407,403,399,395,391,387,383,379,375,370,366,
362,357,353,348,343,339,334,329,324,319,314,310,304,299,294,289,
284,279,273,268,263,257,252,246,241,235,230,224,218,213,207,201,
195,190,184,178,172,166,160,154,148,142,136,130,124,118,112,106,
99,93,87,81,75,68,62,56,50,43,37,31,25,18,12,6,
0,-6,-12,-18,-25,-31,-37,-43,-50,-56,-62,-68,-75,-81,-87,-93,
-99,-106,-112,-118,-124,-130,-136,-142,-148,-154,-160,-166,-172,-178,-184,-190,
-195,-201,-207,-213,-218,-224,-230,-235,-241,-246,-252,-257,-263,-268,-273,-279,
-284,-289,-294,-299,-304,-310,-314,-319,-324,-329,-334,-339,-343,-348,-353,-357,
-362,-366,-370,-375,-379,-383,-387,-391,-395,-399,-403,-407,-411,-414,-418,-422,
-425,-429,-432,-435,-439,-442,-445,-448,-451,-454,-457,-460,-462,-465,-468,-470,
-473,-475,-477,-479,-482,-484,-486,-488,-489,-491,-493,-495,-496,-498,-499,-500,
-502,-503,-504,-505,-506,-507,-508,-508,-509,-510,-510,-511,-511,-511,-511,-511,
-511,-511,-511,-511,-511,-511,-510,-510,-509,-508,-508,-507,-506,-505,-504,-503,
-502,-500,-499,-498,-496,-495,-493,-491,-489,-488,-486,-484,-482,-479,-477,-475,
-473,-470,-468,-465,-462,-460,-457,-454,-451,-448,-445,-442,-439,-435,-432,-429,
-425,-422,-418,-414,-411,-407,-403,-399,-395,-391,-387,-383,-379,-375,-370,-366,
-362,-357,-353,-348,-343,-339,-334,-329,-324,-319,-314,-310,-304,-299,-294,-289,
-284,-279,-273,-268,-263,-257,-252,-246,-241,-235,-230,-224,-218,-213,-207,-201,
-195,-190,-184,-178,-172,-166,-160,-154,-148,-142,-136,-130,-124,-118,-112,-106,
-99,-93,-87,-81,-75,-68,-62,-56,-50,-43,-37,-31,-25,-18,-12,-6,
0,6,12,18,25,31,37,43,50,56,62,68,75,81,87,93,
99,106,112,118,124,130,136,142,148,154,160,166,172,178,184,190,
195,201,207,213,218,224,230,235,241,246,252,257,263,268,273,279,
284,289,294,299,304,310,314,319,324,329,334,339,343,348,353,357,
362,366,370,375,379,383,387,391,395,399,403,407,411,414,418,422,
425,429,432,435,439,442,445,448,451,454,457,460,462,465,468,470,
473,475,477,479,482,484,486,488,489,491,493,495,496,498,499,500,
502,503,504,505,506,507,508,508,509,510,510,511,511,511,511,511,
};


static Int32 fix_sin[512]={
0,6,12,18,25,31,37,43,50,56,62,68,75,81,87,93,
99,106,112,118,124,130,136,142,148,154,160,166,172,178,184,190,
195,201,207,213,218,224,230,235,241,246,252,257,263,268,273,279,
284,289,294,299,304,310,314,319,324,329,334,339,343,348,353,357,
362,366,370,375,379,383,387,391,395,399,403,407,411,414,418,422,
425,429,432,435,439,442,445,448,451,454,457,460,462,465,468,470,
473,475,477,479,482,484,486,488,489,491,493,495,496,498,499,500,
502,503,504,505,506,507,508,508,509,510,510,511,511,511,511,511,
511,511,511,511,511,511,510,510,509,508,508,507,506,505,504,503,
502,500,499,498,496,495,493,491,489,488,486,484,482,479,477,475,
473,470,468,465,462,460,457,454,451,448,445,442,439,435,432,429,
425,422,418,414,411,407,403,399,395,391,387,383,379,375,370,366,
362,357,353,348,343,339,334,329,324,319,314,310,304,299,294,289,
284,279,273,268,263,257,252,246,241,235,230,224,218,213,207,201,
195,190,184,178,172,166,160,154,148,142,136,130,124,118,112,106,
99,93,87,81,75,68,62,56,50,43,37,31,25,18,12,6,
0,-6,-12,-18,-25,-31,-37,-43,-50,-56,-62,-68,-75,-81,-87,-93,
-99,-106,-112,-118,-124,-130,-136,-142,-148,-154,-160,-166,-172,-178,-184,-190,
-195,-201,-207,-213,-218,-224,-230,-235,-241,-246,-252,-257,-263,-268,-273,-279,
-284,-289,-294,-299,-304,-310,-314,-319,-324,-329,-334,-339,-343,-348,-353,-357,
-362,-366,-370,-375,-379,-383,-387,-391,-395,-399,-403,-407,-411,-414,-418,-422,
-425,-429,-432,-435,-439,-442,-445,-448,-451,-454,-457,-460,-462,-465,-468,-470,
-473,-475,-477,-479,-482,-484,-486,-488,-489,-491,-493,-495,-496,-498,-499,-500,
-502,-503,-504,-505,-506,-507,-508,-508,-509,-510,-510,-511,-511,-511,-511,-511,
-511,-511,-511,-511,-511,-511,-510,-510,-509,-508,-508,-507,-506,-505,-504,-503,
-502,-500,-499,-498,-496,-495,-493,-491,-489,-488,-486,-484,-482,-479,-477,-475,
-473,-470,-468,-465,-462,-460,-457,-454,-451,-448,-445,-442,-439,-435,-432,-429,
-425,-422,-418,-414,-411,-407,-403,-399,-395,-391,-387,-383,-379,-375,-370,-366,
-362,-357,-353,-348,-343,-339,-334,-329,-324,-319,-315,-310,-305,-299,-294,-289,
-284,-279,-273,-268,-263,-257,-252,-246,-241,-235,-230,-224,-218,-213,-207,-201,
-195,-190,-184,-178,-172,-166,-160,-154,-148,-142,-136,-130,-124,-118,-112,-106,
-99,-93,-87,-81,-75,-68,-62,-56,-50,-43,-37,-31,-25,-18,-12,-6,
};


