
;@ Dave's Cyclone 68000 Emulator v0.069 - Assembler Output

;@ (c) Copyright 2003 Dave, All rights reserved.
;@ Cyclone 68000 is free for non-commercial use.

;@ For commercial use, separate licencing terms must be obtained.

  area |.text|, code
  export CycloneRun
  export CycloneVer

CycloneVer dcd 0x0069

;@ --------------------------- Framework --------------------------
CycloneRun
  stmdb sp!,{r4-r11,lr}
  mov r7,r0          ;@ r7 = Pointer to Cpu Context
                     ;@ r0-3 = Temporary registers
  ldrb r9,[r7,#0x46] ;@ r9 = Flags (NZCV)
  ldr r6,=JumpTab    ;@ r6 = Opcode Jump table
  add r6,r6,r10	   ;@ r6 = relocated
  mov r12,r10		   ;@ r12 = PALMOS (CW) ARMlet relocation ptr
  ldr r5,[r7,#0x5c]  ;@ r5 = Cycles
  ldr r4,[r7,#0x40]  ;@ r4 = Current PC + Memory Base
                     ;@ r8 = Current Opcode
  mov r9,r9,lsl #28  ;@ r9 = Flags 0xf0000000, cpsr format
                     ;@ r10 = Source value / Memory Base

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

;@ Check if interrupt used up all the cycles:
  subs r5,r5,#0
  blt CycloneEndNoBack
  ldrh r8,[r4],#2 ;@ Fetch first opcode
  ldr r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  add pc,r0,r12 ;@ Jump to opcode handler
  LTORG

;@ We come back here after execution
CycloneEnd
  sub r4,r4,#2
CycloneEndNoBack
  mov r9,r9,lsr #28
  str r4,[r7,#0x40]  ;@ Save Current PC + Memory Base
  str r5,[r7,#0x5c]  ;@ Save Cycles
  strb r9,[r7,#0x46] ;@ Save Flags (NZCV)
  ldmia sp!,{r4-r11,pc}

;@ DoInterrupt - r0=IRQ number
DoInterrupt

  ldrb r1,[r7,#0x44] ;@ Get SR high: T_S__III
  and r1,r1,#7 ;@ Get interrupt mask
  cmp r0,#6 ;@ irq>6 ?
  cmple r0,r1 ;@ irq<=6: Is irq<=mask ?
  movle pc,lr ;@ irq<=6 and mask, not allowed

  ldr r10,[r7,#0x60] ;@ Get Memory base
  mov r11,lr ;@ Preserve ARM return address
  sub r1,r4,r10 ;@ r1 = Old PC
;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Push SR:
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  ldr r0,[r7,#0x3c]
  sub r0,r0,#2 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get IRQ Vector address:
  ldrb r1,[r7,#0x47] ;@ IRQ
  mov r0,r1,asl #2
  add r0,r0,#0x60
;@ Read IRQ Vector:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  add r0,r0,r10 ;@ r0 = Memory Base + New PC
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ldrb r0,[r7,#0x47] ;@ IRQ
  mov lr,pc
  ldr pc,[r7,#0x94] ;@ Call irq_callback()
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ todo - swap OSP and A7 if not in Supervisor mode
  ldrb r0,[r7,#0x47] ;@ IRQ
  orr r0,r0,#0x20 ;@ Supervisor mode + IRQ number
  strb r0,[r7,#0x44] ;@ Put SR high

;@ Clear irq:
  mov r0,#0
  strb r0,[r7,#0x47]
  subs r5,r5,#46 ;@ Subtract cycles
  mov pc,r11 ;@ Return

Exception

  mov r11,lr ;@ Preserve ARM return address
  ;@ Cause an Exception - Vector in [r7,#0x50]
  ldr r10,[r7,#0x60] ;@ Get Memory base
  sub r1,r4,r10 ;@ r1 = Old PC
;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Push SR:
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  ldr r0,[r7,#0x3c]
  sub r0,r0,#2 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldr r0,[r7,#0x50] ;@ Get Vector
;@ Read IRQ Vector:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  add r0,r0,r10 ;@ r0 = Memory Base + New PC
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r4,r0

  mov pc,r11 ;@ Return

;@ ---------------------------- Opcodes ---------------------------
Op____ ;@ Called if an opcode is not recognised
;@ ---------- [ffffffff]  uses Opffffffff ----------
Opffffffff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

  b CycloneEnd

;@ ---------- [0000] ori.b #$33, d0 uses Op0000 ----------
Op0000
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0010] ori.b #$33, (a0) uses Op0010 ----------
Op0010
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0018] ori.b #$33, (a0)+ uses Op0018 ----------
Op0018
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [001f] ori.b #$33, (a7)+ uses Op001f ----------
Op001f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0020] ori.b #$33, -(a0) uses Op0020 ----------
Op0020
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0027] ori.b #$33, -(a7) uses Op0027 ----------
Op0027
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0028] ori.b #$33, ($3333,a0) uses Op0028 ----------
Op0028
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0030] ori.b #$33, ($33,a0,d3.w*2) uses Op0030 ----------
Op0030
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0038] ori.b #$33, $3333.w uses Op0038 ----------
Op0038
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0039] ori.b #$33, $33333333.l uses Op0039 ----------
Op0039
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [003a] ori.b #$33, ($3333,pc); =3337 uses Op003a ----------
Op003a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [003b] ori.b #$33, ($33,pc,d3.w*2); =37 uses Op003b ----------
Op003b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [003c] ori.b #$33, ccr uses Op003c ----------
Op003c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode003c ;@ No

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  orr r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode003c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0040] ori.w #$3333, d0 uses Op0040 ----------
Op0040
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #16
  strh r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0050] ori.w #$3333, (a0) uses Op0050 ----------
Op0050
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0058] ori.w #$3333, (a0)+ uses Op0058 ----------
Op0058
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [005f] ori.w #$3333, (a7)+ uses Op005f ----------
Op005f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0060] ori.w #$3333, -(a0) uses Op0060 ----------
Op0060
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0067] ori.w #$3333, -(a7) uses Op0067 ----------
Op0067
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0068] ori.w #$3333, ($3333,a0) uses Op0068 ----------
Op0068
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0070] ori.w #$3333, ($33,a0,d3.w*2) uses Op0070 ----------
Op0070
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0078] ori.w #$3333, $3333.w uses Op0078 ----------
Op0078
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0079] ori.w #$3333, $33333333.l uses Op0079 ----------
Op0079
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [007a] ori.w #$3333, ($3333,pc); =3337 uses Op007a ----------
Op007a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [007b] ori.w #$3333, ($33,pc,d3.w*2); =37 uses Op007b ----------
Op007b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [007c] ori.w #$3333, sr uses Op007c ----------
Op007c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode007c ;@ No

;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  orr r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode007c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0080] ori.l #$33333333, d0 uses Op0080 ----------
Op0080
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0090] ori.l #$33333333, (a0) uses Op0090 ----------
Op0090
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0098] ori.l #$33333333, (a0)+ uses Op0098 ----------
Op0098
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [009f] ori.l #$33333333, (a7)+ uses Op009f ----------
Op009f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00a0] ori.l #$33333333, -(a0) uses Op00a0 ----------
Op00a0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00a7] ori.l #$33333333, -(a7) uses Op00a7 ----------
Op00a7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00a8] ori.l #$33333333, ($3333,a0) uses Op00a8 ----------
Op00a8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00b0] ori.l #$33333333, ($33,a0,d3.w*2) uses Op00b0 ----------
Op00b0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00b8] ori.l #$33333333, $3333.w uses Op00b8 ----------
Op00b8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00b9] ori.l #$33333333, $33333333.l uses Op00b9 ----------
Op00b9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00ba] ori.l #$33333333, ($3333,pc); =3339 uses Op00ba ----------
Op00ba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [00bb] ori.l #$33333333, ($33,pc,d3.w*2); =39 uses Op00bb ----------
Op00bb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  orr r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0100] btst d0, d0 uses Op0100 ----------
Op0100
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0110] btst d0, (a0) uses Op0110 ----------
Op0110
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0118] btst d0, (a0)+ uses Op0118 ----------
Op0118
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [011f] btst d0, (a7)+ uses Op011f ----------
Op011f
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0120] btst d0, -(a0) uses Op0120 ----------
Op0120
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0127] btst d0, -(a7) uses Op0127 ----------
Op0127
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0128] btst d0, ($3333,a0) uses Op0128 ----------
Op0128
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0130] btst d0, ($33,a0,d3.w*2) uses Op0130 ----------
Op0130
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0138] btst d0, $3333.w uses Op0138 ----------
Op0138
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0139] btst d0, $33333333.l uses Op0139 ----------
Op0139
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [013a] btst d0, ($3333,pc); =3335 uses Op013a ----------
Op013a
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [013b] btst d0, ($33,pc,d3.w*2); =35 uses Op013b ----------
Op013b
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [013c] btst d0, #$33 uses Op013c ----------
Op013c
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '#$33' into r11:
  ldrsb r11,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r11) into r0:
  mov r0,r11,asl #0

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0140] bchg d0, d0 uses Op0140 ----------
Op0140
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0150] bchg d0, (a0) uses Op0150 ----------
Op0150
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0158] bchg d0, (a0)+ uses Op0158 ----------
Op0158
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [015f] bchg d0, (a7)+ uses Op015f ----------
Op015f
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0160] bchg d0, -(a0) uses Op0160 ----------
Op0160
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0167] bchg d0, -(a7) uses Op0167 ----------
Op0167
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0168] bchg d0, ($3333,a0) uses Op0168 ----------
Op0168
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0170] bchg d0, ($33,a0,d3.w*2) uses Op0170 ----------
Op0170
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0178] bchg d0, $3333.w uses Op0178 ----------
Op0178
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0179] bchg d0, $33333333.l uses Op0179 ----------
Op0179
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [017a] bchg d0, ($3333,pc); =3335 uses Op017a ----------
Op017a
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [017b] bchg d0, ($33,pc,d3.w*2); =35 uses Op017b ----------
Op017b
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0180] bclr d0, d0 uses Op0180 ----------
Op0180
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0190] bclr d0, (a0) uses Op0190 ----------
Op0190
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0198] bclr d0, (a0)+ uses Op0198 ----------
Op0198
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [019f] bclr d0, (a7)+ uses Op019f ----------
Op019f
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01a0] bclr d0, -(a0) uses Op01a0 ----------
Op01a0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01a7] bclr d0, -(a7) uses Op01a7 ----------
Op01a7
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01a8] bclr d0, ($3333,a0) uses Op01a8 ----------
Op01a8
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01b0] bclr d0, ($33,a0,d3.w*2) uses Op01b0 ----------
Op01b0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01b8] bclr d0, $3333.w uses Op01b8 ----------
Op01b8
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01b9] bclr d0, $33333333.l uses Op01b9 ----------
Op01b9
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01ba] bclr d0, ($3333,pc); =3335 uses Op01ba ----------
Op01ba
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01bb] bclr d0, ($33,pc,d3.w*2); =35 uses Op01bb ----------
Op01bb
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01c0] bset d0, d0 uses Op01c0 ----------
Op01c0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01d0] bset d0, (a0) uses Op01d0 ----------
Op01d0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01d8] bset d0, (a0)+ uses Op01d8 ----------
Op01d8
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01df] bset d0, (a7)+ uses Op01df ----------
Op01df
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01e0] bset d0, -(a0) uses Op01e0 ----------
Op01e0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01e7] bset d0, -(a7) uses Op01e7 ----------
Op01e7
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01e8] bset d0, ($3333,a0) uses Op01e8 ----------
Op01e8
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01f0] bset d0, ($33,a0,d3.w*2) uses Op01f0 ----------
Op01f0
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01f8] bset d0, $3333.w uses Op01f8 ----------
Op01f8
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01f9] bset d0, $33333333.l uses Op01f9 ----------
Op01f9
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01fa] bset d0, ($3333,pc); =3335 uses Op01fa ----------
Op01fa
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [01fb] bset d0, ($33,pc,d3.w*2); =35 uses Op01fb ----------
Op01fb
  mov r10,#1
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0200] andi.b #$33, d0 uses Op0200 ----------
Op0200
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0210] andi.b #$33, (a0) uses Op0210 ----------
Op0210
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0218] andi.b #$33, (a0)+ uses Op0218 ----------
Op0218
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [021f] andi.b #$33, (a7)+ uses Op021f ----------
Op021f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0220] andi.b #$33, -(a0) uses Op0220 ----------
Op0220
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0227] andi.b #$33, -(a7) uses Op0227 ----------
Op0227
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0228] andi.b #$33, ($3333,a0) uses Op0228 ----------
Op0228
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0230] andi.b #$33, ($33,a0,d3.w*2) uses Op0230 ----------
Op0230
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0238] andi.b #$33, $3333.w uses Op0238 ----------
Op0238
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0239] andi.b #$33, $33333333.l uses Op0239 ----------
Op0239
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [023a] andi.b #$33, ($3333,pc); =3337 uses Op023a ----------
Op023a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [023b] andi.b #$33, ($33,pc,d3.w*2); =37 uses Op023b ----------
Op023b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [023c] andi.b #$33, ccr uses Op023c ----------
Op023c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode023c ;@ No

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  and r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode023c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0240] andi.w #$3333, d0 uses Op0240 ----------
Op0240
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #16
  strh r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0250] andi.w #$3333, (a0) uses Op0250 ----------
Op0250
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0258] andi.w #$3333, (a0)+ uses Op0258 ----------
Op0258
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [025f] andi.w #$3333, (a7)+ uses Op025f ----------
Op025f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0260] andi.w #$3333, -(a0) uses Op0260 ----------
Op0260
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0267] andi.w #$3333, -(a7) uses Op0267 ----------
Op0267
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0268] andi.w #$3333, ($3333,a0) uses Op0268 ----------
Op0268
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0270] andi.w #$3333, ($33,a0,d3.w*2) uses Op0270 ----------
Op0270
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0278] andi.w #$3333, $3333.w uses Op0278 ----------
Op0278
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0279] andi.w #$3333, $33333333.l uses Op0279 ----------
Op0279
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [027a] andi.w #$3333, ($3333,pc); =3337 uses Op027a ----------
Op027a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [027b] andi.w #$3333, ($33,pc,d3.w*2); =37 uses Op027b ----------
Op027b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [027c] andi.w #$3333, sr uses Op027c ----------
Op027c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode027c ;@ No

;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  and r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode027c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0280] andi.l #$33333333, d0 uses Op0280 ----------
Op0280
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0290] andi.l #$33333333, (a0) uses Op0290 ----------
Op0290
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0298] andi.l #$33333333, (a0)+ uses Op0298 ----------
Op0298
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [029f] andi.l #$33333333, (a7)+ uses Op029f ----------
Op029f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02a0] andi.l #$33333333, -(a0) uses Op02a0 ----------
Op02a0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02a7] andi.l #$33333333, -(a7) uses Op02a7 ----------
Op02a7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02a8] andi.l #$33333333, ($3333,a0) uses Op02a8 ----------
Op02a8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02b0] andi.l #$33333333, ($33,a0,d3.w*2) uses Op02b0 ----------
Op02b0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02b8] andi.l #$33333333, $3333.w uses Op02b8 ----------
Op02b8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02b9] andi.l #$33333333, $33333333.l uses Op02b9 ----------
Op02b9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02ba] andi.l #$33333333, ($3333,pc); =3339 uses Op02ba ----------
Op02ba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [02bb] andi.l #$33333333, ($33,pc,d3.w*2); =39 uses Op02bb ----------
Op02bb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  and r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0400] subi.b #$33, d0 uses Op0400 ----------
Op0400
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0410] subi.b #$33, (a0) uses Op0410 ----------
Op0410
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0418] subi.b #$33, (a0)+ uses Op0418 ----------
Op0418
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [041f] subi.b #$33, (a7)+ uses Op041f ----------
Op041f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0420] subi.b #$33, -(a0) uses Op0420 ----------
Op0420
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0427] subi.b #$33, -(a7) uses Op0427 ----------
Op0427
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0428] subi.b #$33, ($3333,a0) uses Op0428 ----------
Op0428
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0430] subi.b #$33, ($33,a0,d3.w*2) uses Op0430 ----------
Op0430
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0438] subi.b #$33, $3333.w uses Op0438 ----------
Op0438
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0439] subi.b #$33, $33333333.l uses Op0439 ----------
Op0439
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [043a] subi.b #$33, ($3333,pc); =3337 uses Op043a ----------
Op043a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [043b] subi.b #$33, ($33,pc,d3.w*2); =37 uses Op043b ----------
Op043b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0440] subi.w #$3333, d0 uses Op0440 ----------
Op0440
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #16
  strh r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0450] subi.w #$3333, (a0) uses Op0450 ----------
Op0450
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0458] subi.w #$3333, (a0)+ uses Op0458 ----------
Op0458
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [045f] subi.w #$3333, (a7)+ uses Op045f ----------
Op045f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0460] subi.w #$3333, -(a0) uses Op0460 ----------
Op0460
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0467] subi.w #$3333, -(a7) uses Op0467 ----------
Op0467
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0468] subi.w #$3333, ($3333,a0) uses Op0468 ----------
Op0468
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0470] subi.w #$3333, ($33,a0,d3.w*2) uses Op0470 ----------
Op0470
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0478] subi.w #$3333, $3333.w uses Op0478 ----------
Op0478
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0479] subi.w #$3333, $33333333.l uses Op0479 ----------
Op0479
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [047a] subi.w #$3333, ($3333,pc); =3337 uses Op047a ----------
Op047a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [047b] subi.w #$3333, ($33,pc,d3.w*2); =37 uses Op047b ----------
Op047b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0480] subi.l #$33333333, d0 uses Op0480 ----------
Op0480
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0490] subi.l #$33333333, (a0) uses Op0490 ----------
Op0490
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0498] subi.l #$33333333, (a0)+ uses Op0498 ----------
Op0498
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [049f] subi.l #$33333333, (a7)+ uses Op049f ----------
Op049f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04a0] subi.l #$33333333, -(a0) uses Op04a0 ----------
Op04a0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04a7] subi.l #$33333333, -(a7) uses Op04a7 ----------
Op04a7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04a8] subi.l #$33333333, ($3333,a0) uses Op04a8 ----------
Op04a8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04b0] subi.l #$33333333, ($33,a0,d3.w*2) uses Op04b0 ----------
Op04b0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04b8] subi.l #$33333333, $3333.w uses Op04b8 ----------
Op04b8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04b9] subi.l #$33333333, $33333333.l uses Op04b9 ----------
Op04b9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04ba] subi.l #$33333333, ($3333,pc); =3339 uses Op04ba ----------
Op04ba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [04bb] subi.l #$33333333, ($33,pc,d3.w*2); =39 uses Op04bb ----------
Op04bb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  subs r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0600] addi.b #$33, d0 uses Op0600 ----------
Op0600
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0610] addi.b #$33, (a0) uses Op0610 ----------
Op0610
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0618] addi.b #$33, (a0)+ uses Op0618 ----------
Op0618
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [061f] addi.b #$33, (a7)+ uses Op061f ----------
Op061f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0620] addi.b #$33, -(a0) uses Op0620 ----------
Op0620
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0627] addi.b #$33, -(a7) uses Op0627 ----------
Op0627
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0628] addi.b #$33, ($3333,a0) uses Op0628 ----------
Op0628
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0630] addi.b #$33, ($33,a0,d3.w*2) uses Op0630 ----------
Op0630
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0638] addi.b #$33, $3333.w uses Op0638 ----------
Op0638
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0639] addi.b #$33, $33333333.l uses Op0639 ----------
Op0639
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [063a] addi.b #$33, ($3333,pc); =3337 uses Op063a ----------
Op063a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [063b] addi.b #$33, ($33,pc,d3.w*2); =37 uses Op063b ----------
Op063b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0640] addi.w #$3333, d0 uses Op0640 ----------
Op0640
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #16
  strh r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0650] addi.w #$3333, (a0) uses Op0650 ----------
Op0650
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0658] addi.w #$3333, (a0)+ uses Op0658 ----------
Op0658
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [065f] addi.w #$3333, (a7)+ uses Op065f ----------
Op065f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0660] addi.w #$3333, -(a0) uses Op0660 ----------
Op0660
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0667] addi.w #$3333, -(a7) uses Op0667 ----------
Op0667
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0668] addi.w #$3333, ($3333,a0) uses Op0668 ----------
Op0668
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0670] addi.w #$3333, ($33,a0,d3.w*2) uses Op0670 ----------
Op0670
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0678] addi.w #$3333, $3333.w uses Op0678 ----------
Op0678
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0679] addi.w #$3333, $33333333.l uses Op0679 ----------
Op0679
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [067a] addi.w #$3333, ($3333,pc); =3337 uses Op067a ----------
Op067a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [067b] addi.w #$3333, ($33,pc,d3.w*2); =37 uses Op067b ----------
Op067b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0680] addi.l #$33333333, d0 uses Op0680 ----------
Op0680
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0690] addi.l #$33333333, (a0) uses Op0690 ----------
Op0690
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0698] addi.l #$33333333, (a0)+ uses Op0698 ----------
Op0698
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [069f] addi.l #$33333333, (a7)+ uses Op069f ----------
Op069f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06a0] addi.l #$33333333, -(a0) uses Op06a0 ----------
Op06a0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06a7] addi.l #$33333333, -(a7) uses Op06a7 ----------
Op06a7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06a8] addi.l #$33333333, ($3333,a0) uses Op06a8 ----------
Op06a8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06b0] addi.l #$33333333, ($33,a0,d3.w*2) uses Op06b0 ----------
Op06b0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06b8] addi.l #$33333333, $3333.w uses Op06b8 ----------
Op06b8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06b9] addi.l #$33333333, $33333333.l uses Op06b9 ----------
Op06b9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06ba] addi.l #$33333333, ($3333,pc); =3339 uses Op06ba ----------
Op06ba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [06bb] addi.l #$33333333, ($33,pc,d3.w*2); =39 uses Op06bb ----------
Op06bb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  adds r1,r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0800] btst #$33, d0 uses Op0800 ----------
Op0800
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0810] btst #$33, (a0) uses Op0810 ----------
Op0810
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0818] btst #$33, (a0)+ uses Op0818 ----------
Op0818
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [081f] btst #$33, (a7)+ uses Op081f ----------
Op081f
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0820] btst #$33, -(a0) uses Op0820 ----------
Op0820
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0827] btst #$33, -(a7) uses Op0827 ----------
Op0827
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0828] btst #$33, ($3333,a0) uses Op0828 ----------
Op0828
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0830] btst #$33, ($33,a0,d3.w*2) uses Op0830 ----------
Op0830
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0838] btst #$33, $3333.w uses Op0838 ----------
Op0838
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0839] btst #$33, $33333333.l uses Op0839 ----------
Op0839
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [083a] btst #$33, ($3333,pc); =3337 uses Op083a ----------
Op083a
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [083b] btst #$33, ($33,pc,d3.w*2); =37 uses Op083b ----------
Op083b
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [083c] btst #$33, #$33 uses Op083c ----------
Op083c
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '#$33' into r11:
  ldrsb r11,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r11) into r0:
  mov r0,r11,asl #0

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0840] bchg #$33, d0 uses Op0840 ----------
Op0840
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0850] bchg #$33, (a0) uses Op0850 ----------
Op0850
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0858] bchg #$33, (a0)+ uses Op0858 ----------
Op0858
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [085f] bchg #$33, (a7)+ uses Op085f ----------
Op085f
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0860] bchg #$33, -(a0) uses Op0860 ----------
Op0860
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0867] bchg #$33, -(a7) uses Op0867 ----------
Op0867
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0868] bchg #$33, ($3333,a0) uses Op0868 ----------
Op0868
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0870] bchg #$33, ($33,a0,d3.w*2) uses Op0870 ----------
Op0870
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0878] bchg #$33, $3333.w uses Op0878 ----------
Op0878
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0879] bchg #$33, $33333333.l uses Op0879 ----------
Op0879
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [087a] bchg #$33, ($3333,pc); =3337 uses Op087a ----------
Op087a
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [087b] bchg #$33, ($33,pc,d3.w*2); =37 uses Op087b ----------
Op087b
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  eor r1,r0,r10 ;@ Toggle bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0880] bclr #$33, d0 uses Op0880 ----------
Op0880
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0890] bclr #$33, (a0) uses Op0890 ----------
Op0890
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0898] bclr #$33, (a0)+ uses Op0898 ----------
Op0898
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [089f] bclr #$33, (a7)+ uses Op089f ----------
Op089f
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08a0] bclr #$33, -(a0) uses Op08a0 ----------
Op08a0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08a7] bclr #$33, -(a7) uses Op08a7 ----------
Op08a7
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08a8] bclr #$33, ($3333,a0) uses Op08a8 ----------
Op08a8
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08b0] bclr #$33, ($33,a0,d3.w*2) uses Op08b0 ----------
Op08b0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08b8] bclr #$33, $3333.w uses Op08b8 ----------
Op08b8
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08b9] bclr #$33, $33333333.l uses Op08b9 ----------
Op08b9
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08ba] bclr #$33, ($3333,pc); =3337 uses Op08ba ----------
Op08ba
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08bb] bclr #$33, ($33,pc,d3.w*2); =37 uses Op08bb ----------
Op08bb
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  bic r1,r0,r10 ;@ Clear bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08c0] bset #$33, d0 uses Op08c0 ----------
Op08c0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08d0] bset #$33, (a0) uses Op08d0 ----------
Op08d0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08d8] bset #$33, (a0)+ uses Op08d8 ----------
Op08d8
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08df] bset #$33, (a7)+ uses Op08df ----------
Op08df
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08e0] bset #$33, -(a0) uses Op08e0 ----------
Op08e0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08e7] bset #$33, -(a7) uses Op08e7 ----------
Op08e7
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08e8] bset #$33, ($3333,a0) uses Op08e8 ----------
Op08e8
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08f0] bset #$33, ($33,a0,d3.w*2) uses Op08f0 ----------
Op08f0
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08f8] bset #$33, $3333.w uses Op08f8 ----------
Op08f8
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08f9] bset #$33, $33333333.l uses Op08f9 ----------
Op08f9
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08fa] bset #$33, ($3333,pc); =3337 uses Op08fa ----------
Op08fa
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [08fb] bset #$33, ($33,pc,d3.w*2); =37 uses Op08fb ----------
Op08fb
  mov r10,#1

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  bic r9,r9,#0x40000000 ;@ Blank Z flag
  mov r10,r10,lsl r0 ;@ Make bit mask

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  tst r0,r10 ;@ Do arithmetic
  orreq r9,r9,#0x40000000 ;@ Get Z flag

  orr r1,r0,r10 ;@ Set bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a00] eori.b #$33, d0 uses Op0a00 ----------
Op0a00
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a10] eori.b #$33, (a0) uses Op0a10 ----------
Op0a10
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a18] eori.b #$33, (a0)+ uses Op0a18 ----------
Op0a18
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a1f] eori.b #$33, (a7)+ uses Op0a1f ----------
Op0a1f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a20] eori.b #$33, -(a0) uses Op0a20 ----------
Op0a20
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a27] eori.b #$33, -(a7) uses Op0a27 ----------
Op0a27
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a28] eori.b #$33, ($3333,a0) uses Op0a28 ----------
Op0a28
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a30] eori.b #$33, ($33,a0,d3.w*2) uses Op0a30 ----------
Op0a30
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a38] eori.b #$33, $3333.w uses Op0a38 ----------
Op0a38
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a39] eori.b #$33, $33333333.l uses Op0a39 ----------
Op0a39
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a3a] eori.b #$33, ($3333,pc); =3337 uses Op0a3a ----------
Op0a3a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a3b] eori.b #$33, ($33,pc,d3.w*2); =37 uses Op0a3b ----------
Op0a3b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a3c] eori.b #$33, ccr uses Op0a3c ----------
Op0a3c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode0a3c ;@ No

;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  eor r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode0a3c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a40] eori.w #$3333, d0 uses Op0a40 ----------
Op0a40
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #16
  strh r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a50] eori.w #$3333, (a0) uses Op0a50 ----------
Op0a50
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a58] eori.w #$3333, (a0)+ uses Op0a58 ----------
Op0a58
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a5f] eori.w #$3333, (a7)+ uses Op0a5f ----------
Op0a5f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a60] eori.w #$3333, -(a0) uses Op0a60 ----------
Op0a60
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a67] eori.w #$3333, -(a7) uses Op0a67 ----------
Op0a67
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a68] eori.w #$3333, ($3333,a0) uses Op0a68 ----------
Op0a68
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a70] eori.w #$3333, ($33,a0,d3.w*2) uses Op0a70 ----------
Op0a70
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a78] eori.w #$3333, $3333.w uses Op0a78 ----------
Op0a78
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a79] eori.w #$3333, $33333333.l uses Op0a79 ----------
Op0a79
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a7a] eori.w #$3333, ($3333,pc); =3337 uses Op0a7a ----------
Op0a7a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a7b] eori.w #$3333, ($33,pc,d3.w*2); =37 uses Op0a7b ----------
Op0a7b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a7c] eori.w #$3333, sr uses Op0a7c ----------
Op0a7c
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode0a7c ;@ No

;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

  eor r0,r1,r10
  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode0a7c
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a80] eori.l #$33333333, d0 uses Op0a80 ----------
Op0a80
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r11]:
  str r1,[r7,r11,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a90] eori.l #$33333333, (a0) uses Op0a90 ----------
Op0a90
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a98] eori.l #$33333333, (a0)+ uses Op0a98 ----------
Op0a98
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0a9f] eori.l #$33333333, (a7)+ uses Op0a9f ----------
Op0a9f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0aa0] eori.l #$33333333, -(a0) uses Op0aa0 ----------
Op0aa0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0aa7] eori.l #$33333333, -(a7) uses Op0aa7 ----------
Op0aa7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0aa8] eori.l #$33333333, ($3333,a0) uses Op0aa8 ----------
Op0aa8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ab0] eori.l #$33333333, ($33,a0,d3.w*2) uses Op0ab0 ----------
Op0ab0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ab8] eori.l #$33333333, $3333.w uses Op0ab8 ----------
Op0ab8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ab9] eori.l #$33333333, $33333333.l uses Op0ab9 ----------
Op0ab9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0aba] eori.l #$33333333, ($3333,pc); =3339 uses Op0aba ----------
Op0aba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0abb] eori.l #$33333333, ($33,pc,d3.w*2); =39 uses Op0abb ----------
Op0abb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  eor r1,r0,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r11):
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c00] cmpi.b #$33, d0 uses Op0c00 ----------
Op0c00
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsb r0,[r7,r11]
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c10] cmpi.b #$33, (a0) uses Op0c10 ----------
Op0c10
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c18] cmpi.b #$33, (a0)+ uses Op0c18 ----------
Op0c18
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c1f] cmpi.b #$33, (a7)+ uses Op0c1f ----------
Op0c1f
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c20] cmpi.b #$33, -(a0) uses Op0c20 ----------
Op0c20
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c27] cmpi.b #$33, -(a7) uses Op0c27 ----------
Op0c27
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c28] cmpi.b #$33, ($3333,a0) uses Op0c28 ----------
Op0c28
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c30] cmpi.b #$33, ($33,a0,d3.w*2) uses Op0c30 ----------
Op0c30
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c38] cmpi.b #$33, $3333.w uses Op0c38 ----------
Op0c38
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c39] cmpi.b #$33, $33333333.l uses Op0c39 ----------
Op0c39
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c3a] cmpi.b #$33, ($3333,pc); =3337 uses Op0c3a ----------
Op0c3a
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c3b] cmpi.b #$33, ($33,pc,d3.w*2); =37 uses Op0c3b ----------
Op0c3b
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c3c]  uses Op0c3c ----------
Op0c3c
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r10:
  mov r10,r10,asl #24

;@ EaCalc : Get '#$33' into r11:
  ldrsb r11,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r11) into r0:
  mov r0,r11,asl #24

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c40] cmpi.w #$3333, d0 uses Op0c40 ----------
Op0c40
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
  mov r11,r11,lsl #2
;@ EaRead : Read register[r11] into r0:
  ldrsh r0,[r7,r11]
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c50] cmpi.w #$3333, (a0) uses Op0c50 ----------
Op0c50
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c58] cmpi.w #$3333, (a0)+ uses Op0c58 ----------
Op0c58
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c5f] cmpi.w #$3333, (a7)+ uses Op0c5f ----------
Op0c5f
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c60] cmpi.w #$3333, -(a0) uses Op0c60 ----------
Op0c60
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c67] cmpi.w #$3333, -(a7) uses Op0c67 ----------
Op0c67
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#2 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c68] cmpi.w #$3333, ($3333,a0) uses Op0c68 ----------
Op0c68
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c70] cmpi.w #$3333, ($33,a0,d3.w*2) uses Op0c70 ----------
Op0c70
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c78] cmpi.w #$3333, $3333.w uses Op0c78 ----------
Op0c78
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c79] cmpi.w #$3333, $33333333.l uses Op0c79 ----------
Op0c79
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c7a] cmpi.w #$3333, ($3333,pc); =3337 uses Op0c7a ----------
Op0c7a
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c7b] cmpi.w #$3333, ($33,pc,d3.w*2); =37 uses Op0c7b ----------
Op0c7b
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c7c]  uses Op0c7c ----------
Op0c7c
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:
  mov r10,r10,asl #16

;@ EaCalc : Get '#$3333' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r11) into r0:
  mov r0,r11,asl #16

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c80] cmpi.l #$33333333, d0 uses Op0c80 ----------
Op0c80
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x000f
;@ EaRead : Read register[r11] into r0:
  ldr r0,[r7,r11,lsl #2]

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c90] cmpi.l #$33333333, (a0) uses Op0c90 ----------
Op0c90
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
;@ EaRead : Read '(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c98] cmpi.l #$33333333, (a0)+ uses Op0c98 ----------
Op0c98
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a0)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0c9f] cmpi.l #$33333333, (a7)+ uses Op0c9f ----------
Op0c9f
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '(a7)+' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r11,[r7,r2]
  add r3,r11,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ca0] cmpi.l #$33333333, -(a0) uses Op0ca0 ----------
Op0ca0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ca7] cmpi.l #$33333333, -(a7) uses Op0ca7 ----------
Op0ca7
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '-(a7)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#4 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0ca8] cmpi.l #$33333333, ($3333,a0) uses Op0ca8 ----------
Op0ca8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,a0)' into r11:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r11,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cb0] cmpi.l #$33333333, ($33,a0,d3.w*2) uses Op0cb0 ----------
Op0cb0
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,a0,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r11,r2,r3 ;@ r11=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cb8] cmpi.l #$33333333, $3333.w uses Op0cb8 ----------
Op0cb8
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$3333.w' into r11:
  ldrsh r11,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cb9] cmpi.l #$33333333, $33333333.l uses Op0cb9 ----------
Op0cb9
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '$33333333.l' into r11:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cba] cmpi.l #$33333333, ($3333,pc); =3339 uses Op0cba ----------
Op0cba
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($3333,pc)' into r11:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r11,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cbb] cmpi.l #$33333333, ($33,pc,d3.w*2); =39 uses Op0cbb ----------
Op0cbb
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '($33,pc,d3.w*2)' into r11:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r11,r2,r0 ;@ r11=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r11) into r0:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [0cbc]  uses Op0cbc ----------
Op0cbc
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r10:

;@ EaCalc : Get '#$33333333' into r11:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r11,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r11) into r0:
  mov r0,r11,asl #0

;@ Do arithmetic:
  cmp r0,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1000] move.b d0, d0 uses Op1000 ----------
Op1000
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1010] move.b (a0), d0 uses Op1010 ----------
Op1010
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1018] move.b (a0)+, d0 uses Op1018 ----------
Op1018
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [101f] move.b (a7)+, d0 uses Op101f ----------
Op101f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1020] move.b -(a0), d0 uses Op1020 ----------
Op1020
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1027] move.b -(a7), d0 uses Op1027 ----------
Op1027
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1028] move.b ($3333,a0), d0 uses Op1028 ----------
Op1028
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1030] move.b ($33,a0,d3.w*2), d0 uses Op1030 ----------
Op1030
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1038] move.b $3333.w, d0 uses Op1038 ----------
Op1038
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1039] move.b $33333333.l, d0 uses Op1039 ----------
Op1039
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [103a] move.b ($3333,pc), d0; =3335 uses Op103a ----------
Op103a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [103b] move.b ($33,pc,d3.w*2), d0; =35 uses Op103b ----------
Op103b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [103c] move.b #$33, d0 uses Op103c ----------
Op103c
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1080] move.b d0, (a0) uses Op1080 ----------
Op1080
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1090] move.b (a0), (a0) uses Op1090 ----------
Op1090
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1098] move.b (a0)+, (a0) uses Op1098 ----------
Op1098
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [109f] move.b (a7)+, (a0) uses Op109f ----------
Op109f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10a0] move.b -(a0), (a0) uses Op10a0 ----------
Op10a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10a7] move.b -(a7), (a0) uses Op10a7 ----------
Op10a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10a8] move.b ($3333,a0), (a0) uses Op10a8 ----------
Op10a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10b0] move.b ($33,a0,d3.w*2), (a0) uses Op10b0 ----------
Op10b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10b8] move.b $3333.w, (a0) uses Op10b8 ----------
Op10b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10b9] move.b $33333333.l, (a0) uses Op10b9 ----------
Op10b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10ba] move.b ($3333,pc), (a0); =3335 uses Op10ba ----------
Op10ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10bb] move.b ($33,pc,d3.w*2), (a0); =35 uses Op10bb ----------
Op10bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10bc] move.b #$33, (a0) uses Op10bc ----------
Op10bc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10c0] move.b d0, (a0)+ uses Op10c0 ----------
Op10c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10d0] move.b (a0), (a0)+ uses Op10d0 ----------
Op10d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10d8] move.b (a0)+, (a0)+ uses Op10d8 ----------
Op10d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10df] move.b (a7)+, (a0)+ uses Op10df ----------
Op10df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10e0] move.b -(a0), (a0)+ uses Op10e0 ----------
Op10e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10e7] move.b -(a7), (a0)+ uses Op10e7 ----------
Op10e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10e8] move.b ($3333,a0), (a0)+ uses Op10e8 ----------
Op10e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10f0] move.b ($33,a0,d3.w*2), (a0)+ uses Op10f0 ----------
Op10f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10f8] move.b $3333.w, (a0)+ uses Op10f8 ----------
Op10f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10f9] move.b $33333333.l, (a0)+ uses Op10f9 ----------
Op10f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10fa] move.b ($3333,pc), (a0)+; =3335 uses Op10fa ----------
Op10fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10fb] move.b ($33,pc,d3.w*2), (a0)+; =35 uses Op10fb ----------
Op10fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [10fc] move.b #$33, (a0)+ uses Op10fc ----------
Op10fc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1100] move.b d0, -(a0) uses Op1100 ----------
Op1100
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1110] move.b (a0), -(a0) uses Op1110 ----------
Op1110
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1118] move.b (a0)+, -(a0) uses Op1118 ----------
Op1118
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [111f] move.b (a7)+, -(a0) uses Op111f ----------
Op111f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1120] move.b -(a0), -(a0) uses Op1120 ----------
Op1120
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1127] move.b -(a7), -(a0) uses Op1127 ----------
Op1127
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1128] move.b ($3333,a0), -(a0) uses Op1128 ----------
Op1128
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1130] move.b ($33,a0,d3.w*2), -(a0) uses Op1130 ----------
Op1130
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1138] move.b $3333.w, -(a0) uses Op1138 ----------
Op1138
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1139] move.b $33333333.l, -(a0) uses Op1139 ----------
Op1139
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [113a] move.b ($3333,pc), -(a0); =3335 uses Op113a ----------
Op113a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [113b] move.b ($33,pc,d3.w*2), -(a0); =35 uses Op113b ----------
Op113b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [113c] move.b #$33, -(a0) uses Op113c ----------
Op113c
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1140] move.b d0, ($3333,a0) uses Op1140 ----------
Op1140
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1150] move.b (a0), ($3333,a0) uses Op1150 ----------
Op1150
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1158] move.b (a0)+, ($3333,a0) uses Op1158 ----------
Op1158
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [115f] move.b (a7)+, ($3333,a0) uses Op115f ----------
Op115f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1160] move.b -(a0), ($3333,a0) uses Op1160 ----------
Op1160
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1167] move.b -(a7), ($3333,a0) uses Op1167 ----------
Op1167
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1168] move.b ($3333,a0), ($3333,a0) uses Op1168 ----------
Op1168
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1170] move.b ($33,a0,d3.w*2), ($3333,a0) uses Op1170 ----------
Op1170
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1178] move.b $3333.w, ($3333,a0) uses Op1178 ----------
Op1178
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1179] move.b $33333333.l, ($3333,a0) uses Op1179 ----------
Op1179
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [117a] move.b ($3333,pc), ($3333,a0); =3335 uses Op117a ----------
Op117a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [117b] move.b ($33,pc,d3.w*2), ($3333,a0); =35 uses Op117b ----------
Op117b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [117c] move.b #$33, ($3333,a0) uses Op117c ----------
Op117c
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1180] move.b d0, ($33,a0,d3.w*2) uses Op1180 ----------
Op1180
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1190] move.b (a0), ($33,a0,d3.w*2) uses Op1190 ----------
Op1190
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1198] move.b (a0)+, ($33,a0,d3.w*2) uses Op1198 ----------
Op1198
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [119f] move.b (a7)+, ($33,a0,d3.w*2) uses Op119f ----------
Op119f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11a0] move.b -(a0), ($33,a0,d3.w*2) uses Op11a0 ----------
Op11a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11a7] move.b -(a7), ($33,a0,d3.w*2) uses Op11a7 ----------
Op11a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11a8] move.b ($3333,a0), ($33,a0,d3.w*2) uses Op11a8 ----------
Op11a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11b0] move.b ($33,a0,d3.w*2), ($33,a0,d3.w*2) uses Op11b0 ----------
Op11b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11b8] move.b $3333.w, ($33,a0,d3.w*2) uses Op11b8 ----------
Op11b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11b9] move.b $33333333.l, ($33,a0,d3.w*2) uses Op11b9 ----------
Op11b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11ba] move.b ($3333,pc), ($33,a0,d3.w*2); =3335 uses Op11ba ----------
Op11ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11bb] move.b ($33,pc,d3.w*2), ($33,a0,d3.w*2); =35 uses Op11bb ----------
Op11bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11bc] move.b #$33, ($33,a0,d3.w*2) uses Op11bc ----------
Op11bc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11c0] move.b d0, $3333.w uses Op11c0 ----------
Op11c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11d0] move.b (a0), $3333.w uses Op11d0 ----------
Op11d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11d8] move.b (a0)+, $3333.w uses Op11d8 ----------
Op11d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11df] move.b (a7)+, $3333.w uses Op11df ----------
Op11df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11e0] move.b -(a0), $3333.w uses Op11e0 ----------
Op11e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11e7] move.b -(a7), $3333.w uses Op11e7 ----------
Op11e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11e8] move.b ($3333,a0), $3333.w uses Op11e8 ----------
Op11e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11f0] move.b ($33,a0,d3.w*2), $3333.w uses Op11f0 ----------
Op11f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11f8] move.b $3333.w, $3333.w uses Op11f8 ----------
Op11f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11f9] move.b $33333333.l, $3333.w uses Op11f9 ----------
Op11f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11fa] move.b ($3333,pc), $3333.w; =3335 uses Op11fa ----------
Op11fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11fb] move.b ($33,pc,d3.w*2), $3333.w; =35 uses Op11fb ----------
Op11fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [11fc] move.b #$33, $3333.w uses Op11fc ----------
Op11fc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13c0] move.b d0, $33333333.l uses Op13c0 ----------
Op13c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13d0] move.b (a0), $33333333.l uses Op13d0 ----------
Op13d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13d8] move.b (a0)+, $33333333.l uses Op13d8 ----------
Op13d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13df] move.b (a7)+, $33333333.l uses Op13df ----------
Op13df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13e0] move.b -(a0), $33333333.l uses Op13e0 ----------
Op13e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13e7] move.b -(a7), $33333333.l uses Op13e7 ----------
Op13e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13e8] move.b ($3333,a0), $33333333.l uses Op13e8 ----------
Op13e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13f0] move.b ($33,a0,d3.w*2), $33333333.l uses Op13f0 ----------
Op13f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13f8] move.b $3333.w, $33333333.l uses Op13f8 ----------
Op13f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13f9] move.b $33333333.l, $33333333.l uses Op13f9 ----------
Op13f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13fa] move.b ($3333,pc), $33333333.l; =3335 uses Op13fa ----------
Op13fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13fb] move.b ($33,pc,d3.w*2), $33333333.l; =35 uses Op13fb ----------
Op13fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [13fc] move.b #$33, $33333333.l uses Op13fc ----------
Op13fc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15c0] move.b d0, ($3333,pc); =3335 uses Op15c0 ----------
Op15c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15d0] move.b (a0), ($3333,pc); =3335 uses Op15d0 ----------
Op15d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15d8] move.b (a0)+, ($3333,pc); =3335 uses Op15d8 ----------
Op15d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15df] move.b (a7)+, ($3333,pc); =3335 uses Op15df ----------
Op15df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15e0] move.b -(a0), ($3333,pc); =3335 uses Op15e0 ----------
Op15e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15e7] move.b -(a7), ($3333,pc); =3335 uses Op15e7 ----------
Op15e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15e8] move.b ($3333,a0), ($3333,pc); =3337 uses Op15e8 ----------
Op15e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15f0] move.b ($33,a0,d3.w*2), ($3333,pc); =3337 uses Op15f0 ----------
Op15f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15f8] move.b $3333.w, ($3333,pc); =3337 uses Op15f8 ----------
Op15f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15f9] move.b $33333333.l, ($3333,pc); =3339 uses Op15f9 ----------
Op15f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15fa] move.b ($3333,pc), ($3333,pc); =3337 uses Op15fa ----------
Op15fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15fb] move.b ($33,pc,d3.w*2), ($3333,pc); =3337 uses Op15fb ----------
Op15fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [15fc] move.b #$33, ($3333,pc); =3337 uses Op15fc ----------
Op15fc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17c0] move.b d0, ($33,pc,d3.w*2); =35 uses Op17c0 ----------
Op17c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17d0] move.b (a0), ($33,pc,d3.w*2); =35 uses Op17d0 ----------
Op17d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17d8] move.b (a0)+, ($33,pc,d3.w*2); =35 uses Op17d8 ----------
Op17d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17df] move.b (a7)+, ($33,pc,d3.w*2); =35 uses Op17df ----------
Op17df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17e0] move.b -(a0), ($33,pc,d3.w*2); =35 uses Op17e0 ----------
Op17e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17e7] move.b -(a7), ($33,pc,d3.w*2); =35 uses Op17e7 ----------
Op17e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17e8] move.b ($3333,a0), ($33,pc,d3.w*2); =37 uses Op17e8 ----------
Op17e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17f0] move.b ($33,a0,d3.w*2), ($33,pc,d3.w*2); =37 uses Op17f0 ----------
Op17f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17f8] move.b $3333.w, ($33,pc,d3.w*2); =37 uses Op17f8 ----------
Op17f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17f9] move.b $33333333.l, ($33,pc,d3.w*2); =39 uses Op17f9 ----------
Op17f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17fa] move.b ($3333,pc), ($33,pc,d3.w*2); =37 uses Op17fa ----------
Op17fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17fb] move.b ($33,pc,d3.w*2), ($33,pc,d3.w*2); =37 uses Op17fb ----------
Op17fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [17fc] move.b #$33, ($33,pc,d3.w*2); =37 uses Op17fc ----------
Op17fc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ec0] move.b d0, (a7)+ uses Op1ec0 ----------
Op1ec0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ed0] move.b (a0), (a7)+ uses Op1ed0 ----------
Op1ed0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ed8] move.b (a0)+, (a7)+ uses Op1ed8 ----------
Op1ed8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1edf] move.b (a7)+, (a7)+ uses Op1edf ----------
Op1edf
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ee0] move.b -(a0), (a7)+ uses Op1ee0 ----------
Op1ee0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ee7] move.b -(a7), (a7)+ uses Op1ee7 ----------
Op1ee7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ee8] move.b ($3333,a0), (a7)+ uses Op1ee8 ----------
Op1ee8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ef0] move.b ($33,a0,d3.w*2), (a7)+ uses Op1ef0 ----------
Op1ef0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ef8] move.b $3333.w, (a7)+ uses Op1ef8 ----------
Op1ef8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1ef9] move.b $33333333.l, (a7)+ uses Op1ef9 ----------
Op1ef9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1efa] move.b ($3333,pc), (a7)+; =3335 uses Op1efa ----------
Op1efa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1efb] move.b ($33,pc,d3.w*2), (a7)+; =35 uses Op1efb ----------
Op1efb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1efc] move.b #$33, (a7)+ uses Op1efc ----------
Op1efc
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f00] move.b d0, -(a7) uses Op1f00 ----------
Op1f00
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f10] move.b (a0), -(a7) uses Op1f10 ----------
Op1f10
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f18] move.b (a0)+, -(a7) uses Op1f18 ----------
Op1f18
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f1f] move.b (a7)+, -(a7) uses Op1f1f ----------
Op1f1f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f20] move.b -(a0), -(a7) uses Op1f20 ----------
Op1f20
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f27] move.b -(a7), -(a7) uses Op1f27 ----------
Op1f27
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f28] move.b ($3333,a0), -(a7) uses Op1f28 ----------
Op1f28
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f30] move.b ($33,a0,d3.w*2), -(a7) uses Op1f30 ----------
Op1f30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f38] move.b $3333.w, -(a7) uses Op1f38 ----------
Op1f38
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f39] move.b $33333333.l, -(a7) uses Op1f39 ----------
Op1f39
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f3a] move.b ($3333,pc), -(a7); =3335 uses Op1f3a ----------
Op1f3a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f3b] move.b ($33,pc,d3.w*2), -(a7); =35 uses Op1f3b ----------
Op1f3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [1f3c] move.b #$33, -(a7) uses Op1f3c ----------
Op1f3c
;@ EaCalc : Get '#$33' into r0:
  ldrsb r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2000] move.l d0, d0 uses Op2000 ----------
Op2000
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2010] move.l (a0), d0 uses Op2010 ----------
Op2010
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2018] move.l (a0)+, d0 uses Op2018 ----------
Op2018
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [201f] move.l (a7)+, d0 uses Op201f ----------
Op201f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2020] move.l -(a0), d0 uses Op2020 ----------
Op2020
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2027] move.l -(a7), d0 uses Op2027 ----------
Op2027
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2028] move.l ($3333,a0), d0 uses Op2028 ----------
Op2028
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2030] move.l ($33,a0,d3.w*2), d0 uses Op2030 ----------
Op2030
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2038] move.l $3333.w, d0 uses Op2038 ----------
Op2038
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2039] move.l $33333333.l, d0 uses Op2039 ----------
Op2039
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [203a] move.l ($3333,pc), d0; =3335 uses Op203a ----------
Op203a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [203b] move.l ($33,pc,d3.w*2), d0; =35 uses Op203b ----------
Op203b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [203c] move.l #$33333333, d0 uses Op203c ----------
Op203c
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2040] movea.l d0, a0 uses Op2040 ----------
Op2040
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2050] movea.l (a0), a0 uses Op2050 ----------
Op2050
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2058] movea.l (a0)+, a0 uses Op2058 ----------
Op2058
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [205f] movea.l (a7)+, a0 uses Op205f ----------
Op205f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2060] movea.l -(a0), a0 uses Op2060 ----------
Op2060
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2067] movea.l -(a7), a0 uses Op2067 ----------
Op2067
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2068] movea.l ($3333,a0), a0 uses Op2068 ----------
Op2068
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2070] movea.l ($33,a0,d3.w*2), a0 uses Op2070 ----------
Op2070
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2078] movea.l $3333.w, a0 uses Op2078 ----------
Op2078
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2079] movea.l $33333333.l, a0 uses Op2079 ----------
Op2079
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [207a] movea.l ($3333,pc), a0; =3335 uses Op207a ----------
Op207a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [207b] movea.l ($33,pc,d3.w*2), a0; =35 uses Op207b ----------
Op207b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [207c] movea.l #$33333333, a0 uses Op207c ----------
Op207c
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2080] move.l d0, (a0) uses Op2080 ----------
Op2080
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2090] move.l (a0), (a0) uses Op2090 ----------
Op2090
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2098] move.l (a0)+, (a0) uses Op2098 ----------
Op2098
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [209f] move.l (a7)+, (a0) uses Op209f ----------
Op209f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20a0] move.l -(a0), (a0) uses Op20a0 ----------
Op20a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20a7] move.l -(a7), (a0) uses Op20a7 ----------
Op20a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20a8] move.l ($3333,a0), (a0) uses Op20a8 ----------
Op20a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20b0] move.l ($33,a0,d3.w*2), (a0) uses Op20b0 ----------
Op20b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20b8] move.l $3333.w, (a0) uses Op20b8 ----------
Op20b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20b9] move.l $33333333.l, (a0) uses Op20b9 ----------
Op20b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20ba] move.l ($3333,pc), (a0); =3335 uses Op20ba ----------
Op20ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20bb] move.l ($33,pc,d3.w*2), (a0); =35 uses Op20bb ----------
Op20bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20bc] move.l #$33333333, (a0) uses Op20bc ----------
Op20bc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20c0] move.l d0, (a0)+ uses Op20c0 ----------
Op20c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20d0] move.l (a0), (a0)+ uses Op20d0 ----------
Op20d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20d8] move.l (a0)+, (a0)+ uses Op20d8 ----------
Op20d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20df] move.l (a7)+, (a0)+ uses Op20df ----------
Op20df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20e0] move.l -(a0), (a0)+ uses Op20e0 ----------
Op20e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20e7] move.l -(a7), (a0)+ uses Op20e7 ----------
Op20e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20e8] move.l ($3333,a0), (a0)+ uses Op20e8 ----------
Op20e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20f0] move.l ($33,a0,d3.w*2), (a0)+ uses Op20f0 ----------
Op20f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20f8] move.l $3333.w, (a0)+ uses Op20f8 ----------
Op20f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20f9] move.l $33333333.l, (a0)+ uses Op20f9 ----------
Op20f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20fa] move.l ($3333,pc), (a0)+; =3335 uses Op20fa ----------
Op20fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20fb] move.l ($33,pc,d3.w*2), (a0)+; =35 uses Op20fb ----------
Op20fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [20fc] move.l #$33333333, (a0)+ uses Op20fc ----------
Op20fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2100] move.l d0, -(a0) uses Op2100 ----------
Op2100
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2110] move.l (a0), -(a0) uses Op2110 ----------
Op2110
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2118] move.l (a0)+, -(a0) uses Op2118 ----------
Op2118
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [211f] move.l (a7)+, -(a0) uses Op211f ----------
Op211f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2120] move.l -(a0), -(a0) uses Op2120 ----------
Op2120
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2127] move.l -(a7), -(a0) uses Op2127 ----------
Op2127
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2128] move.l ($3333,a0), -(a0) uses Op2128 ----------
Op2128
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2130] move.l ($33,a0,d3.w*2), -(a0) uses Op2130 ----------
Op2130
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2138] move.l $3333.w, -(a0) uses Op2138 ----------
Op2138
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2139] move.l $33333333.l, -(a0) uses Op2139 ----------
Op2139
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [213a] move.l ($3333,pc), -(a0); =3335 uses Op213a ----------
Op213a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [213b] move.l ($33,pc,d3.w*2), -(a0); =35 uses Op213b ----------
Op213b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [213c] move.l #$33333333, -(a0) uses Op213c ----------
Op213c
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2140] move.l d0, ($3333,a0) uses Op2140 ----------
Op2140
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2150] move.l (a0), ($3333,a0) uses Op2150 ----------
Op2150
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2158] move.l (a0)+, ($3333,a0) uses Op2158 ----------
Op2158
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [215f] move.l (a7)+, ($3333,a0) uses Op215f ----------
Op215f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2160] move.l -(a0), ($3333,a0) uses Op2160 ----------
Op2160
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2167] move.l -(a7), ($3333,a0) uses Op2167 ----------
Op2167
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2168] move.l ($3333,a0), ($3333,a0) uses Op2168 ----------
Op2168
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2170] move.l ($33,a0,d3.w*2), ($3333,a0) uses Op2170 ----------
Op2170
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2178] move.l $3333.w, ($3333,a0) uses Op2178 ----------
Op2178
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2179] move.l $33333333.l, ($3333,a0) uses Op2179 ----------
Op2179
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [217a] move.l ($3333,pc), ($3333,a0); =3335 uses Op217a ----------
Op217a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [217b] move.l ($33,pc,d3.w*2), ($3333,a0); =35 uses Op217b ----------
Op217b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [217c] move.l #$33333333, ($3333,a0) uses Op217c ----------
Op217c
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2180] move.l d0, ($33,a0,d3.w*2) uses Op2180 ----------
Op2180
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2190] move.l (a0), ($33,a0,d3.w*2) uses Op2190 ----------
Op2190
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2198] move.l (a0)+, ($33,a0,d3.w*2) uses Op2198 ----------
Op2198
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [219f] move.l (a7)+, ($33,a0,d3.w*2) uses Op219f ----------
Op219f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21a0] move.l -(a0), ($33,a0,d3.w*2) uses Op21a0 ----------
Op21a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21a7] move.l -(a7), ($33,a0,d3.w*2) uses Op21a7 ----------
Op21a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21a8] move.l ($3333,a0), ($33,a0,d3.w*2) uses Op21a8 ----------
Op21a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21b0] move.l ($33,a0,d3.w*2), ($33,a0,d3.w*2) uses Op21b0 ----------
Op21b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21b8] move.l $3333.w, ($33,a0,d3.w*2) uses Op21b8 ----------
Op21b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21b9] move.l $33333333.l, ($33,a0,d3.w*2) uses Op21b9 ----------
Op21b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21ba] move.l ($3333,pc), ($33,a0,d3.w*2); =3335 uses Op21ba ----------
Op21ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21bb] move.l ($33,pc,d3.w*2), ($33,a0,d3.w*2); =35 uses Op21bb ----------
Op21bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21bc] move.l #$33333333, ($33,a0,d3.w*2) uses Op21bc ----------
Op21bc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21c0] move.l d0, $3333.w uses Op21c0 ----------
Op21c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21d0] move.l (a0), $3333.w uses Op21d0 ----------
Op21d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21d8] move.l (a0)+, $3333.w uses Op21d8 ----------
Op21d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21df] move.l (a7)+, $3333.w uses Op21df ----------
Op21df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21e0] move.l -(a0), $3333.w uses Op21e0 ----------
Op21e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21e7] move.l -(a7), $3333.w uses Op21e7 ----------
Op21e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21e8] move.l ($3333,a0), $3333.w uses Op21e8 ----------
Op21e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21f0] move.l ($33,a0,d3.w*2), $3333.w uses Op21f0 ----------
Op21f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21f8] move.l $3333.w, $3333.w uses Op21f8 ----------
Op21f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21f9] move.l $33333333.l, $3333.w uses Op21f9 ----------
Op21f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21fa] move.l ($3333,pc), $3333.w; =3335 uses Op21fa ----------
Op21fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21fb] move.l ($33,pc,d3.w*2), $3333.w; =35 uses Op21fb ----------
Op21fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [21fc] move.l #$33333333, $3333.w uses Op21fc ----------
Op21fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23c0] move.l d0, $33333333.l uses Op23c0 ----------
Op23c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23d0] move.l (a0), $33333333.l uses Op23d0 ----------
Op23d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23d8] move.l (a0)+, $33333333.l uses Op23d8 ----------
Op23d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23df] move.l (a7)+, $33333333.l uses Op23df ----------
Op23df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23e0] move.l -(a0), $33333333.l uses Op23e0 ----------
Op23e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23e7] move.l -(a7), $33333333.l uses Op23e7 ----------
Op23e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23e8] move.l ($3333,a0), $33333333.l uses Op23e8 ----------
Op23e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23f0] move.l ($33,a0,d3.w*2), $33333333.l uses Op23f0 ----------
Op23f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23f8] move.l $3333.w, $33333333.l uses Op23f8 ----------
Op23f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23f9] move.l $33333333.l, $33333333.l uses Op23f9 ----------
Op23f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#36 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23fa] move.l ($3333,pc), $33333333.l; =3335 uses Op23fa ----------
Op23fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23fb] move.l ($33,pc,d3.w*2), $33333333.l; =35 uses Op23fb ----------
Op23fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [23fc] move.l #$33333333, $33333333.l uses Op23fc ----------
Op23fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25c0] move.l d0, ($3333,pc); =3335 uses Op25c0 ----------
Op25c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25d0] move.l (a0), ($3333,pc); =3335 uses Op25d0 ----------
Op25d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25d8] move.l (a0)+, ($3333,pc); =3335 uses Op25d8 ----------
Op25d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25df] move.l (a7)+, ($3333,pc); =3335 uses Op25df ----------
Op25df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25e0] move.l -(a0), ($3333,pc); =3335 uses Op25e0 ----------
Op25e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25e7] move.l -(a7), ($3333,pc); =3335 uses Op25e7 ----------
Op25e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25e8] move.l ($3333,a0), ($3333,pc); =3337 uses Op25e8 ----------
Op25e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25f0] move.l ($33,a0,d3.w*2), ($3333,pc); =3337 uses Op25f0 ----------
Op25f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25f8] move.l $3333.w, ($3333,pc); =3337 uses Op25f8 ----------
Op25f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25f9] move.l $33333333.l, ($3333,pc); =3339 uses Op25f9 ----------
Op25f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25fa] move.l ($3333,pc), ($3333,pc); =3337 uses Op25fa ----------
Op25fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25fb] move.l ($33,pc,d3.w*2), ($3333,pc); =3337 uses Op25fb ----------
Op25fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [25fc] move.l #$33333333, ($3333,pc); =3339 uses Op25fc ----------
Op25fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27c0] move.l d0, ($33,pc,d3.w*2); =35 uses Op27c0 ----------
Op27c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27d0] move.l (a0), ($33,pc,d3.w*2); =35 uses Op27d0 ----------
Op27d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27d8] move.l (a0)+, ($33,pc,d3.w*2); =35 uses Op27d8 ----------
Op27d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27df] move.l (a7)+, ($33,pc,d3.w*2); =35 uses Op27df ----------
Op27df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27e0] move.l -(a0), ($33,pc,d3.w*2); =35 uses Op27e0 ----------
Op27e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27e7] move.l -(a7), ($33,pc,d3.w*2); =35 uses Op27e7 ----------
Op27e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27e8] move.l ($3333,a0), ($33,pc,d3.w*2); =37 uses Op27e8 ----------
Op27e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27f0] move.l ($33,a0,d3.w*2), ($33,pc,d3.w*2); =37 uses Op27f0 ----------
Op27f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27f8] move.l $3333.w, ($33,pc,d3.w*2); =37 uses Op27f8 ----------
Op27f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27f9] move.l $33333333.l, ($33,pc,d3.w*2); =39 uses Op27f9 ----------
Op27f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27fa] move.l ($3333,pc), ($33,pc,d3.w*2); =37 uses Op27fa ----------
Op27fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27fb] move.l ($33,pc,d3.w*2), ($33,pc,d3.w*2); =37 uses Op27fb ----------
Op27fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [27fc] move.l #$33333333, ($33,pc,d3.w*2); =39 uses Op27fc ----------
Op27fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ec0] move.l d0, (a7)+ uses Op2ec0 ----------
Op2ec0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ed0] move.l (a0), (a7)+ uses Op2ed0 ----------
Op2ed0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ed8] move.l (a0)+, (a7)+ uses Op2ed8 ----------
Op2ed8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2edf] move.l (a7)+, (a7)+ uses Op2edf ----------
Op2edf
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ee0] move.l -(a0), (a7)+ uses Op2ee0 ----------
Op2ee0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ee7] move.l -(a7), (a7)+ uses Op2ee7 ----------
Op2ee7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ee8] move.l ($3333,a0), (a7)+ uses Op2ee8 ----------
Op2ee8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ef0] move.l ($33,a0,d3.w*2), (a7)+ uses Op2ef0 ----------
Op2ef0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ef8] move.l $3333.w, (a7)+ uses Op2ef8 ----------
Op2ef8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2ef9] move.l $33333333.l, (a7)+ uses Op2ef9 ----------
Op2ef9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2efa] move.l ($3333,pc), (a7)+; =3335 uses Op2efa ----------
Op2efa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2efb] move.l ($33,pc,d3.w*2), (a7)+; =35 uses Op2efb ----------
Op2efb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2efc] move.l #$33333333, (a7)+ uses Op2efc ----------
Op2efc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f00] move.l d0, -(a7) uses Op2f00 ----------
Op2f00
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f10] move.l (a0), -(a7) uses Op2f10 ----------
Op2f10
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f18] move.l (a0)+, -(a7) uses Op2f18 ----------
Op2f18
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f1f] move.l (a7)+, -(a7) uses Op2f1f ----------
Op2f1f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f20] move.l -(a0), -(a7) uses Op2f20 ----------
Op2f20
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f27] move.l -(a7), -(a7) uses Op2f27 ----------
Op2f27
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f28] move.l ($3333,a0), -(a7) uses Op2f28 ----------
Op2f28
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f30] move.l ($33,a0,d3.w*2), -(a7) uses Op2f30 ----------
Op2f30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f38] move.l $3333.w, -(a7) uses Op2f38 ----------
Op2f38
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f39] move.l $33333333.l, -(a7) uses Op2f39 ----------
Op2f39
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f3a] move.l ($3333,pc), -(a7); =3335 uses Op2f3a ----------
Op2f3a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f3b] move.l ($33,pc,d3.w*2), -(a7); =35 uses Op2f3b ----------
Op2f3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [2f3c] move.l #$33333333, -(a7) uses Op2f3c ----------
Op2f3c
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3000] move.w d0, d0 uses Op3000 ----------
Op3000
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3010] move.w (a0), d0 uses Op3010 ----------
Op3010
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3018] move.w (a0)+, d0 uses Op3018 ----------
Op3018
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [301f] move.w (a7)+, d0 uses Op301f ----------
Op301f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3020] move.w -(a0), d0 uses Op3020 ----------
Op3020
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3027] move.w -(a7), d0 uses Op3027 ----------
Op3027
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3028] move.w ($3333,a0), d0 uses Op3028 ----------
Op3028
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3030] move.w ($33,a0,d3.w*2), d0 uses Op3030 ----------
Op3030
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3038] move.w $3333.w, d0 uses Op3038 ----------
Op3038
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3039] move.w $33333333.l, d0 uses Op3039 ----------
Op3039
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [303a] move.w ($3333,pc), d0; =3335 uses Op303a ----------
Op303a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [303b] move.w ($33,pc,d3.w*2), d0; =35 uses Op303b ----------
Op303b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [303c] move.w #$3333, d0 uses Op303c ----------
Op303c
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3040] movea.w d0, a0 uses Op3040 ----------
Op3040
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3050] movea.w (a0), a0 uses Op3050 ----------
Op3050
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3058] movea.w (a0)+, a0 uses Op3058 ----------
Op3058
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [305f] movea.w (a7)+, a0 uses Op305f ----------
Op305f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3060] movea.w -(a0), a0 uses Op3060 ----------
Op3060
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3067] movea.w -(a7), a0 uses Op3067 ----------
Op3067
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3068] movea.w ($3333,a0), a0 uses Op3068 ----------
Op3068
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3070] movea.w ($33,a0,d3.w*2), a0 uses Op3070 ----------
Op3070
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3078] movea.w $3333.w, a0 uses Op3078 ----------
Op3078
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3079] movea.w $33333333.l, a0 uses Op3079 ----------
Op3079
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [307a] movea.w ($3333,pc), a0; =3335 uses Op307a ----------
Op307a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [307b] movea.w ($33,pc,d3.w*2), a0; =35 uses Op307b ----------
Op307b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [307c] movea.w #$3333, a0 uses Op307c ----------
Op307c
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3080] move.w d0, (a0) uses Op3080 ----------
Op3080
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3090] move.w (a0), (a0) uses Op3090 ----------
Op3090
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3098] move.w (a0)+, (a0) uses Op3098 ----------
Op3098
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [309f] move.w (a7)+, (a0) uses Op309f ----------
Op309f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30a0] move.w -(a0), (a0) uses Op30a0 ----------
Op30a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30a7] move.w -(a7), (a0) uses Op30a7 ----------
Op30a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30a8] move.w ($3333,a0), (a0) uses Op30a8 ----------
Op30a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30b0] move.w ($33,a0,d3.w*2), (a0) uses Op30b0 ----------
Op30b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30b8] move.w $3333.w, (a0) uses Op30b8 ----------
Op30b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30b9] move.w $33333333.l, (a0) uses Op30b9 ----------
Op30b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30ba] move.w ($3333,pc), (a0); =3335 uses Op30ba ----------
Op30ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30bb] move.w ($33,pc,d3.w*2), (a0); =35 uses Op30bb ----------
Op30bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30bc] move.w #$3333, (a0) uses Op30bc ----------
Op30bc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30c0] move.w d0, (a0)+ uses Op30c0 ----------
Op30c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30d0] move.w (a0), (a0)+ uses Op30d0 ----------
Op30d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30d8] move.w (a0)+, (a0)+ uses Op30d8 ----------
Op30d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30df] move.w (a7)+, (a0)+ uses Op30df ----------
Op30df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30e0] move.w -(a0), (a0)+ uses Op30e0 ----------
Op30e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30e7] move.w -(a7), (a0)+ uses Op30e7 ----------
Op30e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30e8] move.w ($3333,a0), (a0)+ uses Op30e8 ----------
Op30e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30f0] move.w ($33,a0,d3.w*2), (a0)+ uses Op30f0 ----------
Op30f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30f8] move.w $3333.w, (a0)+ uses Op30f8 ----------
Op30f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30f9] move.w $33333333.l, (a0)+ uses Op30f9 ----------
Op30f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30fa] move.w ($3333,pc), (a0)+; =3335 uses Op30fa ----------
Op30fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30fb] move.w ($33,pc,d3.w*2), (a0)+; =35 uses Op30fb ----------
Op30fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [30fc] move.w #$3333, (a0)+ uses Op30fc ----------
Op30fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3100] move.w d0, -(a0) uses Op3100 ----------
Op3100
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3110] move.w (a0), -(a0) uses Op3110 ----------
Op3110
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3118] move.w (a0)+, -(a0) uses Op3118 ----------
Op3118
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [311f] move.w (a7)+, -(a0) uses Op311f ----------
Op311f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3120] move.w -(a0), -(a0) uses Op3120 ----------
Op3120
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3127] move.w -(a7), -(a0) uses Op3127 ----------
Op3127
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3128] move.w ($3333,a0), -(a0) uses Op3128 ----------
Op3128
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3130] move.w ($33,a0,d3.w*2), -(a0) uses Op3130 ----------
Op3130
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3138] move.w $3333.w, -(a0) uses Op3138 ----------
Op3138
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3139] move.w $33333333.l, -(a0) uses Op3139 ----------
Op3139
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [313a] move.w ($3333,pc), -(a0); =3335 uses Op313a ----------
Op313a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [313b] move.w ($33,pc,d3.w*2), -(a0); =35 uses Op313b ----------
Op313b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [313c] move.w #$3333, -(a0) uses Op313c ----------
Op313c
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3140] move.w d0, ($3333,a0) uses Op3140 ----------
Op3140
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3150] move.w (a0), ($3333,a0) uses Op3150 ----------
Op3150
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3158] move.w (a0)+, ($3333,a0) uses Op3158 ----------
Op3158
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [315f] move.w (a7)+, ($3333,a0) uses Op315f ----------
Op315f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3160] move.w -(a0), ($3333,a0) uses Op3160 ----------
Op3160
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3167] move.w -(a7), ($3333,a0) uses Op3167 ----------
Op3167
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3168] move.w ($3333,a0), ($3333,a0) uses Op3168 ----------
Op3168
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3170] move.w ($33,a0,d3.w*2), ($3333,a0) uses Op3170 ----------
Op3170
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3178] move.w $3333.w, ($3333,a0) uses Op3178 ----------
Op3178
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3179] move.w $33333333.l, ($3333,a0) uses Op3179 ----------
Op3179
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [317a] move.w ($3333,pc), ($3333,a0); =3335 uses Op317a ----------
Op317a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [317b] move.w ($33,pc,d3.w*2), ($3333,a0); =35 uses Op317b ----------
Op317b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [317c] move.w #$3333, ($3333,a0) uses Op317c ----------
Op317c
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3180] move.w d0, ($33,a0,d3.w*2) uses Op3180 ----------
Op3180
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3190] move.w (a0), ($33,a0,d3.w*2) uses Op3190 ----------
Op3190
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3198] move.w (a0)+, ($33,a0,d3.w*2) uses Op3198 ----------
Op3198
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [319f] move.w (a7)+, ($33,a0,d3.w*2) uses Op319f ----------
Op319f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31a0] move.w -(a0), ($33,a0,d3.w*2) uses Op31a0 ----------
Op31a0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31a7] move.w -(a7), ($33,a0,d3.w*2) uses Op31a7 ----------
Op31a7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31a8] move.w ($3333,a0), ($33,a0,d3.w*2) uses Op31a8 ----------
Op31a8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31b0] move.w ($33,a0,d3.w*2), ($33,a0,d3.w*2) uses Op31b0 ----------
Op31b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31b8] move.w $3333.w, ($33,a0,d3.w*2) uses Op31b8 ----------
Op31b8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31b9] move.w $33333333.l, ($33,a0,d3.w*2) uses Op31b9 ----------
Op31b9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31ba] move.w ($3333,pc), ($33,a0,d3.w*2); =3335 uses Op31ba ----------
Op31ba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31bb] move.w ($33,pc,d3.w*2), ($33,a0,d3.w*2); =35 uses Op31bb ----------
Op31bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31bc] move.w #$3333, ($33,a0,d3.w*2) uses Op31bc ----------
Op31bc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31c0] move.w d0, $3333.w uses Op31c0 ----------
Op31c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31d0] move.w (a0), $3333.w uses Op31d0 ----------
Op31d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31d8] move.w (a0)+, $3333.w uses Op31d8 ----------
Op31d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31df] move.w (a7)+, $3333.w uses Op31df ----------
Op31df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31e0] move.w -(a0), $3333.w uses Op31e0 ----------
Op31e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31e7] move.w -(a7), $3333.w uses Op31e7 ----------
Op31e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31e8] move.w ($3333,a0), $3333.w uses Op31e8 ----------
Op31e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31f0] move.w ($33,a0,d3.w*2), $3333.w uses Op31f0 ----------
Op31f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31f8] move.w $3333.w, $3333.w uses Op31f8 ----------
Op31f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31f9] move.w $33333333.l, $3333.w uses Op31f9 ----------
Op31f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31fa] move.w ($3333,pc), $3333.w; =3335 uses Op31fa ----------
Op31fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31fb] move.w ($33,pc,d3.w*2), $3333.w; =35 uses Op31fb ----------
Op31fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [31fc] move.w #$3333, $3333.w uses Op31fc ----------
Op31fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33c0] move.w d0, $33333333.l uses Op33c0 ----------
Op33c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33d0] move.w (a0), $33333333.l uses Op33d0 ----------
Op33d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33d8] move.w (a0)+, $33333333.l uses Op33d8 ----------
Op33d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33df] move.w (a7)+, $33333333.l uses Op33df ----------
Op33df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33e0] move.w -(a0), $33333333.l uses Op33e0 ----------
Op33e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33e7] move.w -(a7), $33333333.l uses Op33e7 ----------
Op33e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33e8] move.w ($3333,a0), $33333333.l uses Op33e8 ----------
Op33e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33f0] move.w ($33,a0,d3.w*2), $33333333.l uses Op33f0 ----------
Op33f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33f8] move.w $3333.w, $33333333.l uses Op33f8 ----------
Op33f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33f9] move.w $33333333.l, $33333333.l uses Op33f9 ----------
Op33f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33fa] move.w ($3333,pc), $33333333.l; =3335 uses Op33fa ----------
Op33fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33fb] move.w ($33,pc,d3.w*2), $33333333.l; =35 uses Op33fb ----------
Op33fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [33fc] move.w #$3333, $33333333.l uses Op33fc ----------
Op33fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35c0] move.w d0, ($3333,pc); =3335 uses Op35c0 ----------
Op35c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35d0] move.w (a0), ($3333,pc); =3335 uses Op35d0 ----------
Op35d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35d8] move.w (a0)+, ($3333,pc); =3335 uses Op35d8 ----------
Op35d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35df] move.w (a7)+, ($3333,pc); =3335 uses Op35df ----------
Op35df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35e0] move.w -(a0), ($3333,pc); =3335 uses Op35e0 ----------
Op35e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35e7] move.w -(a7), ($3333,pc); =3335 uses Op35e7 ----------
Op35e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35e8] move.w ($3333,a0), ($3333,pc); =3337 uses Op35e8 ----------
Op35e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35f0] move.w ($33,a0,d3.w*2), ($3333,pc); =3337 uses Op35f0 ----------
Op35f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35f8] move.w $3333.w, ($3333,pc); =3337 uses Op35f8 ----------
Op35f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35f9] move.w $33333333.l, ($3333,pc); =3339 uses Op35f9 ----------
Op35f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35fa] move.w ($3333,pc), ($3333,pc); =3337 uses Op35fa ----------
Op35fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35fb] move.w ($33,pc,d3.w*2), ($3333,pc); =3337 uses Op35fb ----------
Op35fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [35fc] move.w #$3333, ($3333,pc); =3337 uses Op35fc ----------
Op35fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37c0] move.w d0, ($33,pc,d3.w*2); =35 uses Op37c0 ----------
Op37c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37d0] move.w (a0), ($33,pc,d3.w*2); =35 uses Op37d0 ----------
Op37d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37d8] move.w (a0)+, ($33,pc,d3.w*2); =35 uses Op37d8 ----------
Op37d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37df] move.w (a7)+, ($33,pc,d3.w*2); =35 uses Op37df ----------
Op37df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37e0] move.w -(a0), ($33,pc,d3.w*2); =35 uses Op37e0 ----------
Op37e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37e7] move.w -(a7), ($33,pc,d3.w*2); =35 uses Op37e7 ----------
Op37e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37e8] move.w ($3333,a0), ($33,pc,d3.w*2); =37 uses Op37e8 ----------
Op37e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37f0] move.w ($33,a0,d3.w*2), ($33,pc,d3.w*2); =37 uses Op37f0 ----------
Op37f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37f8] move.w $3333.w, ($33,pc,d3.w*2); =37 uses Op37f8 ----------
Op37f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37f9] move.w $33333333.l, ($33,pc,d3.w*2); =39 uses Op37f9 ----------
Op37f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37fa] move.w ($3333,pc), ($33,pc,d3.w*2); =37 uses Op37fa ----------
Op37fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37fb] move.w ($33,pc,d3.w*2), ($33,pc,d3.w*2); =37 uses Op37fb ----------
Op37fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [37fc] move.w #$3333, ($33,pc,d3.w*2); =37 uses Op37fc ----------
Op37fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ec0] move.w d0, (a7)+ uses Op3ec0 ----------
Op3ec0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ed0] move.w (a0), (a7)+ uses Op3ed0 ----------
Op3ed0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ed8] move.w (a0)+, (a7)+ uses Op3ed8 ----------
Op3ed8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3edf] move.w (a7)+, (a7)+ uses Op3edf ----------
Op3edf
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ee0] move.w -(a0), (a7)+ uses Op3ee0 ----------
Op3ee0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ee7] move.w -(a7), (a7)+ uses Op3ee7 ----------
Op3ee7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ee8] move.w ($3333,a0), (a7)+ uses Op3ee8 ----------
Op3ee8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ef0] move.w ($33,a0,d3.w*2), (a7)+ uses Op3ef0 ----------
Op3ef0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ef8] move.w $3333.w, (a7)+ uses Op3ef8 ----------
Op3ef8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3ef9] move.w $33333333.l, (a7)+ uses Op3ef9 ----------
Op3ef9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3efa] move.w ($3333,pc), (a7)+; =3335 uses Op3efa ----------
Op3efa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3efb] move.w ($33,pc,d3.w*2), (a7)+; =35 uses Op3efb ----------
Op3efb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3efc] move.w #$3333, (a7)+ uses Op3efc ----------
Op3efc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f00] move.w d0, -(a7) uses Op3f00 ----------
Op3f00
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f10] move.w (a0), -(a7) uses Op3f10 ----------
Op3f10
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f18] move.w (a0)+, -(a7) uses Op3f18 ----------
Op3f18
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f1f] move.w (a7)+, -(a7) uses Op3f1f ----------
Op3f1f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f20] move.w -(a0), -(a7) uses Op3f20 ----------
Op3f20
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f27] move.w -(a7), -(a7) uses Op3f27 ----------
Op3f27
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f28] move.w ($3333,a0), -(a7) uses Op3f28 ----------
Op3f28
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f30] move.w ($33,a0,d3.w*2), -(a7) uses Op3f30 ----------
Op3f30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f38] move.w $3333.w, -(a7) uses Op3f38 ----------
Op3f38
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f39] move.w $33333333.l, -(a7) uses Op3f39 ----------
Op3f39
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f3a] move.w ($3333,pc), -(a7); =3335 uses Op3f3a ----------
Op3f3a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f3b] move.w ($33,pc,d3.w*2), -(a7); =35 uses Op3f3b ----------
Op3f3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [3f3c] move.w #$3333, -(a7) uses Op3f3c ----------
Op3f3c
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  adds r1,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=NZCV flags

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40c0] move sr, d0 uses Op40c0 ----------
Op40c0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40c0 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40c0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40d0] move sr, (a0) uses Op40d0 ----------
Op40d0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40d0 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40d0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40d8] move sr, (a0)+ uses Op40d8 ----------
Op40d8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40d8 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40d8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40df] move sr, (a7)+ uses Op40df ----------
Op40df
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40df ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40df
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40e0] move sr, -(a0) uses Op40e0 ----------
Op40e0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40e0 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40e0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40e7] move sr, -(a7) uses Op40e7 ----------
Op40e7
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40e7 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40e7
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40e8] move sr, ($3333,a0) uses Op40e8 ----------
Op40e8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40e8 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40e8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40f0] move sr, ($33,a0,d3.w*2) uses Op40f0 ----------
Op40f0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40f0 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40f0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40f8] move sr, $3333.w uses Op40f8 ----------
Op40f8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40f8 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40f8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40f9] move sr, $33333333.l uses Op40f9 ----------
Op40f9
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40f9 ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40f9
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40fa] move sr, ($3333,pc); =3335 uses Op40fa ----------
Op40fa
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40fa ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40fa
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [40fb] move sr, ($33,pc,d3.w*2); =35 uses Op40fb ----------
Op40fb
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode40fb ;@ No

  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  ldrb r2,[r7,#0x44]  ;@ Include SR high
  and r0,r0,#0x02
  orr r1,r1,r2,lsl #8
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode40fb
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41d0] lea (a0), a0 uses Op41d0 ----------
Op41d0
;@ EaCalc : Get '(a0)' into r1:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r1,[r7,r2]
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41e8] lea ($3333,a0), a0 uses Op41e8 ----------
Op41e8
;@ EaCalc : Get '($3333,a0)' into r1:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r1,r0,r2 ;@ Add on offset
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41f0] lea ($33,a0,d3.w*2), a0 uses Op41f0 ----------
Op41f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r1:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r1,r2,r3 ;@ r1=Disp+An+Rn
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41f8] lea $3333.w, a0 uses Op41f8 ----------
Op41f8
;@ EaCalc : Get '$3333.w' into r1:
  ldrsh r1,[r4],#2 ;@ Fetch Absolute Short address
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41f9] lea $33333333.l, a0 uses Op41f9 ----------
Op41f9
;@ EaCalc : Get '$33333333.l' into r1:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r1,r0,r2,lsl #16
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41fa] lea ($3333,pc), a0; =3335 uses Op41fa ----------
Op41fa
;@ EaCalc : Get '($3333,pc)' into r1:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r1,r0,r2 ;@ ($nn,PC)
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [41fb] lea ($33,pc,d3.w*2), a0; =35 uses Op41fb ----------
Op41fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r1:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r1,r2,r0 ;@ r1=Disp+PC+Rn
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4200] clr.b d0 uses Op4200 ----------
Op4200
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: r1 into register[r10]:
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4210] clr.b (a0) uses Op4210 ----------
Op4210
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4218] clr.b (a0)+ uses Op4218 ----------
Op4218
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [421f] clr.b (a7)+ uses Op421f ----------
Op421f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4220] clr.b -(a0) uses Op4220 ----------
Op4220
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4227] clr.b -(a7) uses Op4227 ----------
Op4227
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4228] clr.b ($3333,a0) uses Op4228 ----------
Op4228
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4230] clr.b ($33,a0,d3.w*2) uses Op4230 ----------
Op4230
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4238] clr.b $3333.w uses Op4238 ----------
Op4238
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4239] clr.b $33333333.l uses Op4239 ----------
Op4239
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [423a] clr.b ($3333,pc); =3335 uses Op423a ----------
Op423a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [423b] clr.b ($33,pc,d3.w*2); =35 uses Op423b ----------
Op423b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4240] clr.w d0 uses Op4240 ----------
Op4240
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: r1 into register[r10]:
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4250] clr.w (a0) uses Op4250 ----------
Op4250
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4258] clr.w (a0)+ uses Op4258 ----------
Op4258
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [425f] clr.w (a7)+ uses Op425f ----------
Op425f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4260] clr.w -(a0) uses Op4260 ----------
Op4260
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4267] clr.w -(a7) uses Op4267 ----------
Op4267
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4268] clr.w ($3333,a0) uses Op4268 ----------
Op4268
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4270] clr.w ($33,a0,d3.w*2) uses Op4270 ----------
Op4270
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4278] clr.w $3333.w uses Op4278 ----------
Op4278
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4279] clr.w $33333333.l uses Op4279 ----------
Op4279
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [427a] clr.w ($3333,pc); =3335 uses Op427a ----------
Op427a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [427b] clr.w ($33,pc,d3.w*2); =35 uses Op427b ----------
Op427b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4280] clr.l d0 uses Op4280 ----------
Op4280
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4290] clr.l (a0) uses Op4290 ----------
Op4290
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4298] clr.l (a0)+ uses Op4298 ----------
Op4298
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [429f] clr.l (a7)+ uses Op429f ----------
Op429f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42a0] clr.l -(a0) uses Op42a0 ----------
Op42a0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42a7] clr.l -(a7) uses Op42a7 ----------
Op42a7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42a8] clr.l ($3333,a0) uses Op42a8 ----------
Op42a8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42b0] clr.l ($33,a0,d3.w*2) uses Op42b0 ----------
Op42b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42b8] clr.l $3333.w uses Op42b8 ----------
Op42b8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42b9] clr.l $33333333.l uses Op42b9 ----------
Op42b9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42ba] clr.l ($3333,pc); =3335 uses Op42ba ----------
Op42ba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42bb] clr.l ($33,pc,d3.w*2); =35 uses Op42bb ----------
Op42bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn

;@ Clear:
  mov r1,#0
  mov r9,#0x40000000 ;@ NZCV=0100

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42c0] move ccr, d0 uses Op42c0 ----------
Op42c0
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42d0] move ccr, (a0) uses Op42d0 ----------
Op42d0
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42d8] move ccr, (a0)+ uses Op42d8 ----------
Op42d8
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42df] move ccr, (a7)+ uses Op42df ----------
Op42df
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42e0] move ccr, -(a0) uses Op42e0 ----------
Op42e0
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42e7] move ccr, -(a7) uses Op42e7 ----------
Op42e7
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42e8] move ccr, ($3333,a0) uses Op42e8 ----------
Op42e8
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42f0] move ccr, ($33,a0,d3.w*2) uses Op42f0 ----------
Op42f0
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42f8] move ccr, $3333.w uses Op42f8 ----------
Op42f8
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42f9] move ccr, $33333333.l uses Op42f9 ----------
Op42f9
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42fa] move ccr, ($3333,pc); =3335 uses Op42fa ----------
Op42fa
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [42fb] move ccr, ($33,pc,d3.w*2); =35 uses Op42fb ----------
Op42fb
  mov r1,r9,lsr #28   ;@ ____NZCV
  eor r0,r1,r1,ror #1 ;@ Bit 0=C^V
  tst r0,#1           ;@ 1 if C!=V
  eorne r1,r1,#3      ;@ ____NZVC

  ldrb r0,[r7,#0x45]  ;@ X bit
  and r0,r0,#0x02
  orr r1,r1,r0,lsl #3 ;@ ___XNZVC

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4400] neg.b d0 uses Op4400 ----------
Op4400
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4410] neg.b (a0) uses Op4410 ----------
Op4410
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4418] neg.b (a0)+ uses Op4418 ----------
Op4418
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [441f] neg.b (a7)+ uses Op441f ----------
Op441f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4420] neg.b -(a0) uses Op4420 ----------
Op4420
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4427] neg.b -(a7) uses Op4427 ----------
Op4427
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4428] neg.b ($3333,a0) uses Op4428 ----------
Op4428
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4430] neg.b ($33,a0,d3.w*2) uses Op4430 ----------
Op4430
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4438] neg.b $3333.w uses Op4438 ----------
Op4438
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4439] neg.b $33333333.l uses Op4439 ----------
Op4439
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [443a] neg.b ($3333,pc); =3335 uses Op443a ----------
Op443a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [443b] neg.b ($33,pc,d3.w*2); =35 uses Op443b ----------
Op443b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4440] neg.w d0 uses Op4440 ----------
Op4440
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4450] neg.w (a0) uses Op4450 ----------
Op4450
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4458] neg.w (a0)+ uses Op4458 ----------
Op4458
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [445f] neg.w (a7)+ uses Op445f ----------
Op445f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4460] neg.w -(a0) uses Op4460 ----------
Op4460
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4467] neg.w -(a7) uses Op4467 ----------
Op4467
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4468] neg.w ($3333,a0) uses Op4468 ----------
Op4468
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4470] neg.w ($33,a0,d3.w*2) uses Op4470 ----------
Op4470
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4478] neg.w $3333.w uses Op4478 ----------
Op4478
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4479] neg.w $33333333.l uses Op4479 ----------
Op4479
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [447a] neg.w ($3333,pc); =3335 uses Op447a ----------
Op447a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [447b] neg.w ($33,pc,d3.w*2); =35 uses Op447b ----------
Op447b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4480] neg.l d0 uses Op4480 ----------
Op4480
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4490] neg.l (a0) uses Op4490 ----------
Op4490
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4498] neg.l (a0)+ uses Op4498 ----------
Op4498
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [449f] neg.l (a7)+ uses Op449f ----------
Op449f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44a0] neg.l -(a0) uses Op44a0 ----------
Op44a0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44a7] neg.l -(a7) uses Op44a7 ----------
Op44a7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44a8] neg.l ($3333,a0) uses Op44a8 ----------
Op44a8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44b0] neg.l ($33,a0,d3.w*2) uses Op44b0 ----------
Op44b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44b8] neg.l $3333.w uses Op44b8 ----------
Op44b8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44b9] neg.l $33333333.l uses Op44b9 ----------
Op44b9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44ba] neg.l ($3333,pc); =3335 uses Op44ba ----------
Op44ba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44bb] neg.l ($33,pc,d3.w*2); =35 uses Op44bb ----------
Op44bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Neg:
  rsbs r1,r0,#0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44c0] move d0, ccr uses Op44c0 ----------
Op44c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44d0] move (a0), ccr uses Op44d0 ----------
Op44d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44d8] move (a0)+, ccr uses Op44d8 ----------
Op44d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44df] move (a7)+, ccr uses Op44df ----------
Op44df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44e0] move -(a0), ccr uses Op44e0 ----------
Op44e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44e7] move -(a7), ccr uses Op44e7 ----------
Op44e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44e8] move ($3333,a0), ccr uses Op44e8 ----------
Op44e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44f0] move ($33,a0,d3.w*2), ccr uses Op44f0 ----------
Op44f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44f8] move $3333.w, ccr uses Op44f8 ----------
Op44f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44f9] move $33333333.l, ccr uses Op44f9 ----------
Op44f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44fa] move ($3333,pc), ccr; =3335 uses Op44fa ----------
Op44fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44fb] move ($33,pc,d3.w*2), ccr; =35 uses Op44fb ----------
Op44fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [44fc] move #$3333, ccr uses Op44fc ----------
Op44fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4600] not.b d0 uses Op4600 ----------
Op4600
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4610] not.b (a0) uses Op4610 ----------
Op4610
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4618] not.b (a0)+ uses Op4618 ----------
Op4618
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [461f] not.b (a7)+ uses Op461f ----------
Op461f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4620] not.b -(a0) uses Op4620 ----------
Op4620
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4627] not.b -(a7) uses Op4627 ----------
Op4627
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4628] not.b ($3333,a0) uses Op4628 ----------
Op4628
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4630] not.b ($33,a0,d3.w*2) uses Op4630 ----------
Op4630
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4638] not.b $3333.w uses Op4638 ----------
Op4638
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4639] not.b $33333333.l uses Op4639 ----------
Op4639
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [463a] not.b ($3333,pc); =3335 uses Op463a ----------
Op463a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [463b] not.b ($33,pc,d3.w*2); =35 uses Op463b ----------
Op463b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4640] not.w d0 uses Op4640 ----------
Op4640
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4650] not.w (a0) uses Op4650 ----------
Op4650
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4658] not.w (a0)+ uses Op4658 ----------
Op4658
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [465f] not.w (a7)+ uses Op465f ----------
Op465f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4660] not.w -(a0) uses Op4660 ----------
Op4660
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4667] not.w -(a7) uses Op4667 ----------
Op4667
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4668] not.w ($3333,a0) uses Op4668 ----------
Op4668
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4670] not.w ($33,a0,d3.w*2) uses Op4670 ----------
Op4670
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4678] not.w $3333.w uses Op4678 ----------
Op4678
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4679] not.w $33333333.l uses Op4679 ----------
Op4679
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [467a] not.w ($3333,pc); =3335 uses Op467a ----------
Op467a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [467b] not.w ($33,pc,d3.w*2); =35 uses Op467b ----------
Op467b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4680] not.l d0 uses Op4680 ----------
Op4680
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4690] not.l (a0) uses Op4690 ----------
Op4690
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4698] not.l (a0)+ uses Op4698 ----------
Op4698
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [469f] not.l (a7)+ uses Op469f ----------
Op469f
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46a0] not.l -(a0) uses Op46a0 ----------
Op46a0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46a7] not.l -(a7) uses Op46a7 ----------
Op46a7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46a8] not.l ($3333,a0) uses Op46a8 ----------
Op46a8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46b0] not.l ($33,a0,d3.w*2) uses Op46b0 ----------
Op46b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46b8] not.l $3333.w uses Op46b8 ----------
Op46b8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46b9] not.l $33333333.l uses Op46b9 ----------
Op46b9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46ba] not.l ($3333,pc); =3335 uses Op46ba ----------
Op46ba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46bb] not.l ($33,pc,d3.w*2); =35 uses Op46bb ----------
Op46bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Not:
  mvn r1,r0
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46c0] move d0, sr uses Op46c0 ----------
Op46c0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46c0 ;@ No

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46c0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46d0] move (a0), sr uses Op46d0 ----------
Op46d0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46d0 ;@ No

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46d0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46d8] move (a0)+, sr uses Op46d8 ----------
Op46d8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46d8 ;@ No

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46d8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46df] move (a7)+, sr uses Op46df ----------
Op46df
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46df ;@ No

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46df
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46e0] move -(a0), sr uses Op46e0 ----------
Op46e0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46e0 ;@ No

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46e0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46e7] move -(a7), sr uses Op46e7 ----------
Op46e7
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46e7 ;@ No

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46e7
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46e8] move ($3333,a0), sr uses Op46e8 ----------
Op46e8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46e8 ;@ No

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46e8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46f0] move ($33,a0,d3.w*2), sr uses Op46f0 ----------
Op46f0
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46f0 ;@ No

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46f0
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46f8] move $3333.w, sr uses Op46f8 ----------
Op46f8
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46f8 ;@ No

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46f8
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46f9] move $33333333.l, sr uses Op46f9 ----------
Op46f9
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46f9 ;@ No

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46f9
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46fa] move ($3333,pc), sr; =3335 uses Op46fa ----------
Op46fa
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46fa ;@ No

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46fa
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46fb] move ($33,pc,d3.w*2), sr; =35 uses Op46fb ----------
Op46fb
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46fb ;@ No

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46fb
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [46fc] move #$3333, sr uses Op46fc ----------
Op46fc
  ldrb r0,[r7,#0x44] ;@ Get SR high
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode46fc ;@ No

;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode46fc
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4840] swap d0 uses Op4840 ----------
Op4840
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r1,r0,ror #16
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4850] pea (a0) uses Op4850 ----------
Op4850
;@ EaCalc : Get '(a0)' into r1:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r1,[r7,r2]

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4868] pea ($3333,a0) uses Op4868 ----------
Op4868
;@ EaCalc : Get '($3333,a0)' into r1:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r1,r0,r2 ;@ Add on offset

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4870] pea ($33,a0,d3.w*2) uses Op4870 ----------
Op4870
;@ EaCalc : Get '($33,a0,d3.w*2)' into r1:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r1,r2,r3 ;@ r1=Disp+An+Rn

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4878] pea $3333.w uses Op4878 ----------
Op4878
;@ EaCalc : Get '$3333.w' into r1:
  ldrsh r1,[r4],#2 ;@ Fetch Absolute Short address

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4879] pea $33333333.l uses Op4879 ----------
Op4879
;@ EaCalc : Get '$33333333.l' into r1:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r1,r0,r2,lsl #16

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [487a] pea ($3333,pc); =3335 uses Op487a ----------
Op487a
;@ EaCalc : Get '($3333,pc)' into r1:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r1,r0,r2 ;@ ($nn,PC)

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [487b] pea ($33,pc,d3.w*2); =35 uses Op487b ----------
Op487b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r1:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r1,r2,r0 ;@ r1=Disp+PC+Rn

  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7

  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4880] ext.w d0 uses Op4880 ----------
Op4880
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]

  mov r0,r0,asl #24
  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r0,asr #24

;@ EaWrite: r1 into register[r10]:
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4890] movem.w d0-d1/d4-d5/a0-a1/a4-a5, (a0) uses Op4890 ----------
Op4890
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4890
  tst r11,#1
  beq SkipReg4890

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4890
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4890

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4898] movem.w d0-d1/d4-d5/a0-a1/a4-a5, (a0)+ uses Op4898 ----------
Op4898
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4898
  tst r11,#1
  beq SkipReg4898

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a0)+' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4898
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4898

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [489f] movem.w d0-d1/d4-d5/a0-a1/a4-a5, (a7)+ uses Op489f ----------
Op489f
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg489f
  tst r11,#1
  beq SkipReg489f

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a7)+' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg489f
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg489f

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48a0] movem.w d2-d3/d6-d7/a2-a3/a6-a7, -(a0) uses Op48a0 ----------
Op48a0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg48a0
  tst r11,#1
  beq SkipReg48a0

  sub r9,r9,#2 ;@ Pre-decrement address
  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '-(a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  sub r5,r5,#4 ;@ Take some cycles

SkipReg48a0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg48a0

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48a7] movem.w d2-d3/d6-d7/a2-a3/a6-a7, -(a7) uses Op48a7 ----------
Op48a7
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg48a7
  tst r11,#1
  beq SkipReg48a7

  sub r9,r9,#2 ;@ Pre-decrement address
  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '-(a7)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  sub r5,r5,#4 ;@ Take some cycles

SkipReg48a7
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg48a7

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48a8] movem.w d0-d1/d4-d5/a0-a1/a4-a5, ($3333,a0) uses Op48a8 ----------
Op48a8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($3333,a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r9,r0,r2 ;@ Add on offset
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48a8
  tst r11,#1
  beq SkipReg48a8

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '($3333,a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg48a8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48a8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48b0] movem.w d0-d1/d4-d5/a0-a1/a4-a5, ($33,a0,d3.w*2) uses Op48b0 ----------
Op48b0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r9:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r9,r2,r3 ;@ r9=Disp+An+Rn
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48b0
  tst r11,#1
  beq SkipReg48b0

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg48b0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48b0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48b8] movem.w d0-d1/d4-d5/a0-a1/a4-a5, $3333.w uses Op48b8 ----------
Op48b8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$3333.w' into r9:
  ldrsh r9,[r4],#2 ;@ Fetch Absolute Short address
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48b8
  tst r11,#1
  beq SkipReg48b8

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '$3333.w' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg48b8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48b8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48b9] movem.w d0-d1/d4-d5/a0-a1/a4-a5, $33333333.l uses Op48b9 ----------
Op48b9
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$33333333.l' into r9:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r9,r0,r2,lsl #16
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48b9
  tst r11,#1
  beq SkipReg48b9

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '$33333333.l' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg48b9
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48b9

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48c0] ext.l d0 uses Op48c0 ----------
Op48c0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r0,r0,asl #16
  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r0,asr #16

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48d0] movem.l d0-d1/d4-d5/a0-a1/a4-a5, (a0) uses Op48d0 ----------
Op48d0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48d0
  tst r11,#1
  beq SkipReg48d0

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48d0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48d0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48d8] movem.l d0-d1/d4-d5/a0-a1/a4-a5, (a0)+ uses Op48d8 ----------
Op48d8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48d8
  tst r11,#1
  beq SkipReg48d8

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a0)+' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48d8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48d8

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48df] movem.l d0-d1/d4-d5/a0-a1/a4-a5, (a7)+ uses Op48df ----------
Op48df
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48df
  tst r11,#1
  beq SkipReg48df

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '(a7)+' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48df
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48df

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48e0] movem.l d2-d3/d6-d7/a2-a3/a6-a7, -(a0) uses Op48e0 ----------
Op48e0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg48e0
  tst r11,#1
  beq SkipReg48e0

  sub r9,r9,#4 ;@ Pre-decrement address
  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '-(a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  sub r5,r5,#8 ;@ Take some cycles

SkipReg48e0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg48e0

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48e7] movem.l d2-d3/d6-d7/a2-a3/a6-a7, -(a7) uses Op48e7 ----------
Op48e7
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg48e7
  tst r11,#1
  beq SkipReg48e7

  sub r9,r9,#4 ;@ Pre-decrement address
  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '-(a7)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  sub r5,r5,#8 ;@ Take some cycles

SkipReg48e7
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg48e7

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48e8] movem.l d0-d1/d4-d5/a0-a1/a4-a5, ($3333,a0) uses Op48e8 ----------
Op48e8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($3333,a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r9,r0,r2 ;@ Add on offset
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48e8
  tst r11,#1
  beq SkipReg48e8

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '($3333,a0)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48e8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48e8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48f0] movem.l d0-d1/d4-d5/a0-a1/a4-a5, ($33,a0,d3.w*2) uses Op48f0 ----------
Op48f0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r9:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r9,r2,r3 ;@ r9=Disp+An+Rn
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48f0
  tst r11,#1
  beq SkipReg48f0

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48f0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48f0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48f8] movem.l d0-d1/d4-d5/a0-a1/a4-a5, $3333.w uses Op48f8 ----------
Op48f8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$3333.w' into r9:
  ldrsh r9,[r4],#2 ;@ Fetch Absolute Short address
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48f8
  tst r11,#1
  beq SkipReg48f8

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '$3333.w' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48f8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48f8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [48f9] movem.l d0-d1/d4-d5/a0-a1/a4-a5, $33333333.l uses Op48f9 ----------
Op48f9
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$33333333.l' into r9:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r9,r0,r2,lsl #16
;@ r10=Register Index*4:
  mov r10,#0

MoreReg48f9
  tst r11,#1
  beq SkipReg48f9

  ;@ Copy register to memory:
  ldr r1,[r7,r10] ;@ Load value from Dn/An
;@ EaWrite: Write r1 into '$33333333.l' (address in r9):
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg48f9
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg48f9

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a00] tst.b d0 uses Op4a00 ----------
Op4a00
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsb r0,[r7,r0]

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a10] tst.b (a0) uses Op4a10 ----------
Op4a10
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a18] tst.b (a0)+ uses Op4a18 ----------
Op4a18
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a1f] tst.b (a7)+ uses Op4a1f ----------
Op4a1f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a20] tst.b -(a0) uses Op4a20 ----------
Op4a20
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a27] tst.b -(a7) uses Op4a27 ----------
Op4a27
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a28] tst.b ($3333,a0) uses Op4a28 ----------
Op4a28
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a30] tst.b ($33,a0,d3.w*2) uses Op4a30 ----------
Op4a30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a38] tst.b $3333.w uses Op4a38 ----------
Op4a38
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a39] tst.b $33333333.l uses Op4a39 ----------
Op4a39
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a3a] tst.b ($3333,pc); =3335 uses Op4a3a ----------
Op4a3a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a3b] tst.b ($33,pc,d3.w*2); =35 uses Op4a3b ----------
Op4a3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24
  mov r0,r0,asr #24

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a40] tst.w d0 uses Op4a40 ----------
Op4a40
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r0:
  ldrsh r0,[r7,r0]

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a50] tst.w (a0) uses Op4a50 ----------
Op4a50
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a58] tst.w (a0)+ uses Op4a58 ----------
Op4a58
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a5f] tst.w (a7)+ uses Op4a5f ----------
Op4a5f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a60] tst.w -(a0) uses Op4a60 ----------
Op4a60
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a67] tst.w -(a7) uses Op4a67 ----------
Op4a67
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a68] tst.w ($3333,a0) uses Op4a68 ----------
Op4a68
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a70] tst.w ($33,a0,d3.w*2) uses Op4a70 ----------
Op4a70
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a78] tst.w $3333.w uses Op4a78 ----------
Op4a78
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a79] tst.w $33333333.l uses Op4a79 ----------
Op4a79
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a7a] tst.w ($3333,pc); =3335 uses Op4a7a ----------
Op4a7a
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a7b] tst.w ($33,pc,d3.w*2); =35 uses Op4a7b ----------
Op4a7b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a80] tst.l d0 uses Op4a80 ----------
Op4a80
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a90] tst.l (a0) uses Op4a90 ----------
Op4a90
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a98] tst.l (a0)+ uses Op4a98 ----------
Op4a98
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4a9f] tst.l (a7)+ uses Op4a9f ----------
Op4a9f
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4aa0] tst.l -(a0) uses Op4aa0 ----------
Op4aa0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4aa7] tst.l -(a7) uses Op4aa7 ----------
Op4aa7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4aa8] tst.l ($3333,a0) uses Op4aa8 ----------
Op4aa8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ab0] tst.l ($33,a0,d3.w*2) uses Op4ab0 ----------
Op4ab0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ab8] tst.l $3333.w uses Op4ab8 ----------
Op4ab8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ab9] tst.l $33333333.l uses Op4ab9 ----------
Op4ab9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4aba] tst.l ($3333,pc); =3335 uses Op4aba ----------
Op4aba
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4abb] tst.l ($33,pc,d3.w*2); =35 uses Op4abb ----------
Op4abb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r0:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r0,r0,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4c90] movem.w (a0), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4c90 ----------
Op4c90
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4c90
  tst r11,#1
  beq SkipReg4c90

  ;@ Copy memory to register:
;@ EaRead : Read '(a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4c90
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4c90

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4c98] movem.w (a0)+, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4c98 ----------
Op4c98
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4c98
  tst r11,#1
  beq SkipReg4c98

  ;@ Copy memory to register:
;@ EaRead : Read '(a0)+' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4c98
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4c98

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4c9f] movem.w (a7)+, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4c9f ----------
Op4c9f
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4c9f
  tst r11,#1
  beq SkipReg4c9f

  ;@ Copy memory to register:
;@ EaRead : Read '(a7)+' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4c9f
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4c9f

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ca0] movem.w -(a0), d2-d3/d6-d7/a2-a3/a6-a7 uses Op4ca0 ----------
Op4ca0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg4ca0
  tst r11,#1
  beq SkipReg4ca0

  sub r9,r9,#2 ;@ Pre-decrement address
  ;@ Copy memory to register:
;@ EaRead : Read '-(a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4ca0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg4ca0

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ca7] movem.w -(a7), d2-d3/d6-d7/a2-a3/a6-a7 uses Op4ca7 ----------
Op4ca7
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg4ca7
  tst r11,#1
  beq SkipReg4ca7

  sub r9,r9,#2 ;@ Pre-decrement address
  ;@ Copy memory to register:
;@ EaRead : Read '-(a7)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4ca7
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg4ca7

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ca8] movem.w ($3333,a0), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4ca8 ----------
Op4ca8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($3333,a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r9,r0,r2 ;@ Add on offset
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4ca8
  tst r11,#1
  beq SkipReg4ca8

  ;@ Copy memory to register:
;@ EaRead : Read '($3333,a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4ca8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4ca8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cb0] movem.w ($33,a0,d3.w*2), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cb0 ----------
Op4cb0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r9:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r9,r2,r3 ;@ r9=Disp+An+Rn
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cb0
  tst r11,#1
  beq SkipReg4cb0

  ;@ Copy memory to register:
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4cb0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cb0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cb8] movem.w $3333.w, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cb8 ----------
Op4cb8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$3333.w' into r9:
  ldrsh r9,[r4],#2 ;@ Fetch Absolute Short address
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cb8
  tst r11,#1
  beq SkipReg4cb8

  ;@ Copy memory to register:
;@ EaRead : Read '$3333.w' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4cb8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cb8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cb9] movem.w $33333333.l, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cb9 ----------
Op4cb9
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$33333333.l' into r9:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r9,r0,r2,lsl #16
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cb9
  tst r11,#1
  beq SkipReg4cb9

  ;@ Copy memory to register:
;@ EaRead : Read '$33333333.l' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16
  mov r0,r0,asr #16

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#2 ;@ Post-increment address
  sub r5,r5,#4 ;@ Take some cycles

SkipReg4cb9
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cb9

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cd0] movem.l (a0), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cd0 ----------
Op4cd0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cd0
  tst r11,#1
  beq SkipReg4cd0

  ;@ Copy memory to register:
;@ EaRead : Read '(a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cd0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cd0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cd8] movem.l (a0)+, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cd8 ----------
Op4cd8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cd8
  tst r11,#1
  beq SkipReg4cd8

  ;@ Copy memory to register:
;@ EaRead : Read '(a0)+' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cd8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cd8

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cdf] movem.l (a7)+, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cdf ----------
Op4cdf
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cdf
  tst r11,#1
  beq SkipReg4cdf

  ;@ Copy memory to register:
;@ EaRead : Read '(a7)+' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cdf
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cdf

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#0 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ce0] movem.l -(a0), d2-d3/d6-d7/a2-a3/a6-a7 uses Op4ce0 ----------
Op4ce0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg4ce0
  tst r11,#1
  beq SkipReg4ce0

  sub r9,r9,#4 ;@ Pre-decrement address
  ;@ Copy memory to register:
;@ EaRead : Read '-(a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4ce0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg4ce0

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ce7] movem.l -(a7), d2-d3/d6-d7/a2-a3/a6-a7 uses Op4ce7 ----------
Op4ce7
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '(a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r9,[r7,r2]
;@ r10=Register Index*4:
  mov r10,#0x3c ;@ order reversed for -(An)

MoreReg4ce7
  tst r11,#1
  beq SkipReg4ce7

  sub r9,r9,#4 ;@ Pre-decrement address
  ;@ Copy memory to register:
;@ EaRead : Read '-(a7)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4ce7
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#-4 ;@ r10=Next Register
  bne MoreReg4ce7

;@ Write back address:
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaWrite: r9 into register[r0]:
  str r9,[r7,r0,lsl #2]

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ce8] movem.l ($3333,a0), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4ce8 ----------
Op4ce8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($3333,a0)' into r9:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r9,r0,r2 ;@ Add on offset
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4ce8
  tst r11,#1
  beq SkipReg4ce8

  ;@ Copy memory to register:
;@ EaRead : Read '($3333,a0)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4ce8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4ce8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cf0] movem.l ($33,a0,d3.w*2), d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cf0 ----------
Op4cf0
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r9:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r9,r2,r3 ;@ r9=Disp+An+Rn
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cf0
  tst r11,#1
  beq SkipReg4cf0

  ;@ Copy memory to register:
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cf0
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cf0

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cf8] movem.l $3333.w, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cf8 ----------
Op4cf8
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$3333.w' into r9:
  ldrsh r9,[r4],#2 ;@ Fetch Absolute Short address
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cf8
  tst r11,#1
  beq SkipReg4cf8

  ;@ Copy memory to register:
;@ EaRead : Read '$3333.w' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cf8
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cf8

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4cf9] movem.l $33333333.l, d0-d1/d4-d5/a0-a1/a4-a5 uses Op4cf9 ----------
Op4cf9
  stmdb sp!,{r9} ;@ Push r9
  ldrh r11,[r4],#2 ;@ r11=register mask

;@ Get the address into r9:
;@ EaCalc : Get '$33333333.l' into r9:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r9,r0,r2,lsl #16
;@ r10=Register Index*4:
  mov r10,#0

MoreReg4cf9
  tst r11,#1
  beq SkipReg4cf9

  ;@ Copy memory to register:
;@ EaRead : Read '$33333333.l' (address in r9) into r0:
  mov r0,r9
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r0,[r7,r10] ;@ Save value into Dn/An
  add r9,r9,#4 ;@ Post-increment address
  sub r5,r5,#8 ;@ Take some cycles

SkipReg4cf9
  movs r11,r11,lsr #1;@ Shift mask:
  add r10,r10,#4 ;@ r10=Next Register
  bne MoreReg4cf9

  ldmia sp!,{r9} ;@ Pop r9

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e40] trap #0 uses Op4e40 ----------
Op4e40
  and r0,r8,#0xf ;@ Get trap number
  orr r0,r0,#0x20
  mov r0,r0,asl #2
  str r0,[r7,#0x50] ;@ Save Exception Vector
  bl Exception

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#38 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e50] link a0,#$3333 uses Op4e50 ----------
Op4e50
;@ Get An
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  orr r10,r10,#0x8 ;@ A0-7
;@ EaRead : Read register[r10] into r1:
  ldr r1,[r7,r10,lsl #2]

  ldr r0,[r7,#0x3c] ;@ Get A7
  sub r0,r0,#4 ;@ A7-=4
  mov r11,r0

;@ Write An to Stack
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
;@ Save to An
;@ EaWrite: r11 into register[r10]:
  str r11,[r7,r10,lsl #2]

;@ Get offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

  add r11,r11,r0 ;@ Add offset to A7
  str r11,[r7,#0x3c]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e58] unlk a0 uses Op4e58 ----------
Op4e58
;@ Get An
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  orr r10,r10,#0x8 ;@ A0-7
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  add r11,r0,#4 ;@ A7+=4

;@ Pop An from stack:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  str r11,[r7,#0x3c] ;@ Save A7

;@ An = value from stack:
;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e60] move a0, usp uses Op4e60 ----------
Op4e60
  ldrb r0,[r7,#0x44] ;@ Get SR
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode4e60 ;@ No

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r0:
  ldr r0,[r7,r0,lsl #2]

  str r0,[r7,#0x48] ;@ Put in USP

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode4e60
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e68] move usp, a0 uses Op4e68 ----------
Op4e68
  ldrb r0,[r7,#0x44] ;@ Get SR
  tst r0,#0x20 ;@ Check we are in supervisor mode
  beq WrongMode4e68 ;@ No

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  orr r0,r0,#0x8 ;@ A0-7
  ldr r1,[r7,#0x48] ;@ Get from USP

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

WrongMode4e68
;@ todo - cause an exception
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e71] nop uses Op4e71 ----------
Op4e71
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e73] rte uses Op4e73 ----------
Op4e73
;@ Pop SR:
  ldr r0,[r7,#0x3c]
  add r1,r0,#2 ;@ Postincrement A7
  str r1,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...
  mov r0,r0,ror #8
  strb r0,[r7,#0x44] ;@ Store SR high

  ldr r10,[r7,#0x60] ;@ Get Memory base
;@ Pop PC:
  ldr r0,[r7,#0x3c]
  add r1,r0,#4 ;@ Postincrement A7
  str r1,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  add r4,r0,r10 ;@ r4=Memory Base+PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ CheckInterrupt:
  ldrb r0,[r7,#0x47] ;@ Get IRQ level
  tst r0,r0
  blne DoInterrupt

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e75] rts uses Op4e75 ----------
Op4e75
  ldr r10,[r7,#0x60] ;@ Get Memory base
;@ Pop PC:
  ldr r0,[r7,#0x3c]
  add r1,r0,#4 ;@ Postincrement A7
  str r1,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  add r4,r0,r10 ;@ r4=Memory Base+PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e77] rtr uses Op4e77 ----------
Op4e77
;@ Pop SR:
  ldr r0,[r7,#0x3c]
  add r1,r0,#2 ;@ Postincrement A7
  str r1,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  eor r1,r0,r0,ror #1 ;@ Bit 0=C^V
  mov r2,r0,lsr #3    ;@ r2=___XN
  tst r1,#1           ;@ 1 if C!=V
  eorne r0,r0,#3      ;@ ___XNZCV
  strb r2,[r7,#0x45]  ;@ Store X bit
  mov r9,r0,lsl #28   ;@ r9=NZCV...

  ldr r10,[r7,#0x60] ;@ Get Memory base
;@ Pop PC:
  ldr r0,[r7,#0x3c]
  add r1,r0,#4 ;@ Postincrement A7
  str r1,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  add r4,r0,r10 ;@ r4=Memory Base+PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4e90] jsr (a0) uses Op4e90 ----------
Op4e90
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ea8] jsr ($3333,a0) uses Op4ea8 ----------
Op4ea8
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4eb0] jsr ($33,a0,d3.w*2) uses Op4eb0 ----------
Op4eb0
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4eb8] jsr $3333.w uses Op4eb8 ----------
Op4eb8
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4eb9] jsr $33333333.l uses Op4eb9 ----------
Op4eb9
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#34 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4eba] jsr ($3333,pc); =3335 uses Op4eba ----------
Op4eba
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#30 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ebb] jsr ($33,pc,d3.w*2); =35 uses Op4ebb ----------
Op4ebb
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ Jump - Get new PC from r0
;@ Jsr - Push old PC first
  sub r1,r4,r10 ;@ r1 = Old PC
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#32 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ed0] jmp (a0) uses Op4ed0 ----------
Op4ed0
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ee8] jmp ($3333,a0) uses Op4ee8 ----------
Op4ee8
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ef0] jmp ($33,a0,d3.w*2) uses Op4ef0 ----------
Op4ef0
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ef8] jmp $3333.w uses Op4ef8 ----------
Op4ef8
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4ef9] jmp $33333333.l uses Op4ef9 ----------
Op4ef9
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4efa] jmp ($3333,pc); =3335 uses Op4efa ----------
Op4efa
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [4efb] jmp ($33,pc,d3.w*2); =35 uses Op4efb ----------
Op4efb
  ldr r10,[r7,#0x60] ;@ Get Memory base

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ Jump - Get new PC from r0
  add r4,r0,r10 ;@ r4 = Memory Base + New PC

;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5000] addq.b #8, d0 uses Op5000 ----------
Op5000
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #24
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5008] addq.b #8, a0 uses Op5008 ----------
Op5008
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  adds r1,r0,#0x0008

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5010] addq.b #8, (a0) uses Op5010 ----------
Op5010
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5018] addq.b #8, (a0)+ uses Op5018 ----------
Op5018
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5020] addq.b #8, -(a0) uses Op5020 ----------
Op5020
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5028] addq.b #8, ($3333,a0) uses Op5028 ----------
Op5028
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5030] addq.b #8, ($33,a0,d3.w*2) uses Op5030 ----------
Op5030
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5038] addq.b #8, $3333.w uses Op5038 ----------
Op5038
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5039] addq.b #8, $33333333.l uses Op5039 ----------
Op5039
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [503a] addq.b #8, ($3333,pc); =3335 uses Op503a ----------
Op503a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [503b] addq.b #8, ($33,pc,d3.w*2); =35 uses Op503b ----------
Op503b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  adds r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5040] addq.w #8, d0 uses Op5040 ----------
Op5040
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #16
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5050] addq.w #8, (a0) uses Op5050 ----------
Op5050
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5058] addq.w #8, (a0)+ uses Op5058 ----------
Op5058
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5060] addq.w #8, -(a0) uses Op5060 ----------
Op5060
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5068] addq.w #8, ($3333,a0) uses Op5068 ----------
Op5068
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5070] addq.w #8, ($33,a0,d3.w*2) uses Op5070 ----------
Op5070
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5078] addq.w #8, $3333.w uses Op5078 ----------
Op5078
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5079] addq.w #8, $33333333.l uses Op5079 ----------
Op5079
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [507a] addq.w #8, ($3333,pc); =3335 uses Op507a ----------
Op507a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [507b] addq.w #8, ($33,pc,d3.w*2); =35 uses Op507b ----------
Op507b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  adds r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5080] addq.l #8, d0 uses Op5080 ----------
Op5080
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5090] addq.l #8, (a0) uses Op5090 ----------
Op5090
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5098] addq.l #8, (a0)+ uses Op5098 ----------
Op5098
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50a0] addq.l #8, -(a0) uses Op50a0 ----------
Op50a0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50a8] addq.l #8, ($3333,a0) uses Op50a8 ----------
Op50a8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50b0] addq.l #8, ($33,a0,d3.w*2) uses Op50b0 ----------
Op50b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50b8] addq.l #8, $3333.w uses Op50b8 ----------
Op50b8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50b9] addq.l #8, $33333333.l uses Op50b9 ----------
Op50b9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50ba] addq.l #8, ($3333,pc); =3335 uses Op50ba ----------
Op50ba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50bb] addq.l #8, ($33,pc,d3.w*2); =35 uses Op50bb ----------
Op50bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  adds r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50c0] st d0 uses Op50c0 ----------
Op50c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1
  subal r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50c8] dbt d0, 3335 uses Op50c8 ----------
Op50c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd50c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd50c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50d0] st (a0) uses Op50d0 ----------
Op50d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50d8] st (a0)+ uses Op50d8 ----------
Op50d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50df] st (a7)+ uses Op50df ----------
Op50df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50e0] st -(a0) uses Op50e0 ----------
Op50e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50e7] st -(a7) uses Op50e7 ----------
Op50e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50e8] st ($3333,a0) uses Op50e8 ----------
Op50e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50f0] st ($33,a0,d3.w*2) uses Op50f0 ----------
Op50f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50f8] st $3333.w uses Op50f8 ----------
Op50f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50f9] st $33333333.l uses Op50f9 ----------
Op50f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50fa] st ($3333,pc); =3335 uses Op50fa ----------
Op50fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [50fb] st ($33,pc,d3.w*2); =35 uses Op50fb ----------
Op50fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnal r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5100] subq.b #8, d0 uses Op5100 ----------
Op5100
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #24
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5108] subq.b #8, a0 uses Op5108 ----------
Op5108
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  subs r1,r0,#0x0008

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5110] subq.b #8, (a0) uses Op5110 ----------
Op5110
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5118] subq.b #8, (a0)+ uses Op5118 ----------
Op5118
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5120] subq.b #8, -(a0) uses Op5120 ----------
Op5120
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5128] subq.b #8, ($3333,a0) uses Op5128 ----------
Op5128
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5130] subq.b #8, ($33,a0,d3.w*2) uses Op5130 ----------
Op5130
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5138] subq.b #8, $3333.w uses Op5138 ----------
Op5138
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5139] subq.b #8, $33333333.l uses Op5139 ----------
Op5139
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [513a] subq.b #8, ($3333,pc); =3335 uses Op513a ----------
Op513a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [513b] subq.b #8, ($33,pc,d3.w*2); =35 uses Op513b ----------
Op513b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  subs r1,r0,#0x8000000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5140] subq.w #8, d0 uses Op5140 ----------
Op5140
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #16
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5150] subq.w #8, (a0) uses Op5150 ----------
Op5150
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5158] subq.w #8, (a0)+ uses Op5158 ----------
Op5158
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5160] subq.w #8, -(a0) uses Op5160 ----------
Op5160
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5168] subq.w #8, ($3333,a0) uses Op5168 ----------
Op5168
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5170] subq.w #8, ($33,a0,d3.w*2) uses Op5170 ----------
Op5170
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5178] subq.w #8, $3333.w uses Op5178 ----------
Op5178
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5179] subq.w #8, $33333333.l uses Op5179 ----------
Op5179
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [517a] subq.w #8, ($3333,pc); =3335 uses Op517a ----------
Op517a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [517b] subq.w #8, ($33,pc,d3.w*2); =35 uses Op517b ----------
Op517b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  subs r1,r0,#0x80000
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5180] subq.l #8, d0 uses Op5180 ----------
Op5180
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5190] subq.l #8, (a0) uses Op5190 ----------
Op5190
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5198] subq.l #8, (a0)+ uses Op5198 ----------
Op5198
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51a0] subq.l #8, -(a0) uses Op51a0 ----------
Op51a0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51a8] subq.l #8, ($3333,a0) uses Op51a8 ----------
Op51a8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51b0] subq.l #8, ($33,a0,d3.w*2) uses Op51b0 ----------
Op51b0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51b8] subq.l #8, $3333.w uses Op51b8 ----------
Op51b8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51b9] subq.l #8, $33333333.l uses Op51b9 ----------
Op51b9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51ba] subq.l #8, ($3333,pc); =3335 uses Op51ba ----------
Op51ba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51bb] subq.l #8, ($33,pc,d3.w*2); =35 uses Op51bb ----------
Op51bb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  subs r1,r0,#0x0008
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51c0] sf d0 uses Op51c0 ----------
Op51c0
  mov r1,#0
  sub r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51c8] dbra d0, 3335 uses Op51c8 ----------
Op51c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd51c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd51c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51d0] sf (a0) uses Op51d0 ----------
Op51d0
  mov r1,#0

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51d8] sf (a0)+ uses Op51d8 ----------
Op51d8
  mov r1,#0

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51df] sf (a7)+ uses Op51df ----------
Op51df
  mov r1,#0

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51e0] sf -(a0) uses Op51e0 ----------
Op51e0
  mov r1,#0

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51e7] sf -(a7) uses Op51e7 ----------
Op51e7
  mov r1,#0

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51e8] sf ($3333,a0) uses Op51e8 ----------
Op51e8
  mov r1,#0

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51f0] sf ($33,a0,d3.w*2) uses Op51f0 ----------
Op51f0
  mov r1,#0

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51f8] sf $3333.w uses Op51f8 ----------
Op51f8
  mov r1,#0

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51f9] sf $33333333.l uses Op51f9 ----------
Op51f9
  mov r1,#0

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51fa] sf ($3333,pc); =3335 uses Op51fa ----------
Op51fa
  mov r1,#0

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [51fb] sf ($33,pc,d3.w*2); =35 uses Op51fb ----------
Op51fb
  mov r1,#0

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52c0] shi d0 uses Op52c0 ----------
Op52c0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1
  subhi r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52c8] dbhi d0, 3335 uses Op52c8 ----------
Op52c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
;@ If so, don't dbra
  bhi DbraEnd52c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd52c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd52c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52d0] shi (a0) uses Op52d0 ----------
Op52d0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52d8] shi (a0)+ uses Op52d8 ----------
Op52d8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52df] shi (a7)+ uses Op52df ----------
Op52df
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52e0] shi -(a0) uses Op52e0 ----------
Op52e0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52e7] shi -(a7) uses Op52e7 ----------
Op52e7
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52e8] shi ($3333,a0) uses Op52e8 ----------
Op52e8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52f0] shi ($33,a0,d3.w*2) uses Op52f0 ----------
Op52f0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52f8] shi $3333.w uses Op52f8 ----------
Op52f8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52f9] shi $33333333.l uses Op52f9 ----------
Op52f9
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52fa] shi ($3333,pc); =3335 uses Op52fa ----------
Op52fa
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [52fb] shi ($33,pc,d3.w*2); =35 uses Op52fb ----------
Op52fb
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnhi r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53c0] sls d0 uses Op53c0 ----------
Op53c0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1
  subls r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53c8] dbls d0, 3335 uses Op53c8 ----------
Op53c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
;@ If so, don't dbra
  bls DbraEnd53c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd53c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd53c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53d0] sls (a0) uses Op53d0 ----------
Op53d0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53d8] sls (a0)+ uses Op53d8 ----------
Op53d8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53df] sls (a7)+ uses Op53df ----------
Op53df
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53e0] sls -(a0) uses Op53e0 ----------
Op53e0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53e7] sls -(a7) uses Op53e7 ----------
Op53e7
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53e8] sls ($3333,a0) uses Op53e8 ----------
Op53e8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53f0] sls ($33,a0,d3.w*2) uses Op53f0 ----------
Op53f0
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53f8] sls $3333.w uses Op53f8 ----------
Op53f8
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53f9] sls $33333333.l uses Op53f9 ----------
Op53f9
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53fa] sls ($3333,pc); =3335 uses Op53fa ----------
Op53fa
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [53fb] sls ($33,pc,d3.w*2); =35 uses Op53fb ----------
Op53fb
  mov r1,#0
;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  mvnls r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54c0] scc d0 uses Op54c0 ----------
Op54c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1
  subcc r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54c8] dbcc d0, 3335 uses Op54c8 ----------
Op54c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bcc DbraEnd54c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd54c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd54c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54d0] scc (a0) uses Op54d0 ----------
Op54d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54d8] scc (a0)+ uses Op54d8 ----------
Op54d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54df] scc (a7)+ uses Op54df ----------
Op54df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54e0] scc -(a0) uses Op54e0 ----------
Op54e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54e7] scc -(a7) uses Op54e7 ----------
Op54e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54e8] scc ($3333,a0) uses Op54e8 ----------
Op54e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54f0] scc ($33,a0,d3.w*2) uses Op54f0 ----------
Op54f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54f8] scc $3333.w uses Op54f8 ----------
Op54f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54f9] scc $33333333.l uses Op54f9 ----------
Op54f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54fa] scc ($3333,pc); =3335 uses Op54fa ----------
Op54fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [54fb] scc ($33,pc,d3.w*2); =35 uses Op54fb ----------
Op54fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncc r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55c0] scs d0 uses Op55c0 ----------
Op55c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1
  subcs r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55c8] dbcs d0, 3335 uses Op55c8 ----------
Op55c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bcs DbraEnd55c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd55c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd55c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55d0] scs (a0) uses Op55d0 ----------
Op55d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55d8] scs (a0)+ uses Op55d8 ----------
Op55d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55df] scs (a7)+ uses Op55df ----------
Op55df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55e0] scs -(a0) uses Op55e0 ----------
Op55e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55e7] scs -(a7) uses Op55e7 ----------
Op55e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55e8] scs ($3333,a0) uses Op55e8 ----------
Op55e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55f0] scs ($33,a0,d3.w*2) uses Op55f0 ----------
Op55f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55f8] scs $3333.w uses Op55f8 ----------
Op55f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55f9] scs $33333333.l uses Op55f9 ----------
Op55f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55fa] scs ($3333,pc); =3335 uses Op55fa ----------
Op55fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [55fb] scs ($33,pc,d3.w*2); =35 uses Op55fb ----------
Op55fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvncs r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56c0] sne d0 uses Op56c0 ----------
Op56c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1
  subne r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56c8] dbne d0, 3335 uses Op56c8 ----------
Op56c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bne DbraEnd56c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd56c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd56c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56d0] sne (a0) uses Op56d0 ----------
Op56d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56d8] sne (a0)+ uses Op56d8 ----------
Op56d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56df] sne (a7)+ uses Op56df ----------
Op56df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56e0] sne -(a0) uses Op56e0 ----------
Op56e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56e7] sne -(a7) uses Op56e7 ----------
Op56e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56e8] sne ($3333,a0) uses Op56e8 ----------
Op56e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56f0] sne ($33,a0,d3.w*2) uses Op56f0 ----------
Op56f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56f8] sne $3333.w uses Op56f8 ----------
Op56f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56f9] sne $33333333.l uses Op56f9 ----------
Op56f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56fa] sne ($3333,pc); =3335 uses Op56fa ----------
Op56fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [56fb] sne ($33,pc,d3.w*2); =35 uses Op56fb ----------
Op56fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnne r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57c0] seq d0 uses Op57c0 ----------
Op57c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1
  subeq r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57c8] dbeq d0, 3335 uses Op57c8 ----------
Op57c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  beq DbraEnd57c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd57c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd57c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57d0] seq (a0) uses Op57d0 ----------
Op57d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57d8] seq (a0)+ uses Op57d8 ----------
Op57d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57df] seq (a7)+ uses Op57df ----------
Op57df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57e0] seq -(a0) uses Op57e0 ----------
Op57e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57e7] seq -(a7) uses Op57e7 ----------
Op57e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57e8] seq ($3333,a0) uses Op57e8 ----------
Op57e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57f0] seq ($33,a0,d3.w*2) uses Op57f0 ----------
Op57f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57f8] seq $3333.w uses Op57f8 ----------
Op57f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57f9] seq $33333333.l uses Op57f9 ----------
Op57f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57fa] seq ($3333,pc); =3335 uses Op57fa ----------
Op57fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [57fb] seq ($33,pc,d3.w*2); =35 uses Op57fb ----------
Op57fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvneq r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58c0] svc d0 uses Op58c0 ----------
Op58c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1
  subvc r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58c8] dbvc d0, 3335 uses Op58c8 ----------
Op58c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bvc DbraEnd58c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd58c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd58c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58d0] svc (a0) uses Op58d0 ----------
Op58d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58d8] svc (a0)+ uses Op58d8 ----------
Op58d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58df] svc (a7)+ uses Op58df ----------
Op58df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58e0] svc -(a0) uses Op58e0 ----------
Op58e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58e7] svc -(a7) uses Op58e7 ----------
Op58e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58e8] svc ($3333,a0) uses Op58e8 ----------
Op58e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58f0] svc ($33,a0,d3.w*2) uses Op58f0 ----------
Op58f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58f8] svc $3333.w uses Op58f8 ----------
Op58f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58f9] svc $33333333.l uses Op58f9 ----------
Op58f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58fa] svc ($3333,pc); =3335 uses Op58fa ----------
Op58fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [58fb] svc ($33,pc,d3.w*2); =35 uses Op58fb ----------
Op58fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvc r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59c0] svs d0 uses Op59c0 ----------
Op59c0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1
  subvs r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59c8] dbvs d0, 3335 uses Op59c8 ----------
Op59c8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bvs DbraEnd59c8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd59c8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd59c8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59d0] svs (a0) uses Op59d0 ----------
Op59d0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59d8] svs (a0)+ uses Op59d8 ----------
Op59d8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59df] svs (a7)+ uses Op59df ----------
Op59df
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59e0] svs -(a0) uses Op59e0 ----------
Op59e0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59e7] svs -(a7) uses Op59e7 ----------
Op59e7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59e8] svs ($3333,a0) uses Op59e8 ----------
Op59e8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59f0] svs ($33,a0,d3.w*2) uses Op59f0 ----------
Op59f0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59f8] svs $3333.w uses Op59f8 ----------
Op59f8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59f9] svs $33333333.l uses Op59f9 ----------
Op59f9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59fa] svs ($3333,pc); =3335 uses Op59fa ----------
Op59fa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [59fb] svs ($33,pc,d3.w*2); =35 uses Op59fb ----------
Op59fb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnvs r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ac0] spl d0 uses Op5ac0 ----------
Op5ac0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1
  subpl r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ac8] dbpl d0, 3335 uses Op5ac8 ----------
Op5ac8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bpl DbraEnd5ac8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5ac8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5ac8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ad0] spl (a0) uses Op5ad0 ----------
Op5ad0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ad8] spl (a0)+ uses Op5ad8 ----------
Op5ad8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5adf] spl (a7)+ uses Op5adf ----------
Op5adf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ae0] spl -(a0) uses Op5ae0 ----------
Op5ae0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ae7] spl -(a7) uses Op5ae7 ----------
Op5ae7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ae8] spl ($3333,a0) uses Op5ae8 ----------
Op5ae8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5af0] spl ($33,a0,d3.w*2) uses Op5af0 ----------
Op5af0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5af8] spl $3333.w uses Op5af8 ----------
Op5af8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5af9] spl $33333333.l uses Op5af9 ----------
Op5af9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5afa] spl ($3333,pc); =3335 uses Op5afa ----------
Op5afa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5afb] spl ($33,pc,d3.w*2); =35 uses Op5afb ----------
Op5afb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnpl r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bc0] smi d0 uses Op5bc0 ----------
Op5bc0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1
  submi r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bc8] dbmi d0, 3335 uses Op5bc8 ----------
Op5bc8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bmi DbraEnd5bc8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5bc8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5bc8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bd0] smi (a0) uses Op5bd0 ----------
Op5bd0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bd8] smi (a0)+ uses Op5bd8 ----------
Op5bd8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bdf] smi (a7)+ uses Op5bdf ----------
Op5bdf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5be0] smi -(a0) uses Op5be0 ----------
Op5be0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5be7] smi -(a7) uses Op5be7 ----------
Op5be7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5be8] smi ($3333,a0) uses Op5be8 ----------
Op5be8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bf0] smi ($33,a0,d3.w*2) uses Op5bf0 ----------
Op5bf0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bf8] smi $3333.w uses Op5bf8 ----------
Op5bf8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bf9] smi $33333333.l uses Op5bf9 ----------
Op5bf9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bfa] smi ($3333,pc); =3335 uses Op5bfa ----------
Op5bfa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5bfb] smi ($33,pc,d3.w*2); =35 uses Op5bfb ----------
Op5bfb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnmi r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cc0] sge d0 uses Op5cc0 ----------
Op5cc0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1
  subge r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cc8] dbge d0, 3335 uses Op5cc8 ----------
Op5cc8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bge DbraEnd5cc8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5cc8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5cc8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cd0] sge (a0) uses Op5cd0 ----------
Op5cd0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cd8] sge (a0)+ uses Op5cd8 ----------
Op5cd8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cdf] sge (a7)+ uses Op5cdf ----------
Op5cdf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ce0] sge -(a0) uses Op5ce0 ----------
Op5ce0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ce7] sge -(a7) uses Op5ce7 ----------
Op5ce7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ce8] sge ($3333,a0) uses Op5ce8 ----------
Op5ce8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cf0] sge ($33,a0,d3.w*2) uses Op5cf0 ----------
Op5cf0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cf8] sge $3333.w uses Op5cf8 ----------
Op5cf8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cf9] sge $33333333.l uses Op5cf9 ----------
Op5cf9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cfa] sge ($3333,pc); =3335 uses Op5cfa ----------
Op5cfa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5cfb] sge ($33,pc,d3.w*2); =35 uses Op5cfb ----------
Op5cfb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnge r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dc0] slt d0 uses Op5dc0 ----------
Op5dc0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1
  sublt r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dc8] dblt d0, 3335 uses Op5dc8 ----------
Op5dc8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  blt DbraEnd5dc8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5dc8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5dc8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dd0] slt (a0) uses Op5dd0 ----------
Op5dd0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dd8] slt (a0)+ uses Op5dd8 ----------
Op5dd8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ddf] slt (a7)+ uses Op5ddf ----------
Op5ddf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5de0] slt -(a0) uses Op5de0 ----------
Op5de0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5de7] slt -(a7) uses Op5de7 ----------
Op5de7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5de8] slt ($3333,a0) uses Op5de8 ----------
Op5de8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5df0] slt ($33,a0,d3.w*2) uses Op5df0 ----------
Op5df0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5df8] slt $3333.w uses Op5df8 ----------
Op5df8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5df9] slt $33333333.l uses Op5df9 ----------
Op5df9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dfa] slt ($3333,pc); =3335 uses Op5dfa ----------
Op5dfa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5dfb] slt ($33,pc,d3.w*2); =35 uses Op5dfb ----------
Op5dfb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnlt r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e00] addq.b #7, d0 uses Op5e00 ----------
Op5e00
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #24
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e08] addq.b #7, a0 uses Op5e08 ----------
Op5e08
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e10] addq.b #7, (a0) uses Op5e10 ----------
Op5e10
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e18] addq.b #7, (a0)+ uses Op5e18 ----------
Op5e18
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e20] addq.b #7, -(a0) uses Op5e20 ----------
Op5e20
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e28] addq.b #7, ($3333,a0) uses Op5e28 ----------
Op5e28
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e30] addq.b #7, ($33,a0,d3.w*2) uses Op5e30 ----------
Op5e30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e38] addq.b #7, $3333.w uses Op5e38 ----------
Op5e38
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e39] addq.b #7, $33333333.l uses Op5e39 ----------
Op5e39
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e3a] addq.b #7, ($3333,pc); =3335 uses Op5e3a ----------
Op5e3a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e3b] addq.b #7, ($33,pc,d3.w*2); =35 uses Op5e3b ----------
Op5e3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e40] addq.w #7, d0 uses Op5e40 ----------
Op5e40
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #16
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e50] addq.w #7, (a0) uses Op5e50 ----------
Op5e50
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e58] addq.w #7, (a0)+ uses Op5e58 ----------
Op5e58
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e60] addq.w #7, -(a0) uses Op5e60 ----------
Op5e60
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e68] addq.w #7, ($3333,a0) uses Op5e68 ----------
Op5e68
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e70] addq.w #7, ($33,a0,d3.w*2) uses Op5e70 ----------
Op5e70
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e78] addq.w #7, $3333.w uses Op5e78 ----------
Op5e78
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e79] addq.w #7, $33333333.l uses Op5e79 ----------
Op5e79
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e7a] addq.w #7, ($3333,pc); =3335 uses Op5e7a ----------
Op5e7a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e7b] addq.w #7, ($33,pc,d3.w*2); =35 uses Op5e7b ----------
Op5e7b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e80] addq.l #7, d0 uses Op5e80 ----------
Op5e80
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e90] addq.l #7, (a0) uses Op5e90 ----------
Op5e90
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5e98] addq.l #7, (a0)+ uses Op5e98 ----------
Op5e98
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ea0] addq.l #7, -(a0) uses Op5ea0 ----------
Op5ea0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ea8] addq.l #7, ($3333,a0) uses Op5ea8 ----------
Op5ea8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5eb0] addq.l #7, ($33,a0,d3.w*2) uses Op5eb0 ----------
Op5eb0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5eb8] addq.l #7, $3333.w uses Op5eb8 ----------
Op5eb8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5eb9] addq.l #7, $33333333.l uses Op5eb9 ----------
Op5eb9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5eba] addq.l #7, ($3333,pc); =3335 uses Op5eba ----------
Op5eba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ebb] addq.l #7, ($33,pc,d3.w*2); =35 uses Op5ebb ----------
Op5ebb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  adds r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ec0] sgt d0 uses Op5ec0 ----------
Op5ec0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1
  subgt r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ec8] dbgt d0, 3335 uses Op5ec8 ----------
Op5ec8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  bgt DbraEnd5ec8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5ec8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5ec8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ed0] sgt (a0) uses Op5ed0 ----------
Op5ed0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ed8] sgt (a0)+ uses Op5ed8 ----------
Op5ed8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5edf] sgt (a7)+ uses Op5edf ----------
Op5edf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ee0] sgt -(a0) uses Op5ee0 ----------
Op5ee0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ee7] sgt -(a7) uses Op5ee7 ----------
Op5ee7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ee8] sgt ($3333,a0) uses Op5ee8 ----------
Op5ee8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ef0] sgt ($33,a0,d3.w*2) uses Op5ef0 ----------
Op5ef0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ef8] sgt $3333.w uses Op5ef8 ----------
Op5ef8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ef9] sgt $33333333.l uses Op5ef9 ----------
Op5ef9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5efa] sgt ($3333,pc); =3335 uses Op5efa ----------
Op5efa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5efb] sgt ($33,pc,d3.w*2); =35 uses Op5efb ----------
Op5efb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvngt r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f00] subq.b #7, d0 uses Op5f00 ----------
Op5f00
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #24
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f08] subq.b #7, a0 uses Op5f08 ----------
Op5f08
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f10] subq.b #7, (a0) uses Op5f10 ----------
Op5f10
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f18] subq.b #7, (a0)+ uses Op5f18 ----------
Op5f18
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f20] subq.b #7, -(a0) uses Op5f20 ----------
Op5f20
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f28] subq.b #7, ($3333,a0) uses Op5f28 ----------
Op5f28
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f30] subq.b #7, ($33,a0,d3.w*2) uses Op5f30 ----------
Op5f30
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f38] subq.b #7, $3333.w uses Op5f38 ----------
Op5f38
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f39] subq.b #7, $33333333.l uses Op5f39 ----------
Op5f39
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f3a] subq.b #7, ($3333,pc); =3335 uses Op5f3a ----------
Op5f3a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f3b] subq.b #7, ($33,pc,d3.w*2); =35 uses Op5f3b ----------
Op5f3b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

  mov r2,r8,lsl #15 ;@ Get quick value
  and r2,r2,#0x7000000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f40] subq.w #7, d0 uses Op5f40 ----------
Op5f40
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #16
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f50] subq.w #7, (a0) uses Op5f50 ----------
Op5f50
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f58] subq.w #7, (a0)+ uses Op5f58 ----------
Op5f58
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f60] subq.w #7, -(a0) uses Op5f60 ----------
Op5f60
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f68] subq.w #7, ($3333,a0) uses Op5f68 ----------
Op5f68
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f70] subq.w #7, ($33,a0,d3.w*2) uses Op5f70 ----------
Op5f70
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f78] subq.w #7, $3333.w uses Op5f78 ----------
Op5f78
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f79] subq.w #7, $33333333.l uses Op5f79 ----------
Op5f79
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f7a] subq.w #7, ($3333,pc); =3335 uses Op5f7a ----------
Op5f7a
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f7b] subq.w #7, ($33,pc,d3.w*2); =35 uses Op5f7b ----------
Op5f7b
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r2,r8,lsl #7 ;@ Get quick value
  and r2,r2,#0x70000

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f80] subq.l #7, d0 uses Op5f80 ----------
Op5f80
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f90] subq.l #7, (a0) uses Op5f90 ----------
Op5f90
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5f98] subq.l #7, (a0)+ uses Op5f98 ----------
Op5f98
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fa0] subq.l #7, -(a0) uses Op5fa0 ----------
Op5fa0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fa8] subq.l #7, ($3333,a0) uses Op5fa8 ----------
Op5fa8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fb0] subq.l #7, ($33,a0,d3.w*2) uses Op5fb0 ----------
Op5fb0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fb8] subq.l #7, $3333.w uses Op5fb8 ----------
Op5fb8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fb9] subq.l #7, $33333333.l uses Op5fb9 ----------
Op5fb9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fba] subq.l #7, ($3333,pc); =3335 uses Op5fba ----------
Op5fba
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fbb] subq.l #7, ($33,pc,d3.w*2); =35 uses Op5fbb ----------
Op5fbb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  mov r2,r8,lsr #9 ;@ Get quick value
  and r2,r2,#0x0007

  subs r1,r0,r2
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fc0] sle d0 uses Op5fc0 ----------
Op5fc0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1
  suble r5,r5,#2 ;@ Extra cycles

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaWrite: r1 into register[r0]:
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fc8] dble d0, 3335 uses Op5fc8 ----------
Op5fc8
;@ Decrement Dn.w
  and r1,r8,#0x0007
  mov r1,r1,lsl #2
  ldrsh r0,[r7,r1]
  sub r0,r0,#1
  strh r0,[r7,r1]

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
;@ If so, don't dbra
  ble DbraEnd5fc8

;@ Check if Dn.w is -1
  cmps r0,#-1
  beq DbraEnd5fc8

;@ Get Branch offset:
  ldrsh r0,[r4]
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DbraEnd5fc8
  add r4,r4,#2 ;@ Skip branch offset

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fd0] sle (a0) uses Op5fd0 ----------
Op5fd0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaWrite: Write r1 into '(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fd8] sle (a0)+ uses Op5fd8 ----------
Op5fd8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a0)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fdf] sle (a7)+ uses Op5fdf ----------
Op5fdf
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaWrite: Write r1 into '(a7)+' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fe0] sle -(a0) uses Op5fe0 ----------
Op5fe0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fe7] sle -(a7) uses Op5fe7 ----------
Op5fe7
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaWrite: Write r1 into '-(a7)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5fe8] sle ($3333,a0) uses Op5fe8 ----------
Op5fe8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaWrite: Write r1 into '($3333,a0)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ff0] sle ($33,a0,d3.w*2) uses Op5ff0 ----------
Op5ff0
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ff8] sle $3333.w uses Op5ff8 ----------
Op5ff8
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaWrite: Write r1 into '$3333.w' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ff9] sle $33333333.l uses Op5ff9 ----------
Op5ff9
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaWrite: Write r1 into '$33333333.l' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ffa] sle ($3333,pc); =3335 uses Op5ffa ----------
Op5ffa
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaWrite: Write r1 into '($3333,pc)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [5ffb] sle ($33,pc,d3.w*2); =35 uses Op5ffb ----------
Op5ffb
  mov r1,#0
;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mvnle r1,r1

;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r0):
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6000] bra 3335 uses Op6000 ----------
Op6000
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6001] bra 3 uses Op6001 ----------
Op6001
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

  mov r0,r0,asr #24 ;@ ...shift down

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [60ff] bra 33333335 uses Op60ff ----------
Op60ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6100] bsr 3335 uses Op6100 ----------
Op6100
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Branch taken - Add on r0 to PC
;@ Bsr - remember old PC
  ldr r10,[r7,#0x60] ;@ Get Memory base
  sub r1,r4,r10 ;@ r1 = Old PC

  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6101] bsr 3 uses Op6101 ----------
Op6101
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

  mov r0,r0,asr #24 ;@ ...shift down

;@ Branch taken - Add on r0 to PC
;@ Bsr - remember old PC
  ldr r10,[r7,#0x60] ;@ Get Memory base
  sub r1,r4,r10 ;@ r1 = Old PC

  add r4,r4,r0 ;@ r4 = New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [61ff] bsr 33333335 uses Op61ff ----------
Op61ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Branch taken - Add on r0 to PC
;@ Bsr - remember old PC
  ldr r10,[r7,#0x60] ;@ Get Memory base
  sub r1,r4,r10 ;@ r1 = Old PC

  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Push r1 onto stack
  ldr r0,[r7,#0x3c]
  sub r0,r0,#4 ;@ Predecrement A7
  str r0,[r7,#0x3c] ;@ Save A7
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6200] bhi 3335 uses Op6200 ----------
Op6200
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  bls DontBranch6200

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6200
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6201] bhi 3 uses Op6201 ----------
Op6201
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  mov r0,r0,asr #24 ;@ ...shift down

  bls DontBranch6201

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6201
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [62ff] bhi 33333335 uses Op62ff ----------
Op62ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  bls DontBranch62ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch62ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6300] bls 3335 uses Op6300 ----------
Op6300
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  bhi DontBranch6300

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6300
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6301] bls 3 uses Op6301 ----------
Op6301
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  mov r0,r0,asr #24 ;@ ...shift down

  bhi DontBranch6301

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6301
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [63ff] bls 33333335 uses Op63ff ----------
Op63ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  eor r9,r9,#0x20000000 ;@ Invert carry for hi/ls
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  eor r9,r9,#0x20000000
  bhi DontBranch63ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch63ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6400] bcc 3335 uses Op6400 ----------
Op6400
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bcs DontBranch6400

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6400
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6401] bcc 3 uses Op6401 ----------
Op6401
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bcs DontBranch6401

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6401
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [64ff] bcc 33333335 uses Op64ff ----------
Op64ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bcs DontBranch64ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch64ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6500] bcs 3335 uses Op6500 ----------
Op6500
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bcc DontBranch6500

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6500
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6501] bcs 3 uses Op6501 ----------
Op6501
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bcc DontBranch6501

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6501
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [65ff] bcs 33333335 uses Op65ff ----------
Op65ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bcc DontBranch65ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch65ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6600] bne 3335 uses Op6600 ----------
Op6600
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  beq DontBranch6600

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6600
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6601] bne 3 uses Op6601 ----------
Op6601
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  beq DontBranch6601

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6601
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [66ff] bne 33333335 uses Op66ff ----------
Op66ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  beq DontBranch66ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch66ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6700] beq 3335 uses Op6700 ----------
Op6700
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bne DontBranch6700

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6700
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6701] beq 3 uses Op6701 ----------
Op6701
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bne DontBranch6701

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6701
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [67ff] beq 33333335 uses Op67ff ----------
Op67ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bne DontBranch67ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch67ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6800] bvc 3335 uses Op6800 ----------
Op6800
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bvs DontBranch6800

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6800
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6801] bvc 3 uses Op6801 ----------
Op6801
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bvs DontBranch6801

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6801
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [68ff] bvc 33333335 uses Op68ff ----------
Op68ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bvs DontBranch68ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch68ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6900] bvs 3335 uses Op6900 ----------
Op6900
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bvc DontBranch6900

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6900
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6901] bvs 3 uses Op6901 ----------
Op6901
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bvc DontBranch6901

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6901
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [69ff] bvs 33333335 uses Op69ff ----------
Op69ff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bvc DontBranch69ff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch69ff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6a00] bpl 3335 uses Op6a00 ----------
Op6a00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bmi DontBranch6a00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6a00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6a01] bpl 3 uses Op6a01 ----------
Op6a01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bmi DontBranch6a01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6a01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6aff] bpl 33333335 uses Op6aff ----------
Op6aff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bmi DontBranch6aff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6aff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6b00] bmi 3335 uses Op6b00 ----------
Op6b00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bpl DontBranch6b00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6b00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6b01] bmi 3 uses Op6b01 ----------
Op6b01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bpl DontBranch6b01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6b01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6bff] bmi 33333335 uses Op6bff ----------
Op6bff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bpl DontBranch6bff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6bff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6c00] bge 3335 uses Op6c00 ----------
Op6c00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  blt DontBranch6c00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6c00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6c01] bge 3 uses Op6c01 ----------
Op6c01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  blt DontBranch6c01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6c01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6cff] bge 33333335 uses Op6cff ----------
Op6cff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  blt DontBranch6cff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6cff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6d00] blt 3335 uses Op6d00 ----------
Op6d00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bge DontBranch6d00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6d00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6d01] blt 3 uses Op6d01 ----------
Op6d01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bge DontBranch6d01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6d01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6dff] blt 33333335 uses Op6dff ----------
Op6dff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bge DontBranch6dff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6dff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6e00] bgt 3335 uses Op6e00 ----------
Op6e00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  ble DontBranch6e00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6e00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6e01] bgt 3 uses Op6e01 ----------
Op6e01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  ble DontBranch6e01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6e01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6eff] bgt 33333335 uses Op6eff ----------
Op6eff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  ble DontBranch6eff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6eff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6f00] ble 3335 uses Op6f00 ----------
Op6f00
;@ Get Branch offset:
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bgt DontBranch6f00

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#2 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6f00
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6f01] ble 3 uses Op6f01 ----------
Op6f01
;@ Get Branch offset:
  mov r0,r8,asl #24 ;@ Shift 8-bit signed offset up...

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  mov r0,r0,asr #24 ;@ ...shift down

  bgt DontBranch6f01

;@ Branch taken - Add on r0 to PC
  add r4,r4,r0 ;@ r4 = New PC

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6f01
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [6fff] ble 33333335 uses Op6fff ----------
Op6fff
;@ Get Branch offset:
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r0:

;@ Is the condition true?
  msr cpsr_flg,r9 ;@ ARM flags = 68000 flags
  bgt DontBranch6fff

;@ Branch taken - Add on r0 to PC
  sub r4,r4,#4 ;@ (Branch is relative to Opcode+2)
  add r4,r4,r0 ;@ r4 = New PC

;@ Branch is quite far, so may be a good idea to check Memory Base+pc
;@ Check Memory Base+pc (r4)
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  add lr,pc,#4
  mov r0,r4
  ldr pc,[r7,#0x64] ;@ Call checkpc()
  mov r4,r0
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

DontBranch6fff
  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [7000] moveq #$0, d0 uses Op7000 ----------
Op7000
  movs r0,r8,asl #24
  and r1,r8,#0x0e00
  mov r0,r0,asr #24 ;@ Sign extended Quick value
  mrs r9,cpsr ;@ r9=NZ flags
  str r0,[r7,r1,lsr #7] ;@ Store into Dn

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8000] or.b d0, d0 uses Op8000 ----------
Op8000
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsb r11,[r7,r10]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8010] or.b (a0), d0 uses Op8010 ----------
Op8010
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8018] or.b (a0)+, d0 uses Op8018 ----------
Op8018
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [801f] or.b (a7)+, d0 uses Op801f ----------
Op801f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8020] or.b -(a0), d0 uses Op8020 ----------
Op8020
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8027] or.b -(a7), d0 uses Op8027 ----------
Op8027
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8028] or.b ($3333,a0), d0 uses Op8028 ----------
Op8028
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8030] or.b ($33,a0,d3.w*2), d0 uses Op8030 ----------
Op8030
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8038] or.b $3333.w, d0 uses Op8038 ----------
Op8038
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8039] or.b $33333333.l, d0 uses Op8039 ----------
Op8039
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803a] or.b ($3333,pc), d0; =3335 uses Op803a ----------
Op803a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803b] or.b ($33,pc,d3.w*2), d0; =35 uses Op803b ----------
Op803b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803c] or.b #$33, d0 uses Op803c ----------
Op803c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r11:
  mov r11,r10,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803d] or.b ea=(111 101), d0 uses Op803d ----------
Op803d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803e] or.b ea=(111 110), d0 uses Op803e ----------
Op803e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [803f] or.b ea=(111 111), d0 uses Op803f ----------
Op803f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8040] or.w d0, d0 uses Op8040 ----------
Op8040
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsh r11,[r7,r10]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8050] or.w (a0), d0 uses Op8050 ----------
Op8050
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8058] or.w (a0)+, d0 uses Op8058 ----------
Op8058
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [805f] or.w (a7)+, d0 uses Op805f ----------
Op805f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8060] or.w -(a0), d0 uses Op8060 ----------
Op8060
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8067] or.w -(a7), d0 uses Op8067 ----------
Op8067
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8068] or.w ($3333,a0), d0 uses Op8068 ----------
Op8068
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8070] or.w ($33,a0,d3.w*2), d0 uses Op8070 ----------
Op8070
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8078] or.w $3333.w, d0 uses Op8078 ----------
Op8078
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8079] or.w $33333333.l, d0 uses Op8079 ----------
Op8079
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807a] or.w ($3333,pc), d0; =3335 uses Op807a ----------
Op807a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807b] or.w ($33,pc,d3.w*2), d0; =35 uses Op807b ----------
Op807b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807c] or.w #$3333, d0 uses Op807c ----------
Op807c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r11:
  mov r11,r10,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807d] or.w ea=(111 101), d0 uses Op807d ----------
Op807d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807e] or.w ea=(111 110), d0 uses Op807e ----------
Op807e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [807f] or.w ea=(111 111), d0 uses Op807f ----------
Op807f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8080] or.l d0, d0 uses Op8080 ----------
Op8080
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r11:
  ldr r11,[r7,r10,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8090] or.l (a0), d0 uses Op8090 ----------
Op8090
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8098] or.l (a0)+, d0 uses Op8098 ----------
Op8098
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [809f] or.l (a7)+, d0 uses Op809f ----------
Op809f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80a0] or.l -(a0), d0 uses Op80a0 ----------
Op80a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80a7] or.l -(a7), d0 uses Op80a7 ----------
Op80a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80a8] or.l ($3333,a0), d0 uses Op80a8 ----------
Op80a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80b0] or.l ($33,a0,d3.w*2), d0 uses Op80b0 ----------
Op80b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80b8] or.l $3333.w, d0 uses Op80b8 ----------
Op80b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80b9] or.l $33333333.l, d0 uses Op80b9 ----------
Op80b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80ba] or.l ($3333,pc), d0; =3335 uses Op80ba ----------
Op80ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80bb] or.l ($33,pc,d3.w*2), d0; =35 uses Op80bb ----------
Op80bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80bc] or.l #$33333333, d0 uses Op80bc ----------
Op80bc
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r11:
  mov r11,r10,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80bd] or.l ea=(111 101), d0 uses Op80bd ----------
Op80bd
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80be] or.l ea=(111 110), d0 uses Op80be ----------
Op80be
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80bf] or.l ea=(111 111), d0 uses Op80bf ----------
Op80bf
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80c0] divu.w d0, d0 uses Op80c0 ----------
Op80c0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r10:
  ldrsh r10,[r7,r10]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80c0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80c0

Divide80c0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80c0

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#133 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80d0] divu.w (a0), d0 uses Op80d0 ----------
Op80d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80d0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80d0

Divide80d0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80d0

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80d8] divu.w (a0)+, d0 uses Op80d8 ----------
Op80d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80d8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80d8

Divide80d8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80d8

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80df] divu.w (a7)+, d0 uses Op80df ----------
Op80df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80df
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80df

Divide80df
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80df

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80e0] divu.w -(a0), d0 uses Op80e0 ----------
Op80e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80e0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80e0

Divide80e0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80e0

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#139 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80e7] divu.w -(a7), d0 uses Op80e7 ----------
Op80e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80e7
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80e7

Divide80e7
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80e7

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#139 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80e8] divu.w ($3333,a0), d0 uses Op80e8 ----------
Op80e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80e8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80e8

Divide80e8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80e8

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80f0] divu.w ($33,a0,d3.w*2), d0 uses Op80f0 ----------
Op80f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80f0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80f0

Divide80f0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80f0

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#143 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80f8] divu.w $3333.w, d0 uses Op80f8 ----------
Op80f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80f8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80f8

Divide80f8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80f8

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80f9] divu.w $33333333.l, d0 uses Op80f9 ----------
Op80f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80f9
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80f9

Divide80f9
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80f9

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#145 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80fa] divu.w ($3333,pc), d0; =3335 uses Op80fa ----------
Op80fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80fa
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80fa

Divide80fa
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80fa

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80fb] divu.w ($33,pc,d3.w*2), d0; =35 uses Op80fb ----------
Op80fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80fb
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80fb

Divide80fb
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80fb

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#143 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [80fc] divu.w #$3333, d0 uses Op80fc ----------
Op80fc
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift80fc
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift80fc

Divide80fc
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide80fc

  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8100]  uses Op8100 ----------
Op8100
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r10:
  ldrsb r10,[r7,r0]
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x0e00
  mov r11,r11,lsr #7
;@ EaRead : Read register[r11] into r1:
  ldrsb r1,[r7,r11]
  mov r1,r1,asl #24

  ldrb r2,[r7,#0x45] ;@ Get X bit
  tst r2,#2
  addne r10,r10,#0x01000000 ;@ Add carry bit
;@ Sub units into r2:
  and r2,r1, #0x0f000000
  and r0,r10,#0x0f000000
  subs r2,r2,r0
  submi r1,r1,#0x06000000 ;@ Decimal adjust units
  subs r1,r1,r10 ;@ Subtract BCD
  submis r1,r1,#0x60000000 ;@ Decimal adjust tens
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8108]  uses Op8108 ----------
Op8108
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0007
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r1:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r1,r0,asl #24

  ldrb r2,[r7,#0x45] ;@ Get X bit
  tst r2,#2
  addne r10,r10,#0x01000000 ;@ Add carry bit
;@ Sub units into r2:
  and r2,r1, #0x0f000000
  and r0,r10,#0x0f000000
  subs r2,r2,r0
  submi r1,r1,#0x06000000 ;@ Decimal adjust units
  subs r1,r1,r10 ;@ Subtract BCD
  submis r1,r1,#0x60000000 ;@ Decimal adjust tens
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8110] or.b d0, (a0) uses Op8110 ----------
Op8110
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8118] or.b d0, (a0)+ uses Op8118 ----------
Op8118
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [811f] or.b d0, (a7)+ uses Op811f ----------
Op811f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8120] or.b d0, -(a0) uses Op8120 ----------
Op8120
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8127] or.b d0, -(a7) uses Op8127 ----------
Op8127
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8128] or.b d0, ($3333,a0) uses Op8128 ----------
Op8128
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8130] or.b d0, ($33,a0,d3.w*2) uses Op8130 ----------
Op8130
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8138] or.b d0, $3333.w uses Op8138 ----------
Op8138
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8139] or.b d0, $33333333.l uses Op8139 ----------
Op8139
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [813a] or.b d0, ($3333,pc); =3335 uses Op813a ----------
Op813a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [813b] or.b d0, ($33,pc,d3.w*2); =35 uses Op813b ----------
Op813b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8150] or.w d0, (a0) uses Op8150 ----------
Op8150
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8158] or.w d0, (a0)+ uses Op8158 ----------
Op8158
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [815f] or.w d0, (a7)+ uses Op815f ----------
Op815f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8160] or.w d0, -(a0) uses Op8160 ----------
Op8160
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8167] or.w d0, -(a7) uses Op8167 ----------
Op8167
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8168] or.w d0, ($3333,a0) uses Op8168 ----------
Op8168
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8170] or.w d0, ($33,a0,d3.w*2) uses Op8170 ----------
Op8170
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8178] or.w d0, $3333.w uses Op8178 ----------
Op8178
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8179] or.w d0, $33333333.l uses Op8179 ----------
Op8179
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [817a] or.w d0, ($3333,pc); =3335 uses Op817a ----------
Op817a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [817b] or.w d0, ($33,pc,d3.w*2); =35 uses Op817b ----------
Op817b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8190] or.l d0, (a0) uses Op8190 ----------
Op8190
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [8198] or.l d0, (a0)+ uses Op8198 ----------
Op8198
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [819f] or.l d0, (a7)+ uses Op819f ----------
Op819f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81a0] or.l d0, -(a0) uses Op81a0 ----------
Op81a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81a7] or.l d0, -(a7) uses Op81a7 ----------
Op81a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81a8] or.l d0, ($3333,a0) uses Op81a8 ----------
Op81a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81b0] or.l d0, ($33,a0,d3.w*2) uses Op81b0 ----------
Op81b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81b8] or.l d0, $3333.w uses Op81b8 ----------
Op81b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81b9] or.l d0, $33333333.l uses Op81b9 ----------
Op81b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81ba] or.l d0, ($3333,pc); =3335 uses Op81ba ----------
Op81ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81bb] or.l d0, ($33,pc,d3.w*2); =35 uses Op81bb ----------
Op81bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  orr  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81c0] divs.w d0, d0 uses Op81c0 ----------
Op81c0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r10:
  ldrsh r10,[r7,r10]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81c0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81c0

Divide81c0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81c0

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#133 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81d0] divs.w (a0), d0 uses Op81d0 ----------
Op81d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81d0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81d0

Divide81d0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81d0

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81d8] divs.w (a0)+, d0 uses Op81d8 ----------
Op81d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81d8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81d8

Divide81d8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81d8

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81df] divs.w (a7)+, d0 uses Op81df ----------
Op81df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81df
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81df

Divide81df
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81df

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81e0] divs.w -(a0), d0 uses Op81e0 ----------
Op81e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81e0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81e0

Divide81e0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81e0

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#139 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81e7] divs.w -(a7), d0 uses Op81e7 ----------
Op81e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81e7
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81e7

Divide81e7
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81e7

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#139 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81e8] divs.w ($3333,a0), d0 uses Op81e8 ----------
Op81e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81e8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81e8

Divide81e8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81e8

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81f0] divs.w ($33,a0,d3.w*2), d0 uses Op81f0 ----------
Op81f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81f0
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81f0

Divide81f0
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81f0

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#143 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81f8] divs.w $3333.w, d0 uses Op81f8 ----------
Op81f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81f8
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81f8

Divide81f8
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81f8

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81f9] divs.w $33333333.l, d0 uses Op81f9 ----------
Op81f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81f9
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81f9

Divide81f9
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81f9

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#145 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81fa] divs.w ($3333,pc), d0; =3335 uses Op81fa ----------
Op81fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81fa
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81fa

Divide81fa
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81fa

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#141 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81fb] divs.w ($33,pc,d3.w*2), d0; =35 uses Op81fb ----------
Op81fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81fb
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81fb

Divide81fb
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81fb

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#143 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [81fc] divs.w #$3333, d0 uses Op81fc ----------
Op81fc
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

  cmp r10,#0
  moveq r10,#1 ;@ Divide by zero

  mov r11,#0 ;@ r11 = 1 if the result is negative
  eorlt r11,r11,#1
  rsblt r10,r10,#0 ;@ Make r10 positive

  cmp r2,#0
  eorlt r11,r11,#1
  rsblt r2,r2,#0 ;@ Make r2 positive

;@ Divide r2 by r10
  mov r3,#0
  mov r1,r10

;@ Shift up divisor till it's just less than numerator
Shift81fc
  cmp r1,r2,lsr #1
  movls r1,r1,lsl #1
  bcc Shift81fc

Divide81fc
  cmp r2,r1
  adc r3,r3,r3 ;@ Double r3 and add 1 if carry set
  subcs r2,r2,r1
  teq r1,r10
  movne r1,r1,lsr #1
  bne Divide81fc

  tst r11,r11
  rsbne r3,r3,#0 ;@ Negate if result is negative
  mov r11,r2 ;@ Remainder
  adds r1,r3,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags
  mov r1,r1,lsl #16 ;@ Clip to 16-bits
  mov r1,r1,lsr #16
  orr r1,r1,r11,lsl #16 ;@ Insert remainder

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#137 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9000] sub.b d0, d0 uses Op9000 ----------
Op9000
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsb r11,[r7,r10]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9010] sub.b (a0), d0 uses Op9010 ----------
Op9010
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9018] sub.b (a0)+, d0 uses Op9018 ----------
Op9018
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [901f] sub.b (a7)+, d0 uses Op901f ----------
Op901f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9020] sub.b -(a0), d0 uses Op9020 ----------
Op9020
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9027] sub.b -(a7), d0 uses Op9027 ----------
Op9027
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9028] sub.b ($3333,a0), d0 uses Op9028 ----------
Op9028
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9030] sub.b ($33,a0,d3.w*2), d0 uses Op9030 ----------
Op9030
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9038] sub.b $3333.w, d0 uses Op9038 ----------
Op9038
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9039] sub.b $33333333.l, d0 uses Op9039 ----------
Op9039
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903a] sub.b ($3333,pc), d0; =3335 uses Op903a ----------
Op903a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903b] sub.b ($33,pc,d3.w*2), d0; =35 uses Op903b ----------
Op903b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903c] sub.b #$33, d0 uses Op903c ----------
Op903c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r11:
  mov r11,r10,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903d] sub.b ea=(111 101), d0 uses Op903d ----------
Op903d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903e] sub.b ea=(111 110), d0 uses Op903e ----------
Op903e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [903f] sub.b ea=(111 111), d0 uses Op903f ----------
Op903f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9040] sub.w d0, d0 uses Op9040 ----------
Op9040
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsh r11,[r7,r10]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9050] sub.w (a0), d0 uses Op9050 ----------
Op9050
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9058] sub.w (a0)+, d0 uses Op9058 ----------
Op9058
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [905f] sub.w (a7)+, d0 uses Op905f ----------
Op905f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9060] sub.w -(a0), d0 uses Op9060 ----------
Op9060
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9067] sub.w -(a7), d0 uses Op9067 ----------
Op9067
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9068] sub.w ($3333,a0), d0 uses Op9068 ----------
Op9068
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9070] sub.w ($33,a0,d3.w*2), d0 uses Op9070 ----------
Op9070
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9078] sub.w $3333.w, d0 uses Op9078 ----------
Op9078
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9079] sub.w $33333333.l, d0 uses Op9079 ----------
Op9079
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907a] sub.w ($3333,pc), d0; =3335 uses Op907a ----------
Op907a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907b] sub.w ($33,pc,d3.w*2), d0; =35 uses Op907b ----------
Op907b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907c] sub.w #$3333, d0 uses Op907c ----------
Op907c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r11:
  mov r11,r10,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907d] sub.w ea=(111 101), d0 uses Op907d ----------
Op907d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907e] sub.w ea=(111 110), d0 uses Op907e ----------
Op907e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [907f] sub.w ea=(111 111), d0 uses Op907f ----------
Op907f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9080] sub.l d0, d0 uses Op9080 ----------
Op9080
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r11:
  ldr r11,[r7,r10,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9090] sub.l (a0), d0 uses Op9090 ----------
Op9090
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9098] sub.l (a0)+, d0 uses Op9098 ----------
Op9098
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [909f] sub.l (a7)+, d0 uses Op909f ----------
Op909f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90a0] sub.l -(a0), d0 uses Op90a0 ----------
Op90a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90a7] sub.l -(a7), d0 uses Op90a7 ----------
Op90a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90a8] sub.l ($3333,a0), d0 uses Op90a8 ----------
Op90a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90b0] sub.l ($33,a0,d3.w*2), d0 uses Op90b0 ----------
Op90b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90b8] sub.l $3333.w, d0 uses Op90b8 ----------
Op90b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90b9] sub.l $33333333.l, d0 uses Op90b9 ----------
Op90b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90ba] sub.l ($3333,pc), d0; =3335 uses Op90ba ----------
Op90ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90bb] sub.l ($33,pc,d3.w*2), d0; =35 uses Op90bb ----------
Op90bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90bc] sub.l #$33333333, d0 uses Op90bc ----------
Op90bc
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r11:
  mov r11,r10,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90bd] sub.l ea=(111 101), d0 uses Op90bd ----------
Op90bd
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90be] sub.l ea=(111 110), d0 uses Op90be ----------
Op90be
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90bf] sub.l ea=(111 111), d0 uses Op90bf ----------
Op90bf
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90c0] suba.w d0, a0 uses Op90c0 ----------
Op90c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r10:
  ldrsh r10,[r7,r0]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90d0] suba.w (a0), a0 uses Op90d0 ----------
Op90d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90d8] suba.w (a0)+, a0 uses Op90d8 ----------
Op90d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90df] suba.w (a7)+, a0 uses Op90df ----------
Op90df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90e0] suba.w -(a0), a0 uses Op90e0 ----------
Op90e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90e7] suba.w -(a7), a0 uses Op90e7 ----------
Op90e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90e8] suba.w ($3333,a0), a0 uses Op90e8 ----------
Op90e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90f0] suba.w ($33,a0,d3.w*2), a0 uses Op90f0 ----------
Op90f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90f8] suba.w $3333.w, a0 uses Op90f8 ----------
Op90f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90f9] suba.w $33333333.l, a0 uses Op90f9 ----------
Op90f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90fa] suba.w ($3333,pc), a0; =3335 uses Op90fa ----------
Op90fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90fb] suba.w ($33,pc,d3.w*2), a0; =35 uses Op90fb ----------
Op90fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [90fc] suba.w #$3333, a0 uses Op90fc ----------
Op90fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9100] subx.b d0, d0 uses Op9100 ----------
Op9100
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r11:
  ldrsb r11,[r7,r0]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mvn r2,r2,lsl #28 ;@ Invert it
  msr cpsr_flg,r2 ;@ Get into Carry

  sbcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9110] sub.b d0, (a0) uses Op9110 ----------
Op9110
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9118] sub.b d0, (a0)+ uses Op9118 ----------
Op9118
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [911f] sub.b d0, (a7)+ uses Op911f ----------
Op911f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9120] sub.b d0, -(a0) uses Op9120 ----------
Op9120
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9127] sub.b d0, -(a7) uses Op9127 ----------
Op9127
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9128] sub.b d0, ($3333,a0) uses Op9128 ----------
Op9128
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9130] sub.b d0, ($33,a0,d3.w*2) uses Op9130 ----------
Op9130
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9138] sub.b d0, $3333.w uses Op9138 ----------
Op9138
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9139] sub.b d0, $33333333.l uses Op9139 ----------
Op9139
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [913a] sub.b d0, ($3333,pc); =3335 uses Op913a ----------
Op913a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [913b] sub.b d0, ($33,pc,d3.w*2); =35 uses Op913b ----------
Op913b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9140] subx.w d0, d0 uses Op9140 ----------
Op9140
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r11:
  ldrsh r11,[r7,r0]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mvn r2,r2,lsl #28 ;@ Invert it
  msr cpsr_flg,r2 ;@ Get into Carry

  sbcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9150] sub.w d0, (a0) uses Op9150 ----------
Op9150
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9158] sub.w d0, (a0)+ uses Op9158 ----------
Op9158
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [915f] sub.w d0, (a7)+ uses Op915f ----------
Op915f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9160] sub.w d0, -(a0) uses Op9160 ----------
Op9160
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9167] sub.w d0, -(a7) uses Op9167 ----------
Op9167
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9168] sub.w d0, ($3333,a0) uses Op9168 ----------
Op9168
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9170] sub.w d0, ($33,a0,d3.w*2) uses Op9170 ----------
Op9170
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9178] sub.w d0, $3333.w uses Op9178 ----------
Op9178
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9179] sub.w d0, $33333333.l uses Op9179 ----------
Op9179
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [917a] sub.w d0, ($3333,pc); =3335 uses Op917a ----------
Op917a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [917b] sub.w d0, ($33,pc,d3.w*2); =35 uses Op917b ----------
Op917b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9180] subx.l d0, d0 uses Op9180 ----------
Op9180
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r11:
  ldr r11,[r7,r0,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mvn r2,r2,lsl #28 ;@ Invert it
  msr cpsr_flg,r2 ;@ Get into Carry

  sbcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9190] sub.l d0, (a0) uses Op9190 ----------
Op9190
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [9198] sub.l d0, (a0)+ uses Op9198 ----------
Op9198
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [919f] sub.l d0, (a7)+ uses Op919f ----------
Op919f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91a0] sub.l d0, -(a0) uses Op91a0 ----------
Op91a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91a7] sub.l d0, -(a7) uses Op91a7 ----------
Op91a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91a8] sub.l d0, ($3333,a0) uses Op91a8 ----------
Op91a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91b0] sub.l d0, ($33,a0,d3.w*2) uses Op91b0 ----------
Op91b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91b8] sub.l d0, $3333.w uses Op91b8 ----------
Op91b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91b9] sub.l d0, $33333333.l uses Op91b9 ----------
Op91b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91ba] sub.l d0, ($3333,pc); =3335 uses Op91ba ----------
Op91ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91bb] sub.l d0, ($33,pc,d3.w*2); =35 uses Op91bb ----------
Op91bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  subs r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91c0] suba.l d0, a0 uses Op91c0 ----------
Op91c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r10:
  ldr r10,[r7,r0,lsl #2]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91d0] suba.l (a0), a0 uses Op91d0 ----------
Op91d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91d8] suba.l (a0)+, a0 uses Op91d8 ----------
Op91d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91df] suba.l (a7)+, a0 uses Op91df ----------
Op91df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91e0] suba.l -(a0), a0 uses Op91e0 ----------
Op91e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91e7] suba.l -(a7), a0 uses Op91e7 ----------
Op91e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91e8] suba.l ($3333,a0), a0 uses Op91e8 ----------
Op91e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91f0] suba.l ($33,a0,d3.w*2), a0 uses Op91f0 ----------
Op91f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91f8] suba.l $3333.w, a0 uses Op91f8 ----------
Op91f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91f9] suba.l $33333333.l, a0 uses Op91f9 ----------
Op91f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91fa] suba.l ($3333,pc), a0; =3335 uses Op91fa ----------
Op91fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91fb] suba.l ($33,pc,d3.w*2), a0; =35 uses Op91fb ----------
Op91fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [91fc] suba.l #$33333333, a0 uses Op91fc ----------
Op91fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  sub r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b000] cmp.b d0, d0 uses Opb000 ----------
Opb000
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b010] cmp.b (a0), d0 uses Opb010 ----------
Opb010
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b018] cmp.b (a0)+, d0 uses Opb018 ----------
Opb018
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b01f] cmp.b (a7)+, d0 uses Opb01f ----------
Opb01f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b020] cmp.b -(a0), d0 uses Opb020 ----------
Opb020
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b027] cmp.b -(a7), d0 uses Opb027 ----------
Opb027
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b028] cmp.b ($3333,a0), d0 uses Opb028 ----------
Opb028
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b030] cmp.b ($33,a0,d3.w*2), d0 uses Opb030 ----------
Opb030
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b038] cmp.b $3333.w, d0 uses Opb038 ----------
Opb038
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b039] cmp.b $33333333.l, d0 uses Opb039 ----------
Opb039
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b03a] cmp.b ($3333,pc), d0; =3335 uses Opb03a ----------
Opb03a
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b03b] cmp.b ($33,pc,d3.w*2), d0; =35 uses Opb03b ----------
Opb03b
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b03c] cmp.b #$33, d0 uses Opb03c ----------
Opb03c
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r0:
  mov r0,r10,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b040] cmp.w d0, d0 uses Opb040 ----------
Opb040
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b050] cmp.w (a0), d0 uses Opb050 ----------
Opb050
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b058] cmp.w (a0)+, d0 uses Opb058 ----------
Opb058
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b05f] cmp.w (a7)+, d0 uses Opb05f ----------
Opb05f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b060] cmp.w -(a0), d0 uses Opb060 ----------
Opb060
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b067] cmp.w -(a7), d0 uses Opb067 ----------
Opb067
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b068] cmp.w ($3333,a0), d0 uses Opb068 ----------
Opb068
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b070] cmp.w ($33,a0,d3.w*2), d0 uses Opb070 ----------
Opb070
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b078] cmp.w $3333.w, d0 uses Opb078 ----------
Opb078
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b079] cmp.w $33333333.l, d0 uses Opb079 ----------
Opb079
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b07a] cmp.w ($3333,pc), d0; =3335 uses Opb07a ----------
Opb07a
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b07b] cmp.w ($33,pc,d3.w*2), d0; =35 uses Opb07b ----------
Opb07b
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b07c] cmp.w #$3333, d0 uses Opb07c ----------
Opb07c
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r0:
  mov r0,r10,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b080] cmp.l d0, d0 uses Opb080 ----------
Opb080
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b090] cmp.l (a0), d0 uses Opb090 ----------
Opb090
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b098] cmp.l (a0)+, d0 uses Opb098 ----------
Opb098
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b09f] cmp.l (a7)+, d0 uses Opb09f ----------
Opb09f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0a0] cmp.l -(a0), d0 uses Opb0a0 ----------
Opb0a0
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0a7] cmp.l -(a7), d0 uses Opb0a7 ----------
Opb0a7
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0a8] cmp.l ($3333,a0), d0 uses Opb0a8 ----------
Opb0a8
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0b0] cmp.l ($33,a0,d3.w*2), d0 uses Opb0b0 ----------
Opb0b0
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0b8] cmp.l $3333.w, d0 uses Opb0b8 ----------
Opb0b8
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0b9] cmp.l $33333333.l, d0 uses Opb0b9 ----------
Opb0b9
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0ba] cmp.l ($3333,pc), d0; =3335 uses Opb0ba ----------
Opb0ba
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0bb] cmp.l ($33,pc,d3.w*2), d0; =35 uses Opb0bb ----------
Opb0bb
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0bc] cmp.l #$33333333, d0 uses Opb0bc ----------
Opb0bc
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r0:
  mov r0,r10,asl #0

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  cmp r1,r0
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0c0] cmpa.w d0, a0 uses Opb0c0 ----------
Opb0c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r10:
  ldrsh r10,[r7,r0]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0d0] cmpa.w (a0), a0 uses Opb0d0 ----------
Opb0d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0d8] cmpa.w (a0)+, a0 uses Opb0d8 ----------
Opb0d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0df] cmpa.w (a7)+, a0 uses Opb0df ----------
Opb0df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0e0] cmpa.w -(a0), a0 uses Opb0e0 ----------
Opb0e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0e7] cmpa.w -(a7), a0 uses Opb0e7 ----------
Opb0e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0e8] cmpa.w ($3333,a0), a0 uses Opb0e8 ----------
Opb0e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0f0] cmpa.w ($33,a0,d3.w*2), a0 uses Opb0f0 ----------
Opb0f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0f8] cmpa.w $3333.w, a0 uses Opb0f8 ----------
Opb0f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0f9] cmpa.w $33333333.l, a0 uses Opb0f9 ----------
Opb0f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0fa] cmpa.w ($3333,pc), a0; =3335 uses Opb0fa ----------
Opb0fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0fb] cmpa.w ($33,pc,d3.w*2), a0; =35 uses Opb0fb ----------
Opb0fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b0fc] cmpa.w #$3333, a0 uses Opb0fc ----------
Opb0fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b100] eor.b d0, d0 uses Opb100 ----------
Opb100
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #24
  strb r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b110] eor.b d0, (a0) uses Opb110 ----------
Opb110
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b118] eor.b d0, (a0)+ uses Opb118 ----------
Opb118
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b11f] eor.b d0, (a7)+ uses Opb11f ----------
Opb11f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b120] eor.b d0, -(a0) uses Opb120 ----------
Opb120
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b127] eor.b d0, -(a7) uses Opb127 ----------
Opb127
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b128] eor.b d0, ($3333,a0) uses Opb128 ----------
Opb128
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b130] eor.b d0, ($33,a0,d3.w*2) uses Opb130 ----------
Opb130
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b138] eor.b d0, $3333.w uses Opb138 ----------
Opb138
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b139] eor.b d0, $33333333.l uses Opb139 ----------
Opb139
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b13a] eor.b d0, ($3333,pc); =3335 uses Opb13a ----------
Opb13a
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b13b] eor.b d0, ($33,pc,d3.w*2); =35 uses Opb13b ----------
Opb13b
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #24

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsb r1,[r7,r1]
  mov r1,r1,asl #24

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b140] eor.w d0, d0 uses Opb140 ----------
Opb140
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  mov r1,r1,asr #16
  strh r1,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b150] eor.w d0, (a0) uses Opb150 ----------
Opb150
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b158] eor.w d0, (a0)+ uses Opb158 ----------
Opb158
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b15f] eor.w d0, (a7)+ uses Opb15f ----------
Opb15f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b160] eor.w d0, -(a0) uses Opb160 ----------
Opb160
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b167] eor.w d0, -(a7) uses Opb167 ----------
Opb167
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b168] eor.w d0, ($3333,a0) uses Opb168 ----------
Opb168
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b170] eor.w d0, ($33,a0,d3.w*2) uses Opb170 ----------
Opb170
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b178] eor.w d0, $3333.w uses Opb178 ----------
Opb178
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b179] eor.w d0, $33333333.l uses Opb179 ----------
Opb179
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b17a] eor.w d0, ($3333,pc); =3335 uses Opb17a ----------
Opb17a
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b17b] eor.w d0, ($33,pc,d3.w*2); =35 uses Opb17b ----------
Opb17b
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #7
;@ EaRead : Read register[r1] into r1:
  ldrsh r1,[r7,r1]
  mov r1,r1,asl #16

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b180] eor.l d0, d0 uses Opb180 ----------
Opb180
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r10]:
  str r1,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b190] eor.l d0, (a0) uses Opb190 ----------
Opb190
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b198] eor.l d0, (a0)+ uses Opb198 ----------
Opb198
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b19f] eor.l d0, (a7)+ uses Opb19f ----------
Opb19f
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1a0] eor.l d0, -(a0) uses Opb1a0 ----------
Opb1a0
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1a7] eor.l d0, -(a7) uses Opb1a7 ----------
Opb1a7
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1a8] eor.l d0, ($3333,a0) uses Opb1a8 ----------
Opb1a8
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1b0] eor.l d0, ($33,a0,d3.w*2) uses Opb1b0 ----------
Opb1b0
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1b8] eor.l d0, $3333.w uses Opb1b8 ----------
Opb1b8
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1b9] eor.l d0, $33333333.l uses Opb1b9 ----------
Opb1b9
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#28 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1ba] eor.l d0, ($3333,pc); =3335 uses Opb1ba ----------
Opb1ba
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1bb] eor.l d0, ($33,pc,d3.w*2); =35 uses Opb1bb ----------
Opb1bb
;@ Get EA into r10 and value into r0:
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

;@ Get register operand into r1:
;@ EaCalc : Get register index into r1:
  and r1,r8,#0x0e00
  mov r1,r1,lsr #9
;@ EaRead : Read register[r1] into r1:
  ldr r1,[r7,r1,lsl #2]

;@ Do arithmetic:
  eor r1,r0,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#26 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1c0] cmpa.l d0, a0 uses Opb1c0 ----------
Opb1c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r10:
  ldr r10,[r7,r0,lsl #2]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1d0] cmpa.l (a0), a0 uses Opb1d0 ----------
Opb1d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1d8] cmpa.l (a0)+, a0 uses Opb1d8 ----------
Opb1d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1df] cmpa.l (a7)+, a0 uses Opb1df ----------
Opb1df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1e0] cmpa.l -(a0), a0 uses Opb1e0 ----------
Opb1e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1e7] cmpa.l -(a7), a0 uses Opb1e7 ----------
Opb1e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1e8] cmpa.l ($3333,a0), a0 uses Opb1e8 ----------
Opb1e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1f0] cmpa.l ($33,a0,d3.w*2), a0 uses Opb1f0 ----------
Opb1f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1f8] cmpa.l $3333.w, a0 uses Opb1f8 ----------
Opb1f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1f9] cmpa.l $33333333.l, a0 uses Opb1f9 ----------
Opb1f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1fa] cmpa.l ($3333,pc), a0; =3335 uses Opb1fa ----------
Opb1fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1fb] cmpa.l ($33,pc,d3.w*2), a0; =35 uses Opb1fb ----------
Opb1fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [b1fc] cmpa.l #$33333333, a0 uses Opb1fc ----------
Opb1fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  cmp r1,r10 ;@ Defines NZCV
  mrs r9,cpsr ;@ r9=flags
  eor r9,r9,#0x20000000 ;@ Invert carry

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c000] and.b d0, d0 uses Opc000 ----------
Opc000
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsb r11,[r7,r10]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c010] and.b (a0), d0 uses Opc010 ----------
Opc010
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c018] and.b (a0)+, d0 uses Opc018 ----------
Opc018
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c01f] and.b (a7)+, d0 uses Opc01f ----------
Opc01f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c020] and.b -(a0), d0 uses Opc020 ----------
Opc020
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c027] and.b -(a7), d0 uses Opc027 ----------
Opc027
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c028] and.b ($3333,a0), d0 uses Opc028 ----------
Opc028
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c030] and.b ($33,a0,d3.w*2), d0 uses Opc030 ----------
Opc030
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c038] and.b $3333.w, d0 uses Opc038 ----------
Opc038
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c039] and.b $33333333.l, d0 uses Opc039 ----------
Opc039
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03a] and.b ($3333,pc), d0; =3335 uses Opc03a ----------
Opc03a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03b] and.b ($33,pc,d3.w*2), d0; =35 uses Opc03b ----------
Opc03b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03c] and.b #$33, d0 uses Opc03c ----------
Opc03c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r11:
  mov r11,r10,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03d] and.b ea=(111 101), d0 uses Opc03d ----------
Opc03d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03e] and.b ea=(111 110), d0 uses Opc03e ----------
Opc03e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c03f] and.b ea=(111 111), d0 uses Opc03f ----------
Opc03f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c040] and.w d0, d0 uses Opc040 ----------
Opc040
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsh r11,[r7,r10]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c050] and.w (a0), d0 uses Opc050 ----------
Opc050
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c058] and.w (a0)+, d0 uses Opc058 ----------
Opc058
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c05f] and.w (a7)+, d0 uses Opc05f ----------
Opc05f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c060] and.w -(a0), d0 uses Opc060 ----------
Opc060
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c067] and.w -(a7), d0 uses Opc067 ----------
Opc067
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c068] and.w ($3333,a0), d0 uses Opc068 ----------
Opc068
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c070] and.w ($33,a0,d3.w*2), d0 uses Opc070 ----------
Opc070
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c078] and.w $3333.w, d0 uses Opc078 ----------
Opc078
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c079] and.w $33333333.l, d0 uses Opc079 ----------
Opc079
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07a] and.w ($3333,pc), d0; =3335 uses Opc07a ----------
Opc07a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07b] and.w ($33,pc,d3.w*2), d0; =35 uses Opc07b ----------
Opc07b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07c] and.w #$3333, d0 uses Opc07c ----------
Opc07c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r11:
  mov r11,r10,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07d] and.w ea=(111 101), d0 uses Opc07d ----------
Opc07d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07e] and.w ea=(111 110), d0 uses Opc07e ----------
Opc07e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c07f] and.w ea=(111 111), d0 uses Opc07f ----------
Opc07f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c080] and.l d0, d0 uses Opc080 ----------
Opc080
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r11:
  ldr r11,[r7,r10,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c090] and.l (a0), d0 uses Opc090 ----------
Opc090
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c098] and.l (a0)+, d0 uses Opc098 ----------
Opc098
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c09f] and.l (a7)+, d0 uses Opc09f ----------
Opc09f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0a0] and.l -(a0), d0 uses Opc0a0 ----------
Opc0a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0a7] and.l -(a7), d0 uses Opc0a7 ----------
Opc0a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0a8] and.l ($3333,a0), d0 uses Opc0a8 ----------
Opc0a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0b0] and.l ($33,a0,d3.w*2), d0 uses Opc0b0 ----------
Opc0b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0b8] and.l $3333.w, d0 uses Opc0b8 ----------
Opc0b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0b9] and.l $33333333.l, d0 uses Opc0b9 ----------
Opc0b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0ba] and.l ($3333,pc), d0; =3335 uses Opc0ba ----------
Opc0ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0bb] and.l ($33,pc,d3.w*2), d0; =35 uses Opc0bb ----------
Opc0bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0bc] and.l #$33333333, d0 uses Opc0bc ----------
Opc0bc
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r11:
  mov r11,r10,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0bd] and.l ea=(111 101), d0 uses Opc0bd ----------
Opc0bd
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0be] and.l ea=(111 110), d0 uses Opc0be ----------
Opc0be
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0bf] and.l ea=(111 111), d0 uses Opc0bf ----------
Opc0bf
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r1,r11
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0c0] mulu.w d0, d0 uses Opc0c0 ----------
Opc0c0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r10:
  ldrsh r10,[r7,r10]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#70 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0d0] mulu.w (a0), d0 uses Opc0d0 ----------
Opc0d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0d8] mulu.w (a0)+, d0 uses Opc0d8 ----------
Opc0d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0df] mulu.w (a7)+, d0 uses Opc0df ----------
Opc0df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0e0] mulu.w -(a0), d0 uses Opc0e0 ----------
Opc0e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#76 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0e7] mulu.w -(a7), d0 uses Opc0e7 ----------
Opc0e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#76 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0e8] mulu.w ($3333,a0), d0 uses Opc0e8 ----------
Opc0e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0f0] mulu.w ($33,a0,d3.w*2), d0 uses Opc0f0 ----------
Opc0f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#80 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0f8] mulu.w $3333.w, d0 uses Opc0f8 ----------
Opc0f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0f9] mulu.w $33333333.l, d0 uses Opc0f9 ----------
Opc0f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#82 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0fa] mulu.w ($3333,pc), d0; =3335 uses Opc0fa ----------
Opc0fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0fb] mulu.w ($33,pc,d3.w*2), d0; =35 uses Opc0fb ----------
Opc0fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#80 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c0fc] mulu.w #$3333, d0 uses Opc0fc ----------
Opc0fc
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r10,r10,lsl #16
  mov r2,r2,lsl #16
  mov r10,r10,lsr #16
  mov r2,r2,lsr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#70 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c100]  uses Opc100 ----------
Opc100
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0007
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r10:
  ldrsb r10,[r7,r0]
  mov r10,r10,asl #24

;@ EaCalc : Get register index into r11:
  and r11,r8,#0x0e00
  mov r11,r11,lsr #7
;@ EaRead : Read register[r11] into r1:
  ldrsb r1,[r7,r11]
  mov r1,r1,asl #24

  ldrb r2,[r7,#0x45] ;@ Get X bit
  tst r2,#2
  addne r10,r10,#0x01000000 ;@ Add carry bit
;@ Add units into r2:
  and r2,r1, #0x0f000000
  and r0,r10,#0x0f000000
  add r2,r2,r0
  cmp r2,#0x0a000000
  addpl r1,r1,#0x06000000 ;@ Decimal adjust units
  add r1,r1,r10 ;@ Add BCD
  mov r0,r1,lsr #24
  cmp r0,#0xa0
  addpl r1,r1,#0x60000000 ;@ Decimal adjust tens
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r1 into register[r11]:
  mov r1,r1,asr #24
  strb r1,[r7,r11]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c108]  uses Opc108 ----------
Opc108
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x0007
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#1 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #24

;@ EaCalc : Get '-(a0)' into r11:
  and r2,r8,#0x0e00
  mov r2,r2,lsr #7
  orr r2,r2,#0x20 ;@ A0-7
  ldr r11,[r7,r2]
  sub r11,r11,#1 ;@ Pre-decrement An
  str r11,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r11) into r1:
  mov r0,r11
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r1,r0,asl #24

  ldrb r2,[r7,#0x45] ;@ Get X bit
  tst r2,#2
  addne r10,r10,#0x01000000 ;@ Add carry bit
;@ Add units into r2:
  and r2,r1, #0x0f000000
  and r0,r10,#0x0f000000
  add r2,r2,r0
  cmp r2,#0x0a000000
  addpl r1,r1,#0x06000000 ;@ Decimal adjust units
  add r1,r1,r10 ;@ Add BCD
  mov r0,r1,lsr #24
  cmp r0,#0xa0
  addpl r1,r1,#0x60000000 ;@ Decimal adjust tens
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: Write r1 into '-(a0)' (address in r11):
  mov r0,r11
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c110] and.b d0, (a0) uses Opc110 ----------
Opc110
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c118] and.b d0, (a0)+ uses Opc118 ----------
Opc118
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c11f] and.b d0, (a7)+ uses Opc11f ----------
Opc11f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c120] and.b d0, -(a0) uses Opc120 ----------
Opc120
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c127] and.b d0, -(a7) uses Opc127 ----------
Opc127
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c128] and.b d0, ($3333,a0) uses Opc128 ----------
Opc128
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c130] and.b d0, ($33,a0,d3.w*2) uses Opc130 ----------
Opc130
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c138] and.b d0, $3333.w uses Opc138 ----------
Opc138
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c139] and.b d0, $33333333.l uses Opc139 ----------
Opc139
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c13a] and.b d0, ($3333,pc); =3335 uses Opc13a ----------
Opc13a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c13b] and.b d0, ($33,pc,d3.w*2); =35 uses Opc13b ----------
Opc13b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c140] exg d0, d0 uses Opc140 ----------
Opc140
  and r10,r8,#0x0e00 ;@ Find T register
  and r11,r8,#0x000f ;@ Find S register

  ldr r0,[r7,r10,lsr #7] ;@ Get T
  ldr r1,[r7,r11,lsl #2] ;@ Get S

  str r0,[r7,r11,lsl #2] ;@ T->S
  str r1,[r7,r10,lsr #7] ;@ S->T

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c148] exg a0, a0 uses Opc148 ----------
Opc148
  and r10,r8,#0x0e00 ;@ Find T register
  and r11,r8,#0x000f ;@ Find S register
  orr r10,r10,#0x1000 ;@ T is an address register

  ldr r0,[r7,r10,lsr #7] ;@ Get T
  ldr r1,[r7,r11,lsl #2] ;@ Get S

  str r0,[r7,r11,lsl #2] ;@ T->S
  str r1,[r7,r10,lsr #7] ;@ S->T

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c150] and.w d0, (a0) uses Opc150 ----------
Opc150
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c158] and.w d0, (a0)+ uses Opc158 ----------
Opc158
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c15f] and.w d0, (a7)+ uses Opc15f ----------
Opc15f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c160] and.w d0, -(a0) uses Opc160 ----------
Opc160
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c167] and.w d0, -(a7) uses Opc167 ----------
Opc167
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c168] and.w d0, ($3333,a0) uses Opc168 ----------
Opc168
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c170] and.w d0, ($33,a0,d3.w*2) uses Opc170 ----------
Opc170
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c178] and.w d0, $3333.w uses Opc178 ----------
Opc178
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c179] and.w d0, $33333333.l uses Opc179 ----------
Opc179
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c17a] and.w d0, ($3333,pc); =3335 uses Opc17a ----------
Opc17a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c17b] and.w d0, ($33,pc,d3.w*2); =35 uses Opc17b ----------
Opc17b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c188] exg a0, d0 uses Opc188 ----------
Opc188
  and r10,r8,#0x0e00 ;@ Find T register
  and r11,r8,#0x000f ;@ Find S register

  ldr r0,[r7,r10,lsr #7] ;@ Get T
  ldr r1,[r7,r11,lsl #2] ;@ Get S

  str r0,[r7,r11,lsl #2] ;@ T->S
  str r1,[r7,r10,lsr #7] ;@ S->T

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c190] and.l d0, (a0) uses Opc190 ----------
Opc190
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c198] and.l d0, (a0)+ uses Opc198 ----------
Opc198
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c19f] and.l d0, (a7)+ uses Opc19f ----------
Opc19f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1a0] and.l d0, -(a0) uses Opc1a0 ----------
Opc1a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1a7] and.l d0, -(a7) uses Opc1a7 ----------
Opc1a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1a8] and.l d0, ($3333,a0) uses Opc1a8 ----------
Opc1a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1b0] and.l d0, ($33,a0,d3.w*2) uses Opc1b0 ----------
Opc1b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1b8] and.l d0, $3333.w uses Opc1b8 ----------
Opc1b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1b9] and.l d0, $33333333.l uses Opc1b9 ----------
Opc1b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1ba] and.l d0, ($3333,pc); =3335 uses Opc1ba ----------
Opc1ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1bb] and.l d0, ($33,pc,d3.w*2); =35 uses Opc1bb ----------
Opc1bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  and  r1,r11,r1
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1c0] muls.w d0, d0 uses Opc1c0 ----------
Opc1c0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r10:
  ldrsh r10,[r7,r10]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#70 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1d0] muls.w (a0), d0 uses Opc1d0 ----------
Opc1d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1d8] muls.w (a0)+, d0 uses Opc1d8 ----------
Opc1d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1df] muls.w (a7)+, d0 uses Opc1df ----------
Opc1df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#74 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1e0] muls.w -(a0), d0 uses Opc1e0 ----------
Opc1e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#76 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1e7] muls.w -(a7), d0 uses Opc1e7 ----------
Opc1e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#76 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1e8] muls.w ($3333,a0), d0 uses Opc1e8 ----------
Opc1e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1f0] muls.w ($33,a0,d3.w*2), d0 uses Opc1f0 ----------
Opc1f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#80 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1f8] muls.w $3333.w, d0 uses Opc1f8 ----------
Opc1f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1f9] muls.w $33333333.l, d0 uses Opc1f9 ----------
Opc1f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#82 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1fa] muls.w ($3333,pc), d0; =3335 uses Opc1fa ----------
Opc1fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#78 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1fb] muls.w ($33,pc,d3.w*2), d0; =35 uses Opc1fb ----------
Opc1fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r10:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#80 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [c1fc] muls.w #$3333, d0 uses Opc1fc ----------
Opc1fc
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r10:

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r2:
  ldr r2,[r7,r0,lsl #2]

;@ Get 16-bit signs right:
  mov r2,r2,lsl #16
  mov r2,r2,asr #16

  mul r1,r2,r10
  adds r1,r1,#0 ;@ Defines NZ, clears CV
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#70 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d000] add.b d0, d0 uses Opd000 ----------
Opd000
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsb r11,[r7,r10]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d010] add.b (a0), d0 uses Opd010 ----------
Opd010
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d018] add.b (a0)+, d0 uses Opd018 ----------
Opd018
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d01f] add.b (a7)+, d0 uses Opd01f ----------
Opd01f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d020] add.b -(a0), d0 uses Opd020 ----------
Opd020
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d027] add.b -(a7), d0 uses Opd027 ----------
Opd027
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d028] add.b ($3333,a0), d0 uses Opd028 ----------
Opd028
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d030] add.b ($33,a0,d3.w*2), d0 uses Opd030 ----------
Opd030
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d038] add.b $3333.w, d0 uses Opd038 ----------
Opd038
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d039] add.b $33333333.l, d0 uses Opd039 ----------
Opd039
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03a] add.b ($3333,pc), d0; =3335 uses Opd03a ----------
Opd03a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03b] add.b ($33,pc,d3.w*2), d0; =35 uses Opd03b ----------
Opd03b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03c] add.b #$33, d0 uses Opd03c ----------
Opd03c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33' into r10:
  ldrsb r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$33' (address in r10) into r11:
  mov r11,r10,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03d] add.b ea=(111 101), d0 uses Opd03d ----------
Opd03d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03e] add.b ea=(111 110), d0 uses Opd03e ----------
Opd03e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d03f] add.b ea=(111 111), d0 uses Opd03f ----------
Opd03f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d040] add.w d0, d0 uses Opd040 ----------
Opd040
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r11:
  ldrsh r11,[r7,r10]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d050] add.w (a0), d0 uses Opd050 ----------
Opd050
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d058] add.w (a0)+, d0 uses Opd058 ----------
Opd058
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d05f] add.w (a7)+, d0 uses Opd05f ----------
Opd05f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d060] add.w -(a0), d0 uses Opd060 ----------
Opd060
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d067] add.w -(a7), d0 uses Opd067 ----------
Opd067
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d068] add.w ($3333,a0), d0 uses Opd068 ----------
Opd068
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d070] add.w ($33,a0,d3.w*2), d0 uses Opd070 ----------
Opd070
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d078] add.w $3333.w, d0 uses Opd078 ----------
Opd078
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d079] add.w $33333333.l, d0 uses Opd079 ----------
Opd079
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07a] add.w ($3333,pc), d0; =3335 uses Opd07a ----------
Opd07a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07b] add.w ($33,pc,d3.w*2), d0; =35 uses Opd07b ----------
Opd07b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07c] add.w #$3333, d0 uses Opd07c ----------
Opd07c
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$3333' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r10) into r11:
  mov r11,r10,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07d] add.w ea=(111 101), d0 uses Opd07d ----------
Opd07d
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07e] add.w ea=(111 110), d0 uses Opd07e ----------
Opd07e
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d07f] add.w ea=(111 111), d0 uses Opd07f ----------
Opd07f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d080] add.l d0, d0 uses Opd080 ----------
Opd080
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x000f
;@ EaRead : Read register[r10] into r11:
  ldr r11,[r7,r10,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d090] add.l (a0), d0 uses Opd090 ----------
Opd090
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d098] add.l (a0)+, d0 uses Opd098 ----------
Opd098
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d09f] add.l (a7)+, d0 uses Opd09f ----------
Opd09f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0a0] add.l -(a0), d0 uses Opd0a0 ----------
Opd0a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0a7] add.l -(a7), d0 uses Opd0a7 ----------
Opd0a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0a8] add.l ($3333,a0), d0 uses Opd0a8 ----------
Opd0a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0b0] add.l ($33,a0,d3.w*2), d0 uses Opd0b0 ----------
Opd0b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0b8] add.l $3333.w, d0 uses Opd0b8 ----------
Opd0b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0b9] add.l $33333333.l, d0 uses Opd0b9 ----------
Opd0b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0ba] add.l ($3333,pc), d0; =3335 uses Opd0ba ----------
Opd0ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0bb] add.l ($33,pc,d3.w*2), d0; =35 uses Opd0bb ----------
Opd0bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0bc] add.l #$33333333, d0 uses Opd0bc ----------
Opd0bc
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '#$33333333' into r10:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r10) into r11:
  mov r11,r10,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0bd] add.l ea=(111 101), d0 uses Opd0bd ----------
Opd0bd
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 101)' into r10:
;@ EaRead : Read 'ea=(111 101)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0be] add.l ea=(111 110), d0 uses Opd0be ----------
Opd0be
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 110)' into r10:
;@ EaRead : Read 'ea=(111 110)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0bf] add.l ea=(111 111), d0 uses Opd0bf ----------
Opd0bf
;@ Get r10=EA r11=EA value
;@ EaCalc : Get 'ea=(111 111)' into r10:
;@ EaRead : Read 'ea=(111 111)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0c0] adda.w d0, a0 uses Opd0c0 ----------
Opd0c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r10:
  ldrsh r10,[r7,r0]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0d0] adda.w (a0), a0 uses Opd0d0 ----------
Opd0d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0d8] adda.w (a0)+, a0 uses Opd0d8 ----------
Opd0d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0df] adda.w (a7)+, a0 uses Opd0df ----------
Opd0df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0e0] adda.w -(a0), a0 uses Opd0e0 ----------
Opd0e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0e7] adda.w -(a7), a0 uses Opd0e7 ----------
Opd0e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#2 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0e8] adda.w ($3333,a0), a0 uses Opd0e8 ----------
Opd0e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0f0] adda.w ($33,a0,d3.w*2), a0 uses Opd0f0 ----------
Opd0f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0f8] adda.w $3333.w, a0 uses Opd0f8 ----------
Opd0f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0f9] adda.w $33333333.l, a0 uses Opd0f9 ----------
Opd0f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0fa] adda.w ($3333,pc), a0; =3335 uses Opd0fa ----------
Opd0fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0fb] adda.w ($33,pc,d3.w*2), a0; =35 uses Opd0fb ----------
Opd0fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #16
  mov r10,r10,asr #16

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d0fc] adda.w #$3333, a0 uses Opd0fc ----------
Opd0fc
;@ EaCalc : Get '#$3333' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch immediate value
;@ EaRead : Read '#$3333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#4 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d100] addx.b d0, d0 uses Opd100 ----------
Opd100
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r11:
  ldrsb r11,[r7,r0]
  mov r11,r11,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mov r2,r2,lsl #28
  msr cpsr_flg,r2 ;@ Get into Carry

;@ Make sure the carry bit will tip the balance:
  ldr r2,=0x00ffffff
  orr r11,r11,r2

  adcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #24
  strb r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d110] add.b d0, (a0) uses Opd110 ----------
Opd110
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d118] add.b d0, (a0)+ uses Opd118 ----------
Opd118
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#1 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d11f] add.b d0, (a7)+ uses Opd11f ----------
Opd11f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d120] add.b d0, -(a0) uses Opd120 ----------
Opd120
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#1 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d127] add.b d0, -(a7) uses Opd127 ----------
Opd127
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#10 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d128] add.b d0, ($3333,a0) uses Opd128 ----------
Opd128
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d130] add.b d0, ($33,a0,d3.w*2) uses Opd130 ----------
Opd130
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d138] add.b d0, $3333.w uses Opd138 ----------
Opd138
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d139] add.b d0, $33333333.l uses Opd139 ----------
Opd139
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x68] ;@ Call read8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d13a] add.b d0, ($3333,pc); =3335 uses Opd13a ----------
Opd13a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d13b] add.b d0, ($33,pc,d3.w*2); =35 uses Opd13b ----------
Opd13b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x80] ;@ Call fetch8(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #24

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsb r1,[r7,r0]
  mov r1,r1,asl #24

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #24
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x74] ;@ Call write8(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d140] addx.w d0, d0 uses Opd140 ----------
Opd140
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
  mov r0,r0,lsl #2
;@ EaRead : Read register[r0] into r11:
  ldrsh r11,[r7,r0]
  mov r11,r11,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mov r2,r2,lsl #28
  msr cpsr_flg,r2 ;@ Get into Carry

;@ Make sure the carry bit will tip the balance:
  ldr r2,=0x0000ffff
  orr r11,r11,r2

  adcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  mov r1,r1,asr #16
  strh r1,[r7,r0]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d150] add.w d0, (a0) uses Opd150 ----------
Opd150
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d158] add.w d0, (a0)+ uses Opd158 ----------
Opd158
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d15f] add.w d0, (a7)+ uses Opd15f ----------
Opd15f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#12 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d160] add.w d0, -(a0) uses Opd160 ----------
Opd160
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d167] add.w d0, -(a7) uses Opd167 ----------
Opd167
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d168] add.w d0, ($3333,a0) uses Opd168 ----------
Opd168
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d170] add.w d0, ($33,a0,d3.w*2) uses Opd170 ----------
Opd170
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d178] add.w d0, $3333.w uses Opd178 ----------
Opd178
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d179] add.w d0, $33333333.l uses Opd179 ----------
Opd179
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d17a] add.w d0, ($3333,pc); =3335 uses Opd17a ----------
Opd17a
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d17b] add.w d0, ($33,pc,d3.w*2); =35 uses Opd17b ----------
Opd17b
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #16

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #7
;@ EaRead : Read register[r0] into r1:
  ldrsh r1,[r7,r0]
  mov r1,r1,asl #16

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d180] addx.l d0, d0 uses Opd180 ----------
Opd180
;@ Get r10=EA r11=EA value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r11:
  ldr r11,[r7,r0,lsl #2]

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
;@ Get X bit:
  ldrb r2,[r7,#0x45]
  mov r2,r2,lsl #28
  msr cpsr_flg,r2 ;@ Get into Carry

  adcs r1,r1,r11
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d190] add.l d0, (a0) uses Opd190 ----------
Opd190
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d198] add.l d0, (a0)+ uses Opd198 ----------
Opd198
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d19f] add.l d0, (a7)+ uses Opd19f ----------
Opd19f
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1a0] add.l d0, -(a0) uses Opd1a0 ----------
Opd1a0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1a7] add.l d0, -(a7) uses Opd1a7 ----------
Opd1a7
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#4 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1a8] add.l d0, ($3333,a0) uses Opd1a8 ----------
Opd1a8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1b0] add.l d0, ($33,a0,d3.w*2) uses Opd1b0 ----------
Opd1b0
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1b8] add.l d0, $3333.w uses Opd1b8 ----------
Opd1b8
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1b9] add.l d0, $33333333.l uses Opd1b9 ----------
Opd1b9
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1ba] add.l d0, ($3333,pc); =3335 uses Opd1ba ----------
Opd1ba
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1bb] add.l d0, ($33,pc,d3.w*2); =35 uses Opd1bb ----------
Opd1bb
;@ Get r10=EA r11=EA value
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r11:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r11,r0,asl #0

;@ Get r0=Register r1=Register value
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

;@ Do arithmetic:
  adds r1,r11,r1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save result:
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x7c] ;@ Call write32(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1c0] adda.l d0, a0 uses Opd1c0 ----------
Opd1c0
;@ EaCalc : Get register index into r0:
  and r0,r8,#0x000f
;@ EaRead : Read register[r0] into r10:
  ldr r10,[r7,r0,lsl #2]

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1d0] adda.l (a0), a0 uses Opd1d0 ----------
Opd1d0
;@ EaCalc : Get '(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
;@ EaRead : Read '(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1d8] adda.l (a0)+, a0 uses Opd1d8 ----------
Opd1d8
;@ EaCalc : Get '(a0)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1df] adda.l (a7)+, a0 uses Opd1df ----------
Opd1df
;@ EaCalc : Get '(a7)+' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r0,[r7,r2]
  add r3,r0,#4 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1e0] adda.l -(a0), a0 uses Opd1e0 ----------
Opd1e0
;@ EaCalc : Get '-(a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1e7] adda.l -(a7), a0 uses Opd1e7 ----------
Opd1e7
;@ EaCalc : Get '-(a7)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r0,[r7,r2]
  sub r0,r0,#4 ;@ Pre-decrement An
  str r0,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1e8] adda.l ($3333,a0), a0 uses Opd1e8 ----------
Opd1e8
;@ EaCalc : Get '($3333,a0)' into r0:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r0,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1f0] adda.l ($33,a0,d3.w*2), a0 uses Opd1f0 ----------
Opd1f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r0,r2,r3 ;@ r0=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1f8] adda.l $3333.w, a0 uses Opd1f8 ----------
Opd1f8
;@ EaCalc : Get '$3333.w' into r0:
  ldrsh r0,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1f9] adda.l $33333333.l, a0 uses Opd1f9 ----------
Opd1f9
;@ EaCalc : Get '$33333333.l' into r0:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x70] ;@ Call read32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1fa] adda.l ($3333,pc), a0; =3335 uses Opd1fa ----------
Opd1fa
;@ EaCalc : Get '($3333,pc)' into r0:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r0,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1fb] adda.l ($33,pc,d3.w*2), a0; =35 uses Opd1fb ----------
Opd1fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r0:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r0,r2,r0 ;@ r0=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r0) into r10:
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x88] ;@ Call fetch32(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [d1fc] adda.l #$33333333, a0 uses Opd1fc ----------
Opd1fc
;@ EaCalc : Get '#$33333333' into r0:
  ldrh r2,[r4],#2 ;@ Fetch immediate value
  ldrh r0,[r4],#2
  orr r0,r0,r2,lsl #16
;@ EaRead : Read '#$33333333' (address in r0) into r10:
  mov r10,r0,asl #0

;@ EaCalc : Get register index into r0:
  and r0,r8,#0x0e00
  mov r0,r0,lsr #9
  orr r0,r0,#0x8 ;@ A0-7
;@ EaRead : Read register[r0] into r1:
  ldr r1,[r7,r0,lsl #2]

  add r1,r1,r10

;@ EaWrite: r1 into register[r0]:
  str r1,[r7,r0,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e000] asr.b #8, d0 uses Ope000 ----------
Ope000
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,asr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e008] lsr.b #8, d0 uses Ope008 ----------
Ope008
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,lsr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e010] roxr.b #8, d0 uses Ope010 ----------
Ope010
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
  mov r2,#8
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e018] ror.b #8, d0 uses Ope018 ----------
Ope018
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #8
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e020] asr.b d0, d0 uses Ope020 ----------
Ope020
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e028] lsr.b d0, d0 uses Ope028 ----------
Ope028
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e030] roxr.b d0, d0 uses Ope030 ----------
Ope030
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e030
  subs r2,r2,#9
  bpl Reduce_e030
  add r2,r2,#9 ;@ Now r2=0-8

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e038] ror.b d0, d0 uses Ope038 ----------
Ope038
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e040] asr.w #8, d0 uses Ope040 ----------
Ope040
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e048] lsr.w #8, d0 uses Ope048 ----------
Ope048
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e050] roxr.w #8, d0 uses Ope050 ----------
Ope050
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#8
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e058] ror.w #8, d0 uses Ope058 ----------
Ope058
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #8
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e060] asr.w d0, d0 uses Ope060 ----------
Ope060
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e068] lsr.w d0, d0 uses Ope068 ----------
Ope068
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e070] roxr.w d0, d0 uses Ope070 ----------
Ope070
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e070
  subs r2,r2,#17
  bpl Reduce_e070
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e078] ror.w d0, d0 uses Ope078 ----------
Ope078
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e080] asr.l #8, d0 uses Ope080 ----------
Ope080
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Shift register:
  movs r0,r0,asr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e088] lsr.l #8, d0 uses Ope088 ----------
Ope088
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Shift register:
  movs r0,r0,lsr #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e090] roxr.l #8, d0 uses Ope090 ----------
Ope090
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Rotate register through X:
  mov r2,#8
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e098] ror.l #8, d0 uses Ope098 ----------
Ope098
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Rotate register:
  movs r0,r0,ror #8
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0a0] asr.l d0, d0 uses Ope0a0 ----------
Ope0a0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0a8] lsr.l d0, d0 uses Ope0a8 ----------
Ope0a8
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0b0] roxr.l d0, d0 uses Ope0b0 ----------
Ope0b0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e0b0
  subs r2,r2,#33
  bpl Reduce_e0b0
  add r2,r2,#33 ;@ Now r2=0-32

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0b8] ror.l d0, d0 uses Ope0b8 ----------
Ope0b8
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0d0] asr.w (a0) uses Ope0d0 ----------
Ope0d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0d8] asr.w (a0)+ uses Ope0d8 ----------
Ope0d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0df] asr.w (a7)+ uses Ope0df ----------
Ope0df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0e0] asr.w -(a0) uses Ope0e0 ----------
Ope0e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0e7] asr.w -(a7) uses Ope0e7 ----------
Ope0e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0e8] asr.w ($3333,a0) uses Ope0e8 ----------
Ope0e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0f0] asr.w ($33,a0,d3.w*2) uses Ope0f0 ----------
Ope0f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0f8] asr.w $3333.w uses Ope0f8 ----------
Ope0f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0f9] asr.w $33333333.l uses Ope0f9 ----------
Ope0f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0fa] asr.w ($3333,pc); =3335 uses Ope0fa ----------
Ope0fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e0fb] asr.w ($33,pc,d3.w*2); =35 uses Ope0fb ----------
Ope0fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e100] asl.b #8, d0 uses Ope100 ----------
Ope100
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Shift register:
  movs r0,r0,asl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e108] lsl.b #8, d0 uses Ope108 ----------
Ope108
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Shift register:
  movs r0,r0,lsl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e110] roxl.b #8, d0 uses Ope110 ----------
Ope110
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
  mov r2,#8
  rsb r2,r2,#9 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e118] rol.b #8, d0 uses Ope118 ----------
Ope118
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #24
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e120] asl.b d0, d0 uses Ope120 ----------
Ope120
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e128] lsl.b d0, d0 uses Ope128 ----------
Ope128
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e130] roxl.b d0, d0 uses Ope130 ----------
Ope130
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e130
  subs r2,r2,#9
  bpl Reduce_e130
  add r2,r2,#9 ;@ Now r2=0-8

  rsb r2,r2,#9 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e138] rol.b d0, d0 uses Ope138 ----------
Ope138
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e140] asl.w #8, d0 uses Ope140 ----------
Ope140
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e148] lsl.w #8, d0 uses Ope148 ----------
Ope148
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e150] roxl.w #8, d0 uses Ope150 ----------
Ope150
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#8
  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e158] rol.w #8, d0 uses Ope158 ----------
Ope158
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #24
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e160] asl.w d0, d0 uses Ope160 ----------
Ope160
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e168] lsl.w d0, d0 uses Ope168 ----------
Ope168
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e170] roxl.w d0, d0 uses Ope170 ----------
Ope170
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e170
  subs r2,r2,#17
  bpl Reduce_e170
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e178] rol.w d0, d0 uses Ope178 ----------
Ope178
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e180] asl.l #8, d0 uses Ope180 ----------
Ope180
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Shift register:
  movs r0,r0,asl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e188] lsl.l #8, d0 uses Ope188 ----------
Ope188
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Shift register:
  movs r0,r0,lsl #8
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e190] roxl.l #8, d0 uses Ope190 ----------
Ope190
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Rotate register through X:
  mov r2,#8
  rsb r2,r2,#33 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e198] rol.l #8, d0 uses Ope198 ----------
Ope198
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

;@ Rotate register:
  movs r0,r0,ror #24
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#24 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1a0] asl.l d0, d0 uses Ope1a0 ----------
Ope1a0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1a8] lsl.l d0, d0 uses Ope1a8 ----------
Ope1a8
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1b0] roxl.l d0, d0 uses Ope1b0 ----------
Ope1b0
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_e1b0
  subs r2,r2,#33
  bpl Reduce_e1b0
  add r2,r2,#33 ;@ Now r2=0-32

  rsb r2,r2,#33 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1b8] rol.l d0, d0 uses Ope1b8 ----------
Ope1b8
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

;@ Use Dn for count:
  ldr r2,[r7,r2,lsl #2]
  and r2,r2,#63

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1d0] asl.w (a0) uses Ope1d0 ----------
Ope1d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1d8] asl.w (a0)+ uses Ope1d8 ----------
Ope1d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1df] asl.w (a7)+ uses Ope1df ----------
Ope1df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1e0] asl.w -(a0) uses Ope1e0 ----------
Ope1e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1e7] asl.w -(a7) uses Ope1e7 ----------
Ope1e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1e8] asl.w ($3333,a0) uses Ope1e8 ----------
Ope1e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1f0] asl.w ($33,a0,d3.w*2) uses Ope1f0 ----------
Ope1f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1f8] asl.w $3333.w uses Ope1f8 ----------
Ope1f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1f9] asl.w $33333333.l uses Ope1f9 ----------
Ope1f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1fa] asl.w ($3333,pc); =3335 uses Ope1fa ----------
Ope1fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e1fb] asl.w ($33,pc,d3.w*2); =35 uses Ope1fb ----------
Ope1fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,asl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2d0] lsr.w (a0) uses Ope2d0 ----------
Ope2d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2d8] lsr.w (a0)+ uses Ope2d8 ----------
Ope2d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2df] lsr.w (a7)+ uses Ope2df ----------
Ope2df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2e0] lsr.w -(a0) uses Ope2e0 ----------
Ope2e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2e7] lsr.w -(a7) uses Ope2e7 ----------
Ope2e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2e8] lsr.w ($3333,a0) uses Ope2e8 ----------
Ope2e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2f0] lsr.w ($33,a0,d3.w*2) uses Ope2f0 ----------
Ope2f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2f8] lsr.w $3333.w uses Ope2f8 ----------
Ope2f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2f9] lsr.w $33333333.l uses Ope2f9 ----------
Ope2f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2fa] lsr.w ($3333,pc); =3335 uses Ope2fa ----------
Ope2fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e2fb] lsr.w ($33,pc,d3.w*2); =35 uses Ope2fb ----------
Ope2fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3d0] lsl.w (a0) uses Ope3d0 ----------
Ope3d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3d8] lsl.w (a0)+ uses Ope3d8 ----------
Ope3d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3df] lsl.w (a7)+ uses Ope3df ----------
Ope3df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3e0] lsl.w -(a0) uses Ope3e0 ----------
Ope3e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3e7] lsl.w -(a7) uses Ope3e7 ----------
Ope3e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3e8] lsl.w ($3333,a0) uses Ope3e8 ----------
Ope3e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3f0] lsl.w ($33,a0,d3.w*2) uses Ope3f0 ----------
Ope3f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3f8] lsl.w $3333.w uses Ope3f8 ----------
Ope3f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3f9] lsl.w $33333333.l uses Ope3f9 ----------
Ope3f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3fa] lsl.w ($3333,pc); =3335 uses Ope3fa ----------
Ope3fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e3fb] lsl.w ($33,pc,d3.w*2); =35 uses Ope3fb ----------
Ope3fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Shift register:
  movs r0,r0,lsl #1
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4d0] roxr.w (a0) uses Ope4d0 ----------
Ope4d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4d0
  subs r2,r2,#17
  bpl Reduce_e4d0
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4d8] roxr.w (a0)+ uses Ope4d8 ----------
Ope4d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4d8
  subs r2,r2,#17
  bpl Reduce_e4d8
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4df] roxr.w (a7)+ uses Ope4df ----------
Ope4df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4df
  subs r2,r2,#17
  bpl Reduce_e4df
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4e0] roxr.w -(a0) uses Ope4e0 ----------
Ope4e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4e0
  subs r2,r2,#17
  bpl Reduce_e4e0
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4e7] roxr.w -(a7) uses Ope4e7 ----------
Ope4e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4e7
  subs r2,r2,#17
  bpl Reduce_e4e7
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4e8] roxr.w ($3333,a0) uses Ope4e8 ----------
Ope4e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4e8
  subs r2,r2,#17
  bpl Reduce_e4e8
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4f0] roxr.w ($33,a0,d3.w*2) uses Ope4f0 ----------
Ope4f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4f0
  subs r2,r2,#17
  bpl Reduce_e4f0
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4f8] roxr.w $3333.w uses Ope4f8 ----------
Ope4f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4f8
  subs r2,r2,#17
  bpl Reduce_e4f8
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4f9] roxr.w $33333333.l uses Ope4f9 ----------
Ope4f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4f9
  subs r2,r2,#17
  bpl Reduce_e4f9
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4fa] roxr.w ($3333,pc); =3335 uses Ope4fa ----------
Ope4fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4fa
  subs r2,r2,#17
  bpl Reduce_e4fa
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e4fb] roxr.w ($33,pc,d3.w*2); =35 uses Ope4fb ----------
Ope4fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e4fb
  subs r2,r2,#17
  bpl Reduce_e4fb
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5d0] roxl.w (a0) uses Ope5d0 ----------
Ope5d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5d0
  subs r2,r2,#17
  bpl Reduce_e5d0
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5d8] roxl.w (a0)+ uses Ope5d8 ----------
Ope5d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5d8
  subs r2,r2,#17
  bpl Reduce_e5d8
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5df] roxl.w (a7)+ uses Ope5df ----------
Ope5df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5df
  subs r2,r2,#17
  bpl Reduce_e5df
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5e0] roxl.w -(a0) uses Ope5e0 ----------
Ope5e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5e0
  subs r2,r2,#17
  bpl Reduce_e5e0
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5e7] roxl.w -(a7) uses Ope5e7 ----------
Ope5e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5e7
  subs r2,r2,#17
  bpl Reduce_e5e7
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5e8] roxl.w ($3333,a0) uses Ope5e8 ----------
Ope5e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5e8
  subs r2,r2,#17
  bpl Reduce_e5e8
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5f0] roxl.w ($33,a0,d3.w*2) uses Ope5f0 ----------
Ope5f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5f0
  subs r2,r2,#17
  bpl Reduce_e5f0
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5f8] roxl.w $3333.w uses Ope5f8 ----------
Ope5f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5f8
  subs r2,r2,#17
  bpl Reduce_e5f8
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5f9] roxl.w $33333333.l uses Ope5f9 ----------
Ope5f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5f9
  subs r2,r2,#17
  bpl Reduce_e5f9
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5fa] roxl.w ($3333,pc); =3335 uses Ope5fa ----------
Ope5fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5fa
  subs r2,r2,#17
  bpl Reduce_e5fa
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e5fb] roxl.w ($33,pc,d3.w*2); =35 uses Ope5fb ----------
Ope5fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
  mov r2,#1
;@ Reduce r2 until <0:
Reduce_e5fb
  subs r2,r2,#17
  bpl Reduce_e5fb
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6d0] ror.w (a0) uses Ope6d0 ----------
Ope6d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6d8] ror.w (a0)+ uses Ope6d8 ----------
Ope6d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6df] ror.w (a7)+ uses Ope6df ----------
Ope6df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6e0] ror.w -(a0) uses Ope6e0 ----------
Ope6e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6e7] ror.w -(a7) uses Ope6e7 ----------
Ope6e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6e8] ror.w ($3333,a0) uses Ope6e8 ----------
Ope6e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6f0] ror.w ($33,a0,d3.w*2) uses Ope6f0 ----------
Ope6f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6f8] ror.w $3333.w uses Ope6f8 ----------
Ope6f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6f9] ror.w $33333333.l uses Ope6f9 ----------
Ope6f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6fa] ror.w ($3333,pc); =3335 uses Ope6fa ----------
Ope6fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e6fb] ror.w ($33,pc,d3.w*2); =35 uses Ope6fb ----------
Ope6fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #1
  mrs r9,cpsr ;@ r9=flags

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7d0] rol.w (a0) uses Ope7d0 ----------
Ope7d0
;@ EaCalc : Get '(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
;@ EaRead : Read '(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7d8] rol.w (a0)+ uses Ope7d8 ----------
Ope7d8
;@ EaCalc : Get '(a0)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a0)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a0)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7df] rol.w (a7)+ uses Ope7df ----------
Ope7df
;@ EaCalc : Get '(a7)+' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r10,[r7,r2]
  add r3,r10,#2 ;@ Post-increment An
  str r3,[r7,r2]
;@ EaRead : Read '(a7)+' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '(a7)+' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#14 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7e0] rol.w -(a0) uses Ope7e0 ----------
Ope7e0
;@ EaCalc : Get '-(a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7e7] rol.w -(a7) uses Ope7e7 ----------
Ope7e7
;@ EaCalc : Get '-(a7)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r10,[r7,r2]
  sub r10,r10,#2 ;@ Pre-decrement An
  str r10,[r7,r2]
;@ EaRead : Read '-(a7)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '-(a7)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#16 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7e8] rol.w ($3333,a0) uses Ope7e8 ----------
Ope7e8
;@ EaCalc : Get '($3333,a0)' into r10:
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  ldr r2,[r7,r2]
  ldrsh r0,[r4],#2 ;@ Fetch offset
  add r10,r0,r2 ;@ Add on offset
;@ EaRead : Read '($3333,a0)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,a0)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7f0] rol.w ($33,a0,d3.w*2) uses Ope7f0 ----------
Ope7f0
;@ EaCalc : Get '($33,a0,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4],#2 ;@ ($Disp,PC,Rn)
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r3,r2,r0,asr #24 ;@ r3=Disp+Rn
  and r2,r8,#0x000f
  mov r2,r2,lsl #2
  orr r2,r2,#0x20 ;@ A0-7
  ldr r2,[r7,r2]
  add r10,r2,r3 ;@ r10=Disp+An+Rn
;@ EaRead : Read '($33,a0,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,a0,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7f8] rol.w $3333.w uses Ope7f8 ----------
Ope7f8
;@ EaCalc : Get '$3333.w' into r10:
  ldrsh r10,[r4],#2 ;@ Fetch Absolute Short address
;@ EaRead : Read '$3333.w' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$3333.w' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7f9] rol.w $33333333.l uses Ope7f9 ----------
Ope7f9
;@ EaCalc : Get '$33333333.l' into r10:
  ldrh r2,[r4],#2 ;@ Fetch Absolute Long address
  ldrh r0,[r4],#2
  orr r10,r0,r2,lsl #16
;@ EaRead : Read '$33333333.l' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x6c] ;@ Call read16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '$33333333.l' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#22 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7fa] rol.w ($3333,pc); =3335 uses Ope7fa ----------
Ope7fa
;@ EaCalc : Get '($3333,pc)' into r10:
  ldr r0,[r7,#0x60] ;@ Get Memory base
  sub r0,r4,r0 ;@ Real PC
  ldrsh r2,[r4],#2 ;@ Fetch extension
  add r10,r0,r2 ;@ ($nn,PC)
;@ EaRead : Read '($3333,pc)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($3333,pc)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#18 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [e7fb] rol.w ($33,pc,d3.w*2); =35 uses Ope7fb ----------
Ope7fb
;@ EaCalc : Get '($33,pc,d3.w*2)' into r10:
;@ Get extension word into r3:
  ldrh r3,[r4]
  mov r2,r3,lsr #10
  tst r3,#0x0800 ;@ Is Rn Word or Long
  and r2,r2,#0x3c ;@ r2=Index of Rn
  mov r0,r3,asl #24 ;@ r0=Get 8-bit signed Disp
  ldreqsh r2,[r7,r2] ;@ r2=Rn.w
  ldrne   r2,[r7,r2] ;@ r2=Rn.l
  add r2,r2,r0,asr #24 ;@ r2=Disp+Rn
  ldr r0,[r7,#0x60] ;@ Get Memory base
  add r2,r2,r4 ;@ r2=Disp+Rn + Base+PC
  add r4,r4,#2 ;@ Increase PC
  sub r10,r2,r0 ;@ r10=Disp+PC+Rn
;@ EaRead : Read '($33,pc,d3.w*2)' (address in r10) into r0:
  mov r0,r10
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x84] ;@ Call fetch16(r0) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}
  mov r0,r0,asl #16

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror #31
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ Save shifted value back to EA:
  mov r1,r0
;@ EaWrite: Write r1 into '($33,pc,d3.w*2)' (address in r10):
  mov r0,r10
  mov r1,r1,asr #16
  ;@ Save R9,R10
  stmfd sp!,{r9,r10}
  ;@ Restore R9(PALMOS) & R10(GLOBAL VAR ACCESS with CW)
  ldr r9,[r7,#0x8C]
  mov r10,r12
  mov lr,pc
  ldr pc,[r7,#0x78] ;@ Call write16(r0,r1) handler
  ;@ Restore R9,R10
  mov	r12,r10
  ldmfd sp!,{r9,r10}

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#20 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee00] asr.b #7, d0 uses Opee00 ----------
Opee00
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee08] lsr.b #7, d0 uses Opee08 ----------
Opee08
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #24

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee10] roxr.b #7, d0 uses Opee10 ----------
Opee10
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ee10
  subs r2,r2,#9
  bpl Reduce_ee10
  add r2,r2,#9 ;@ Now r2=0-8

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee18] ror.b #7, d0 uses Opee18 ----------
Opee18
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee40] asr.w #7, d0 uses Opee40 ----------
Opee40
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee48] lsr.w #7, d0 uses Opee48 ----------
Opee48
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ For shift right, also copy to lowest bits (to get carry bit):
  orr r0,r0,r0,lsr #16

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee50] roxr.w #7, d0 uses Opee50 ----------
Opee50
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ee50
  subs r2,r2,#17
  bpl Reduce_ee50
  add r2,r2,#17 ;@ Now r2=0-16

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee58] ror.w #7, d0 uses Opee58 ----------
Opee58
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee80] asr.l #7, d0 uses Opee80 ----------
Opee80
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee88] lsr.l #7, d0 uses Opee88 ----------
Opee88
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsr r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee90] roxr.l #7, d0 uses Opee90 ----------
Opee90
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ee90
  subs r2,r2,#33
  bpl Reduce_ee90
  add r2,r2,#33 ;@ Now r2=0-32

;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ee98] ror.l #7, d0 uses Opee98 ----------
Opee98
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register:
  movs r0,r0,ror r2
  mrs r9,cpsr ;@ r9=flags

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef00] asl.b #7, d0 uses Opef00 ----------
Opef00
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef08] lsl.b #7, d0 uses Opef08 ----------
Opef08
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #24
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef10] roxl.b #7, d0 uses Opef10 ----------
Opef10
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #24 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ef10
  subs r2,r2,#9
  bpl Reduce_ef10
  add r2,r2,#9 ;@ Now r2=0-8

  rsb r2,r2,#9 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#9
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #24 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef18] rol.b #7, d0 uses Opef18 ----------
Opef18
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsb r0,[r7,r10]
  mov r0,r0,asl #24

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #8
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #24
  strb r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef40] asl.w #7, d0 uses Opef40 ----------
Opef40
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef48] lsl.w #7, d0 uses Opef48 ----------
Opef48
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ Check if result is zero:
  movs r2,r0,lsr #16
  orreq r9,r9,#0x40000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef50] roxl.w #7, d0 uses Opef50 ----------
Opef50
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

  mov r0,r0,lsr #16 ;@ Shift down
;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ef50
  subs r2,r2,#17
  bpl Reduce_ef50
  add r2,r2,#17 ;@ Now r2=0-16

  rsb r2,r2,#17 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#17
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  movs r0,r0,lsl #16 ;@ Shift up and get correct NC flags
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef58] rol.w #7, d0 uses Opef58 ----------
Opef58
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
  mov r10,r10,lsl #2
;@ EaRead : Read register[r10] into r0:
  ldrsh r0,[r7,r10]
  mov r0,r0,asl #16

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Mirror value in whole 32 bits:
  orr r0,r0,r0,lsr #16

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  mov r0,r0,asr #16
  strh r0,[r7,r10]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#6 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef80] asl.l #7, d0 uses Opef80 ----------
Opef80
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,asl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef88] lsl.l #7, d0 uses Opef88 ----------
Opef88
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Shift register:
  movs r0,r0,lsl r2
  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef90] roxl.l #7, d0 uses Opef90 ----------
Opef90
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register through X:
;@ Reduce r2 until <0:
Reduce_ef90
  subs r2,r2,#33
  bpl Reduce_ef90
  add r2,r2,#33 ;@ Now r2=0-32

  rsb r2,r2,#33 ;@ Reverse direction
;@ Rotate bits:
  mov r3,r0,lsr r2 ;@ Get right part
  rsb r2,r2,#33
  movs r0,r0,lsl r2 ;@ Get left part
  orr r0,r3,r0 ;@ r0=Rotated value
;@ Insert X bit into r2-1:
  ldrb r3,[r7,#0x45]
  sub r2,r2,#1
  and r3,r3,#2
  mov r3,r3,lsr #1
  orr r0,r0,r3,lsl r2

  mrs r9,cpsr ;@ r9=flags
  mov r2,r9,lsr #28
  strb r2,[r7,#0x45] ;@ Save X bit

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG

;@ ---------- [ef98] rol.l #7, d0 uses Opef98 ----------
Opef98
;@ EaCalc : Get register index into r10:
  and r10,r8,#0x0007
;@ EaRead : Read register[r10] into r0:
  ldr r0,[r7,r10,lsl #2]

  mov r2,r8,lsr #9 ;@ Get 'n'
  and r2,r2,#7

  sub r5,r5,r2,asl #1 ;@ Take 2*n cycles

;@ Rotate register:
  rsb r2,r2,#32
  movs r0,r0,ror r2
;@ Get carry bit from bit 0:
  mov r9,#0
  ands r2,r0,#1
  orrne r9,r9,#0x20000000

;@ EaWrite: r0 into register[r10]:
  str r0,[r7,r10,lsl #2]

  ldrh r8,[r4],#2 ;@ Fetch next opcode
  subs r5,r5,#8 ;@ Subtract cycles
  ldrge r0,[r6,r8,asl #2] ;@ Jump to opcode handler
  addge pc,r0,r12         ;@ Jump to opcode handler
  b CycloneEnd
  LTORG


;@ -------------------------- Jump Table --------------------------
JumpTab
  dcd Op0000,Op0000,Op0000,Op0000,Op0000,Op0000,Op0000,Op0000 ;@ 0000
  dcd Op0000,Op0000,Op0000,Op0000,Op0000,Op0000,Op0000,Op0000 ;@ 0008
  dcd Op0010,Op0010,Op0010,Op0010,Op0010,Op0010,Op0010,Op0010 ;@ 0010
  dcd Op0018,Op0018,Op0018,Op0018,Op0018,Op0018,Op0018,Op001f ;@ 0018
  dcd Op0020,Op0020,Op0020,Op0020,Op0020,Op0020,Op0020,Op0027 ;@ 0020
  dcd Op0028,Op0028,Op0028,Op0028,Op0028,Op0028,Op0028,Op0028 ;@ 0028
  dcd Op0030,Op0030,Op0030,Op0030,Op0030,Op0030,Op0030,Op0030 ;@ 0030
  dcd Op0038,Op0039,Op003a,Op003b,Op003c,Op____,Op____,Op____ ;@ 0038
  dcd Op0040,Op0040,Op0040,Op0040,Op0040,Op0040,Op0040,Op0040 ;@ 0040
  dcd Op0040,Op0040,Op0040,Op0040,Op0040,Op0040,Op0040,Op0040 ;@ 0048
  dcd Op0050,Op0050,Op0050,Op0050,Op0050,Op0050,Op0050,Op0050 ;@ 0050
  dcd Op0058,Op0058,Op0058,Op0058,Op0058,Op0058,Op0058,Op005f ;@ 0058
  dcd Op0060,Op0060,Op0060,Op0060,Op0060,Op0060,Op0060,Op0067 ;@ 0060
  dcd Op0068,Op0068,Op0068,Op0068,Op0068,Op0068,Op0068,Op0068 ;@ 0068
  dcd Op0070,Op0070,Op0070,Op0070,Op0070,Op0070,Op0070,Op0070 ;@ 0070
  dcd Op0078,Op0079,Op007a,Op007b,Op007c,Op____,Op____,Op____ ;@ 0078
  dcd Op0080,Op0080,Op0080,Op0080,Op0080,Op0080,Op0080,Op0080 ;@ 0080
  dcd Op0080,Op0080,Op0080,Op0080,Op0080,Op0080,Op0080,Op0080 ;@ 0088
  dcd Op0090,Op0090,Op0090,Op0090,Op0090,Op0090,Op0090,Op0090 ;@ 0090
  dcd Op0098,Op0098,Op0098,Op0098,Op0098,Op0098,Op0098,Op009f ;@ 0098
  dcd Op00a0,Op00a0,Op00a0,Op00a0,Op00a0,Op00a0,Op00a0,Op00a7 ;@ 00a0
  dcd Op00a8,Op00a8,Op00a8,Op00a8,Op00a8,Op00a8,Op00a8,Op00a8 ;@ 00a8
  dcd Op00b0,Op00b0,Op00b0,Op00b0,Op00b0,Op00b0,Op00b0,Op00b0 ;@ 00b0
  dcd Op00b8,Op00b9,Op00ba,Op00bb,Op____,Op____,Op____,Op____ ;@ 00b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 00f8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0108
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0110
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0118
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0120
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0128
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0130
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0138
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0148
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0150
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0158
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0160
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0168
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0170
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0178
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0188
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0190
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0198
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 01a0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 01a8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 01b0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 01b8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 01c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 01c8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 01d0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 01d8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 01e0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 01e8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 01f0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 01f8
  dcd Op0200,Op0200,Op0200,Op0200,Op0200,Op0200,Op0200,Op0200 ;@ 0200
  dcd Op0200,Op0200,Op0200,Op0200,Op0200,Op0200,Op0200,Op0200 ;@ 0208
  dcd Op0210,Op0210,Op0210,Op0210,Op0210,Op0210,Op0210,Op0210 ;@ 0210
  dcd Op0218,Op0218,Op0218,Op0218,Op0218,Op0218,Op0218,Op021f ;@ 0218
  dcd Op0220,Op0220,Op0220,Op0220,Op0220,Op0220,Op0220,Op0227 ;@ 0220
  dcd Op0228,Op0228,Op0228,Op0228,Op0228,Op0228,Op0228,Op0228 ;@ 0228
  dcd Op0230,Op0230,Op0230,Op0230,Op0230,Op0230,Op0230,Op0230 ;@ 0230
  dcd Op0238,Op0239,Op023a,Op023b,Op023c,Op____,Op____,Op____ ;@ 0238
  dcd Op0240,Op0240,Op0240,Op0240,Op0240,Op0240,Op0240,Op0240 ;@ 0240
  dcd Op0240,Op0240,Op0240,Op0240,Op0240,Op0240,Op0240,Op0240 ;@ 0248
  dcd Op0250,Op0250,Op0250,Op0250,Op0250,Op0250,Op0250,Op0250 ;@ 0250
  dcd Op0258,Op0258,Op0258,Op0258,Op0258,Op0258,Op0258,Op025f ;@ 0258
  dcd Op0260,Op0260,Op0260,Op0260,Op0260,Op0260,Op0260,Op0267 ;@ 0260
  dcd Op0268,Op0268,Op0268,Op0268,Op0268,Op0268,Op0268,Op0268 ;@ 0268
  dcd Op0270,Op0270,Op0270,Op0270,Op0270,Op0270,Op0270,Op0270 ;@ 0270
  dcd Op0278,Op0279,Op027a,Op027b,Op027c,Op____,Op____,Op____ ;@ 0278
  dcd Op0280,Op0280,Op0280,Op0280,Op0280,Op0280,Op0280,Op0280 ;@ 0280
  dcd Op0280,Op0280,Op0280,Op0280,Op0280,Op0280,Op0280,Op0280 ;@ 0288
  dcd Op0290,Op0290,Op0290,Op0290,Op0290,Op0290,Op0290,Op0290 ;@ 0290
  dcd Op0298,Op0298,Op0298,Op0298,Op0298,Op0298,Op0298,Op029f ;@ 0298
  dcd Op02a0,Op02a0,Op02a0,Op02a0,Op02a0,Op02a0,Op02a0,Op02a7 ;@ 02a0
  dcd Op02a8,Op02a8,Op02a8,Op02a8,Op02a8,Op02a8,Op02a8,Op02a8 ;@ 02a8
  dcd Op02b0,Op02b0,Op02b0,Op02b0,Op02b0,Op02b0,Op02b0,Op02b0 ;@ 02b0
  dcd Op02b8,Op02b9,Op02ba,Op02bb,Op____,Op____,Op____,Op____ ;@ 02b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 02f8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0308
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0310
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0318
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0320
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0328
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0330
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0338
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0348
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0350
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0358
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0360
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0368
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0370
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0378
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0388
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0390
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0398
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 03a0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 03a8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 03b0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 03b8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 03c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 03c8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 03d0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 03d8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 03e0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 03e8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 03f0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 03f8
  dcd Op0400,Op0400,Op0400,Op0400,Op0400,Op0400,Op0400,Op0400 ;@ 0400
  dcd Op0400,Op0400,Op0400,Op0400,Op0400,Op0400,Op0400,Op0400 ;@ 0408
  dcd Op0410,Op0410,Op0410,Op0410,Op0410,Op0410,Op0410,Op0410 ;@ 0410
  dcd Op0418,Op0418,Op0418,Op0418,Op0418,Op0418,Op0418,Op041f ;@ 0418
  dcd Op0420,Op0420,Op0420,Op0420,Op0420,Op0420,Op0420,Op0427 ;@ 0420
  dcd Op0428,Op0428,Op0428,Op0428,Op0428,Op0428,Op0428,Op0428 ;@ 0428
  dcd Op0430,Op0430,Op0430,Op0430,Op0430,Op0430,Op0430,Op0430 ;@ 0430
  dcd Op0438,Op0439,Op043a,Op043b,Op____,Op____,Op____,Op____ ;@ 0438
  dcd Op0440,Op0440,Op0440,Op0440,Op0440,Op0440,Op0440,Op0440 ;@ 0440
  dcd Op0440,Op0440,Op0440,Op0440,Op0440,Op0440,Op0440,Op0440 ;@ 0448
  dcd Op0450,Op0450,Op0450,Op0450,Op0450,Op0450,Op0450,Op0450 ;@ 0450
  dcd Op0458,Op0458,Op0458,Op0458,Op0458,Op0458,Op0458,Op045f ;@ 0458
  dcd Op0460,Op0460,Op0460,Op0460,Op0460,Op0460,Op0460,Op0467 ;@ 0460
  dcd Op0468,Op0468,Op0468,Op0468,Op0468,Op0468,Op0468,Op0468 ;@ 0468
  dcd Op0470,Op0470,Op0470,Op0470,Op0470,Op0470,Op0470,Op0470 ;@ 0470
  dcd Op0478,Op0479,Op047a,Op047b,Op____,Op____,Op____,Op____ ;@ 0478
  dcd Op0480,Op0480,Op0480,Op0480,Op0480,Op0480,Op0480,Op0480 ;@ 0480
  dcd Op0480,Op0480,Op0480,Op0480,Op0480,Op0480,Op0480,Op0480 ;@ 0488
  dcd Op0490,Op0490,Op0490,Op0490,Op0490,Op0490,Op0490,Op0490 ;@ 0490
  dcd Op0498,Op0498,Op0498,Op0498,Op0498,Op0498,Op0498,Op049f ;@ 0498
  dcd Op04a0,Op04a0,Op04a0,Op04a0,Op04a0,Op04a0,Op04a0,Op04a7 ;@ 04a0
  dcd Op04a8,Op04a8,Op04a8,Op04a8,Op04a8,Op04a8,Op04a8,Op04a8 ;@ 04a8
  dcd Op04b0,Op04b0,Op04b0,Op04b0,Op04b0,Op04b0,Op04b0,Op04b0 ;@ 04b0
  dcd Op04b8,Op04b9,Op04ba,Op04bb,Op____,Op____,Op____,Op____ ;@ 04b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 04f8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0508
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0510
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0518
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0520
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0528
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0530
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0538
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0548
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0550
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0558
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0560
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0568
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0570
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0578
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0588
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0590
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0598
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 05a0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 05a8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 05b0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 05b8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 05c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 05c8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 05d0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 05d8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 05e0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 05e8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 05f0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 05f8
  dcd Op0600,Op0600,Op0600,Op0600,Op0600,Op0600,Op0600,Op0600 ;@ 0600
  dcd Op0600,Op0600,Op0600,Op0600,Op0600,Op0600,Op0600,Op0600 ;@ 0608
  dcd Op0610,Op0610,Op0610,Op0610,Op0610,Op0610,Op0610,Op0610 ;@ 0610
  dcd Op0618,Op0618,Op0618,Op0618,Op0618,Op0618,Op0618,Op061f ;@ 0618
  dcd Op0620,Op0620,Op0620,Op0620,Op0620,Op0620,Op0620,Op0627 ;@ 0620
  dcd Op0628,Op0628,Op0628,Op0628,Op0628,Op0628,Op0628,Op0628 ;@ 0628
  dcd Op0630,Op0630,Op0630,Op0630,Op0630,Op0630,Op0630,Op0630 ;@ 0630
  dcd Op0638,Op0639,Op063a,Op063b,Op____,Op____,Op____,Op____ ;@ 0638
  dcd Op0640,Op0640,Op0640,Op0640,Op0640,Op0640,Op0640,Op0640 ;@ 0640
  dcd Op0640,Op0640,Op0640,Op0640,Op0640,Op0640,Op0640,Op0640 ;@ 0648
  dcd Op0650,Op0650,Op0650,Op0650,Op0650,Op0650,Op0650,Op0650 ;@ 0650
  dcd Op0658,Op0658,Op0658,Op0658,Op0658,Op0658,Op0658,Op065f ;@ 0658
  dcd Op0660,Op0660,Op0660,Op0660,Op0660,Op0660,Op0660,Op0667 ;@ 0660
  dcd Op0668,Op0668,Op0668,Op0668,Op0668,Op0668,Op0668,Op0668 ;@ 0668
  dcd Op0670,Op0670,Op0670,Op0670,Op0670,Op0670,Op0670,Op0670 ;@ 0670
  dcd Op0678,Op0679,Op067a,Op067b,Op____,Op____,Op____,Op____ ;@ 0678
  dcd Op0680,Op0680,Op0680,Op0680,Op0680,Op0680,Op0680,Op0680 ;@ 0680
  dcd Op0680,Op0680,Op0680,Op0680,Op0680,Op0680,Op0680,Op0680 ;@ 0688
  dcd Op0690,Op0690,Op0690,Op0690,Op0690,Op0690,Op0690,Op0690 ;@ 0690
  dcd Op0698,Op0698,Op0698,Op0698,Op0698,Op0698,Op0698,Op069f ;@ 0698
  dcd Op06a0,Op06a0,Op06a0,Op06a0,Op06a0,Op06a0,Op06a0,Op06a7 ;@ 06a0
  dcd Op06a8,Op06a8,Op06a8,Op06a8,Op06a8,Op06a8,Op06a8,Op06a8 ;@ 06a8
  dcd Op06b0,Op06b0,Op06b0,Op06b0,Op06b0,Op06b0,Op06b0,Op06b0 ;@ 06b0
  dcd Op06b8,Op06b9,Op06ba,Op06bb,Op____,Op____,Op____,Op____ ;@ 06b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 06f8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0708
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0710
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0718
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0720
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0728
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0730
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0738
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0748
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0750
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0758
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0760
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0768
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0770
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0778
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0788
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0790
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0798
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 07a0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 07a8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 07b0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 07b8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 07c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 07c8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 07d0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 07d8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 07e0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 07e8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 07f0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 07f8
  dcd Op0800,Op0800,Op0800,Op0800,Op0800,Op0800,Op0800,Op0800 ;@ 0800
  dcd Op0800,Op0800,Op0800,Op0800,Op0800,Op0800,Op0800,Op0800 ;@ 0808
  dcd Op0810,Op0810,Op0810,Op0810,Op0810,Op0810,Op0810,Op0810 ;@ 0810
  dcd Op0818,Op0818,Op0818,Op0818,Op0818,Op0818,Op0818,Op081f ;@ 0818
  dcd Op0820,Op0820,Op0820,Op0820,Op0820,Op0820,Op0820,Op0827 ;@ 0820
  dcd Op0828,Op0828,Op0828,Op0828,Op0828,Op0828,Op0828,Op0828 ;@ 0828
  dcd Op0830,Op0830,Op0830,Op0830,Op0830,Op0830,Op0830,Op0830 ;@ 0830
  dcd Op0838,Op0839,Op083a,Op083b,Op083c,Op____,Op____,Op____ ;@ 0838
  dcd Op0840,Op0840,Op0840,Op0840,Op0840,Op0840,Op0840,Op0840 ;@ 0840
  dcd Op0840,Op0840,Op0840,Op0840,Op0840,Op0840,Op0840,Op0840 ;@ 0848
  dcd Op0850,Op0850,Op0850,Op0850,Op0850,Op0850,Op0850,Op0850 ;@ 0850
  dcd Op0858,Op0858,Op0858,Op0858,Op0858,Op0858,Op0858,Op085f ;@ 0858
  dcd Op0860,Op0860,Op0860,Op0860,Op0860,Op0860,Op0860,Op0867 ;@ 0860
  dcd Op0868,Op0868,Op0868,Op0868,Op0868,Op0868,Op0868,Op0868 ;@ 0868
  dcd Op0870,Op0870,Op0870,Op0870,Op0870,Op0870,Op0870,Op0870 ;@ 0870
  dcd Op0878,Op0879,Op087a,Op087b,Op____,Op____,Op____,Op____ ;@ 0878
  dcd Op0880,Op0880,Op0880,Op0880,Op0880,Op0880,Op0880,Op0880 ;@ 0880
  dcd Op0880,Op0880,Op0880,Op0880,Op0880,Op0880,Op0880,Op0880 ;@ 0888
  dcd Op0890,Op0890,Op0890,Op0890,Op0890,Op0890,Op0890,Op0890 ;@ 0890
  dcd Op0898,Op0898,Op0898,Op0898,Op0898,Op0898,Op0898,Op089f ;@ 0898
  dcd Op08a0,Op08a0,Op08a0,Op08a0,Op08a0,Op08a0,Op08a0,Op08a7 ;@ 08a0
  dcd Op08a8,Op08a8,Op08a8,Op08a8,Op08a8,Op08a8,Op08a8,Op08a8 ;@ 08a8
  dcd Op08b0,Op08b0,Op08b0,Op08b0,Op08b0,Op08b0,Op08b0,Op08b0 ;@ 08b0
  dcd Op08b8,Op08b9,Op08ba,Op08bb,Op____,Op____,Op____,Op____ ;@ 08b8
  dcd Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0 ;@ 08c0
  dcd Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0,Op08c0 ;@ 08c8
  dcd Op08d0,Op08d0,Op08d0,Op08d0,Op08d0,Op08d0,Op08d0,Op08d0 ;@ 08d0
  dcd Op08d8,Op08d8,Op08d8,Op08d8,Op08d8,Op08d8,Op08d8,Op08df ;@ 08d8
  dcd Op08e0,Op08e0,Op08e0,Op08e0,Op08e0,Op08e0,Op08e0,Op08e7 ;@ 08e0
  dcd Op08e8,Op08e8,Op08e8,Op08e8,Op08e8,Op08e8,Op08e8,Op08e8 ;@ 08e8
  dcd Op08f0,Op08f0,Op08f0,Op08f0,Op08f0,Op08f0,Op08f0,Op08f0 ;@ 08f0
  dcd Op08f8,Op08f9,Op08fa,Op08fb,Op____,Op____,Op____,Op____ ;@ 08f8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0908
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0910
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0918
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0920
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0928
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0930
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0938
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0948
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0950
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0958
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0960
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0968
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0970
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0978
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0988
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0990
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0998
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 09a0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 09a8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 09b0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 09b8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 09c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 09c8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 09d0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 09d8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 09e0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 09e8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 09f0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 09f8
  dcd Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00 ;@ 0a00
  dcd Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00,Op0a00 ;@ 0a08
  dcd Op0a10,Op0a10,Op0a10,Op0a10,Op0a10,Op0a10,Op0a10,Op0a10 ;@ 0a10
  dcd Op0a18,Op0a18,Op0a18,Op0a18,Op0a18,Op0a18,Op0a18,Op0a1f ;@ 0a18
  dcd Op0a20,Op0a20,Op0a20,Op0a20,Op0a20,Op0a20,Op0a20,Op0a27 ;@ 0a20
  dcd Op0a28,Op0a28,Op0a28,Op0a28,Op0a28,Op0a28,Op0a28,Op0a28 ;@ 0a28
  dcd Op0a30,Op0a30,Op0a30,Op0a30,Op0a30,Op0a30,Op0a30,Op0a30 ;@ 0a30
  dcd Op0a38,Op0a39,Op0a3a,Op0a3b,Op0a3c,Op____,Op____,Op____ ;@ 0a38
  dcd Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40 ;@ 0a40
  dcd Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40,Op0a40 ;@ 0a48
  dcd Op0a50,Op0a50,Op0a50,Op0a50,Op0a50,Op0a50,Op0a50,Op0a50 ;@ 0a50
  dcd Op0a58,Op0a58,Op0a58,Op0a58,Op0a58,Op0a58,Op0a58,Op0a5f ;@ 0a58
  dcd Op0a60,Op0a60,Op0a60,Op0a60,Op0a60,Op0a60,Op0a60,Op0a67 ;@ 0a60
  dcd Op0a68,Op0a68,Op0a68,Op0a68,Op0a68,Op0a68,Op0a68,Op0a68 ;@ 0a68
  dcd Op0a70,Op0a70,Op0a70,Op0a70,Op0a70,Op0a70,Op0a70,Op0a70 ;@ 0a70
  dcd Op0a78,Op0a79,Op0a7a,Op0a7b,Op0a7c,Op____,Op____,Op____ ;@ 0a78
  dcd Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80 ;@ 0a80
  dcd Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80,Op0a80 ;@ 0a88
  dcd Op0a90,Op0a90,Op0a90,Op0a90,Op0a90,Op0a90,Op0a90,Op0a90 ;@ 0a90
  dcd Op0a98,Op0a98,Op0a98,Op0a98,Op0a98,Op0a98,Op0a98,Op0a9f ;@ 0a98
  dcd Op0aa0,Op0aa0,Op0aa0,Op0aa0,Op0aa0,Op0aa0,Op0aa0,Op0aa7 ;@ 0aa0
  dcd Op0aa8,Op0aa8,Op0aa8,Op0aa8,Op0aa8,Op0aa8,Op0aa8,Op0aa8 ;@ 0aa8
  dcd Op0ab0,Op0ab0,Op0ab0,Op0ab0,Op0ab0,Op0ab0,Op0ab0,Op0ab0 ;@ 0ab0
  dcd Op0ab8,Op0ab9,Op0aba,Op0abb,Op____,Op____,Op____,Op____ ;@ 0ab8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ac0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ac8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ad0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ad8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ae0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ae8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0af0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0af8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0b00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0b08
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0b10
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0b18
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0b20
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0b28
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0b30
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0b38
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0b40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0b48
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0b50
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0b58
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0b60
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0b68
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0b70
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0b78
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0b80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0b88
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0b90
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0b98
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 0ba0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 0ba8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 0bb0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 0bb8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 0bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0bc8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 0bd0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 0bd8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 0be0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 0be8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 0bf0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 0bf8
  dcd Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00 ;@ 0c00
  dcd Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00,Op0c00 ;@ 0c08
  dcd Op0c10,Op0c10,Op0c10,Op0c10,Op0c10,Op0c10,Op0c10,Op0c10 ;@ 0c10
  dcd Op0c18,Op0c18,Op0c18,Op0c18,Op0c18,Op0c18,Op0c18,Op0c1f ;@ 0c18
  dcd Op0c20,Op0c20,Op0c20,Op0c20,Op0c20,Op0c20,Op0c20,Op0c27 ;@ 0c20
  dcd Op0c28,Op0c28,Op0c28,Op0c28,Op0c28,Op0c28,Op0c28,Op0c28 ;@ 0c28
  dcd Op0c30,Op0c30,Op0c30,Op0c30,Op0c30,Op0c30,Op0c30,Op0c30 ;@ 0c30
  dcd Op0c38,Op0c39,Op0c3a,Op0c3b,Op0c3c,Op____,Op____,Op____ ;@ 0c38
  dcd Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40 ;@ 0c40
  dcd Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40,Op0c40 ;@ 0c48
  dcd Op0c50,Op0c50,Op0c50,Op0c50,Op0c50,Op0c50,Op0c50,Op0c50 ;@ 0c50
  dcd Op0c58,Op0c58,Op0c58,Op0c58,Op0c58,Op0c58,Op0c58,Op0c5f ;@ 0c58
  dcd Op0c60,Op0c60,Op0c60,Op0c60,Op0c60,Op0c60,Op0c60,Op0c67 ;@ 0c60
  dcd Op0c68,Op0c68,Op0c68,Op0c68,Op0c68,Op0c68,Op0c68,Op0c68 ;@ 0c68
  dcd Op0c70,Op0c70,Op0c70,Op0c70,Op0c70,Op0c70,Op0c70,Op0c70 ;@ 0c70
  dcd Op0c78,Op0c79,Op0c7a,Op0c7b,Op0c7c,Op____,Op____,Op____ ;@ 0c78
  dcd Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80 ;@ 0c80
  dcd Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80,Op0c80 ;@ 0c88
  dcd Op0c90,Op0c90,Op0c90,Op0c90,Op0c90,Op0c90,Op0c90,Op0c90 ;@ 0c90
  dcd Op0c98,Op0c98,Op0c98,Op0c98,Op0c98,Op0c98,Op0c98,Op0c9f ;@ 0c98
  dcd Op0ca0,Op0ca0,Op0ca0,Op0ca0,Op0ca0,Op0ca0,Op0ca0,Op0ca7 ;@ 0ca0
  dcd Op0ca8,Op0ca8,Op0ca8,Op0ca8,Op0ca8,Op0ca8,Op0ca8,Op0ca8 ;@ 0ca8
  dcd Op0cb0,Op0cb0,Op0cb0,Op0cb0,Op0cb0,Op0cb0,Op0cb0,Op0cb0 ;@ 0cb0
  dcd Op0cb8,Op0cb9,Op0cba,Op0cbb,Op0cbc,Op____,Op____,Op____ ;@ 0cb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ce0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ce8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0cf8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0d00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0d08
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0d10
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0d18
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0d20
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0d28
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0d30
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0d38
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0d40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0d48
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0d50
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0d58
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0d60
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0d68
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0d70
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0d78
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0d80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0d88
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0d90
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0d98
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 0da0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 0da8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 0db0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 0db8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 0dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0dc8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 0dd0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 0dd8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 0de0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 0de8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 0df0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 0df8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0e98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ea0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ea8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0eb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0eb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ec0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ec8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ed0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ed8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ee0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ee8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ef0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0ef8
  dcd Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100,Op0100 ;@ 0f00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0f08
  dcd Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110,Op0110 ;@ 0f10
  dcd Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op0118,Op011f ;@ 0f18
  dcd Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0120,Op0127 ;@ 0f20
  dcd Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128,Op0128 ;@ 0f28
  dcd Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130,Op0130 ;@ 0f30
  dcd Op0138,Op0139,Op013a,Op013b,Op013c,Op____,Op____,Op____ ;@ 0f38
  dcd Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140,Op0140 ;@ 0f40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0f48
  dcd Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150,Op0150 ;@ 0f50
  dcd Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op0158,Op015f ;@ 0f58
  dcd Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0160,Op0167 ;@ 0f60
  dcd Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168,Op0168 ;@ 0f68
  dcd Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170,Op0170 ;@ 0f70
  dcd Op0178,Op0179,Op017a,Op017b,Op____,Op____,Op____,Op____ ;@ 0f78
  dcd Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180,Op0180 ;@ 0f80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0f88
  dcd Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190,Op0190 ;@ 0f90
  dcd Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op0198,Op019f ;@ 0f98
  dcd Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a0,Op01a7 ;@ 0fa0
  dcd Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8,Op01a8 ;@ 0fa8
  dcd Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0,Op01b0 ;@ 0fb0
  dcd Op01b8,Op01b9,Op01ba,Op01bb,Op____,Op____,Op____,Op____ ;@ 0fb8
  dcd Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0,Op01c0 ;@ 0fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 0fc8
  dcd Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0,Op01d0 ;@ 0fd0
  dcd Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01d8,Op01df ;@ 0fd8
  dcd Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e0,Op01e7 ;@ 0fe0
  dcd Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8,Op01e8 ;@ 0fe8
  dcd Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0,Op01f0 ;@ 0ff0
  dcd Op01f8,Op01f9,Op01fa,Op01fb,Op____,Op____,Op____,Op____ ;@ 0ff8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1000
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1008
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1010
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1018
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1020
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1028
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1030
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1038
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1040
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1048
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1050
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1058
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1060
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1068
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1070
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1078
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1080
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1088
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1090
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1098
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 10a0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 10a8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 10b0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 10b8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 10c0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 10c8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 10d0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 10d8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 10e0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 10e8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 10f0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 10f8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1100
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1108
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1110
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1118
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1120
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1128
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1130
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1138
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1140
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1148
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1150
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1158
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1160
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1168
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1170
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1178
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1180
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1188
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1190
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1198
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 11a0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 11a8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 11b0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 11b8
  dcd Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0 ;@ 11c0
  dcd Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0,Op11c0 ;@ 11c8
  dcd Op11d0,Op11d0,Op11d0,Op11d0,Op11d0,Op11d0,Op11d0,Op11d0 ;@ 11d0
  dcd Op11d8,Op11d8,Op11d8,Op11d8,Op11d8,Op11d8,Op11d8,Op11df ;@ 11d8
  dcd Op11e0,Op11e0,Op11e0,Op11e0,Op11e0,Op11e0,Op11e0,Op11e7 ;@ 11e0
  dcd Op11e8,Op11e8,Op11e8,Op11e8,Op11e8,Op11e8,Op11e8,Op11e8 ;@ 11e8
  dcd Op11f0,Op11f0,Op11f0,Op11f0,Op11f0,Op11f0,Op11f0,Op11f0 ;@ 11f0
  dcd Op11f8,Op11f9,Op11fa,Op11fb,Op11fc,Op____,Op____,Op____ ;@ 11f8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1200
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1208
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1210
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1218
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1220
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1228
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1230
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1238
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1240
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1248
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1250
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1258
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1260
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1268
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1270
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1278
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1280
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1288
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1290
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1298
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 12a0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 12a8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 12b0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 12b8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 12c0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 12c8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 12d0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 12d8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 12e0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 12e8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 12f0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 12f8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1300
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1308
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1310
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1318
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1320
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1328
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1330
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1338
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1340
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1348
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1350
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1358
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1360
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1368
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1370
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1378
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1380
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1388
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1390
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1398
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 13a0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 13a8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 13b0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 13b8
  dcd Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0 ;@ 13c0
  dcd Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0,Op13c0 ;@ 13c8
  dcd Op13d0,Op13d0,Op13d0,Op13d0,Op13d0,Op13d0,Op13d0,Op13d0 ;@ 13d0
  dcd Op13d8,Op13d8,Op13d8,Op13d8,Op13d8,Op13d8,Op13d8,Op13df ;@ 13d8
  dcd Op13e0,Op13e0,Op13e0,Op13e0,Op13e0,Op13e0,Op13e0,Op13e7 ;@ 13e0
  dcd Op13e8,Op13e8,Op13e8,Op13e8,Op13e8,Op13e8,Op13e8,Op13e8 ;@ 13e8
  dcd Op13f0,Op13f0,Op13f0,Op13f0,Op13f0,Op13f0,Op13f0,Op13f0 ;@ 13f0
  dcd Op13f8,Op13f9,Op13fa,Op13fb,Op13fc,Op____,Op____,Op____ ;@ 13f8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1400
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1408
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1410
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1418
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1420
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1428
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1430
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1438
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1440
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1448
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1450
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1458
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1460
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1468
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1470
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1478
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1480
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1488
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1490
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1498
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 14a0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 14a8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 14b0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 14b8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 14c0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 14c8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 14d0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 14d8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 14e0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 14e8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 14f0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 14f8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1500
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1508
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1510
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1518
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1520
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1528
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1530
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1538
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1540
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1548
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1550
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1558
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1560
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1568
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1570
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1578
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1580
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1588
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1590
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1598
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 15a0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 15a8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 15b0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 15b8
  dcd Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0 ;@ 15c0
  dcd Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0,Op15c0 ;@ 15c8
  dcd Op15d0,Op15d0,Op15d0,Op15d0,Op15d0,Op15d0,Op15d0,Op15d0 ;@ 15d0
  dcd Op15d8,Op15d8,Op15d8,Op15d8,Op15d8,Op15d8,Op15d8,Op15df ;@ 15d8
  dcd Op15e0,Op15e0,Op15e0,Op15e0,Op15e0,Op15e0,Op15e0,Op15e7 ;@ 15e0
  dcd Op15e8,Op15e8,Op15e8,Op15e8,Op15e8,Op15e8,Op15e8,Op15e8 ;@ 15e8
  dcd Op15f0,Op15f0,Op15f0,Op15f0,Op15f0,Op15f0,Op15f0,Op15f0 ;@ 15f0
  dcd Op15f8,Op15f9,Op15fa,Op15fb,Op15fc,Op____,Op____,Op____ ;@ 15f8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1600
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1608
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1610
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1618
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1620
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1628
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1630
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1638
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1640
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1648
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1650
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1658
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1660
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1668
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1670
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1678
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1680
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1688
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1690
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1698
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 16a0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 16a8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 16b0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 16b8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 16c0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 16c8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 16d0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 16d8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 16e0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 16e8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 16f0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 16f8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1700
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1708
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1710
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1718
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1720
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1728
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1730
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1738
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1740
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1748
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1750
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1758
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1760
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1768
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1770
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1778
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1780
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1788
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1790
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1798
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 17a0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 17a8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 17b0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 17b8
  dcd Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0 ;@ 17c0
  dcd Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0,Op17c0 ;@ 17c8
  dcd Op17d0,Op17d0,Op17d0,Op17d0,Op17d0,Op17d0,Op17d0,Op17d0 ;@ 17d0
  dcd Op17d8,Op17d8,Op17d8,Op17d8,Op17d8,Op17d8,Op17d8,Op17df ;@ 17d8
  dcd Op17e0,Op17e0,Op17e0,Op17e0,Op17e0,Op17e0,Op17e0,Op17e7 ;@ 17e0
  dcd Op17e8,Op17e8,Op17e8,Op17e8,Op17e8,Op17e8,Op17e8,Op17e8 ;@ 17e8
  dcd Op17f0,Op17f0,Op17f0,Op17f0,Op17f0,Op17f0,Op17f0,Op17f0 ;@ 17f0
  dcd Op17f8,Op17f9,Op17fa,Op17fb,Op17fc,Op____,Op____,Op____ ;@ 17f8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1800
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1808
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1810
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1818
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1820
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1828
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1830
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1838
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1840
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1848
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1850
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1858
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1860
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1868
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1870
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1878
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1880
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1888
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1890
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1898
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 18a0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 18a8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 18b0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 18b8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 18c0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 18c8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 18d0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 18d8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 18e0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 18e8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 18f0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 18f8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1900
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1908
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1910
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1918
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1920
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1928
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1930
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1938
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1940
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1948
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1950
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1958
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1960
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1968
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1970
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1978
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1980
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1988
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1990
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1998
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 19a0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 19a8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 19b0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 19b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 19f8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1a00
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1a08
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1a10
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1a18
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1a20
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1a28
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1a30
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1a38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1a78
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1a80
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1a88
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1a90
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1a98
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 1aa0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 1aa8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 1ab0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 1ab8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 1ac0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 1ac8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 1ad0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 1ad8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 1ae0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 1ae8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 1af0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 1af8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1b00
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1b08
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1b10
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1b18
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1b20
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1b28
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1b30
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1b38
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1b40
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1b48
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1b50
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1b58
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1b60
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1b68
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1b70
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1b78
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1b80
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1b88
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1b90
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1b98
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 1ba0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 1ba8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 1bb0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 1bb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1be0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1be8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1bf8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1c00
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1c08
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1c10
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1c18
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1c20
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1c28
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1c30
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1c38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1c78
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1c80
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1c88
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1c90
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1c98
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 1ca0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 1ca8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 1cb0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 1cb8
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 1cc0
  dcd Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0,Op10c0 ;@ 1cc8
  dcd Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0,Op10d0 ;@ 1cd0
  dcd Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10d8,Op10df ;@ 1cd8
  dcd Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e0,Op10e7 ;@ 1ce0
  dcd Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8,Op10e8 ;@ 1ce8
  dcd Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0,Op10f0 ;@ 1cf0
  dcd Op10f8,Op10f9,Op10fa,Op10fb,Op10fc,Op____,Op____,Op____ ;@ 1cf8
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1d00
  dcd Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100,Op1100 ;@ 1d08
  dcd Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110,Op1110 ;@ 1d10
  dcd Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op1118,Op111f ;@ 1d18
  dcd Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1120,Op1127 ;@ 1d20
  dcd Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128,Op1128 ;@ 1d28
  dcd Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130,Op1130 ;@ 1d30
  dcd Op1138,Op1139,Op113a,Op113b,Op113c,Op____,Op____,Op____ ;@ 1d38
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1d40
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1d48
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1d50
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1d58
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1d60
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1d68
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1d70
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1d78
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1d80
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1d88
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1d90
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1d98
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 1da0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 1da8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 1db0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 1db8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1dc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1dd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1dd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1de0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1de8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1df0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1df8
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1e00
  dcd Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000,Op1000 ;@ 1e08
  dcd Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010,Op1010 ;@ 1e10
  dcd Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op1018,Op101f ;@ 1e18
  dcd Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1020,Op1027 ;@ 1e20
  dcd Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028,Op1028 ;@ 1e28
  dcd Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030,Op1030 ;@ 1e30
  dcd Op1038,Op1039,Op103a,Op103b,Op103c,Op____,Op____,Op____ ;@ 1e38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1e78
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1e80
  dcd Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080,Op1080 ;@ 1e88
  dcd Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090,Op1090 ;@ 1e90
  dcd Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op1098,Op109f ;@ 1e98
  dcd Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a0,Op10a7 ;@ 1ea0
  dcd Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8,Op10a8 ;@ 1ea8
  dcd Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0,Op10b0 ;@ 1eb0
  dcd Op10b8,Op10b9,Op10ba,Op10bb,Op10bc,Op____,Op____,Op____ ;@ 1eb8
  dcd Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0 ;@ 1ec0
  dcd Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0,Op1ec0 ;@ 1ec8
  dcd Op1ed0,Op1ed0,Op1ed0,Op1ed0,Op1ed0,Op1ed0,Op1ed0,Op1ed0 ;@ 1ed0
  dcd Op1ed8,Op1ed8,Op1ed8,Op1ed8,Op1ed8,Op1ed8,Op1ed8,Op1edf ;@ 1ed8
  dcd Op1ee0,Op1ee0,Op1ee0,Op1ee0,Op1ee0,Op1ee0,Op1ee0,Op1ee7 ;@ 1ee0
  dcd Op1ee8,Op1ee8,Op1ee8,Op1ee8,Op1ee8,Op1ee8,Op1ee8,Op1ee8 ;@ 1ee8
  dcd Op1ef0,Op1ef0,Op1ef0,Op1ef0,Op1ef0,Op1ef0,Op1ef0,Op1ef0 ;@ 1ef0
  dcd Op1ef8,Op1ef9,Op1efa,Op1efb,Op1efc,Op____,Op____,Op____ ;@ 1ef8
  dcd Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00 ;@ 1f00
  dcd Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00,Op1f00 ;@ 1f08
  dcd Op1f10,Op1f10,Op1f10,Op1f10,Op1f10,Op1f10,Op1f10,Op1f10 ;@ 1f10
  dcd Op1f18,Op1f18,Op1f18,Op1f18,Op1f18,Op1f18,Op1f18,Op1f1f ;@ 1f18
  dcd Op1f20,Op1f20,Op1f20,Op1f20,Op1f20,Op1f20,Op1f20,Op1f27 ;@ 1f20
  dcd Op1f28,Op1f28,Op1f28,Op1f28,Op1f28,Op1f28,Op1f28,Op1f28 ;@ 1f28
  dcd Op1f30,Op1f30,Op1f30,Op1f30,Op1f30,Op1f30,Op1f30,Op1f30 ;@ 1f30
  dcd Op1f38,Op1f39,Op1f3a,Op1f3b,Op1f3c,Op____,Op____,Op____ ;@ 1f38
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1f40
  dcd Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140,Op1140 ;@ 1f48
  dcd Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150,Op1150 ;@ 1f50
  dcd Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op1158,Op115f ;@ 1f58
  dcd Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1160,Op1167 ;@ 1f60
  dcd Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168,Op1168 ;@ 1f68
  dcd Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170,Op1170 ;@ 1f70
  dcd Op1178,Op1179,Op117a,Op117b,Op117c,Op____,Op____,Op____ ;@ 1f78
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1f80
  dcd Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180,Op1180 ;@ 1f88
  dcd Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190,Op1190 ;@ 1f90
  dcd Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op1198,Op119f ;@ 1f98
  dcd Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a0,Op11a7 ;@ 1fa0
  dcd Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8,Op11a8 ;@ 1fa8
  dcd Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0,Op11b0 ;@ 1fb0
  dcd Op11b8,Op11b9,Op11ba,Op11bb,Op11bc,Op____,Op____,Op____ ;@ 1fb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1fe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1ff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 1ff8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2000
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2008
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2010
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2018
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2020
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2028
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2030
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2038
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2040
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2048
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2050
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2058
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2060
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2068
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2070
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2078
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2080
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2088
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2090
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2098
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 20a0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 20a8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 20b0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 20b8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 20c0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 20c8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 20d0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 20d8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 20e0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 20e8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 20f0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 20f8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2100
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2108
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2110
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2118
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2120
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2128
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2130
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2138
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2140
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2148
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2150
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2158
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2160
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2168
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2170
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2178
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2180
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2188
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2190
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2198
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 21a0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 21a8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 21b0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 21b8
  dcd Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0 ;@ 21c0
  dcd Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0,Op21c0 ;@ 21c8
  dcd Op21d0,Op21d0,Op21d0,Op21d0,Op21d0,Op21d0,Op21d0,Op21d0 ;@ 21d0
  dcd Op21d8,Op21d8,Op21d8,Op21d8,Op21d8,Op21d8,Op21d8,Op21df ;@ 21d8
  dcd Op21e0,Op21e0,Op21e0,Op21e0,Op21e0,Op21e0,Op21e0,Op21e7 ;@ 21e0
  dcd Op21e8,Op21e8,Op21e8,Op21e8,Op21e8,Op21e8,Op21e8,Op21e8 ;@ 21e8
  dcd Op21f0,Op21f0,Op21f0,Op21f0,Op21f0,Op21f0,Op21f0,Op21f0 ;@ 21f0
  dcd Op21f8,Op21f9,Op21fa,Op21fb,Op21fc,Op____,Op____,Op____ ;@ 21f8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2200
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2208
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2210
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2218
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2220
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2228
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2230
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2238
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2240
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2248
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2250
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2258
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2260
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2268
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2270
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2278
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2280
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2288
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2290
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2298
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 22a0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 22a8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 22b0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 22b8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 22c0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 22c8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 22d0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 22d8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 22e0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 22e8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 22f0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 22f8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2300
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2308
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2310
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2318
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2320
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2328
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2330
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2338
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2340
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2348
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2350
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2358
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2360
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2368
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2370
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2378
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2380
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2388
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2390
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2398
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 23a0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 23a8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 23b0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 23b8
  dcd Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0 ;@ 23c0
  dcd Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0,Op23c0 ;@ 23c8
  dcd Op23d0,Op23d0,Op23d0,Op23d0,Op23d0,Op23d0,Op23d0,Op23d0 ;@ 23d0
  dcd Op23d8,Op23d8,Op23d8,Op23d8,Op23d8,Op23d8,Op23d8,Op23df ;@ 23d8
  dcd Op23e0,Op23e0,Op23e0,Op23e0,Op23e0,Op23e0,Op23e0,Op23e7 ;@ 23e0
  dcd Op23e8,Op23e8,Op23e8,Op23e8,Op23e8,Op23e8,Op23e8,Op23e8 ;@ 23e8
  dcd Op23f0,Op23f0,Op23f0,Op23f0,Op23f0,Op23f0,Op23f0,Op23f0 ;@ 23f0
  dcd Op23f8,Op23f9,Op23fa,Op23fb,Op23fc,Op____,Op____,Op____ ;@ 23f8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2400
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2408
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2410
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2418
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2420
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2428
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2430
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2438
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2440
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2448
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2450
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2458
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2460
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2468
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2470
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2478
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2480
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2488
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2490
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2498
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 24a0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 24a8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 24b0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 24b8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 24c0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 24c8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 24d0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 24d8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 24e0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 24e8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 24f0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 24f8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2500
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2508
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2510
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2518
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2520
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2528
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2530
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2538
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2540
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2548
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2550
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2558
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2560
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2568
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2570
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2578
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2580
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2588
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2590
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2598
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 25a0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 25a8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 25b0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 25b8
  dcd Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0 ;@ 25c0
  dcd Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0,Op25c0 ;@ 25c8
  dcd Op25d0,Op25d0,Op25d0,Op25d0,Op25d0,Op25d0,Op25d0,Op25d0 ;@ 25d0
  dcd Op25d8,Op25d8,Op25d8,Op25d8,Op25d8,Op25d8,Op25d8,Op25df ;@ 25d8
  dcd Op25e0,Op25e0,Op25e0,Op25e0,Op25e0,Op25e0,Op25e0,Op25e7 ;@ 25e0
  dcd Op25e8,Op25e8,Op25e8,Op25e8,Op25e8,Op25e8,Op25e8,Op25e8 ;@ 25e8
  dcd Op25f0,Op25f0,Op25f0,Op25f0,Op25f0,Op25f0,Op25f0,Op25f0 ;@ 25f0
  dcd Op25f8,Op25f9,Op25fa,Op25fb,Op25fc,Op____,Op____,Op____ ;@ 25f8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2600
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2608
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2610
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2618
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2620
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2628
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2630
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2638
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2640
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2648
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2650
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2658
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2660
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2668
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2670
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2678
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2680
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2688
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2690
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2698
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 26a0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 26a8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 26b0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 26b8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 26c0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 26c8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 26d0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 26d8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 26e0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 26e8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 26f0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 26f8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2700
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2708
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2710
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2718
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2720
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2728
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2730
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2738
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2740
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2748
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2750
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2758
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2760
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2768
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2770
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2778
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2780
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2788
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2790
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2798
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 27a0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 27a8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 27b0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 27b8
  dcd Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0 ;@ 27c0
  dcd Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0,Op27c0 ;@ 27c8
  dcd Op27d0,Op27d0,Op27d0,Op27d0,Op27d0,Op27d0,Op27d0,Op27d0 ;@ 27d0
  dcd Op27d8,Op27d8,Op27d8,Op27d8,Op27d8,Op27d8,Op27d8,Op27df ;@ 27d8
  dcd Op27e0,Op27e0,Op27e0,Op27e0,Op27e0,Op27e0,Op27e0,Op27e7 ;@ 27e0
  dcd Op27e8,Op27e8,Op27e8,Op27e8,Op27e8,Op27e8,Op27e8,Op27e8 ;@ 27e8
  dcd Op27f0,Op27f0,Op27f0,Op27f0,Op27f0,Op27f0,Op27f0,Op27f0 ;@ 27f0
  dcd Op27f8,Op27f9,Op27fa,Op27fb,Op27fc,Op____,Op____,Op____ ;@ 27f8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2800
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2808
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2810
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2818
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2820
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2828
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2830
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2838
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2840
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2848
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2850
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2858
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2860
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2868
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2870
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2878
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2880
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2888
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2890
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2898
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 28a0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 28a8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 28b0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 28b8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 28c0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 28c8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 28d0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 28d8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 28e0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 28e8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 28f0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 28f8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2900
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2908
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2910
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2918
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2920
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2928
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2930
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2938
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2940
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2948
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2950
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2958
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2960
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2968
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2970
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2978
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2980
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2988
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2990
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2998
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 29a0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 29a8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 29b0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 29b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 29f8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2a00
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2a08
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2a10
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2a18
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2a20
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2a28
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2a30
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2a38
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2a40
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2a48
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2a50
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2a58
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2a60
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2a68
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2a70
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2a78
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2a80
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2a88
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2a90
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2a98
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 2aa0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 2aa8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 2ab0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 2ab8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 2ac0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 2ac8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 2ad0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 2ad8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 2ae0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 2ae8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 2af0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 2af8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2b00
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2b08
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2b10
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2b18
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2b20
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2b28
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2b30
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2b38
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2b40
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2b48
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2b50
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2b58
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2b60
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2b68
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2b70
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2b78
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2b80
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2b88
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2b90
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2b98
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 2ba0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 2ba8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 2bb0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 2bb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2be0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2be8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2bf8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2c00
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2c08
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2c10
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2c18
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2c20
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2c28
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2c30
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2c38
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2c40
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2c48
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2c50
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2c58
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2c60
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2c68
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2c70
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2c78
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2c80
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2c88
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2c90
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2c98
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 2ca0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 2ca8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 2cb0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 2cb8
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 2cc0
  dcd Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0,Op20c0 ;@ 2cc8
  dcd Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0,Op20d0 ;@ 2cd0
  dcd Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20d8,Op20df ;@ 2cd8
  dcd Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e0,Op20e7 ;@ 2ce0
  dcd Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8,Op20e8 ;@ 2ce8
  dcd Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0,Op20f0 ;@ 2cf0
  dcd Op20f8,Op20f9,Op20fa,Op20fb,Op20fc,Op____,Op____,Op____ ;@ 2cf8
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2d00
  dcd Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100,Op2100 ;@ 2d08
  dcd Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110,Op2110 ;@ 2d10
  dcd Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op2118,Op211f ;@ 2d18
  dcd Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2120,Op2127 ;@ 2d20
  dcd Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128,Op2128 ;@ 2d28
  dcd Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130,Op2130 ;@ 2d30
  dcd Op2138,Op2139,Op213a,Op213b,Op213c,Op____,Op____,Op____ ;@ 2d38
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2d40
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2d48
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2d50
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2d58
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2d60
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2d68
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2d70
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2d78
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2d80
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2d88
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2d90
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2d98
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 2da0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 2da8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 2db0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 2db8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2dc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2dd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2dd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2de0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2de8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2df0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2df8
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2e00
  dcd Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000,Op2000 ;@ 2e08
  dcd Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010,Op2010 ;@ 2e10
  dcd Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op2018,Op201f ;@ 2e18
  dcd Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2020,Op2027 ;@ 2e20
  dcd Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028,Op2028 ;@ 2e28
  dcd Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030,Op2030 ;@ 2e30
  dcd Op2038,Op2039,Op203a,Op203b,Op203c,Op____,Op____,Op____ ;@ 2e38
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2e40
  dcd Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040,Op2040 ;@ 2e48
  dcd Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050,Op2050 ;@ 2e50
  dcd Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op2058,Op205f ;@ 2e58
  dcd Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2060,Op2067 ;@ 2e60
  dcd Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068,Op2068 ;@ 2e68
  dcd Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070,Op2070 ;@ 2e70
  dcd Op2078,Op2079,Op207a,Op207b,Op207c,Op____,Op____,Op____ ;@ 2e78
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2e80
  dcd Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080,Op2080 ;@ 2e88
  dcd Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090,Op2090 ;@ 2e90
  dcd Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op2098,Op209f ;@ 2e98
  dcd Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a0,Op20a7 ;@ 2ea0
  dcd Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8,Op20a8 ;@ 2ea8
  dcd Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0,Op20b0 ;@ 2eb0
  dcd Op20b8,Op20b9,Op20ba,Op20bb,Op20bc,Op____,Op____,Op____ ;@ 2eb8
  dcd Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0 ;@ 2ec0
  dcd Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0,Op2ec0 ;@ 2ec8
  dcd Op2ed0,Op2ed0,Op2ed0,Op2ed0,Op2ed0,Op2ed0,Op2ed0,Op2ed0 ;@ 2ed0
  dcd Op2ed8,Op2ed8,Op2ed8,Op2ed8,Op2ed8,Op2ed8,Op2ed8,Op2edf ;@ 2ed8
  dcd Op2ee0,Op2ee0,Op2ee0,Op2ee0,Op2ee0,Op2ee0,Op2ee0,Op2ee7 ;@ 2ee0
  dcd Op2ee8,Op2ee8,Op2ee8,Op2ee8,Op2ee8,Op2ee8,Op2ee8,Op2ee8 ;@ 2ee8
  dcd Op2ef0,Op2ef0,Op2ef0,Op2ef0,Op2ef0,Op2ef0,Op2ef0,Op2ef0 ;@ 2ef0
  dcd Op2ef8,Op2ef9,Op2efa,Op2efb,Op2efc,Op____,Op____,Op____ ;@ 2ef8
  dcd Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00 ;@ 2f00
  dcd Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00,Op2f00 ;@ 2f08
  dcd Op2f10,Op2f10,Op2f10,Op2f10,Op2f10,Op2f10,Op2f10,Op2f10 ;@ 2f10
  dcd Op2f18,Op2f18,Op2f18,Op2f18,Op2f18,Op2f18,Op2f18,Op2f1f ;@ 2f18
  dcd Op2f20,Op2f20,Op2f20,Op2f20,Op2f20,Op2f20,Op2f20,Op2f27 ;@ 2f20
  dcd Op2f28,Op2f28,Op2f28,Op2f28,Op2f28,Op2f28,Op2f28,Op2f28 ;@ 2f28
  dcd Op2f30,Op2f30,Op2f30,Op2f30,Op2f30,Op2f30,Op2f30,Op2f30 ;@ 2f30
  dcd Op2f38,Op2f39,Op2f3a,Op2f3b,Op2f3c,Op____,Op____,Op____ ;@ 2f38
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2f40
  dcd Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140,Op2140 ;@ 2f48
  dcd Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150,Op2150 ;@ 2f50
  dcd Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op2158,Op215f ;@ 2f58
  dcd Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2160,Op2167 ;@ 2f60
  dcd Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168,Op2168 ;@ 2f68
  dcd Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170,Op2170 ;@ 2f70
  dcd Op2178,Op2179,Op217a,Op217b,Op217c,Op____,Op____,Op____ ;@ 2f78
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2f80
  dcd Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180,Op2180 ;@ 2f88
  dcd Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190,Op2190 ;@ 2f90
  dcd Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op2198,Op219f ;@ 2f98
  dcd Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a0,Op21a7 ;@ 2fa0
  dcd Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8,Op21a8 ;@ 2fa8
  dcd Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0,Op21b0 ;@ 2fb0
  dcd Op21b8,Op21b9,Op21ba,Op21bb,Op21bc,Op____,Op____,Op____ ;@ 2fb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2fe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2ff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 2ff8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3000
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3008
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3010
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3018
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3020
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3028
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3030
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3038
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3040
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3048
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3050
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3058
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3060
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3068
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3070
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3078
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3080
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3088
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3090
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3098
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 30a0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 30a8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 30b0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 30b8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 30c0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 30c8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 30d0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 30d8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 30e0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 30e8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 30f0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 30f8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3100
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3108
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3110
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3118
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3120
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3128
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3130
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3138
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3140
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3148
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3150
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3158
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3160
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3168
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3170
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3178
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3180
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3188
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3190
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3198
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 31a0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 31a8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 31b0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 31b8
  dcd Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0 ;@ 31c0
  dcd Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0,Op31c0 ;@ 31c8
  dcd Op31d0,Op31d0,Op31d0,Op31d0,Op31d0,Op31d0,Op31d0,Op31d0 ;@ 31d0
  dcd Op31d8,Op31d8,Op31d8,Op31d8,Op31d8,Op31d8,Op31d8,Op31df ;@ 31d8
  dcd Op31e0,Op31e0,Op31e0,Op31e0,Op31e0,Op31e0,Op31e0,Op31e7 ;@ 31e0
  dcd Op31e8,Op31e8,Op31e8,Op31e8,Op31e8,Op31e8,Op31e8,Op31e8 ;@ 31e8
  dcd Op31f0,Op31f0,Op31f0,Op31f0,Op31f0,Op31f0,Op31f0,Op31f0 ;@ 31f0
  dcd Op31f8,Op31f9,Op31fa,Op31fb,Op31fc,Op____,Op____,Op____ ;@ 31f8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3200
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3208
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3210
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3218
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3220
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3228
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3230
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3238
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3240
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3248
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3250
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3258
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3260
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3268
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3270
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3278
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3280
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3288
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3290
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3298
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 32a0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 32a8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 32b0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 32b8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 32c0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 32c8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 32d0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 32d8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 32e0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 32e8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 32f0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 32f8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3300
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3308
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3310
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3318
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3320
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3328
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3330
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3338
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3340
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3348
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3350
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3358
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3360
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3368
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3370
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3378
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3380
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3388
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3390
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3398
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 33a0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 33a8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 33b0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 33b8
  dcd Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0 ;@ 33c0
  dcd Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0,Op33c0 ;@ 33c8
  dcd Op33d0,Op33d0,Op33d0,Op33d0,Op33d0,Op33d0,Op33d0,Op33d0 ;@ 33d0
  dcd Op33d8,Op33d8,Op33d8,Op33d8,Op33d8,Op33d8,Op33d8,Op33df ;@ 33d8
  dcd Op33e0,Op33e0,Op33e0,Op33e0,Op33e0,Op33e0,Op33e0,Op33e7 ;@ 33e0
  dcd Op33e8,Op33e8,Op33e8,Op33e8,Op33e8,Op33e8,Op33e8,Op33e8 ;@ 33e8
  dcd Op33f0,Op33f0,Op33f0,Op33f0,Op33f0,Op33f0,Op33f0,Op33f0 ;@ 33f0
  dcd Op33f8,Op33f9,Op33fa,Op33fb,Op33fc,Op____,Op____,Op____ ;@ 33f8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3400
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3408
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3410
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3418
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3420
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3428
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3430
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3438
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3440
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3448
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3450
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3458
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3460
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3468
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3470
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3478
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3480
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3488
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3490
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3498
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 34a0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 34a8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 34b0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 34b8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 34c0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 34c8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 34d0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 34d8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 34e0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 34e8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 34f0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 34f8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3500
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3508
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3510
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3518
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3520
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3528
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3530
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3538
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3540
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3548
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3550
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3558
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3560
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3568
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3570
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3578
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3580
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3588
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3590
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3598
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 35a0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 35a8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 35b0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 35b8
  dcd Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0 ;@ 35c0
  dcd Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0,Op35c0 ;@ 35c8
  dcd Op35d0,Op35d0,Op35d0,Op35d0,Op35d0,Op35d0,Op35d0,Op35d0 ;@ 35d0
  dcd Op35d8,Op35d8,Op35d8,Op35d8,Op35d8,Op35d8,Op35d8,Op35df ;@ 35d8
  dcd Op35e0,Op35e0,Op35e0,Op35e0,Op35e0,Op35e0,Op35e0,Op35e7 ;@ 35e0
  dcd Op35e8,Op35e8,Op35e8,Op35e8,Op35e8,Op35e8,Op35e8,Op35e8 ;@ 35e8
  dcd Op35f0,Op35f0,Op35f0,Op35f0,Op35f0,Op35f0,Op35f0,Op35f0 ;@ 35f0
  dcd Op35f8,Op35f9,Op35fa,Op35fb,Op35fc,Op____,Op____,Op____ ;@ 35f8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3600
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3608
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3610
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3618
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3620
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3628
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3630
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3638
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3640
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3648
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3650
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3658
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3660
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3668
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3670
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3678
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3680
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3688
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3690
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3698
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 36a0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 36a8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 36b0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 36b8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 36c0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 36c8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 36d0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 36d8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 36e0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 36e8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 36f0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 36f8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3700
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3708
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3710
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3718
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3720
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3728
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3730
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3738
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3740
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3748
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3750
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3758
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3760
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3768
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3770
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3778
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3780
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3788
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3790
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3798
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 37a0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 37a8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 37b0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 37b8
  dcd Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0 ;@ 37c0
  dcd Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0,Op37c0 ;@ 37c8
  dcd Op37d0,Op37d0,Op37d0,Op37d0,Op37d0,Op37d0,Op37d0,Op37d0 ;@ 37d0
  dcd Op37d8,Op37d8,Op37d8,Op37d8,Op37d8,Op37d8,Op37d8,Op37df ;@ 37d8
  dcd Op37e0,Op37e0,Op37e0,Op37e0,Op37e0,Op37e0,Op37e0,Op37e7 ;@ 37e0
  dcd Op37e8,Op37e8,Op37e8,Op37e8,Op37e8,Op37e8,Op37e8,Op37e8 ;@ 37e8
  dcd Op37f0,Op37f0,Op37f0,Op37f0,Op37f0,Op37f0,Op37f0,Op37f0 ;@ 37f0
  dcd Op37f8,Op37f9,Op37fa,Op37fb,Op37fc,Op____,Op____,Op____ ;@ 37f8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3800
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3808
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3810
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3818
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3820
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3828
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3830
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3838
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3840
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3848
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3850
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3858
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3860
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3868
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3870
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3878
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3880
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3888
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3890
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3898
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 38a0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 38a8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 38b0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 38b8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 38c0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 38c8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 38d0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 38d8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 38e0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 38e8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 38f0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 38f8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3900
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3908
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3910
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3918
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3920
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3928
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3930
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3938
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3940
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3948
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3950
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3958
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3960
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3968
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3970
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3978
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3980
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3988
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3990
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3998
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 39a0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 39a8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 39b0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 39b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 39f8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3a00
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3a08
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3a10
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3a18
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3a20
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3a28
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3a30
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3a38
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3a40
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3a48
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3a50
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3a58
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3a60
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3a68
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3a70
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3a78
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3a80
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3a88
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3a90
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3a98
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 3aa0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 3aa8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 3ab0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 3ab8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 3ac0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 3ac8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 3ad0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 3ad8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 3ae0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 3ae8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 3af0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 3af8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3b00
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3b08
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3b10
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3b18
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3b20
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3b28
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3b30
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3b38
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3b40
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3b48
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3b50
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3b58
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3b60
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3b68
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3b70
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3b78
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3b80
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3b88
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3b90
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3b98
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 3ba0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 3ba8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 3bb0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 3bb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3be0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3be8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3bf8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3c00
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3c08
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3c10
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3c18
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3c20
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3c28
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3c30
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3c38
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3c40
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3c48
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3c50
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3c58
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3c60
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3c68
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3c70
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3c78
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3c80
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3c88
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3c90
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3c98
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 3ca0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 3ca8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 3cb0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 3cb8
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 3cc0
  dcd Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0,Op30c0 ;@ 3cc8
  dcd Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0,Op30d0 ;@ 3cd0
  dcd Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30d8,Op30df ;@ 3cd8
  dcd Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e0,Op30e7 ;@ 3ce0
  dcd Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8,Op30e8 ;@ 3ce8
  dcd Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0,Op30f0 ;@ 3cf0
  dcd Op30f8,Op30f9,Op30fa,Op30fb,Op30fc,Op____,Op____,Op____ ;@ 3cf8
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3d00
  dcd Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100,Op3100 ;@ 3d08
  dcd Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110,Op3110 ;@ 3d10
  dcd Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op3118,Op311f ;@ 3d18
  dcd Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3120,Op3127 ;@ 3d20
  dcd Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128,Op3128 ;@ 3d28
  dcd Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130,Op3130 ;@ 3d30
  dcd Op3138,Op3139,Op313a,Op313b,Op313c,Op____,Op____,Op____ ;@ 3d38
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3d40
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3d48
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3d50
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3d58
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3d60
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3d68
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3d70
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3d78
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3d80
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3d88
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3d90
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3d98
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 3da0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 3da8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 3db0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 3db8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3dc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3dd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3dd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3de0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3de8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3df0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3df8
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3e00
  dcd Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000,Op3000 ;@ 3e08
  dcd Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010,Op3010 ;@ 3e10
  dcd Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op3018,Op301f ;@ 3e18
  dcd Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3020,Op3027 ;@ 3e20
  dcd Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028,Op3028 ;@ 3e28
  dcd Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030,Op3030 ;@ 3e30
  dcd Op3038,Op3039,Op303a,Op303b,Op303c,Op____,Op____,Op____ ;@ 3e38
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3e40
  dcd Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040,Op3040 ;@ 3e48
  dcd Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050,Op3050 ;@ 3e50
  dcd Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op3058,Op305f ;@ 3e58
  dcd Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3060,Op3067 ;@ 3e60
  dcd Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068,Op3068 ;@ 3e68
  dcd Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070,Op3070 ;@ 3e70
  dcd Op3078,Op3079,Op307a,Op307b,Op307c,Op____,Op____,Op____ ;@ 3e78
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3e80
  dcd Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080,Op3080 ;@ 3e88
  dcd Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090,Op3090 ;@ 3e90
  dcd Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op3098,Op309f ;@ 3e98
  dcd Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a0,Op30a7 ;@ 3ea0
  dcd Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8,Op30a8 ;@ 3ea8
  dcd Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0,Op30b0 ;@ 3eb0
  dcd Op30b8,Op30b9,Op30ba,Op30bb,Op30bc,Op____,Op____,Op____ ;@ 3eb8
  dcd Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0 ;@ 3ec0
  dcd Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0,Op3ec0 ;@ 3ec8
  dcd Op3ed0,Op3ed0,Op3ed0,Op3ed0,Op3ed0,Op3ed0,Op3ed0,Op3ed0 ;@ 3ed0
  dcd Op3ed8,Op3ed8,Op3ed8,Op3ed8,Op3ed8,Op3ed8,Op3ed8,Op3edf ;@ 3ed8
  dcd Op3ee0,Op3ee0,Op3ee0,Op3ee0,Op3ee0,Op3ee0,Op3ee0,Op3ee7 ;@ 3ee0
  dcd Op3ee8,Op3ee8,Op3ee8,Op3ee8,Op3ee8,Op3ee8,Op3ee8,Op3ee8 ;@ 3ee8
  dcd Op3ef0,Op3ef0,Op3ef0,Op3ef0,Op3ef0,Op3ef0,Op3ef0,Op3ef0 ;@ 3ef0
  dcd Op3ef8,Op3ef9,Op3efa,Op3efb,Op3efc,Op____,Op____,Op____ ;@ 3ef8
  dcd Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00 ;@ 3f00
  dcd Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00,Op3f00 ;@ 3f08
  dcd Op3f10,Op3f10,Op3f10,Op3f10,Op3f10,Op3f10,Op3f10,Op3f10 ;@ 3f10
  dcd Op3f18,Op3f18,Op3f18,Op3f18,Op3f18,Op3f18,Op3f18,Op3f1f ;@ 3f18
  dcd Op3f20,Op3f20,Op3f20,Op3f20,Op3f20,Op3f20,Op3f20,Op3f27 ;@ 3f20
  dcd Op3f28,Op3f28,Op3f28,Op3f28,Op3f28,Op3f28,Op3f28,Op3f28 ;@ 3f28
  dcd Op3f30,Op3f30,Op3f30,Op3f30,Op3f30,Op3f30,Op3f30,Op3f30 ;@ 3f30
  dcd Op3f38,Op3f39,Op3f3a,Op3f3b,Op3f3c,Op____,Op____,Op____ ;@ 3f38
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3f40
  dcd Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140,Op3140 ;@ 3f48
  dcd Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150,Op3150 ;@ 3f50
  dcd Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op3158,Op315f ;@ 3f58
  dcd Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3160,Op3167 ;@ 3f60
  dcd Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168,Op3168 ;@ 3f68
  dcd Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170,Op3170 ;@ 3f70
  dcd Op3178,Op3179,Op317a,Op317b,Op317c,Op____,Op____,Op____ ;@ 3f78
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3f80
  dcd Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180,Op3180 ;@ 3f88
  dcd Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190,Op3190 ;@ 3f90
  dcd Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op3198,Op319f ;@ 3f98
  dcd Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a0,Op31a7 ;@ 3fa0
  dcd Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8,Op31a8 ;@ 3fa8
  dcd Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0,Op31b0 ;@ 3fb0
  dcd Op31b8,Op31b9,Op31ba,Op31bb,Op31bc,Op____,Op____,Op____ ;@ 3fb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3fe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3ff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 3ff8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4000
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4008
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4010
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4018
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4020
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4028
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4030
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4038
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4040
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4048
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4050
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4058
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4060
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4068
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4070
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4078
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4080
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4088
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4090
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4098
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 40a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 40a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 40b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 40b8
  dcd Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0 ;@ 40c0
  dcd Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0,Op40c0 ;@ 40c8
  dcd Op40d0,Op40d0,Op40d0,Op40d0,Op40d0,Op40d0,Op40d0,Op40d0 ;@ 40d0
  dcd Op40d8,Op40d8,Op40d8,Op40d8,Op40d8,Op40d8,Op40d8,Op40df ;@ 40d8
  dcd Op40e0,Op40e0,Op40e0,Op40e0,Op40e0,Op40e0,Op40e0,Op40e7 ;@ 40e0
  dcd Op40e8,Op40e8,Op40e8,Op40e8,Op40e8,Op40e8,Op40e8,Op40e8 ;@ 40e8
  dcd Op40f0,Op40f0,Op40f0,Op40f0,Op40f0,Op40f0,Op40f0,Op40f0 ;@ 40f0
  dcd Op40f8,Op40f9,Op40fa,Op40fb,Op____,Op____,Op____,Op____ ;@ 40f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4108
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4110
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4118
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4120
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4128
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4130
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4138
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4148
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4150
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4158
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4160
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4168
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4170
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4188
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4190
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4198
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41c8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 41d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 41e0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 41e8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 41f0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 41f8
  dcd Op4200,Op4200,Op4200,Op4200,Op4200,Op4200,Op4200,Op4200 ;@ 4200
  dcd Op4200,Op4200,Op4200,Op4200,Op4200,Op4200,Op4200,Op4200 ;@ 4208
  dcd Op4210,Op4210,Op4210,Op4210,Op4210,Op4210,Op4210,Op4210 ;@ 4210
  dcd Op4218,Op4218,Op4218,Op4218,Op4218,Op4218,Op4218,Op421f ;@ 4218
  dcd Op4220,Op4220,Op4220,Op4220,Op4220,Op4220,Op4220,Op4227 ;@ 4220
  dcd Op4228,Op4228,Op4228,Op4228,Op4228,Op4228,Op4228,Op4228 ;@ 4228
  dcd Op4230,Op4230,Op4230,Op4230,Op4230,Op4230,Op4230,Op4230 ;@ 4230
  dcd Op4238,Op4239,Op423a,Op423b,Op____,Op____,Op____,Op____ ;@ 4238
  dcd Op4240,Op4240,Op4240,Op4240,Op4240,Op4240,Op4240,Op4240 ;@ 4240
  dcd Op4240,Op4240,Op4240,Op4240,Op4240,Op4240,Op4240,Op4240 ;@ 4248
  dcd Op4250,Op4250,Op4250,Op4250,Op4250,Op4250,Op4250,Op4250 ;@ 4250
  dcd Op4258,Op4258,Op4258,Op4258,Op4258,Op4258,Op4258,Op425f ;@ 4258
  dcd Op4260,Op4260,Op4260,Op4260,Op4260,Op4260,Op4260,Op4267 ;@ 4260
  dcd Op4268,Op4268,Op4268,Op4268,Op4268,Op4268,Op4268,Op4268 ;@ 4268
  dcd Op4270,Op4270,Op4270,Op4270,Op4270,Op4270,Op4270,Op4270 ;@ 4270
  dcd Op4278,Op4279,Op427a,Op427b,Op____,Op____,Op____,Op____ ;@ 4278
  dcd Op4280,Op4280,Op4280,Op4280,Op4280,Op4280,Op4280,Op4280 ;@ 4280
  dcd Op4280,Op4280,Op4280,Op4280,Op4280,Op4280,Op4280,Op4280 ;@ 4288
  dcd Op4290,Op4290,Op4290,Op4290,Op4290,Op4290,Op4290,Op4290 ;@ 4290
  dcd Op4298,Op4298,Op4298,Op4298,Op4298,Op4298,Op4298,Op429f ;@ 4298
  dcd Op42a0,Op42a0,Op42a0,Op42a0,Op42a0,Op42a0,Op42a0,Op42a7 ;@ 42a0
  dcd Op42a8,Op42a8,Op42a8,Op42a8,Op42a8,Op42a8,Op42a8,Op42a8 ;@ 42a8
  dcd Op42b0,Op42b0,Op42b0,Op42b0,Op42b0,Op42b0,Op42b0,Op42b0 ;@ 42b0
  dcd Op42b8,Op42b9,Op42ba,Op42bb,Op____,Op____,Op____,Op____ ;@ 42b8
  dcd Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0 ;@ 42c0
  dcd Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0,Op42c0 ;@ 42c8
  dcd Op42d0,Op42d0,Op42d0,Op42d0,Op42d0,Op42d0,Op42d0,Op42d0 ;@ 42d0
  dcd Op42d8,Op42d8,Op42d8,Op42d8,Op42d8,Op42d8,Op42d8,Op42df ;@ 42d8
  dcd Op42e0,Op42e0,Op42e0,Op42e0,Op42e0,Op42e0,Op42e0,Op42e7 ;@ 42e0
  dcd Op42e8,Op42e8,Op42e8,Op42e8,Op42e8,Op42e8,Op42e8,Op42e8 ;@ 42e8
  dcd Op42f0,Op42f0,Op42f0,Op42f0,Op42f0,Op42f0,Op42f0,Op42f0 ;@ 42f0
  dcd Op42f8,Op42f9,Op42fa,Op42fb,Op____,Op____,Op____,Op____ ;@ 42f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4308
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4310
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4318
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4320
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4328
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4330
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4338
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4348
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4350
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4358
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4360
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4368
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4370
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4388
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4390
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4398
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43c8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 43d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 43e0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 43e8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 43f0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 43f8
  dcd Op4400,Op4400,Op4400,Op4400,Op4400,Op4400,Op4400,Op4400 ;@ 4400
  dcd Op4400,Op4400,Op4400,Op4400,Op4400,Op4400,Op4400,Op4400 ;@ 4408
  dcd Op4410,Op4410,Op4410,Op4410,Op4410,Op4410,Op4410,Op4410 ;@ 4410
  dcd Op4418,Op4418,Op4418,Op4418,Op4418,Op4418,Op4418,Op441f ;@ 4418
  dcd Op4420,Op4420,Op4420,Op4420,Op4420,Op4420,Op4420,Op4427 ;@ 4420
  dcd Op4428,Op4428,Op4428,Op4428,Op4428,Op4428,Op4428,Op4428 ;@ 4428
  dcd Op4430,Op4430,Op4430,Op4430,Op4430,Op4430,Op4430,Op4430 ;@ 4430
  dcd Op4438,Op4439,Op443a,Op443b,Op____,Op____,Op____,Op____ ;@ 4438
  dcd Op4440,Op4440,Op4440,Op4440,Op4440,Op4440,Op4440,Op4440 ;@ 4440
  dcd Op4440,Op4440,Op4440,Op4440,Op4440,Op4440,Op4440,Op4440 ;@ 4448
  dcd Op4450,Op4450,Op4450,Op4450,Op4450,Op4450,Op4450,Op4450 ;@ 4450
  dcd Op4458,Op4458,Op4458,Op4458,Op4458,Op4458,Op4458,Op445f ;@ 4458
  dcd Op4460,Op4460,Op4460,Op4460,Op4460,Op4460,Op4460,Op4467 ;@ 4460
  dcd Op4468,Op4468,Op4468,Op4468,Op4468,Op4468,Op4468,Op4468 ;@ 4468
  dcd Op4470,Op4470,Op4470,Op4470,Op4470,Op4470,Op4470,Op4470 ;@ 4470
  dcd Op4478,Op4479,Op447a,Op447b,Op____,Op____,Op____,Op____ ;@ 4478
  dcd Op4480,Op4480,Op4480,Op4480,Op4480,Op4480,Op4480,Op4480 ;@ 4480
  dcd Op4480,Op4480,Op4480,Op4480,Op4480,Op4480,Op4480,Op4480 ;@ 4488
  dcd Op4490,Op4490,Op4490,Op4490,Op4490,Op4490,Op4490,Op4490 ;@ 4490
  dcd Op4498,Op4498,Op4498,Op4498,Op4498,Op4498,Op4498,Op449f ;@ 4498
  dcd Op44a0,Op44a0,Op44a0,Op44a0,Op44a0,Op44a0,Op44a0,Op44a7 ;@ 44a0
  dcd Op44a8,Op44a8,Op44a8,Op44a8,Op44a8,Op44a8,Op44a8,Op44a8 ;@ 44a8
  dcd Op44b0,Op44b0,Op44b0,Op44b0,Op44b0,Op44b0,Op44b0,Op44b0 ;@ 44b0
  dcd Op44b8,Op44b9,Op44ba,Op44bb,Op____,Op____,Op____,Op____ ;@ 44b8
  dcd Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0 ;@ 44c0
  dcd Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0,Op44c0 ;@ 44c8
  dcd Op44d0,Op44d0,Op44d0,Op44d0,Op44d0,Op44d0,Op44d0,Op44d0 ;@ 44d0
  dcd Op44d8,Op44d8,Op44d8,Op44d8,Op44d8,Op44d8,Op44d8,Op44df ;@ 44d8
  dcd Op44e0,Op44e0,Op44e0,Op44e0,Op44e0,Op44e0,Op44e0,Op44e7 ;@ 44e0
  dcd Op44e8,Op44e8,Op44e8,Op44e8,Op44e8,Op44e8,Op44e8,Op44e8 ;@ 44e8
  dcd Op44f0,Op44f0,Op44f0,Op44f0,Op44f0,Op44f0,Op44f0,Op44f0 ;@ 44f0
  dcd Op44f8,Op44f9,Op44fa,Op44fb,Op44fc,Op____,Op____,Op____ ;@ 44f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4508
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4510
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4518
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4520
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4528
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4530
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4538
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4548
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4550
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4558
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4560
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4568
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4570
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4588
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4590
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4598
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45c8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 45d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 45e0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 45e8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 45f0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 45f8
  dcd Op4600,Op4600,Op4600,Op4600,Op4600,Op4600,Op4600,Op4600 ;@ 4600
  dcd Op4600,Op4600,Op4600,Op4600,Op4600,Op4600,Op4600,Op4600 ;@ 4608
  dcd Op4610,Op4610,Op4610,Op4610,Op4610,Op4610,Op4610,Op4610 ;@ 4610
  dcd Op4618,Op4618,Op4618,Op4618,Op4618,Op4618,Op4618,Op461f ;@ 4618
  dcd Op4620,Op4620,Op4620,Op4620,Op4620,Op4620,Op4620,Op4627 ;@ 4620
  dcd Op4628,Op4628,Op4628,Op4628,Op4628,Op4628,Op4628,Op4628 ;@ 4628
  dcd Op4630,Op4630,Op4630,Op4630,Op4630,Op4630,Op4630,Op4630 ;@ 4630
  dcd Op4638,Op4639,Op463a,Op463b,Op____,Op____,Op____,Op____ ;@ 4638
  dcd Op4640,Op4640,Op4640,Op4640,Op4640,Op4640,Op4640,Op4640 ;@ 4640
  dcd Op4640,Op4640,Op4640,Op4640,Op4640,Op4640,Op4640,Op4640 ;@ 4648
  dcd Op4650,Op4650,Op4650,Op4650,Op4650,Op4650,Op4650,Op4650 ;@ 4650
  dcd Op4658,Op4658,Op4658,Op4658,Op4658,Op4658,Op4658,Op465f ;@ 4658
  dcd Op4660,Op4660,Op4660,Op4660,Op4660,Op4660,Op4660,Op4667 ;@ 4660
  dcd Op4668,Op4668,Op4668,Op4668,Op4668,Op4668,Op4668,Op4668 ;@ 4668
  dcd Op4670,Op4670,Op4670,Op4670,Op4670,Op4670,Op4670,Op4670 ;@ 4670
  dcd Op4678,Op4679,Op467a,Op467b,Op____,Op____,Op____,Op____ ;@ 4678
  dcd Op4680,Op4680,Op4680,Op4680,Op4680,Op4680,Op4680,Op4680 ;@ 4680
  dcd Op4680,Op4680,Op4680,Op4680,Op4680,Op4680,Op4680,Op4680 ;@ 4688
  dcd Op4690,Op4690,Op4690,Op4690,Op4690,Op4690,Op4690,Op4690 ;@ 4690
  dcd Op4698,Op4698,Op4698,Op4698,Op4698,Op4698,Op4698,Op469f ;@ 4698
  dcd Op46a0,Op46a0,Op46a0,Op46a0,Op46a0,Op46a0,Op46a0,Op46a7 ;@ 46a0
  dcd Op46a8,Op46a8,Op46a8,Op46a8,Op46a8,Op46a8,Op46a8,Op46a8 ;@ 46a8
  dcd Op46b0,Op46b0,Op46b0,Op46b0,Op46b0,Op46b0,Op46b0,Op46b0 ;@ 46b0
  dcd Op46b8,Op46b9,Op46ba,Op46bb,Op____,Op____,Op____,Op____ ;@ 46b8
  dcd Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0 ;@ 46c0
  dcd Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0,Op46c0 ;@ 46c8
  dcd Op46d0,Op46d0,Op46d0,Op46d0,Op46d0,Op46d0,Op46d0,Op46d0 ;@ 46d0
  dcd Op46d8,Op46d8,Op46d8,Op46d8,Op46d8,Op46d8,Op46d8,Op46df ;@ 46d8
  dcd Op46e0,Op46e0,Op46e0,Op46e0,Op46e0,Op46e0,Op46e0,Op46e7 ;@ 46e0
  dcd Op46e8,Op46e8,Op46e8,Op46e8,Op46e8,Op46e8,Op46e8,Op46e8 ;@ 46e8
  dcd Op46f0,Op46f0,Op46f0,Op46f0,Op46f0,Op46f0,Op46f0,Op46f0 ;@ 46f0
  dcd Op46f8,Op46f9,Op46fa,Op46fb,Op46fc,Op____,Op____,Op____ ;@ 46f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4708
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4710
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4718
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4720
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4728
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4730
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4738
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4748
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4750
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4758
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4760
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4768
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4770
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4788
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4790
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4798
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47c8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 47d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 47e0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 47e8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 47f0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 47f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4800
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4808
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4810
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4818
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4820
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4828
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4830
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4838
  dcd Op4840,Op4840,Op4840,Op4840,Op4840,Op4840,Op4840,Op4840 ;@ 4840
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4848
  dcd Op4850,Op4850,Op4850,Op4850,Op4850,Op4850,Op4850,Op4850 ;@ 4850
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4858
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4860
  dcd Op4868,Op4868,Op4868,Op4868,Op4868,Op4868,Op4868,Op4868 ;@ 4868
  dcd Op4870,Op4870,Op4870,Op4870,Op4870,Op4870,Op4870,Op4870 ;@ 4870
  dcd Op4878,Op4879,Op487a,Op487b,Op____,Op____,Op____,Op____ ;@ 4878
  dcd Op4880,Op4880,Op4880,Op4880,Op4880,Op4880,Op4880,Op4880 ;@ 4880
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4888
  dcd Op4890,Op4890,Op4890,Op4890,Op4890,Op4890,Op4890,Op4890 ;@ 4890
  dcd Op4898,Op4898,Op4898,Op4898,Op4898,Op4898,Op4898,Op489f ;@ 4898
  dcd Op48a0,Op48a0,Op48a0,Op48a0,Op48a0,Op48a0,Op48a0,Op48a7 ;@ 48a0
  dcd Op48a8,Op48a8,Op48a8,Op48a8,Op48a8,Op48a8,Op48a8,Op48a8 ;@ 48a8
  dcd Op48b0,Op48b0,Op48b0,Op48b0,Op48b0,Op48b0,Op48b0,Op48b0 ;@ 48b0
  dcd Op48b8,Op48b9,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 48b8
  dcd Op48c0,Op48c0,Op48c0,Op48c0,Op48c0,Op48c0,Op48c0,Op48c0 ;@ 48c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 48c8
  dcd Op48d0,Op48d0,Op48d0,Op48d0,Op48d0,Op48d0,Op48d0,Op48d0 ;@ 48d0
  dcd Op48d8,Op48d8,Op48d8,Op48d8,Op48d8,Op48d8,Op48d8,Op48df ;@ 48d8
  dcd Op48e0,Op48e0,Op48e0,Op48e0,Op48e0,Op48e0,Op48e0,Op48e7 ;@ 48e0
  dcd Op48e8,Op48e8,Op48e8,Op48e8,Op48e8,Op48e8,Op48e8,Op48e8 ;@ 48e8
  dcd Op48f0,Op48f0,Op48f0,Op48f0,Op48f0,Op48f0,Op48f0,Op48f0 ;@ 48f0
  dcd Op48f8,Op48f9,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 48f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4908
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4910
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4918
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4920
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4928
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4930
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4938
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4948
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4950
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4958
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4960
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4968
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4970
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4988
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4990
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4998
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49c8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 49d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 49e0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 49e8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 49f0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 49f8
  dcd Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00 ;@ 4a00
  dcd Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00,Op4a00 ;@ 4a08
  dcd Op4a10,Op4a10,Op4a10,Op4a10,Op4a10,Op4a10,Op4a10,Op4a10 ;@ 4a10
  dcd Op4a18,Op4a18,Op4a18,Op4a18,Op4a18,Op4a18,Op4a18,Op4a1f ;@ 4a18
  dcd Op4a20,Op4a20,Op4a20,Op4a20,Op4a20,Op4a20,Op4a20,Op4a27 ;@ 4a20
  dcd Op4a28,Op4a28,Op4a28,Op4a28,Op4a28,Op4a28,Op4a28,Op4a28 ;@ 4a28
  dcd Op4a30,Op4a30,Op4a30,Op4a30,Op4a30,Op4a30,Op4a30,Op4a30 ;@ 4a30
  dcd Op4a38,Op4a39,Op4a3a,Op4a3b,Op____,Op____,Op____,Op____ ;@ 4a38
  dcd Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40 ;@ 4a40
  dcd Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40,Op4a40 ;@ 4a48
  dcd Op4a50,Op4a50,Op4a50,Op4a50,Op4a50,Op4a50,Op4a50,Op4a50 ;@ 4a50
  dcd Op4a58,Op4a58,Op4a58,Op4a58,Op4a58,Op4a58,Op4a58,Op4a5f ;@ 4a58
  dcd Op4a60,Op4a60,Op4a60,Op4a60,Op4a60,Op4a60,Op4a60,Op4a67 ;@ 4a60
  dcd Op4a68,Op4a68,Op4a68,Op4a68,Op4a68,Op4a68,Op4a68,Op4a68 ;@ 4a68
  dcd Op4a70,Op4a70,Op4a70,Op4a70,Op4a70,Op4a70,Op4a70,Op4a70 ;@ 4a70
  dcd Op4a78,Op4a79,Op4a7a,Op4a7b,Op____,Op____,Op____,Op____ ;@ 4a78
  dcd Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80 ;@ 4a80
  dcd Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80,Op4a80 ;@ 4a88
  dcd Op4a90,Op4a90,Op4a90,Op4a90,Op4a90,Op4a90,Op4a90,Op4a90 ;@ 4a90
  dcd Op4a98,Op4a98,Op4a98,Op4a98,Op4a98,Op4a98,Op4a98,Op4a9f ;@ 4a98
  dcd Op4aa0,Op4aa0,Op4aa0,Op4aa0,Op4aa0,Op4aa0,Op4aa0,Op4aa7 ;@ 4aa0
  dcd Op4aa8,Op4aa8,Op4aa8,Op4aa8,Op4aa8,Op4aa8,Op4aa8,Op4aa8 ;@ 4aa8
  dcd Op4ab0,Op4ab0,Op4ab0,Op4ab0,Op4ab0,Op4ab0,Op4ab0,Op4ab0 ;@ 4ab0
  dcd Op4ab8,Op4ab9,Op4aba,Op4abb,Op____,Op____,Op____,Op____ ;@ 4ab8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ac0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ac8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ad0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ad8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ae0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ae8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4af0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4af8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4b98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ba0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ba8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4bb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4bb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4bc8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 4bd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4bd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4be0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 4be8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 4bf0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 4bf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4c88
  dcd Op4c90,Op4c90,Op4c90,Op4c90,Op4c90,Op4c90,Op4c90,Op4c90 ;@ 4c90
  dcd Op4c98,Op4c98,Op4c98,Op4c98,Op4c98,Op4c98,Op4c98,Op4c9f ;@ 4c98
  dcd Op4ca0,Op4ca0,Op4ca0,Op4ca0,Op4ca0,Op4ca0,Op4ca0,Op4ca7 ;@ 4ca0
  dcd Op4ca8,Op4ca8,Op4ca8,Op4ca8,Op4ca8,Op4ca8,Op4ca8,Op4ca8 ;@ 4ca8
  dcd Op4cb0,Op4cb0,Op4cb0,Op4cb0,Op4cb0,Op4cb0,Op4cb0,Op4cb0 ;@ 4cb0
  dcd Op4cb8,Op4cb9,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4cb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4cc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4cc8
  dcd Op4cd0,Op4cd0,Op4cd0,Op4cd0,Op4cd0,Op4cd0,Op4cd0,Op4cd0 ;@ 4cd0
  dcd Op4cd8,Op4cd8,Op4cd8,Op4cd8,Op4cd8,Op4cd8,Op4cd8,Op4cdf ;@ 4cd8
  dcd Op4ce0,Op4ce0,Op4ce0,Op4ce0,Op4ce0,Op4ce0,Op4ce0,Op4ce7 ;@ 4ce0
  dcd Op4ce8,Op4ce8,Op4ce8,Op4ce8,Op4ce8,Op4ce8,Op4ce8,Op4ce8 ;@ 4ce8
  dcd Op4cf0,Op4cf0,Op4cf0,Op4cf0,Op4cf0,Op4cf0,Op4cf0,Op4cf0 ;@ 4cf0
  dcd Op4cf8,Op4cf9,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4cf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4d98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4da0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4da8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4db0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4db8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4dc8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 4dd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4dd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4de0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 4de8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 4df0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 4df8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e38
  dcd Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40 ;@ 4e40
  dcd Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40,Op4e40 ;@ 4e48
  dcd Op4e50,Op4e50,Op4e50,Op4e50,Op4e50,Op4e50,Op4e50,Op4e50 ;@ 4e50
  dcd Op4e58,Op4e58,Op4e58,Op4e58,Op4e58,Op4e58,Op4e58,Op4e58 ;@ 4e58
  dcd Op4e60,Op4e60,Op4e60,Op4e60,Op4e60,Op4e60,Op4e60,Op4e60 ;@ 4e60
  dcd Op4e68,Op4e68,Op4e68,Op4e68,Op4e68,Op4e68,Op4e68,Op4e68 ;@ 4e68
  dcd Op____,Op4e71,Op____,Op4e73,Op____,Op4e75,Op____,Op4e77 ;@ 4e70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e88
  dcd Op4e90,Op4e90,Op4e90,Op4e90,Op4e90,Op4e90,Op4e90,Op4e90 ;@ 4e90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4e98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ea0
  dcd Op4ea8,Op4ea8,Op4ea8,Op4ea8,Op4ea8,Op4ea8,Op4ea8,Op4ea8 ;@ 4ea8
  dcd Op4eb0,Op4eb0,Op4eb0,Op4eb0,Op4eb0,Op4eb0,Op4eb0,Op4eb0 ;@ 4eb0
  dcd Op4eb8,Op4eb9,Op4eba,Op4ebb,Op____,Op____,Op____,Op____ ;@ 4eb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ec0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ec8
  dcd Op4ed0,Op4ed0,Op4ed0,Op4ed0,Op4ed0,Op4ed0,Op4ed0,Op4ed0 ;@ 4ed0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ed8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4ee0
  dcd Op4ee8,Op4ee8,Op4ee8,Op4ee8,Op4ee8,Op4ee8,Op4ee8,Op4ee8 ;@ 4ee8
  dcd Op4ef0,Op4ef0,Op4ef0,Op4ef0,Op4ef0,Op4ef0,Op4ef0,Op4ef0 ;@ 4ef0
  dcd Op4ef8,Op4ef9,Op4efa,Op4efb,Op____,Op____,Op____,Op____ ;@ 4ef8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4f98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fc8
  dcd Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0,Op41d0 ;@ 4fd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 4fe0
  dcd Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8,Op41e8 ;@ 4fe8
  dcd Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0,Op41f0 ;@ 4ff0
  dcd Op41f8,Op41f9,Op41fa,Op41fb,Op____,Op____,Op____,Op____ ;@ 4ff8
  dcd Op5000,Op5000,Op5000,Op5000,Op5000,Op5000,Op5000,Op5000 ;@ 5000
  dcd Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008 ;@ 5008
  dcd Op5010,Op5010,Op5010,Op5010,Op5010,Op5010,Op5010,Op5010 ;@ 5010
  dcd Op5018,Op5018,Op5018,Op5018,Op5018,Op5018,Op5018,Op5018 ;@ 5018
  dcd Op5020,Op5020,Op5020,Op5020,Op5020,Op5020,Op5020,Op5020 ;@ 5020
  dcd Op5028,Op5028,Op5028,Op5028,Op5028,Op5028,Op5028,Op5028 ;@ 5028
  dcd Op5030,Op5030,Op5030,Op5030,Op5030,Op5030,Op5030,Op5030 ;@ 5030
  dcd Op5038,Op5039,Op503a,Op503b,Op____,Op____,Op____,Op____ ;@ 5038
  dcd Op5040,Op5040,Op5040,Op5040,Op5040,Op5040,Op5040,Op5040 ;@ 5040
  dcd Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008 ;@ 5048
  dcd Op5050,Op5050,Op5050,Op5050,Op5050,Op5050,Op5050,Op5050 ;@ 5050
  dcd Op5058,Op5058,Op5058,Op5058,Op5058,Op5058,Op5058,Op5058 ;@ 5058
  dcd Op5060,Op5060,Op5060,Op5060,Op5060,Op5060,Op5060,Op5060 ;@ 5060
  dcd Op5068,Op5068,Op5068,Op5068,Op5068,Op5068,Op5068,Op5068 ;@ 5068
  dcd Op5070,Op5070,Op5070,Op5070,Op5070,Op5070,Op5070,Op5070 ;@ 5070
  dcd Op5078,Op5079,Op507a,Op507b,Op____,Op____,Op____,Op____ ;@ 5078
  dcd Op5080,Op5080,Op5080,Op5080,Op5080,Op5080,Op5080,Op5080 ;@ 5080
  dcd Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008,Op5008 ;@ 5088
  dcd Op5090,Op5090,Op5090,Op5090,Op5090,Op5090,Op5090,Op5090 ;@ 5090
  dcd Op5098,Op5098,Op5098,Op5098,Op5098,Op5098,Op5098,Op5098 ;@ 5098
  dcd Op50a0,Op50a0,Op50a0,Op50a0,Op50a0,Op50a0,Op50a0,Op50a0 ;@ 50a0
  dcd Op50a8,Op50a8,Op50a8,Op50a8,Op50a8,Op50a8,Op50a8,Op50a8 ;@ 50a8
  dcd Op50b0,Op50b0,Op50b0,Op50b0,Op50b0,Op50b0,Op50b0,Op50b0 ;@ 50b0
  dcd Op50b8,Op50b9,Op50ba,Op50bb,Op____,Op____,Op____,Op____ ;@ 50b8
  dcd Op50c0,Op50c0,Op50c0,Op50c0,Op50c0,Op50c0,Op50c0,Op50c0 ;@ 50c0
  dcd Op50c8,Op50c8,Op50c8,Op50c8,Op50c8,Op50c8,Op50c8,Op50c8 ;@ 50c8
  dcd Op50d0,Op50d0,Op50d0,Op50d0,Op50d0,Op50d0,Op50d0,Op50d0 ;@ 50d0
  dcd Op50d8,Op50d8,Op50d8,Op50d8,Op50d8,Op50d8,Op50d8,Op50df ;@ 50d8
  dcd Op50e0,Op50e0,Op50e0,Op50e0,Op50e0,Op50e0,Op50e0,Op50e7 ;@ 50e0
  dcd Op50e8,Op50e8,Op50e8,Op50e8,Op50e8,Op50e8,Op50e8,Op50e8 ;@ 50e8
  dcd Op50f0,Op50f0,Op50f0,Op50f0,Op50f0,Op50f0,Op50f0,Op50f0 ;@ 50f0
  dcd Op50f8,Op50f9,Op50fa,Op50fb,Op____,Op____,Op____,Op____ ;@ 50f8
  dcd Op5100,Op5100,Op5100,Op5100,Op5100,Op5100,Op5100,Op5100 ;@ 5100
  dcd Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108 ;@ 5108
  dcd Op5110,Op5110,Op5110,Op5110,Op5110,Op5110,Op5110,Op5110 ;@ 5110
  dcd Op5118,Op5118,Op5118,Op5118,Op5118,Op5118,Op5118,Op5118 ;@ 5118
  dcd Op5120,Op5120,Op5120,Op5120,Op5120,Op5120,Op5120,Op5120 ;@ 5120
  dcd Op5128,Op5128,Op5128,Op5128,Op5128,Op5128,Op5128,Op5128 ;@ 5128
  dcd Op5130,Op5130,Op5130,Op5130,Op5130,Op5130,Op5130,Op5130 ;@ 5130
  dcd Op5138,Op5139,Op513a,Op513b,Op____,Op____,Op____,Op____ ;@ 5138
  dcd Op5140,Op5140,Op5140,Op5140,Op5140,Op5140,Op5140,Op5140 ;@ 5140
  dcd Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108 ;@ 5148
  dcd Op5150,Op5150,Op5150,Op5150,Op5150,Op5150,Op5150,Op5150 ;@ 5150
  dcd Op5158,Op5158,Op5158,Op5158,Op5158,Op5158,Op5158,Op5158 ;@ 5158
  dcd Op5160,Op5160,Op5160,Op5160,Op5160,Op5160,Op5160,Op5160 ;@ 5160
  dcd Op5168,Op5168,Op5168,Op5168,Op5168,Op5168,Op5168,Op5168 ;@ 5168
  dcd Op5170,Op5170,Op5170,Op5170,Op5170,Op5170,Op5170,Op5170 ;@ 5170
  dcd Op5178,Op5179,Op517a,Op517b,Op____,Op____,Op____,Op____ ;@ 5178
  dcd Op5180,Op5180,Op5180,Op5180,Op5180,Op5180,Op5180,Op5180 ;@ 5180
  dcd Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108,Op5108 ;@ 5188
  dcd Op5190,Op5190,Op5190,Op5190,Op5190,Op5190,Op5190,Op5190 ;@ 5190
  dcd Op5198,Op5198,Op5198,Op5198,Op5198,Op5198,Op5198,Op5198 ;@ 5198
  dcd Op51a0,Op51a0,Op51a0,Op51a0,Op51a0,Op51a0,Op51a0,Op51a0 ;@ 51a0
  dcd Op51a8,Op51a8,Op51a8,Op51a8,Op51a8,Op51a8,Op51a8,Op51a8 ;@ 51a8
  dcd Op51b0,Op51b0,Op51b0,Op51b0,Op51b0,Op51b0,Op51b0,Op51b0 ;@ 51b0
  dcd Op51b8,Op51b9,Op51ba,Op51bb,Op____,Op____,Op____,Op____ ;@ 51b8
  dcd Op51c0,Op51c0,Op51c0,Op51c0,Op51c0,Op51c0,Op51c0,Op51c0 ;@ 51c0
  dcd Op51c8,Op51c8,Op51c8,Op51c8,Op51c8,Op51c8,Op51c8,Op51c8 ;@ 51c8
  dcd Op51d0,Op51d0,Op51d0,Op51d0,Op51d0,Op51d0,Op51d0,Op51d0 ;@ 51d0
  dcd Op51d8,Op51d8,Op51d8,Op51d8,Op51d8,Op51d8,Op51d8,Op51df ;@ 51d8
  dcd Op51e0,Op51e0,Op51e0,Op51e0,Op51e0,Op51e0,Op51e0,Op51e7 ;@ 51e0
  dcd Op51e8,Op51e8,Op51e8,Op51e8,Op51e8,Op51e8,Op51e8,Op51e8 ;@ 51e8
  dcd Op51f0,Op51f0,Op51f0,Op51f0,Op51f0,Op51f0,Op51f0,Op51f0 ;@ 51f0
  dcd Op51f8,Op51f9,Op51fa,Op51fb,Op____,Op____,Op____,Op____ ;@ 51f8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5200
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5208
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5210
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5218
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5220
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5228
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5230
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5238
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5240
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5248
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5250
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5258
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5260
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5268
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5270
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5278
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5280
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5288
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5290
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5298
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 52a0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 52a8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 52b0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 52b8
  dcd Op52c0,Op52c0,Op52c0,Op52c0,Op52c0,Op52c0,Op52c0,Op52c0 ;@ 52c0
  dcd Op52c8,Op52c8,Op52c8,Op52c8,Op52c8,Op52c8,Op52c8,Op52c8 ;@ 52c8
  dcd Op52d0,Op52d0,Op52d0,Op52d0,Op52d0,Op52d0,Op52d0,Op52d0 ;@ 52d0
  dcd Op52d8,Op52d8,Op52d8,Op52d8,Op52d8,Op52d8,Op52d8,Op52df ;@ 52d8
  dcd Op52e0,Op52e0,Op52e0,Op52e0,Op52e0,Op52e0,Op52e0,Op52e7 ;@ 52e0
  dcd Op52e8,Op52e8,Op52e8,Op52e8,Op52e8,Op52e8,Op52e8,Op52e8 ;@ 52e8
  dcd Op52f0,Op52f0,Op52f0,Op52f0,Op52f0,Op52f0,Op52f0,Op52f0 ;@ 52f0
  dcd Op52f8,Op52f9,Op52fa,Op52fb,Op____,Op____,Op____,Op____ ;@ 52f8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5300
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5308
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5310
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5318
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5320
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5328
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5330
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5338
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5340
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5348
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5350
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5358
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5360
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5368
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5370
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5378
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5380
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5388
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5390
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5398
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 53a0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 53a8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 53b0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 53b8
  dcd Op53c0,Op53c0,Op53c0,Op53c0,Op53c0,Op53c0,Op53c0,Op53c0 ;@ 53c0
  dcd Op53c8,Op53c8,Op53c8,Op53c8,Op53c8,Op53c8,Op53c8,Op53c8 ;@ 53c8
  dcd Op53d0,Op53d0,Op53d0,Op53d0,Op53d0,Op53d0,Op53d0,Op53d0 ;@ 53d0
  dcd Op53d8,Op53d8,Op53d8,Op53d8,Op53d8,Op53d8,Op53d8,Op53df ;@ 53d8
  dcd Op53e0,Op53e0,Op53e0,Op53e0,Op53e0,Op53e0,Op53e0,Op53e7 ;@ 53e0
  dcd Op53e8,Op53e8,Op53e8,Op53e8,Op53e8,Op53e8,Op53e8,Op53e8 ;@ 53e8
  dcd Op53f0,Op53f0,Op53f0,Op53f0,Op53f0,Op53f0,Op53f0,Op53f0 ;@ 53f0
  dcd Op53f8,Op53f9,Op53fa,Op53fb,Op____,Op____,Op____,Op____ ;@ 53f8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5400
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5408
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5410
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5418
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5420
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5428
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5430
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5438
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5440
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5448
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5450
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5458
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5460
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5468
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5470
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5478
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5480
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5488
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5490
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5498
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 54a0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 54a8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 54b0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 54b8
  dcd Op54c0,Op54c0,Op54c0,Op54c0,Op54c0,Op54c0,Op54c0,Op54c0 ;@ 54c0
  dcd Op54c8,Op54c8,Op54c8,Op54c8,Op54c8,Op54c8,Op54c8,Op54c8 ;@ 54c8
  dcd Op54d0,Op54d0,Op54d0,Op54d0,Op54d0,Op54d0,Op54d0,Op54d0 ;@ 54d0
  dcd Op54d8,Op54d8,Op54d8,Op54d8,Op54d8,Op54d8,Op54d8,Op54df ;@ 54d8
  dcd Op54e0,Op54e0,Op54e0,Op54e0,Op54e0,Op54e0,Op54e0,Op54e7 ;@ 54e0
  dcd Op54e8,Op54e8,Op54e8,Op54e8,Op54e8,Op54e8,Op54e8,Op54e8 ;@ 54e8
  dcd Op54f0,Op54f0,Op54f0,Op54f0,Op54f0,Op54f0,Op54f0,Op54f0 ;@ 54f0
  dcd Op54f8,Op54f9,Op54fa,Op54fb,Op____,Op____,Op____,Op____ ;@ 54f8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5500
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5508
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5510
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5518
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5520
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5528
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5530
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5538
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5540
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5548
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5550
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5558
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5560
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5568
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5570
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5578
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5580
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5588
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5590
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5598
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 55a0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 55a8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 55b0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 55b8
  dcd Op55c0,Op55c0,Op55c0,Op55c0,Op55c0,Op55c0,Op55c0,Op55c0 ;@ 55c0
  dcd Op55c8,Op55c8,Op55c8,Op55c8,Op55c8,Op55c8,Op55c8,Op55c8 ;@ 55c8
  dcd Op55d0,Op55d0,Op55d0,Op55d0,Op55d0,Op55d0,Op55d0,Op55d0 ;@ 55d0
  dcd Op55d8,Op55d8,Op55d8,Op55d8,Op55d8,Op55d8,Op55d8,Op55df ;@ 55d8
  dcd Op55e0,Op55e0,Op55e0,Op55e0,Op55e0,Op55e0,Op55e0,Op55e7 ;@ 55e0
  dcd Op55e8,Op55e8,Op55e8,Op55e8,Op55e8,Op55e8,Op55e8,Op55e8 ;@ 55e8
  dcd Op55f0,Op55f0,Op55f0,Op55f0,Op55f0,Op55f0,Op55f0,Op55f0 ;@ 55f0
  dcd Op55f8,Op55f9,Op55fa,Op55fb,Op____,Op____,Op____,Op____ ;@ 55f8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5600
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5608
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5610
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5618
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5620
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5628
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5630
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5638
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5640
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5648
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5650
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5658
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5660
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5668
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5670
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5678
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5680
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5688
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5690
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5698
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 56a0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 56a8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 56b0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 56b8
  dcd Op56c0,Op56c0,Op56c0,Op56c0,Op56c0,Op56c0,Op56c0,Op56c0 ;@ 56c0
  dcd Op56c8,Op56c8,Op56c8,Op56c8,Op56c8,Op56c8,Op56c8,Op56c8 ;@ 56c8
  dcd Op56d0,Op56d0,Op56d0,Op56d0,Op56d0,Op56d0,Op56d0,Op56d0 ;@ 56d0
  dcd Op56d8,Op56d8,Op56d8,Op56d8,Op56d8,Op56d8,Op56d8,Op56df ;@ 56d8
  dcd Op56e0,Op56e0,Op56e0,Op56e0,Op56e0,Op56e0,Op56e0,Op56e7 ;@ 56e0
  dcd Op56e8,Op56e8,Op56e8,Op56e8,Op56e8,Op56e8,Op56e8,Op56e8 ;@ 56e8
  dcd Op56f0,Op56f0,Op56f0,Op56f0,Op56f0,Op56f0,Op56f0,Op56f0 ;@ 56f0
  dcd Op56f8,Op56f9,Op56fa,Op56fb,Op____,Op____,Op____,Op____ ;@ 56f8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5700
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5708
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5710
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5718
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5720
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5728
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5730
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5738
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5740
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5748
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5750
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5758
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5760
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5768
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5770
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5778
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5780
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5788
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5790
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5798
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 57a0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 57a8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 57b0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 57b8
  dcd Op57c0,Op57c0,Op57c0,Op57c0,Op57c0,Op57c0,Op57c0,Op57c0 ;@ 57c0
  dcd Op57c8,Op57c8,Op57c8,Op57c8,Op57c8,Op57c8,Op57c8,Op57c8 ;@ 57c8
  dcd Op57d0,Op57d0,Op57d0,Op57d0,Op57d0,Op57d0,Op57d0,Op57d0 ;@ 57d0
  dcd Op57d8,Op57d8,Op57d8,Op57d8,Op57d8,Op57d8,Op57d8,Op57df ;@ 57d8
  dcd Op57e0,Op57e0,Op57e0,Op57e0,Op57e0,Op57e0,Op57e0,Op57e7 ;@ 57e0
  dcd Op57e8,Op57e8,Op57e8,Op57e8,Op57e8,Op57e8,Op57e8,Op57e8 ;@ 57e8
  dcd Op57f0,Op57f0,Op57f0,Op57f0,Op57f0,Op57f0,Op57f0,Op57f0 ;@ 57f0
  dcd Op57f8,Op57f9,Op57fa,Op57fb,Op____,Op____,Op____,Op____ ;@ 57f8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5800
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5808
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5810
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5818
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5820
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5828
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5830
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5838
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5840
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5848
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5850
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5858
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5860
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5868
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5870
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5878
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5880
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5888
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5890
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5898
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 58a0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 58a8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 58b0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 58b8
  dcd Op58c0,Op58c0,Op58c0,Op58c0,Op58c0,Op58c0,Op58c0,Op58c0 ;@ 58c0
  dcd Op58c8,Op58c8,Op58c8,Op58c8,Op58c8,Op58c8,Op58c8,Op58c8 ;@ 58c8
  dcd Op58d0,Op58d0,Op58d0,Op58d0,Op58d0,Op58d0,Op58d0,Op58d0 ;@ 58d0
  dcd Op58d8,Op58d8,Op58d8,Op58d8,Op58d8,Op58d8,Op58d8,Op58df ;@ 58d8
  dcd Op58e0,Op58e0,Op58e0,Op58e0,Op58e0,Op58e0,Op58e0,Op58e7 ;@ 58e0
  dcd Op58e8,Op58e8,Op58e8,Op58e8,Op58e8,Op58e8,Op58e8,Op58e8 ;@ 58e8
  dcd Op58f0,Op58f0,Op58f0,Op58f0,Op58f0,Op58f0,Op58f0,Op58f0 ;@ 58f0
  dcd Op58f8,Op58f9,Op58fa,Op58fb,Op____,Op____,Op____,Op____ ;@ 58f8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5900
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5908
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5910
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5918
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5920
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5928
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5930
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5938
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5940
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5948
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5950
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5958
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5960
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5968
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5970
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5978
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5980
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5988
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5990
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5998
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 59a0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 59a8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 59b0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 59b8
  dcd Op59c0,Op59c0,Op59c0,Op59c0,Op59c0,Op59c0,Op59c0,Op59c0 ;@ 59c0
  dcd Op59c8,Op59c8,Op59c8,Op59c8,Op59c8,Op59c8,Op59c8,Op59c8 ;@ 59c8
  dcd Op59d0,Op59d0,Op59d0,Op59d0,Op59d0,Op59d0,Op59d0,Op59d0 ;@ 59d0
  dcd Op59d8,Op59d8,Op59d8,Op59d8,Op59d8,Op59d8,Op59d8,Op59df ;@ 59d8
  dcd Op59e0,Op59e0,Op59e0,Op59e0,Op59e0,Op59e0,Op59e0,Op59e7 ;@ 59e0
  dcd Op59e8,Op59e8,Op59e8,Op59e8,Op59e8,Op59e8,Op59e8,Op59e8 ;@ 59e8
  dcd Op59f0,Op59f0,Op59f0,Op59f0,Op59f0,Op59f0,Op59f0,Op59f0 ;@ 59f0
  dcd Op59f8,Op59f9,Op59fa,Op59fb,Op____,Op____,Op____,Op____ ;@ 59f8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5a00
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5a08
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5a10
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5a18
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5a20
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5a28
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5a30
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5a38
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5a40
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5a48
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5a50
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5a58
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5a60
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5a68
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5a70
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5a78
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5a80
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5a88
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5a90
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5a98
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 5aa0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 5aa8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 5ab0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 5ab8
  dcd Op5ac0,Op5ac0,Op5ac0,Op5ac0,Op5ac0,Op5ac0,Op5ac0,Op5ac0 ;@ 5ac0
  dcd Op5ac8,Op5ac8,Op5ac8,Op5ac8,Op5ac8,Op5ac8,Op5ac8,Op5ac8 ;@ 5ac8
  dcd Op5ad0,Op5ad0,Op5ad0,Op5ad0,Op5ad0,Op5ad0,Op5ad0,Op5ad0 ;@ 5ad0
  dcd Op5ad8,Op5ad8,Op5ad8,Op5ad8,Op5ad8,Op5ad8,Op5ad8,Op5adf ;@ 5ad8
  dcd Op5ae0,Op5ae0,Op5ae0,Op5ae0,Op5ae0,Op5ae0,Op5ae0,Op5ae7 ;@ 5ae0
  dcd Op5ae8,Op5ae8,Op5ae8,Op5ae8,Op5ae8,Op5ae8,Op5ae8,Op5ae8 ;@ 5ae8
  dcd Op5af0,Op5af0,Op5af0,Op5af0,Op5af0,Op5af0,Op5af0,Op5af0 ;@ 5af0
  dcd Op5af8,Op5af9,Op5afa,Op5afb,Op____,Op____,Op____,Op____ ;@ 5af8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5b00
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5b08
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5b10
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5b18
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5b20
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5b28
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5b30
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5b38
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5b40
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5b48
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5b50
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5b58
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5b60
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5b68
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5b70
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5b78
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5b80
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5b88
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5b90
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5b98
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 5ba0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 5ba8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 5bb0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 5bb8
  dcd Op5bc0,Op5bc0,Op5bc0,Op5bc0,Op5bc0,Op5bc0,Op5bc0,Op5bc0 ;@ 5bc0
  dcd Op5bc8,Op5bc8,Op5bc8,Op5bc8,Op5bc8,Op5bc8,Op5bc8,Op5bc8 ;@ 5bc8
  dcd Op5bd0,Op5bd0,Op5bd0,Op5bd0,Op5bd0,Op5bd0,Op5bd0,Op5bd0 ;@ 5bd0
  dcd Op5bd8,Op5bd8,Op5bd8,Op5bd8,Op5bd8,Op5bd8,Op5bd8,Op5bdf ;@ 5bd8
  dcd Op5be0,Op5be0,Op5be0,Op5be0,Op5be0,Op5be0,Op5be0,Op5be7 ;@ 5be0
  dcd Op5be8,Op5be8,Op5be8,Op5be8,Op5be8,Op5be8,Op5be8,Op5be8 ;@ 5be8
  dcd Op5bf0,Op5bf0,Op5bf0,Op5bf0,Op5bf0,Op5bf0,Op5bf0,Op5bf0 ;@ 5bf0
  dcd Op5bf8,Op5bf9,Op5bfa,Op5bfb,Op____,Op____,Op____,Op____ ;@ 5bf8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5c00
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5c08
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5c10
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5c18
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5c20
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5c28
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5c30
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5c38
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5c40
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5c48
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5c50
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5c58
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5c60
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5c68
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5c70
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5c78
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5c80
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5c88
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5c90
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5c98
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 5ca0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 5ca8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 5cb0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 5cb8
  dcd Op5cc0,Op5cc0,Op5cc0,Op5cc0,Op5cc0,Op5cc0,Op5cc0,Op5cc0 ;@ 5cc0
  dcd Op5cc8,Op5cc8,Op5cc8,Op5cc8,Op5cc8,Op5cc8,Op5cc8,Op5cc8 ;@ 5cc8
  dcd Op5cd0,Op5cd0,Op5cd0,Op5cd0,Op5cd0,Op5cd0,Op5cd0,Op5cd0 ;@ 5cd0
  dcd Op5cd8,Op5cd8,Op5cd8,Op5cd8,Op5cd8,Op5cd8,Op5cd8,Op5cdf ;@ 5cd8
  dcd Op5ce0,Op5ce0,Op5ce0,Op5ce0,Op5ce0,Op5ce0,Op5ce0,Op5ce7 ;@ 5ce0
  dcd Op5ce8,Op5ce8,Op5ce8,Op5ce8,Op5ce8,Op5ce8,Op5ce8,Op5ce8 ;@ 5ce8
  dcd Op5cf0,Op5cf0,Op5cf0,Op5cf0,Op5cf0,Op5cf0,Op5cf0,Op5cf0 ;@ 5cf0
  dcd Op5cf8,Op5cf9,Op5cfa,Op5cfb,Op____,Op____,Op____,Op____ ;@ 5cf8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5d00
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5d08
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5d10
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5d18
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5d20
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5d28
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5d30
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5d38
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5d40
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5d48
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5d50
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5d58
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5d60
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5d68
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5d70
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5d78
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5d80
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5d88
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5d90
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5d98
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 5da0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 5da8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 5db0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 5db8
  dcd Op5dc0,Op5dc0,Op5dc0,Op5dc0,Op5dc0,Op5dc0,Op5dc0,Op5dc0 ;@ 5dc0
  dcd Op5dc8,Op5dc8,Op5dc8,Op5dc8,Op5dc8,Op5dc8,Op5dc8,Op5dc8 ;@ 5dc8
  dcd Op5dd0,Op5dd0,Op5dd0,Op5dd0,Op5dd0,Op5dd0,Op5dd0,Op5dd0 ;@ 5dd0
  dcd Op5dd8,Op5dd8,Op5dd8,Op5dd8,Op5dd8,Op5dd8,Op5dd8,Op5ddf ;@ 5dd8
  dcd Op5de0,Op5de0,Op5de0,Op5de0,Op5de0,Op5de0,Op5de0,Op5de7 ;@ 5de0
  dcd Op5de8,Op5de8,Op5de8,Op5de8,Op5de8,Op5de8,Op5de8,Op5de8 ;@ 5de8
  dcd Op5df0,Op5df0,Op5df0,Op5df0,Op5df0,Op5df0,Op5df0,Op5df0 ;@ 5df0
  dcd Op5df8,Op5df9,Op5dfa,Op5dfb,Op____,Op____,Op____,Op____ ;@ 5df8
  dcd Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00,Op5e00 ;@ 5e00
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5e08
  dcd Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10,Op5e10 ;@ 5e10
  dcd Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18,Op5e18 ;@ 5e18
  dcd Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20,Op5e20 ;@ 5e20
  dcd Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28,Op5e28 ;@ 5e28
  dcd Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30,Op5e30 ;@ 5e30
  dcd Op5e38,Op5e39,Op5e3a,Op5e3b,Op____,Op____,Op____,Op____ ;@ 5e38
  dcd Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40,Op5e40 ;@ 5e40
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5e48
  dcd Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50,Op5e50 ;@ 5e50
  dcd Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58,Op5e58 ;@ 5e58
  dcd Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60,Op5e60 ;@ 5e60
  dcd Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68,Op5e68 ;@ 5e68
  dcd Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70,Op5e70 ;@ 5e70
  dcd Op5e78,Op5e79,Op5e7a,Op5e7b,Op____,Op____,Op____,Op____ ;@ 5e78
  dcd Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80,Op5e80 ;@ 5e80
  dcd Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08,Op5e08 ;@ 5e88
  dcd Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90,Op5e90 ;@ 5e90
  dcd Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98,Op5e98 ;@ 5e98
  dcd Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0,Op5ea0 ;@ 5ea0
  dcd Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8,Op5ea8 ;@ 5ea8
  dcd Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0,Op5eb0 ;@ 5eb0
  dcd Op5eb8,Op5eb9,Op5eba,Op5ebb,Op____,Op____,Op____,Op____ ;@ 5eb8
  dcd Op5ec0,Op5ec0,Op5ec0,Op5ec0,Op5ec0,Op5ec0,Op5ec0,Op5ec0 ;@ 5ec0
  dcd Op5ec8,Op5ec8,Op5ec8,Op5ec8,Op5ec8,Op5ec8,Op5ec8,Op5ec8 ;@ 5ec8
  dcd Op5ed0,Op5ed0,Op5ed0,Op5ed0,Op5ed0,Op5ed0,Op5ed0,Op5ed0 ;@ 5ed0
  dcd Op5ed8,Op5ed8,Op5ed8,Op5ed8,Op5ed8,Op5ed8,Op5ed8,Op5edf ;@ 5ed8
  dcd Op5ee0,Op5ee0,Op5ee0,Op5ee0,Op5ee0,Op5ee0,Op5ee0,Op5ee7 ;@ 5ee0
  dcd Op5ee8,Op5ee8,Op5ee8,Op5ee8,Op5ee8,Op5ee8,Op5ee8,Op5ee8 ;@ 5ee8
  dcd Op5ef0,Op5ef0,Op5ef0,Op5ef0,Op5ef0,Op5ef0,Op5ef0,Op5ef0 ;@ 5ef0
  dcd Op5ef8,Op5ef9,Op5efa,Op5efb,Op____,Op____,Op____,Op____ ;@ 5ef8
  dcd Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00,Op5f00 ;@ 5f00
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5f08
  dcd Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10,Op5f10 ;@ 5f10
  dcd Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18,Op5f18 ;@ 5f18
  dcd Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20,Op5f20 ;@ 5f20
  dcd Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28,Op5f28 ;@ 5f28
  dcd Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30,Op5f30 ;@ 5f30
  dcd Op5f38,Op5f39,Op5f3a,Op5f3b,Op____,Op____,Op____,Op____ ;@ 5f38
  dcd Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40,Op5f40 ;@ 5f40
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5f48
  dcd Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50,Op5f50 ;@ 5f50
  dcd Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58,Op5f58 ;@ 5f58
  dcd Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60,Op5f60 ;@ 5f60
  dcd Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68,Op5f68 ;@ 5f68
  dcd Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70,Op5f70 ;@ 5f70
  dcd Op5f78,Op5f79,Op5f7a,Op5f7b,Op____,Op____,Op____,Op____ ;@ 5f78
  dcd Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80,Op5f80 ;@ 5f80
  dcd Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08,Op5f08 ;@ 5f88
  dcd Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90,Op5f90 ;@ 5f90
  dcd Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98,Op5f98 ;@ 5f98
  dcd Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0,Op5fa0 ;@ 5fa0
  dcd Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8,Op5fa8 ;@ 5fa8
  dcd Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0,Op5fb0 ;@ 5fb0
  dcd Op5fb8,Op5fb9,Op5fba,Op5fbb,Op____,Op____,Op____,Op____ ;@ 5fb8
  dcd Op5fc0,Op5fc0,Op5fc0,Op5fc0,Op5fc0,Op5fc0,Op5fc0,Op5fc0 ;@ 5fc0
  dcd Op5fc8,Op5fc8,Op5fc8,Op5fc8,Op5fc8,Op5fc8,Op5fc8,Op5fc8 ;@ 5fc8
  dcd Op5fd0,Op5fd0,Op5fd0,Op5fd0,Op5fd0,Op5fd0,Op5fd0,Op5fd0 ;@ 5fd0
  dcd Op5fd8,Op5fd8,Op5fd8,Op5fd8,Op5fd8,Op5fd8,Op5fd8,Op5fdf ;@ 5fd8
  dcd Op5fe0,Op5fe0,Op5fe0,Op5fe0,Op5fe0,Op5fe0,Op5fe0,Op5fe7 ;@ 5fe0
  dcd Op5fe8,Op5fe8,Op5fe8,Op5fe8,Op5fe8,Op5fe8,Op5fe8,Op5fe8 ;@ 5fe8
  dcd Op5ff0,Op5ff0,Op5ff0,Op5ff0,Op5ff0,Op5ff0,Op5ff0,Op5ff0 ;@ 5ff0
  dcd Op5ff8,Op5ff9,Op5ffa,Op5ffb,Op____,Op____,Op____,Op____ ;@ 5ff8
  dcd Op6000,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6000
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6008
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6010
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6018
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6020
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6028
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6030
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6038
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6040
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6048
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6050
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6058
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6060
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6068
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6070
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6078
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6080
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6088
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6090
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 6098
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60a0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60a8
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60b0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60b8
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60c0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60c8
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60d0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60d8
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60e0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60e8
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001 ;@ 60f0
  dcd Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op6001,Op60ff ;@ 60f8
  dcd Op6100,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6100
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6108
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6110
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6118
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6120
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6128
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6130
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6138
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6140
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6148
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6150
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6158
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6160
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6168
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6170
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6178
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6180
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6188
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6190
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 6198
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61a0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61a8
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61b0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61b8
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61c0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61c8
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61d0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61d8
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61e0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61e8
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101 ;@ 61f0
  dcd Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op6101,Op61ff ;@ 61f8
  dcd Op6200,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6200
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6208
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6210
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6218
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6220
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6228
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6230
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6238
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6240
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6248
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6250
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6258
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6260
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6268
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6270
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6278
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6280
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6288
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6290
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 6298
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62a0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62a8
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62b0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62b8
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62c0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62c8
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62d0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62d8
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62e0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62e8
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201 ;@ 62f0
  dcd Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op6201,Op62ff ;@ 62f8
  dcd Op6300,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6300
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6308
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6310
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6318
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6320
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6328
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6330
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6338
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6340
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6348
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6350
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6358
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6360
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6368
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6370
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6378
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6380
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6388
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6390
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 6398
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63a0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63a8
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63b0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63b8
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63c0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63c8
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63d0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63d8
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63e0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63e8
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301 ;@ 63f0
  dcd Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op6301,Op63ff ;@ 63f8
  dcd Op6400,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6400
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6408
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6410
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6418
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6420
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6428
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6430
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6438
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6440
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6448
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6450
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6458
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6460
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6468
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6470
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6478
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6480
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6488
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6490
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 6498
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64a0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64a8
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64b0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64b8
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64c0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64c8
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64d0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64d8
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64e0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64e8
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401 ;@ 64f0
  dcd Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op6401,Op64ff ;@ 64f8
  dcd Op6500,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6500
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6508
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6510
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6518
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6520
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6528
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6530
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6538
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6540
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6548
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6550
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6558
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6560
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6568
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6570
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6578
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6580
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6588
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6590
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 6598
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65a0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65a8
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65b0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65b8
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65c0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65c8
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65d0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65d8
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65e0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65e8
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501 ;@ 65f0
  dcd Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op6501,Op65ff ;@ 65f8
  dcd Op6600,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6600
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6608
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6610
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6618
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6620
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6628
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6630
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6638
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6640
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6648
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6650
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6658
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6660
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6668
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6670
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6678
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6680
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6688
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6690
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 6698
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66a0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66a8
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66b0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66b8
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66c0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66c8
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66d0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66d8
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66e0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66e8
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601 ;@ 66f0
  dcd Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op6601,Op66ff ;@ 66f8
  dcd Op6700,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6700
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6708
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6710
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6718
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6720
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6728
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6730
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6738
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6740
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6748
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6750
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6758
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6760
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6768
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6770
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6778
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6780
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6788
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6790
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 6798
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67a0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67a8
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67b0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67b8
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67c0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67c8
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67d0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67d8
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67e0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67e8
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701 ;@ 67f0
  dcd Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op6701,Op67ff ;@ 67f8
  dcd Op6800,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6800
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6808
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6810
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6818
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6820
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6828
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6830
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6838
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6840
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6848
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6850
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6858
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6860
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6868
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6870
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6878
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6880
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6888
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6890
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 6898
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68a0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68a8
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68b0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68b8
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68c0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68c8
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68d0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68d8
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68e0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68e8
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801 ;@ 68f0
  dcd Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op6801,Op68ff ;@ 68f8
  dcd Op6900,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6900
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6908
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6910
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6918
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6920
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6928
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6930
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6938
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6940
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6948
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6950
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6958
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6960
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6968
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6970
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6978
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6980
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6988
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6990
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 6998
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69a0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69a8
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69b0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69b8
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69c0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69c8
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69d0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69d8
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69e0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69e8
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901 ;@ 69f0
  dcd Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op6901,Op69ff ;@ 69f8
  dcd Op6a00,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a00
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a08
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a10
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a18
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a20
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a28
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a30
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a38
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a40
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a48
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a50
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a58
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a60
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a68
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a70
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a78
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a80
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a88
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a90
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6a98
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6aa0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6aa8
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ab0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ab8
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ac0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ac8
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ad0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ad8
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ae0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6ae8
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01 ;@ 6af0
  dcd Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6a01,Op6aff ;@ 6af8
  dcd Op6b00,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b00
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b08
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b10
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b18
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b20
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b28
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b30
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b38
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b40
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b48
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b50
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b58
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b60
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b68
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b70
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b78
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b80
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b88
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b90
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6b98
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6ba0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6ba8
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bb0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bb8
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bc0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bc8
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bd0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bd8
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6be0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6be8
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01 ;@ 6bf0
  dcd Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6b01,Op6bff ;@ 6bf8
  dcd Op6c00,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c00
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c08
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c10
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c18
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c20
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c28
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c30
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c38
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c40
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c48
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c50
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c58
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c60
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c68
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c70
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c78
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c80
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c88
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c90
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6c98
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6ca0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6ca8
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cb0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cb8
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cc0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cc8
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cd0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cd8
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6ce0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6ce8
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01 ;@ 6cf0
  dcd Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6c01,Op6cff ;@ 6cf8
  dcd Op6d00,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d00
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d08
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d10
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d18
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d20
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d28
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d30
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d38
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d40
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d48
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d50
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d58
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d60
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d68
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d70
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d78
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d80
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d88
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d90
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6d98
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6da0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6da8
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6db0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6db8
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6dc0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6dc8
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6dd0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6dd8
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6de0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6de8
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01 ;@ 6df0
  dcd Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6d01,Op6dff ;@ 6df8
  dcd Op6e00,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e00
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e08
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e10
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e18
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e20
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e28
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e30
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e38
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e40
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e48
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e50
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e58
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e60
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e68
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e70
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e78
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e80
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e88
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e90
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6e98
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ea0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ea8
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6eb0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6eb8
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ec0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ec8
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ed0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ed8
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ee0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ee8
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01 ;@ 6ef0
  dcd Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6e01,Op6eff ;@ 6ef8
  dcd Op6f00,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f00
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f08
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f10
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f18
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f20
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f28
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f30
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f38
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f40
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f48
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f50
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f58
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f60
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f68
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f70
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f78
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f80
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f88
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f90
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6f98
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fa0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fa8
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fb0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fb8
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fc0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fc8
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fd0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fd8
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fe0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6fe8
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01 ;@ 6ff0
  dcd Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6f01,Op6fff ;@ 6ff8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7000
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7008
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7010
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7018
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7020
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7028
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7030
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7038
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7040
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7048
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7050
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7058
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7060
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7068
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7070
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7078
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7080
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7088
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7090
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7098
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70a0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70a8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70b0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70b8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70c0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70c8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70d0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70d8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70e0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70e8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70f0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 70f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7108
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7110
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7118
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7120
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7128
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7130
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7138
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7148
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7150
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7158
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7160
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7168
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7170
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7188
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7190
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7198
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 71f8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7200
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7208
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7210
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7218
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7220
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7228
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7230
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7238
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7240
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7248
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7250
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7258
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7260
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7268
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7270
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7278
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7280
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7288
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7290
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7298
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72a0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72a8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72b0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72b8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72c0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72c8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72d0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72d8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72e0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72e8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72f0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 72f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7308
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7310
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7318
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7320
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7328
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7330
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7338
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7348
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7350
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7358
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7360
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7368
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7370
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7388
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7390
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7398
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 73f8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7400
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7408
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7410
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7418
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7420
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7428
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7430
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7438
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7440
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7448
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7450
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7458
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7460
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7468
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7470
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7478
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7480
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7488
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7490
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7498
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74a0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74a8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74b0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74b8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74c0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74c8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74d0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74d8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74e0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74e8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74f0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 74f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7508
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7510
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7518
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7520
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7528
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7530
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7538
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7548
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7550
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7558
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7560
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7568
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7570
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7588
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7590
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7598
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 75f8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7600
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7608
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7610
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7618
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7620
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7628
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7630
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7638
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7640
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7648
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7650
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7658
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7660
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7668
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7670
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7678
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7680
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7688
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7690
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7698
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76a0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76a8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76b0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76b8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76c0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76c8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76d0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76d8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76e0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76e8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76f0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 76f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7708
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7710
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7718
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7720
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7728
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7730
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7738
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7748
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7750
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7758
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7760
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7768
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7770
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7788
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7790
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7798
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 77f8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7800
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7808
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7810
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7818
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7820
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7828
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7830
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7838
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7840
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7848
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7850
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7858
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7860
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7868
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7870
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7878
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7880
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7888
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7890
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7898
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78a0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78a8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78b0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78b8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78c0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78c8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78d0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78d8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78e0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78e8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78f0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 78f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7908
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7910
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7918
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7920
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7928
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7930
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7938
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7948
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7950
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7958
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7960
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7968
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7970
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7988
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7990
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7998
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 79f8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a00
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a08
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a10
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a18
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a20
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a28
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a30
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a38
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a40
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a48
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a50
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a58
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a60
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a68
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a70
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a78
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a80
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a88
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a90
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7a98
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7aa0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7aa8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ab0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ab8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ac0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ac8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ad0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ad8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ae0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ae8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7af0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7af8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7b98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7ba0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7ba8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7be0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7be8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7bf8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c00
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c08
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c10
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c18
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c20
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c28
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c30
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c38
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c40
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c48
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c50
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c58
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c60
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c68
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c70
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c78
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c80
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c88
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c90
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7c98
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ca0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ca8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cb0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cb8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cc0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cc8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cd0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cd8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ce0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ce8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cf0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7cf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7d98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7da0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7da8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7db0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7db8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7dc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7dc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7dd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7dd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7de0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7de8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7df0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7df8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e00
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e08
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e10
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e18
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e20
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e28
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e30
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e38
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e40
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e48
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e50
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e58
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e60
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e68
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e70
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e78
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e80
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e88
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e90
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7e98
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ea0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ea8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7eb0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7eb8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ec0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ec8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ed0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ed8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ee0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ee8
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ef0
  dcd Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000,Op7000 ;@ 7ef8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7f98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7fe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7ff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 7ff8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8000
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8008
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8010
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8018
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8020
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8028
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8030
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8038
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8040
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8048
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8050
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8058
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8060
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8068
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8070
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8078
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8080
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8088
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8090
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8098
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 80a0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 80a8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 80b0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 80b8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 80c0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 80c8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 80d0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 80d8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 80e0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 80e8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 80f0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 80f8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8100
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8108
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8110
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8118
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8120
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8128
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8130
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8138
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8148
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8150
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8158
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8160
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8168
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8170
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8188
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8190
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8198
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 81a0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 81a8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 81b0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 81b8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 81c0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 81c8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 81d0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 81d8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 81e0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 81e8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 81f0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 81f8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8200
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8208
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8210
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8218
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8220
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8228
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8230
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8238
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8240
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8248
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8250
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8258
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8260
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8268
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8270
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8278
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8280
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8288
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8290
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8298
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 82a0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 82a8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 82b0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 82b8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 82c0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 82c8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 82d0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 82d8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 82e0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 82e8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 82f0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 82f8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8300
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8308
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8310
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8318
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8320
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8328
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8330
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8338
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8348
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8350
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8358
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8360
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8368
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8370
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8388
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8390
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8398
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 83a0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 83a8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 83b0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 83b8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 83c0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 83c8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 83d0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 83d8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 83e0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 83e8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 83f0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 83f8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8400
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8408
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8410
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8418
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8420
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8428
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8430
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8438
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8440
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8448
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8450
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8458
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8460
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8468
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8470
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8478
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8480
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8488
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8490
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8498
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 84a0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 84a8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 84b0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 84b8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 84c0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 84c8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 84d0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 84d8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 84e0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 84e8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 84f0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 84f8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8500
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8508
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8510
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8518
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8520
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8528
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8530
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8538
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8548
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8550
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8558
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8560
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8568
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8570
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8588
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8590
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8598
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 85a0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 85a8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 85b0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 85b8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 85c0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 85c8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 85d0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 85d8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 85e0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 85e8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 85f0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 85f8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8600
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8608
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8610
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8618
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8620
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8628
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8630
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8638
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8640
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8648
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8650
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8658
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8660
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8668
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8670
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8678
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8680
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8688
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8690
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8698
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 86a0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 86a8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 86b0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 86b8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 86c0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 86c8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 86d0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 86d8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 86e0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 86e8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 86f0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 86f8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8700
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8708
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8710
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8718
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8720
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8728
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8730
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8738
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8748
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8750
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8758
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8760
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8768
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8770
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8788
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8790
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8798
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 87a0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 87a8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 87b0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 87b8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 87c0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 87c8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 87d0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 87d8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 87e0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 87e8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 87f0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 87f8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8800
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8808
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8810
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8818
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8820
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8828
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8830
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8838
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8840
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8848
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8850
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8858
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8860
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8868
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8870
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8878
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8880
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8888
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8890
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8898
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 88a0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 88a8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 88b0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 88b8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 88c0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 88c8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 88d0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 88d8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 88e0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 88e8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 88f0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 88f8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8900
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8908
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8910
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8918
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8920
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8928
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8930
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8938
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8948
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8950
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8958
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8960
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8968
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8970
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8988
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8990
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8998
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 89a0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 89a8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 89b0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 89b8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 89c0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 89c8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 89d0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 89d8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 89e0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 89e8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 89f0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 89f8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8a00
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8a08
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8a10
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8a18
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8a20
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8a28
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8a30
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8a38
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8a40
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8a48
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8a50
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8a58
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8a60
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8a68
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8a70
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8a78
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8a80
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8a88
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8a90
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8a98
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 8aa0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 8aa8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 8ab0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 8ab8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8ac0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8ac8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 8ad0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 8ad8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 8ae0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 8ae8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 8af0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 8af8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8b00
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8b08
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8b10
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8b18
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8b20
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8b28
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8b30
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8b38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8b40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8b48
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8b50
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8b58
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8b60
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8b68
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8b70
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8b78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8b80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8b88
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8b90
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8b98
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 8ba0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 8ba8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 8bb0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 8bb8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8bc0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8bc8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 8bd0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 8bd8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 8be0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 8be8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 8bf0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 8bf8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8c00
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8c08
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8c10
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8c18
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8c20
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8c28
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8c30
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8c38
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8c40
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8c48
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8c50
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8c58
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8c60
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8c68
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8c70
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8c78
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8c80
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8c88
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8c90
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8c98
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 8ca0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 8ca8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 8cb0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 8cb8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8cc0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8cc8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 8cd0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 8cd8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 8ce0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 8ce8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 8cf0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 8cf8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8d00
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8d08
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8d10
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8d18
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8d20
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8d28
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8d30
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8d38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8d40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8d48
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8d50
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8d58
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8d60
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8d68
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8d70
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8d78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8d80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8d88
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8d90
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8d98
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 8da0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 8da8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 8db0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 8db8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8dc0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8dc8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 8dd0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 8dd8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 8de0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 8de8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 8df0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 8df8
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8e00
  dcd Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000,Op8000 ;@ 8e08
  dcd Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010,Op8010 ;@ 8e10
  dcd Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op8018,Op801f ;@ 8e18
  dcd Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8020,Op8027 ;@ 8e20
  dcd Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028,Op8028 ;@ 8e28
  dcd Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030,Op8030 ;@ 8e30
  dcd Op8038,Op8039,Op803a,Op803b,Op803c,Op803d,Op803e,Op803f ;@ 8e38
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8e40
  dcd Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040,Op8040 ;@ 8e48
  dcd Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050,Op8050 ;@ 8e50
  dcd Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op8058,Op805f ;@ 8e58
  dcd Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8060,Op8067 ;@ 8e60
  dcd Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068,Op8068 ;@ 8e68
  dcd Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070,Op8070 ;@ 8e70
  dcd Op8078,Op8079,Op807a,Op807b,Op807c,Op807d,Op807e,Op807f ;@ 8e78
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8e80
  dcd Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080,Op8080 ;@ 8e88
  dcd Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090,Op8090 ;@ 8e90
  dcd Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op8098,Op809f ;@ 8e98
  dcd Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a0,Op80a7 ;@ 8ea0
  dcd Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8,Op80a8 ;@ 8ea8
  dcd Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0,Op80b0 ;@ 8eb0
  dcd Op80b8,Op80b9,Op80ba,Op80bb,Op80bc,Op80bd,Op80be,Op80bf ;@ 8eb8
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8ec0
  dcd Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0,Op80c0 ;@ 8ec8
  dcd Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0,Op80d0 ;@ 8ed0
  dcd Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80d8,Op80df ;@ 8ed8
  dcd Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e0,Op80e7 ;@ 8ee0
  dcd Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8,Op80e8 ;@ 8ee8
  dcd Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0,Op80f0 ;@ 8ef0
  dcd Op80f8,Op80f9,Op80fa,Op80fb,Op80fc,Op____,Op____,Op____ ;@ 8ef8
  dcd Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100,Op8100 ;@ 8f00
  dcd Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108,Op8108 ;@ 8f08
  dcd Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110,Op8110 ;@ 8f10
  dcd Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op8118,Op811f ;@ 8f18
  dcd Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8120,Op8127 ;@ 8f20
  dcd Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128,Op8128 ;@ 8f28
  dcd Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130,Op8130 ;@ 8f30
  dcd Op8138,Op8139,Op813a,Op813b,Op____,Op____,Op____,Op____ ;@ 8f38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8f40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8f48
  dcd Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150,Op8150 ;@ 8f50
  dcd Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op8158,Op815f ;@ 8f58
  dcd Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8160,Op8167 ;@ 8f60
  dcd Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168,Op8168 ;@ 8f68
  dcd Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170,Op8170 ;@ 8f70
  dcd Op8178,Op8179,Op817a,Op817b,Op____,Op____,Op____,Op____ ;@ 8f78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8f80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 8f88
  dcd Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190,Op8190 ;@ 8f90
  dcd Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op8198,Op819f ;@ 8f98
  dcd Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a0,Op81a7 ;@ 8fa0
  dcd Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8,Op81a8 ;@ 8fa8
  dcd Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0,Op81b0 ;@ 8fb0
  dcd Op81b8,Op81b9,Op81ba,Op81bb,Op____,Op____,Op____,Op____ ;@ 8fb8
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8fc0
  dcd Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0,Op81c0 ;@ 8fc8
  dcd Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0,Op81d0 ;@ 8fd0
  dcd Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81d8,Op81df ;@ 8fd8
  dcd Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e0,Op81e7 ;@ 8fe0
  dcd Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8,Op81e8 ;@ 8fe8
  dcd Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0,Op81f0 ;@ 8ff0
  dcd Op81f8,Op81f9,Op81fa,Op81fb,Op81fc,Op____,Op____,Op____ ;@ 8ff8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9000
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9008
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9010
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9018
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9020
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9028
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9030
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9038
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9040
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9048
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9050
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9058
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9060
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9068
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9070
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9078
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9080
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9088
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9090
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9098
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 90a0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 90a8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 90b0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 90b8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 90c0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 90c8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 90d0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 90d8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 90e0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 90e8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 90f0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 90f8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9108
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9110
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9118
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9120
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9128
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9130
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9138
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9148
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9150
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9158
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9160
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9168
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9170
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9178
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9188
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9190
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9198
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 91a0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 91a8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 91b0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 91b8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 91c0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 91c8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 91d0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 91d8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 91e0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 91e8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 91f0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 91f8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9200
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9208
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9210
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9218
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9220
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9228
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9230
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9238
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9240
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9248
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9250
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9258
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9260
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9268
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9270
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9278
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9280
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9288
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9290
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9298
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 92a0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 92a8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 92b0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 92b8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 92c0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 92c8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 92d0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 92d8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 92e0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 92e8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 92f0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 92f8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9308
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9310
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9318
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9320
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9328
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9330
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9338
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9348
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9350
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9358
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9360
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9368
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9370
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9378
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9388
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9390
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9398
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 93a0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 93a8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 93b0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 93b8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 93c0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 93c8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 93d0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 93d8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 93e0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 93e8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 93f0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 93f8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9400
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9408
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9410
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9418
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9420
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9428
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9430
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9438
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9440
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9448
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9450
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9458
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9460
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9468
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9470
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9478
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9480
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9488
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9490
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9498
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 94a0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 94a8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 94b0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 94b8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 94c0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 94c8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 94d0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 94d8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 94e0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 94e8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 94f0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 94f8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9508
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9510
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9518
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9520
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9528
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9530
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9538
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9548
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9550
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9558
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9560
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9568
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9570
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9578
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9588
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9590
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9598
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 95a0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 95a8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 95b0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 95b8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 95c0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 95c8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 95d0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 95d8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 95e0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 95e8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 95f0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 95f8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9600
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9608
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9610
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9618
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9620
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9628
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9630
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9638
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9640
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9648
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9650
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9658
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9660
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9668
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9670
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9678
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9680
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9688
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9690
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9698
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 96a0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 96a8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 96b0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 96b8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 96c0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 96c8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 96d0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 96d8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 96e0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 96e8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 96f0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 96f8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9708
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9710
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9718
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9720
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9728
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9730
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9738
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9748
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9750
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9758
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9760
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9768
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9770
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9778
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9788
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9790
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9798
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 97a0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 97a8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 97b0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 97b8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 97c0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 97c8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 97d0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 97d8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 97e0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 97e8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 97f0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 97f8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9800
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9808
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9810
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9818
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9820
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9828
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9830
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9838
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9840
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9848
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9850
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9858
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9860
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9868
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9870
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9878
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9880
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9888
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9890
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9898
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 98a0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 98a8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 98b0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 98b8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 98c0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 98c8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 98d0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 98d8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 98e0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 98e8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 98f0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 98f8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9908
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9910
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9918
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9920
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9928
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9930
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9938
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9948
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9950
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9958
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9960
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9968
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9970
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9978
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9988
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9990
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9998
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 99a0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 99a8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 99b0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 99b8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 99c0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 99c8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 99d0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 99d8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 99e0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 99e8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 99f0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 99f8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9a00
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9a08
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9a10
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9a18
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9a20
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9a28
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9a30
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9a38
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9a40
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9a48
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9a50
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9a58
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9a60
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9a68
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9a70
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9a78
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9a80
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9a88
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9a90
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9a98
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 9aa0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 9aa8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 9ab0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 9ab8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9ac0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9ac8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 9ad0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 9ad8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 9ae0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 9ae8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 9af0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 9af8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9b00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9b08
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9b10
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9b18
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9b20
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9b28
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9b30
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9b38
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9b40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9b48
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9b50
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9b58
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9b60
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9b68
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9b70
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9b78
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9b80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9b88
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9b90
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9b98
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 9ba0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 9ba8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 9bb0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 9bb8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9bc0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9bc8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 9bd0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 9bd8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 9be0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 9be8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 9bf0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 9bf8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9c00
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9c08
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9c10
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9c18
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9c20
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9c28
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9c30
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9c38
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9c40
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9c48
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9c50
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9c58
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9c60
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9c68
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9c70
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9c78
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9c80
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9c88
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9c90
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9c98
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 9ca0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 9ca8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 9cb0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 9cb8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9cc0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9cc8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 9cd0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 9cd8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 9ce0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 9ce8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 9cf0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 9cf8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9d00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9d08
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9d10
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9d18
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9d20
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9d28
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9d30
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9d38
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9d40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9d48
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9d50
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9d58
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9d60
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9d68
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9d70
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9d78
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9d80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9d88
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9d90
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9d98
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 9da0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 9da8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 9db0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 9db8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9dc0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9dc8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 9dd0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 9dd8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 9de0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 9de8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 9df0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 9df8
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9e00
  dcd Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000,Op9000 ;@ 9e08
  dcd Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010,Op9010 ;@ 9e10
  dcd Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op9018,Op901f ;@ 9e18
  dcd Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9020,Op9027 ;@ 9e20
  dcd Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028,Op9028 ;@ 9e28
  dcd Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030,Op9030 ;@ 9e30
  dcd Op9038,Op9039,Op903a,Op903b,Op903c,Op903d,Op903e,Op903f ;@ 9e38
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9e40
  dcd Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040,Op9040 ;@ 9e48
  dcd Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050,Op9050 ;@ 9e50
  dcd Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op9058,Op905f ;@ 9e58
  dcd Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9060,Op9067 ;@ 9e60
  dcd Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068,Op9068 ;@ 9e68
  dcd Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070,Op9070 ;@ 9e70
  dcd Op9078,Op9079,Op907a,Op907b,Op907c,Op907d,Op907e,Op907f ;@ 9e78
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9e80
  dcd Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080,Op9080 ;@ 9e88
  dcd Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090,Op9090 ;@ 9e90
  dcd Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op9098,Op909f ;@ 9e98
  dcd Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a0,Op90a7 ;@ 9ea0
  dcd Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8,Op90a8 ;@ 9ea8
  dcd Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0,Op90b0 ;@ 9eb0
  dcd Op90b8,Op90b9,Op90ba,Op90bb,Op90bc,Op90bd,Op90be,Op90bf ;@ 9eb8
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9ec0
  dcd Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0,Op90c0 ;@ 9ec8
  dcd Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0,Op90d0 ;@ 9ed0
  dcd Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90d8,Op90df ;@ 9ed8
  dcd Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e0,Op90e7 ;@ 9ee0
  dcd Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8,Op90e8 ;@ 9ee8
  dcd Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0,Op90f0 ;@ 9ef0
  dcd Op90f8,Op90f9,Op90fa,Op90fb,Op90fc,Op____,Op____,Op____ ;@ 9ef8
  dcd Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100,Op9100 ;@ 9f00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9f08
  dcd Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110,Op9110 ;@ 9f10
  dcd Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op9118,Op911f ;@ 9f18
  dcd Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9120,Op9127 ;@ 9f20
  dcd Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128,Op9128 ;@ 9f28
  dcd Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130,Op9130 ;@ 9f30
  dcd Op9138,Op9139,Op913a,Op913b,Op____,Op____,Op____,Op____ ;@ 9f38
  dcd Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140,Op9140 ;@ 9f40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9f48
  dcd Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150,Op9150 ;@ 9f50
  dcd Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op9158,Op915f ;@ 9f58
  dcd Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9160,Op9167 ;@ 9f60
  dcd Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168,Op9168 ;@ 9f68
  dcd Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170,Op9170 ;@ 9f70
  dcd Op9178,Op9179,Op917a,Op917b,Op____,Op____,Op____,Op____ ;@ 9f78
  dcd Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180,Op9180 ;@ 9f80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ 9f88
  dcd Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190,Op9190 ;@ 9f90
  dcd Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op9198,Op919f ;@ 9f98
  dcd Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a0,Op91a7 ;@ 9fa0
  dcd Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8,Op91a8 ;@ 9fa8
  dcd Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0,Op91b0 ;@ 9fb0
  dcd Op91b8,Op91b9,Op91ba,Op91bb,Op____,Op____,Op____,Op____ ;@ 9fb8
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9fc0
  dcd Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0,Op91c0 ;@ 9fc8
  dcd Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0,Op91d0 ;@ 9fd0
  dcd Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91d8,Op91df ;@ 9fd8
  dcd Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e0,Op91e7 ;@ 9fe0
  dcd Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8,Op91e8 ;@ 9fe8
  dcd Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0,Op91f0 ;@ 9ff0
  dcd Op91f8,Op91f9,Op91fa,Op91fb,Op91fc,Op____,Op____,Op____ ;@ 9ff8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a000
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a008
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a010
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a018
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a020
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a028
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a030
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a038
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a040
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a048
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a050
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a058
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a060
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a068
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a070
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a078
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a080
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a088
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a090
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a098
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a0f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a108
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a110
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a118
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a120
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a128
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a130
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a138
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a148
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a150
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a158
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a160
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a168
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a170
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a188
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a190
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a198
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a1f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a200
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a208
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a210
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a218
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a220
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a228
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a230
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a238
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a240
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a248
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a250
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a258
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a260
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a268
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a270
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a278
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a280
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a288
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a290
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a298
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a2f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a308
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a310
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a318
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a320
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a328
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a330
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a338
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a348
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a350
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a358
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a360
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a368
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a370
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a388
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a390
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a398
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a3f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a400
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a408
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a410
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a418
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a420
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a428
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a430
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a438
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a440
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a448
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a450
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a458
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a460
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a468
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a470
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a478
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a480
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a488
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a490
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a498
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a4f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a508
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a510
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a518
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a520
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a528
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a530
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a538
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a548
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a550
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a558
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a560
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a568
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a570
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a588
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a590
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a598
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a5f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a600
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a608
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a610
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a618
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a620
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a628
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a630
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a638
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a640
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a648
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a650
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a658
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a660
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a668
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a670
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a678
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a680
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a688
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a690
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a698
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a6f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a708
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a710
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a718
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a720
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a728
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a730
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a738
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a748
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a750
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a758
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a760
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a768
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a770
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a788
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a790
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a798
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a7f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a800
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a808
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a810
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a818
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a820
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a828
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a830
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a838
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a840
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a848
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a850
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a858
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a860
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a868
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a870
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a878
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a880
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a888
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a890
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a898
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a8f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a908
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a910
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a918
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a920
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a928
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a930
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a938
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a948
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a950
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a958
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a960
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a968
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a970
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a988
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a990
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a998
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ a9f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aa98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aaa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aaa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aab0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aab8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aac0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aac8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aad0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aad8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aae0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aae8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aaf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aaf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ab98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aba0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aba8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ abf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ac98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aca0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aca8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ace0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ace8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ acf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ad98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ada0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ada8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ add0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ add8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ade0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ade8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ adf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ae98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aea0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aea8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aeb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aeb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aec0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aec8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aed0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aed8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aee0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aee8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aef0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aef8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ af98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ afe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ aff8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b000
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b008
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ b010
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ b018
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ b020
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ b028
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ b030
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ b038
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b040
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b048
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ b050
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ b058
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ b060
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ b068
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ b070
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ b078
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b080
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b088
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ b090
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ b098
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ b0a0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ b0a8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ b0b0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ b0b8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b0c0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b0c8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ b0d0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ b0d8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ b0e0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ b0e8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ b0f0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ b0f8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b100
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b108
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ b110
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ b118
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ b120
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ b128
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ b130
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ b138
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b140
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b148
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ b150
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ b158
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ b160
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ b168
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ b170
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ b178
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b180
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b188
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ b190
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ b198
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ b1a0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ b1a8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ b1b0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ b1b8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b1c0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b1c8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ b1d0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ b1d8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ b1e0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ b1e8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ b1f0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ b1f8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b200
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b208
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ b210
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ b218
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ b220
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ b228
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ b230
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ b238
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b240
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b248
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ b250
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ b258
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ b260
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ b268
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ b270
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ b278
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b280
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b288
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ b290
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ b298
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ b2a0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ b2a8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ b2b0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ b2b8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b2c0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b2c8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ b2d0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ b2d8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ b2e0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ b2e8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ b2f0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ b2f8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b300
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b308
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ b310
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ b318
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ b320
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ b328
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ b330
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ b338
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b340
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b348
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ b350
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ b358
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ b360
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ b368
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ b370
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ b378
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b380
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b388
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ b390
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ b398
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ b3a0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ b3a8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ b3b0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ b3b8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b3c0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b3c8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ b3d0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ b3d8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ b3e0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ b3e8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ b3f0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ b3f8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b400
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b408
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ b410
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ b418
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ b420
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ b428
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ b430
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ b438
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b440
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b448
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ b450
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ b458
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ b460
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ b468
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ b470
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ b478
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b480
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b488
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ b490
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ b498
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ b4a0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ b4a8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ b4b0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ b4b8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b4c0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b4c8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ b4d0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ b4d8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ b4e0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ b4e8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ b4f0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ b4f8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b500
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b508
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ b510
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ b518
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ b520
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ b528
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ b530
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ b538
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b540
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b548
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ b550
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ b558
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ b560
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ b568
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ b570
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ b578
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b580
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b588
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ b590
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ b598
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ b5a0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ b5a8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ b5b0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ b5b8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b5c0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b5c8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ b5d0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ b5d8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ b5e0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ b5e8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ b5f0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ b5f8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b600
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b608
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ b610
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ b618
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ b620
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ b628
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ b630
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ b638
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b640
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b648
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ b650
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ b658
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ b660
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ b668
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ b670
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ b678
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b680
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b688
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ b690
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ b698
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ b6a0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ b6a8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ b6b0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ b6b8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b6c0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b6c8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ b6d0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ b6d8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ b6e0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ b6e8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ b6f0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ b6f8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b700
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b708
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ b710
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ b718
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ b720
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ b728
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ b730
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ b738
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b740
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b748
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ b750
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ b758
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ b760
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ b768
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ b770
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ b778
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b780
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b788
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ b790
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ b798
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ b7a0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ b7a8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ b7b0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ b7b8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b7c0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b7c8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ b7d0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ b7d8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ b7e0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ b7e8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ b7f0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ b7f8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b800
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ b808
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ b810
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ b818
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ b820
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ b828
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ b830
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ b838
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b840
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ b848
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ b850
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ b858
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ b860
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ b868
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ b870
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ b878
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b880
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ b888
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ b890
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ b898
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ b8a0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ b8a8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ b8b0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ b8b8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b8c0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ b8c8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ b8d0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ b8d8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ b8e0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ b8e8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ b8f0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ b8f8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b900
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ b908
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ b910
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ b918
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ b920
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ b928
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ b930
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ b938
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b940
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ b948
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ b950
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ b958
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ b960
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ b968
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ b970
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ b978
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b980
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ b988
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ b990
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ b998
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ b9a0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ b9a8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ b9b0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ b9b8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b9c0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ b9c8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ b9d0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ b9d8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ b9e0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ b9e8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ b9f0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ b9f8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ ba00
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ ba08
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ ba10
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ ba18
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ ba20
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ ba28
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ ba30
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ ba38
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ ba40
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ ba48
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ ba50
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ ba58
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ ba60
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ ba68
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ ba70
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ ba78
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ ba80
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ ba88
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ ba90
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ ba98
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ baa0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ baa8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ bab0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ bab8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bac0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bac8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ bad0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ bad8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ bae0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ bae8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ baf0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ baf8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bb00
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bb08
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ bb10
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ bb18
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ bb20
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ bb28
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ bb30
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ bb38
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bb40
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bb48
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ bb50
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ bb58
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ bb60
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ bb68
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ bb70
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ bb78
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bb80
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bb88
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ bb90
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ bb98
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ bba0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ bba8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ bbb0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ bbb8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bbc0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bbc8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ bbd0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ bbd8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ bbe0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ bbe8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ bbf0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ bbf8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ bc00
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ bc08
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ bc10
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ bc18
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ bc20
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ bc28
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ bc30
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ bc38
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ bc40
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ bc48
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ bc50
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ bc58
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ bc60
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ bc68
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ bc70
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ bc78
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ bc80
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ bc88
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ bc90
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ bc98
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ bca0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ bca8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ bcb0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ bcb8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bcc0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bcc8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ bcd0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ bcd8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ bce0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ bce8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ bcf0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ bcf8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bd00
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bd08
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ bd10
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ bd18
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ bd20
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ bd28
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ bd30
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ bd38
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bd40
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bd48
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ bd50
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ bd58
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ bd60
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ bd68
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ bd70
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ bd78
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bd80
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bd88
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ bd90
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ bd98
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ bda0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ bda8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ bdb0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ bdb8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bdc0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bdc8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ bdd0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ bdd8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ bde0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ bde8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ bdf0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ bdf8
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ be00
  dcd Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000,Opb000 ;@ be08
  dcd Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010,Opb010 ;@ be10
  dcd Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb018,Opb01f ;@ be18
  dcd Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb020,Opb027 ;@ be20
  dcd Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028,Opb028 ;@ be28
  dcd Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030,Opb030 ;@ be30
  dcd Opb038,Opb039,Opb03a,Opb03b,Opb03c,Op____,Op____,Op____ ;@ be38
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ be40
  dcd Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040,Opb040 ;@ be48
  dcd Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050,Opb050 ;@ be50
  dcd Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb058,Opb05f ;@ be58
  dcd Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb060,Opb067 ;@ be60
  dcd Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068,Opb068 ;@ be68
  dcd Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070,Opb070 ;@ be70
  dcd Opb078,Opb079,Opb07a,Opb07b,Opb07c,Op____,Op____,Op____ ;@ be78
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ be80
  dcd Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080,Opb080 ;@ be88
  dcd Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090,Opb090 ;@ be90
  dcd Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb098,Opb09f ;@ be98
  dcd Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a0,Opb0a7 ;@ bea0
  dcd Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8,Opb0a8 ;@ bea8
  dcd Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0,Opb0b0 ;@ beb0
  dcd Opb0b8,Opb0b9,Opb0ba,Opb0bb,Opb0bc,Op____,Op____,Op____ ;@ beb8
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bec0
  dcd Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0,Opb0c0 ;@ bec8
  dcd Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0,Opb0d0 ;@ bed0
  dcd Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0d8,Opb0df ;@ bed8
  dcd Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e0,Opb0e7 ;@ bee0
  dcd Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8,Opb0e8 ;@ bee8
  dcd Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0,Opb0f0 ;@ bef0
  dcd Opb0f8,Opb0f9,Opb0fa,Opb0fb,Opb0fc,Op____,Op____,Op____ ;@ bef8
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bf00
  dcd Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100,Opb100 ;@ bf08
  dcd Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110,Opb110 ;@ bf10
  dcd Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb118,Opb11f ;@ bf18
  dcd Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb120,Opb127 ;@ bf20
  dcd Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128,Opb128 ;@ bf28
  dcd Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130,Opb130 ;@ bf30
  dcd Opb138,Opb139,Opb13a,Opb13b,Op____,Op____,Op____,Op____ ;@ bf38
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bf40
  dcd Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140,Opb140 ;@ bf48
  dcd Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150,Opb150 ;@ bf50
  dcd Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb158,Opb15f ;@ bf58
  dcd Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb160,Opb167 ;@ bf60
  dcd Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168,Opb168 ;@ bf68
  dcd Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170,Opb170 ;@ bf70
  dcd Opb178,Opb179,Opb17a,Opb17b,Op____,Op____,Op____,Op____ ;@ bf78
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bf80
  dcd Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180,Opb180 ;@ bf88
  dcd Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190,Opb190 ;@ bf90
  dcd Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb198,Opb19f ;@ bf98
  dcd Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a0,Opb1a7 ;@ bfa0
  dcd Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8,Opb1a8 ;@ bfa8
  dcd Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0,Opb1b0 ;@ bfb0
  dcd Opb1b8,Opb1b9,Opb1ba,Opb1bb,Op____,Op____,Op____,Op____ ;@ bfb8
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bfc0
  dcd Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0,Opb1c0 ;@ bfc8
  dcd Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0,Opb1d0 ;@ bfd0
  dcd Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1d8,Opb1df ;@ bfd8
  dcd Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e0,Opb1e7 ;@ bfe0
  dcd Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8,Opb1e8 ;@ bfe8
  dcd Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0,Opb1f0 ;@ bff0
  dcd Opb1f8,Opb1f9,Opb1fa,Opb1fb,Opb1fc,Op____,Op____,Op____ ;@ bff8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c000
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c008
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ c010
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ c018
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ c020
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ c028
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ c030
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ c038
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c040
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c048
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ c050
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ c058
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ c060
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ c068
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ c070
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ c078
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c080
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c088
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ c090
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ c098
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ c0a0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ c0a8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ c0b0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ c0b8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c0c0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c0c8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ c0d0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ c0d8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ c0e0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ c0e8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ c0f0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ c0f8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ c100
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ c108
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ c110
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ c118
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ c120
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ c128
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ c130
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ c138
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ c140
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ c148
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ c150
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ c158
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ c160
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ c168
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ c170
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ c178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ c180
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ c188
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ c190
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ c198
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ c1a0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ c1a8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ c1b0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ c1b8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c1c0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c1c8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ c1d0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ c1d8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ c1e0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ c1e8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ c1f0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ c1f8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c200
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c208
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ c210
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ c218
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ c220
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ c228
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ c230
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ c238
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c240
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c248
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ c250
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ c258
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ c260
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ c268
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ c270
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ c278
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c280
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c288
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ c290
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ c298
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ c2a0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ c2a8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ c2b0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ c2b8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c2c0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c2c8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ c2d0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ c2d8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ c2e0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ c2e8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ c2f0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ c2f8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ c300
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ c308
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ c310
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ c318
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ c320
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ c328
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ c330
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ c338
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ c340
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ c348
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ c350
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ c358
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ c360
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ c368
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ c370
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ c378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ c380
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ c388
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ c390
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ c398
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ c3a0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ c3a8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ c3b0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ c3b8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c3c0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c3c8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ c3d0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ c3d8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ c3e0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ c3e8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ c3f0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ c3f8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c400
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c408
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ c410
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ c418
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ c420
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ c428
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ c430
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ c438
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c440
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c448
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ c450
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ c458
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ c460
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ c468
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ c470
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ c478
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c480
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c488
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ c490
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ c498
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ c4a0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ c4a8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ c4b0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ c4b8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c4c0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c4c8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ c4d0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ c4d8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ c4e0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ c4e8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ c4f0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ c4f8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ c500
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ c508
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ c510
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ c518
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ c520
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ c528
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ c530
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ c538
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ c540
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ c548
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ c550
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ c558
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ c560
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ c568
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ c570
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ c578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ c580
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ c588
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ c590
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ c598
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ c5a0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ c5a8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ c5b0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ c5b8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c5c0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c5c8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ c5d0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ c5d8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ c5e0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ c5e8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ c5f0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ c5f8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c600
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c608
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ c610
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ c618
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ c620
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ c628
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ c630
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ c638
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c640
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c648
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ c650
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ c658
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ c660
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ c668
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ c670
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ c678
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c680
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c688
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ c690
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ c698
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ c6a0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ c6a8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ c6b0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ c6b8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c6c0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c6c8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ c6d0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ c6d8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ c6e0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ c6e8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ c6f0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ c6f8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ c700
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ c708
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ c710
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ c718
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ c720
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ c728
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ c730
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ c738
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ c740
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ c748
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ c750
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ c758
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ c760
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ c768
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ c770
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ c778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ c780
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ c788
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ c790
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ c798
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ c7a0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ c7a8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ c7b0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ c7b8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c7c0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c7c8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ c7d0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ c7d8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ c7e0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ c7e8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ c7f0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ c7f8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c800
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ c808
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ c810
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ c818
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ c820
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ c828
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ c830
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ c838
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c840
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ c848
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ c850
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ c858
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ c860
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ c868
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ c870
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ c878
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c880
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ c888
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ c890
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ c898
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ c8a0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ c8a8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ c8b0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ c8b8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c8c0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ c8c8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ c8d0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ c8d8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ c8e0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ c8e8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ c8f0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ c8f8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ c900
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ c908
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ c910
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ c918
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ c920
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ c928
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ c930
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ c938
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ c940
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ c948
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ c950
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ c958
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ c960
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ c968
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ c970
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ c978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ c980
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ c988
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ c990
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ c998
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ c9a0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ c9a8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ c9b0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ c9b8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c9c0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ c9c8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ c9d0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ c9d8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ c9e0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ c9e8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ c9f0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ c9f8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ ca00
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ ca08
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ ca10
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ ca18
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ ca20
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ ca28
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ ca30
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ ca38
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ ca40
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ ca48
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ ca50
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ ca58
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ ca60
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ ca68
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ ca70
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ ca78
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ ca80
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ ca88
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ ca90
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ ca98
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ caa0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ caa8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ cab0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ cab8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ cac0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ cac8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ cad0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ cad8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ cae0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ cae8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ caf0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ caf8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ cb00
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ cb08
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ cb10
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ cb18
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ cb20
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ cb28
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ cb30
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ cb38
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ cb40
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ cb48
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ cb50
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ cb58
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ cb60
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ cb68
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ cb70
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ cb78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ cb80
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ cb88
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ cb90
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ cb98
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ cba0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ cba8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ cbb0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ cbb8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cbc0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cbc8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ cbd0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ cbd8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ cbe0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ cbe8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ cbf0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ cbf8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ cc00
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ cc08
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ cc10
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ cc18
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ cc20
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ cc28
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ cc30
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ cc38
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ cc40
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ cc48
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ cc50
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ cc58
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ cc60
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ cc68
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ cc70
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ cc78
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ cc80
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ cc88
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ cc90
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ cc98
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ cca0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ cca8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ ccb0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ ccb8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ ccc0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ ccc8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ ccd0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ ccd8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ cce0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ cce8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ ccf0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ ccf8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ cd00
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ cd08
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ cd10
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ cd18
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ cd20
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ cd28
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ cd30
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ cd38
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ cd40
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ cd48
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ cd50
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ cd58
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ cd60
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ cd68
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ cd70
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ cd78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ cd80
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ cd88
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ cd90
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ cd98
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ cda0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ cda8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ cdb0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ cdb8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cdc0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cdc8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ cdd0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ cdd8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ cde0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ cde8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ cdf0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ cdf8
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ ce00
  dcd Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000,Opc000 ;@ ce08
  dcd Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010,Opc010 ;@ ce10
  dcd Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc018,Opc01f ;@ ce18
  dcd Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc020,Opc027 ;@ ce20
  dcd Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028,Opc028 ;@ ce28
  dcd Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030,Opc030 ;@ ce30
  dcd Opc038,Opc039,Opc03a,Opc03b,Opc03c,Opc03d,Opc03e,Opc03f ;@ ce38
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ ce40
  dcd Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040,Opc040 ;@ ce48
  dcd Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050,Opc050 ;@ ce50
  dcd Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc058,Opc05f ;@ ce58
  dcd Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc060,Opc067 ;@ ce60
  dcd Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068,Opc068 ;@ ce68
  dcd Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070,Opc070 ;@ ce70
  dcd Opc078,Opc079,Opc07a,Opc07b,Opc07c,Opc07d,Opc07e,Opc07f ;@ ce78
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ ce80
  dcd Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080,Opc080 ;@ ce88
  dcd Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090,Opc090 ;@ ce90
  dcd Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc098,Opc09f ;@ ce98
  dcd Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a0,Opc0a7 ;@ cea0
  dcd Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8,Opc0a8 ;@ cea8
  dcd Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0,Opc0b0 ;@ ceb0
  dcd Opc0b8,Opc0b9,Opc0ba,Opc0bb,Opc0bc,Opc0bd,Opc0be,Opc0bf ;@ ceb8
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ cec0
  dcd Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0,Opc0c0 ;@ cec8
  dcd Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0,Opc0d0 ;@ ced0
  dcd Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0d8,Opc0df ;@ ced8
  dcd Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e0,Opc0e7 ;@ cee0
  dcd Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8,Opc0e8 ;@ cee8
  dcd Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0,Opc0f0 ;@ cef0
  dcd Opc0f8,Opc0f9,Opc0fa,Opc0fb,Opc0fc,Op____,Op____,Op____ ;@ cef8
  dcd Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100,Opc100 ;@ cf00
  dcd Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108,Opc108 ;@ cf08
  dcd Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110,Opc110 ;@ cf10
  dcd Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc118,Opc11f ;@ cf18
  dcd Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc120,Opc127 ;@ cf20
  dcd Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128,Opc128 ;@ cf28
  dcd Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130,Opc130 ;@ cf30
  dcd Opc138,Opc139,Opc13a,Opc13b,Op____,Op____,Op____,Op____ ;@ cf38
  dcd Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140,Opc140 ;@ cf40
  dcd Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148,Opc148 ;@ cf48
  dcd Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150,Opc150 ;@ cf50
  dcd Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc158,Opc15f ;@ cf58
  dcd Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc160,Opc167 ;@ cf60
  dcd Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168,Opc168 ;@ cf68
  dcd Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170,Opc170 ;@ cf70
  dcd Opc178,Opc179,Opc17a,Opc17b,Op____,Op____,Op____,Op____ ;@ cf78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ cf80
  dcd Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188,Opc188 ;@ cf88
  dcd Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190,Opc190 ;@ cf90
  dcd Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc198,Opc19f ;@ cf98
  dcd Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a0,Opc1a7 ;@ cfa0
  dcd Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8,Opc1a8 ;@ cfa8
  dcd Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0,Opc1b0 ;@ cfb0
  dcd Opc1b8,Opc1b9,Opc1ba,Opc1bb,Op____,Op____,Op____,Op____ ;@ cfb8
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cfc0
  dcd Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0,Opc1c0 ;@ cfc8
  dcd Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0,Opc1d0 ;@ cfd0
  dcd Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1d8,Opc1df ;@ cfd8
  dcd Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e0,Opc1e7 ;@ cfe0
  dcd Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8,Opc1e8 ;@ cfe8
  dcd Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0,Opc1f0 ;@ cff0
  dcd Opc1f8,Opc1f9,Opc1fa,Opc1fb,Opc1fc,Op____,Op____,Op____ ;@ cff8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d000
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d008
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ d010
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ d018
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ d020
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ d028
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ d030
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ d038
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d040
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d048
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ d050
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ d058
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ d060
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ d068
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ d070
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ d078
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d080
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d088
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ d090
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ d098
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ d0a0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ d0a8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ d0b0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ d0b8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d0c0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d0c8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ d0d0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ d0d8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ d0e0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ d0e8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ d0f0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ d0f8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ d100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d108
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ d110
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ d118
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ d120
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ d128
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ d130
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ d138
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ d140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d148
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ d150
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ d158
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ d160
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ d168
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ d170
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ d178
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ d180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d188
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ d190
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ d198
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ d1a0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ d1a8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ d1b0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ d1b8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d1c0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d1c8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ d1d0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ d1d8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ d1e0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ d1e8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ d1f0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ d1f8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d200
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d208
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ d210
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ d218
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ d220
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ d228
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ d230
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ d238
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d240
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d248
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ d250
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ d258
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ d260
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ d268
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ d270
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ d278
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d280
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d288
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ d290
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ d298
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ d2a0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ d2a8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ d2b0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ d2b8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d2c0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d2c8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ d2d0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ d2d8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ d2e0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ d2e8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ d2f0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ d2f8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ d300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d308
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ d310
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ d318
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ d320
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ d328
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ d330
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ d338
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ d340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d348
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ d350
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ d358
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ d360
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ d368
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ d370
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ d378
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ d380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d388
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ d390
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ d398
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ d3a0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ d3a8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ d3b0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ d3b8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d3c0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d3c8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ d3d0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ d3d8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ d3e0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ d3e8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ d3f0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ d3f8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d400
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d408
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ d410
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ d418
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ d420
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ d428
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ d430
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ d438
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d440
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d448
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ d450
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ d458
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ d460
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ d468
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ d470
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ d478
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d480
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d488
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ d490
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ d498
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ d4a0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ d4a8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ d4b0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ d4b8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d4c0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d4c8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ d4d0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ d4d8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ d4e0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ d4e8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ d4f0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ d4f8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ d500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d508
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ d510
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ d518
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ d520
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ d528
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ d530
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ d538
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ d540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d548
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ d550
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ d558
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ d560
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ d568
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ d570
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ d578
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ d580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d588
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ d590
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ d598
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ d5a0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ d5a8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ d5b0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ d5b8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d5c0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d5c8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ d5d0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ d5d8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ d5e0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ d5e8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ d5f0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ d5f8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d600
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d608
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ d610
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ d618
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ d620
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ d628
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ d630
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ d638
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d640
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d648
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ d650
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ d658
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ d660
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ d668
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ d670
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ d678
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d680
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d688
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ d690
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ d698
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ d6a0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ d6a8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ d6b0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ d6b8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d6c0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d6c8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ d6d0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ d6d8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ d6e0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ d6e8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ d6f0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ d6f8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ d700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d708
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ d710
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ d718
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ d720
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ d728
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ d730
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ d738
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ d740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d748
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ d750
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ d758
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ d760
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ d768
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ d770
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ d778
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ d780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d788
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ d790
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ d798
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ d7a0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ d7a8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ d7b0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ d7b8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d7c0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d7c8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ d7d0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ d7d8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ d7e0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ d7e8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ d7f0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ d7f8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d800
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ d808
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ d810
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ d818
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ d820
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ d828
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ d830
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ d838
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d840
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ d848
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ d850
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ d858
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ d860
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ d868
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ d870
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ d878
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d880
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ d888
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ d890
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ d898
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ d8a0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ d8a8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ d8b0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ d8b8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d8c0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ d8c8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ d8d0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ d8d8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ d8e0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ d8e8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ d8f0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ d8f8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ d900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d908
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ d910
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ d918
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ d920
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ d928
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ d930
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ d938
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ d940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d948
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ d950
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ d958
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ d960
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ d968
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ d970
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ d978
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ d980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ d988
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ d990
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ d998
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ d9a0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ d9a8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ d9b0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ d9b8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d9c0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ d9c8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ d9d0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ d9d8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ d9e0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ d9e8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ d9f0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ d9f8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ da00
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ da08
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ da10
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ da18
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ da20
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ da28
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ da30
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ da38
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ da40
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ da48
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ da50
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ da58
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ da60
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ da68
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ da70
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ da78
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ da80
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ da88
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ da90
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ da98
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ daa0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ daa8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ dab0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ dab8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dac0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dac8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ dad0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ dad8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ dae0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ dae8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ daf0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ daf8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ db00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ db08
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ db10
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ db18
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ db20
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ db28
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ db30
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ db38
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ db40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ db48
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ db50
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ db58
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ db60
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ db68
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ db70
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ db78
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ db80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ db88
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ db90
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ db98
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ dba0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ dba8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ dbb0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ dbb8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ dbc0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ dbc8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ dbd0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ dbd8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ dbe0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ dbe8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ dbf0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ dbf8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ dc00
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ dc08
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ dc10
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ dc18
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ dc20
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ dc28
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ dc30
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ dc38
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ dc40
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ dc48
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ dc50
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ dc58
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ dc60
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ dc68
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ dc70
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ dc78
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ dc80
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ dc88
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ dc90
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ dc98
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ dca0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ dca8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ dcb0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ dcb8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dcc0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dcc8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ dcd0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ dcd8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ dce0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ dce8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ dcf0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ dcf8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ dd00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ dd08
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ dd10
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ dd18
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ dd20
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ dd28
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ dd30
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ dd38
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ dd40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ dd48
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ dd50
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ dd58
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ dd60
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ dd68
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ dd70
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ dd78
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ dd80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ dd88
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ dd90
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ dd98
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ dda0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ dda8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ ddb0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ ddb8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ ddc0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ ddc8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ ddd0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ ddd8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ dde0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ dde8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ ddf0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ ddf8
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ de00
  dcd Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000,Opd000 ;@ de08
  dcd Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010,Opd010 ;@ de10
  dcd Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd018,Opd01f ;@ de18
  dcd Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd020,Opd027 ;@ de20
  dcd Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028,Opd028 ;@ de28
  dcd Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030,Opd030 ;@ de30
  dcd Opd038,Opd039,Opd03a,Opd03b,Opd03c,Opd03d,Opd03e,Opd03f ;@ de38
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ de40
  dcd Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040,Opd040 ;@ de48
  dcd Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050,Opd050 ;@ de50
  dcd Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd058,Opd05f ;@ de58
  dcd Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd060,Opd067 ;@ de60
  dcd Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068,Opd068 ;@ de68
  dcd Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070,Opd070 ;@ de70
  dcd Opd078,Opd079,Opd07a,Opd07b,Opd07c,Opd07d,Opd07e,Opd07f ;@ de78
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ de80
  dcd Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080,Opd080 ;@ de88
  dcd Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090,Opd090 ;@ de90
  dcd Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd098,Opd09f ;@ de98
  dcd Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a0,Opd0a7 ;@ dea0
  dcd Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8,Opd0a8 ;@ dea8
  dcd Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0,Opd0b0 ;@ deb0
  dcd Opd0b8,Opd0b9,Opd0ba,Opd0bb,Opd0bc,Opd0bd,Opd0be,Opd0bf ;@ deb8
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dec0
  dcd Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0,Opd0c0 ;@ dec8
  dcd Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0,Opd0d0 ;@ ded0
  dcd Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0d8,Opd0df ;@ ded8
  dcd Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e0,Opd0e7 ;@ dee0
  dcd Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8,Opd0e8 ;@ dee8
  dcd Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0,Opd0f0 ;@ def0
  dcd Opd0f8,Opd0f9,Opd0fa,Opd0fb,Opd0fc,Op____,Op____,Op____ ;@ def8
  dcd Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100,Opd100 ;@ df00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ df08
  dcd Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110,Opd110 ;@ df10
  dcd Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd118,Opd11f ;@ df18
  dcd Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd120,Opd127 ;@ df20
  dcd Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128,Opd128 ;@ df28
  dcd Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130,Opd130 ;@ df30
  dcd Opd138,Opd139,Opd13a,Opd13b,Op____,Op____,Op____,Op____ ;@ df38
  dcd Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140,Opd140 ;@ df40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ df48
  dcd Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150,Opd150 ;@ df50
  dcd Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd158,Opd15f ;@ df58
  dcd Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd160,Opd167 ;@ df60
  dcd Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168,Opd168 ;@ df68
  dcd Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170,Opd170 ;@ df70
  dcd Opd178,Opd179,Opd17a,Opd17b,Op____,Op____,Op____,Op____ ;@ df78
  dcd Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180,Opd180 ;@ df80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ df88
  dcd Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190,Opd190 ;@ df90
  dcd Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd198,Opd19f ;@ df98
  dcd Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a0,Opd1a7 ;@ dfa0
  dcd Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8,Opd1a8 ;@ dfa8
  dcd Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0,Opd1b0 ;@ dfb0
  dcd Opd1b8,Opd1b9,Opd1ba,Opd1bb,Op____,Op____,Op____,Op____ ;@ dfb8
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ dfc0
  dcd Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0,Opd1c0 ;@ dfc8
  dcd Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0,Opd1d0 ;@ dfd0
  dcd Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1d8,Opd1df ;@ dfd8
  dcd Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e0,Opd1e7 ;@ dfe0
  dcd Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8,Opd1e8 ;@ dfe8
  dcd Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0,Opd1f0 ;@ dff0
  dcd Opd1f8,Opd1f9,Opd1fa,Opd1fb,Opd1fc,Op____,Op____,Op____ ;@ dff8
  dcd Ope000,Ope000,Ope000,Ope000,Ope000,Ope000,Ope000,Ope000 ;@ e000
  dcd Ope008,Ope008,Ope008,Ope008,Ope008,Ope008,Ope008,Ope008 ;@ e008
  dcd Ope010,Ope010,Ope010,Ope010,Ope010,Ope010,Ope010,Ope010 ;@ e010
  dcd Ope018,Ope018,Ope018,Ope018,Ope018,Ope018,Ope018,Ope018 ;@ e018
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ e020
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ e028
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ e030
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ e038
  dcd Ope040,Ope040,Ope040,Ope040,Ope040,Ope040,Ope040,Ope040 ;@ e040
  dcd Ope048,Ope048,Ope048,Ope048,Ope048,Ope048,Ope048,Ope048 ;@ e048
  dcd Ope050,Ope050,Ope050,Ope050,Ope050,Ope050,Ope050,Ope050 ;@ e050
  dcd Ope058,Ope058,Ope058,Ope058,Ope058,Ope058,Ope058,Ope058 ;@ e058
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ e060
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ e068
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ e070
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ e078
  dcd Ope080,Ope080,Ope080,Ope080,Ope080,Ope080,Ope080,Ope080 ;@ e080
  dcd Ope088,Ope088,Ope088,Ope088,Ope088,Ope088,Ope088,Ope088 ;@ e088
  dcd Ope090,Ope090,Ope090,Ope090,Ope090,Ope090,Ope090,Ope090 ;@ e090
  dcd Ope098,Ope098,Ope098,Ope098,Ope098,Ope098,Ope098,Ope098 ;@ e098
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ e0a0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ e0a8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ e0b0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ e0b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e0c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e0c8
  dcd Ope0d0,Ope0d0,Ope0d0,Ope0d0,Ope0d0,Ope0d0,Ope0d0,Ope0d0 ;@ e0d0
  dcd Ope0d8,Ope0d8,Ope0d8,Ope0d8,Ope0d8,Ope0d8,Ope0d8,Ope0df ;@ e0d8
  dcd Ope0e0,Ope0e0,Ope0e0,Ope0e0,Ope0e0,Ope0e0,Ope0e0,Ope0e7 ;@ e0e0
  dcd Ope0e8,Ope0e8,Ope0e8,Ope0e8,Ope0e8,Ope0e8,Ope0e8,Ope0e8 ;@ e0e8
  dcd Ope0f0,Ope0f0,Ope0f0,Ope0f0,Ope0f0,Ope0f0,Ope0f0,Ope0f0 ;@ e0f0
  dcd Ope0f8,Ope0f9,Ope0fa,Ope0fb,Op____,Op____,Op____,Op____ ;@ e0f8
  dcd Ope100,Ope100,Ope100,Ope100,Ope100,Ope100,Ope100,Ope100 ;@ e100
  dcd Ope108,Ope108,Ope108,Ope108,Ope108,Ope108,Ope108,Ope108 ;@ e108
  dcd Ope110,Ope110,Ope110,Ope110,Ope110,Ope110,Ope110,Ope110 ;@ e110
  dcd Ope118,Ope118,Ope118,Ope118,Ope118,Ope118,Ope118,Ope118 ;@ e118
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ e120
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ e128
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ e130
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ e138
  dcd Ope140,Ope140,Ope140,Ope140,Ope140,Ope140,Ope140,Ope140 ;@ e140
  dcd Ope148,Ope148,Ope148,Ope148,Ope148,Ope148,Ope148,Ope148 ;@ e148
  dcd Ope150,Ope150,Ope150,Ope150,Ope150,Ope150,Ope150,Ope150 ;@ e150
  dcd Ope158,Ope158,Ope158,Ope158,Ope158,Ope158,Ope158,Ope158 ;@ e158
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ e160
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ e168
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ e170
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ e178
  dcd Ope180,Ope180,Ope180,Ope180,Ope180,Ope180,Ope180,Ope180 ;@ e180
  dcd Ope188,Ope188,Ope188,Ope188,Ope188,Ope188,Ope188,Ope188 ;@ e188
  dcd Ope190,Ope190,Ope190,Ope190,Ope190,Ope190,Ope190,Ope190 ;@ e190
  dcd Ope198,Ope198,Ope198,Ope198,Ope198,Ope198,Ope198,Ope198 ;@ e198
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ e1a0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ e1a8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ e1b0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ e1b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e1c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e1c8
  dcd Ope1d0,Ope1d0,Ope1d0,Ope1d0,Ope1d0,Ope1d0,Ope1d0,Ope1d0 ;@ e1d0
  dcd Ope1d8,Ope1d8,Ope1d8,Ope1d8,Ope1d8,Ope1d8,Ope1d8,Ope1df ;@ e1d8
  dcd Ope1e0,Ope1e0,Ope1e0,Ope1e0,Ope1e0,Ope1e0,Ope1e0,Ope1e7 ;@ e1e0
  dcd Ope1e8,Ope1e8,Ope1e8,Ope1e8,Ope1e8,Ope1e8,Ope1e8,Ope1e8 ;@ e1e8
  dcd Ope1f0,Ope1f0,Ope1f0,Ope1f0,Ope1f0,Ope1f0,Ope1f0,Ope1f0 ;@ e1f0
  dcd Ope1f8,Ope1f9,Ope1fa,Ope1fb,Op____,Op____,Op____,Op____ ;@ e1f8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ e200
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ e208
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ e210
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ e218
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ e220
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ e228
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ e230
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ e238
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ e240
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ e248
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ e250
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ e258
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ e260
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ e268
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ e270
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ e278
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ e280
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ e288
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ e290
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ e298
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ e2a0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ e2a8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ e2b0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ e2b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e2c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e2c8
  dcd Ope2d0,Ope2d0,Ope2d0,Ope2d0,Ope2d0,Ope2d0,Ope2d0,Ope2d0 ;@ e2d0
  dcd Ope2d8,Ope2d8,Ope2d8,Ope2d8,Ope2d8,Ope2d8,Ope2d8,Ope2df ;@ e2d8
  dcd Ope2e0,Ope2e0,Ope2e0,Ope2e0,Ope2e0,Ope2e0,Ope2e0,Ope2e7 ;@ e2e0
  dcd Ope2e8,Ope2e8,Ope2e8,Ope2e8,Ope2e8,Ope2e8,Ope2e8,Ope2e8 ;@ e2e8
  dcd Ope2f0,Ope2f0,Ope2f0,Ope2f0,Ope2f0,Ope2f0,Ope2f0,Ope2f0 ;@ e2f0
  dcd Ope2f8,Ope2f9,Ope2fa,Ope2fb,Op____,Op____,Op____,Op____ ;@ e2f8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ e300
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ e308
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ e310
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ e318
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ e320
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ e328
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ e330
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ e338
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ e340
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ e348
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ e350
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ e358
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ e360
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ e368
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ e370
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ e378
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ e380
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ e388
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ e390
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ e398
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ e3a0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ e3a8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ e3b0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ e3b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e3c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e3c8
  dcd Ope3d0,Ope3d0,Ope3d0,Ope3d0,Ope3d0,Ope3d0,Ope3d0,Ope3d0 ;@ e3d0
  dcd Ope3d8,Ope3d8,Ope3d8,Ope3d8,Ope3d8,Ope3d8,Ope3d8,Ope3df ;@ e3d8
  dcd Ope3e0,Ope3e0,Ope3e0,Ope3e0,Ope3e0,Ope3e0,Ope3e0,Ope3e7 ;@ e3e0
  dcd Ope3e8,Ope3e8,Ope3e8,Ope3e8,Ope3e8,Ope3e8,Ope3e8,Ope3e8 ;@ e3e8
  dcd Ope3f0,Ope3f0,Ope3f0,Ope3f0,Ope3f0,Ope3f0,Ope3f0,Ope3f0 ;@ e3f0
  dcd Ope3f8,Ope3f9,Ope3fa,Ope3fb,Op____,Op____,Op____,Op____ ;@ e3f8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ e400
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ e408
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ e410
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ e418
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ e420
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ e428
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ e430
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ e438
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ e440
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ e448
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ e450
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ e458
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ e460
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ e468
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ e470
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ e478
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ e480
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ e488
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ e490
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ e498
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ e4a0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ e4a8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ e4b0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ e4b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e4c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e4c8
  dcd Ope4d0,Ope4d0,Ope4d0,Ope4d0,Ope4d0,Ope4d0,Ope4d0,Ope4d0 ;@ e4d0
  dcd Ope4d8,Ope4d8,Ope4d8,Ope4d8,Ope4d8,Ope4d8,Ope4d8,Ope4df ;@ e4d8
  dcd Ope4e0,Ope4e0,Ope4e0,Ope4e0,Ope4e0,Ope4e0,Ope4e0,Ope4e7 ;@ e4e0
  dcd Ope4e8,Ope4e8,Ope4e8,Ope4e8,Ope4e8,Ope4e8,Ope4e8,Ope4e8 ;@ e4e8
  dcd Ope4f0,Ope4f0,Ope4f0,Ope4f0,Ope4f0,Ope4f0,Ope4f0,Ope4f0 ;@ e4f0
  dcd Ope4f8,Ope4f9,Ope4fa,Ope4fb,Op____,Op____,Op____,Op____ ;@ e4f8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ e500
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ e508
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ e510
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ e518
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ e520
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ e528
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ e530
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ e538
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ e540
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ e548
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ e550
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ e558
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ e560
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ e568
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ e570
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ e578
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ e580
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ e588
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ e590
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ e598
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ e5a0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ e5a8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ e5b0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ e5b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e5c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e5c8
  dcd Ope5d0,Ope5d0,Ope5d0,Ope5d0,Ope5d0,Ope5d0,Ope5d0,Ope5d0 ;@ e5d0
  dcd Ope5d8,Ope5d8,Ope5d8,Ope5d8,Ope5d8,Ope5d8,Ope5d8,Ope5df ;@ e5d8
  dcd Ope5e0,Ope5e0,Ope5e0,Ope5e0,Ope5e0,Ope5e0,Ope5e0,Ope5e7 ;@ e5e0
  dcd Ope5e8,Ope5e8,Ope5e8,Ope5e8,Ope5e8,Ope5e8,Ope5e8,Ope5e8 ;@ e5e8
  dcd Ope5f0,Ope5f0,Ope5f0,Ope5f0,Ope5f0,Ope5f0,Ope5f0,Ope5f0 ;@ e5f0
  dcd Ope5f8,Ope5f9,Ope5fa,Ope5fb,Op____,Op____,Op____,Op____ ;@ e5f8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ e600
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ e608
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ e610
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ e618
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ e620
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ e628
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ e630
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ e638
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ e640
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ e648
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ e650
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ e658
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ e660
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ e668
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ e670
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ e678
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ e680
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ e688
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ e690
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ e698
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ e6a0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ e6a8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ e6b0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ e6b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e6c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e6c8
  dcd Ope6d0,Ope6d0,Ope6d0,Ope6d0,Ope6d0,Ope6d0,Ope6d0,Ope6d0 ;@ e6d0
  dcd Ope6d8,Ope6d8,Ope6d8,Ope6d8,Ope6d8,Ope6d8,Ope6d8,Ope6df ;@ e6d8
  dcd Ope6e0,Ope6e0,Ope6e0,Ope6e0,Ope6e0,Ope6e0,Ope6e0,Ope6e7 ;@ e6e0
  dcd Ope6e8,Ope6e8,Ope6e8,Ope6e8,Ope6e8,Ope6e8,Ope6e8,Ope6e8 ;@ e6e8
  dcd Ope6f0,Ope6f0,Ope6f0,Ope6f0,Ope6f0,Ope6f0,Ope6f0,Ope6f0 ;@ e6f0
  dcd Ope6f8,Ope6f9,Ope6fa,Ope6fb,Op____,Op____,Op____,Op____ ;@ e6f8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ e700
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ e708
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ e710
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ e718
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ e720
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ e728
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ e730
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ e738
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ e740
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ e748
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ e750
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ e758
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ e760
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ e768
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ e770
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ e778
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ e780
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ e788
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ e790
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ e798
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ e7a0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ e7a8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ e7b0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ e7b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e7c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e7c8
  dcd Ope7d0,Ope7d0,Ope7d0,Ope7d0,Ope7d0,Ope7d0,Ope7d0,Ope7d0 ;@ e7d0
  dcd Ope7d8,Ope7d8,Ope7d8,Ope7d8,Ope7d8,Ope7d8,Ope7d8,Ope7df ;@ e7d8
  dcd Ope7e0,Ope7e0,Ope7e0,Ope7e0,Ope7e0,Ope7e0,Ope7e0,Ope7e7 ;@ e7e0
  dcd Ope7e8,Ope7e8,Ope7e8,Ope7e8,Ope7e8,Ope7e8,Ope7e8,Ope7e8 ;@ e7e8
  dcd Ope7f0,Ope7f0,Ope7f0,Ope7f0,Ope7f0,Ope7f0,Ope7f0,Ope7f0 ;@ e7f0
  dcd Ope7f8,Ope7f9,Ope7fa,Ope7fb,Op____,Op____,Op____,Op____ ;@ e7f8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ e800
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ e808
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ e810
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ e818
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ e820
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ e828
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ e830
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ e838
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ e840
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ e848
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ e850
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ e858
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ e860
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ e868
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ e870
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ e878
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ e880
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ e888
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ e890
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ e898
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ e8a0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ e8a8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ e8b0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ e8b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e8f8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ e900
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ e908
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ e910
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ e918
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ e920
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ e928
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ e930
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ e938
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ e940
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ e948
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ e950
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ e958
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ e960
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ e968
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ e970
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ e978
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ e980
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ e988
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ e990
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ e998
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ e9a0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ e9a8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ e9b0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ e9b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ e9f8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ ea00
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ ea08
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ ea10
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ ea18
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ ea20
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ ea28
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ ea30
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ ea38
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ ea40
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ ea48
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ ea50
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ ea58
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ ea60
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ ea68
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ ea70
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ ea78
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ ea80
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ ea88
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ ea90
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ ea98
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ eaa0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ eaa8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ eab0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ eab8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eac0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eac8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ead0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ead8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eae0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eae8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eaf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eaf8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ eb00
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ eb08
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ eb10
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ eb18
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ eb20
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ eb28
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ eb30
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ eb38
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ eb40
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ eb48
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ eb50
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ eb58
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ eb60
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ eb68
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ eb70
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ eb78
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ eb80
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ eb88
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ eb90
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ eb98
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ eba0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ eba8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ ebb0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ ebb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ebf8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ ec00
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ ec08
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ ec10
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ ec18
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ ec20
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ ec28
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ ec30
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ ec38
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ ec40
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ ec48
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ ec50
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ ec58
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ ec60
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ ec68
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ ec70
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ ec78
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ ec80
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ ec88
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ ec90
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ ec98
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ eca0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ eca8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ ecb0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ ecb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ece0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ece8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ecf8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ ed00
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ ed08
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ ed10
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ ed18
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ ed20
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ ed28
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ ed30
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ ed38
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ ed40
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ ed48
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ ed50
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ ed58
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ ed60
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ ed68
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ ed70
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ ed78
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ ed80
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ ed88
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ ed90
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ ed98
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ eda0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ eda8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ edb0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ edb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ede0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ede8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ edf8
  dcd Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00,Opee00 ;@ ee00
  dcd Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08,Opee08 ;@ ee08
  dcd Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10,Opee10 ;@ ee10
  dcd Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18,Opee18 ;@ ee18
  dcd Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020,Ope020 ;@ ee20
  dcd Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028,Ope028 ;@ ee28
  dcd Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030,Ope030 ;@ ee30
  dcd Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038,Ope038 ;@ ee38
  dcd Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40,Opee40 ;@ ee40
  dcd Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48,Opee48 ;@ ee48
  dcd Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50,Opee50 ;@ ee50
  dcd Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58,Opee58 ;@ ee58
  dcd Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060,Ope060 ;@ ee60
  dcd Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068,Ope068 ;@ ee68
  dcd Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070,Ope070 ;@ ee70
  dcd Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078,Ope078 ;@ ee78
  dcd Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80,Opee80 ;@ ee80
  dcd Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88,Opee88 ;@ ee88
  dcd Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90,Opee90 ;@ ee90
  dcd Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98,Opee98 ;@ ee98
  dcd Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0,Ope0a0 ;@ eea0
  dcd Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8,Ope0a8 ;@ eea8
  dcd Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0,Ope0b0 ;@ eeb0
  dcd Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8,Ope0b8 ;@ eeb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eec0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eec8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eed0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eed8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eee0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eee8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eef0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eef8
  dcd Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00,Opef00 ;@ ef00
  dcd Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08,Opef08 ;@ ef08
  dcd Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10,Opef10 ;@ ef10
  dcd Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18,Opef18 ;@ ef18
  dcd Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120,Ope120 ;@ ef20
  dcd Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128,Ope128 ;@ ef28
  dcd Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130,Ope130 ;@ ef30
  dcd Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138,Ope138 ;@ ef38
  dcd Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40,Opef40 ;@ ef40
  dcd Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48,Opef48 ;@ ef48
  dcd Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50,Opef50 ;@ ef50
  dcd Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58,Opef58 ;@ ef58
  dcd Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160,Ope160 ;@ ef60
  dcd Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168,Ope168 ;@ ef68
  dcd Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170,Ope170 ;@ ef70
  dcd Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178,Ope178 ;@ ef78
  dcd Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80,Opef80 ;@ ef80
  dcd Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88,Opef88 ;@ ef88
  dcd Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90,Opef90 ;@ ef90
  dcd Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98,Opef98 ;@ ef98
  dcd Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0,Ope1a0 ;@ efa0
  dcd Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8,Ope1a8 ;@ efa8
  dcd Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0,Ope1b0 ;@ efb0
  dcd Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8,Ope1b8 ;@ efb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ efe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ eff8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f000
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f008
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f010
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f018
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f020
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f028
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f030
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f038
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f040
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f048
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f050
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f058
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f060
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f068
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f070
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f078
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f080
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f088
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f090
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f098
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f0f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f100
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f108
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f110
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f118
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f120
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f128
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f130
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f138
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f140
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f148
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f150
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f158
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f160
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f168
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f170
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f178
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f180
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f188
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f190
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f198
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f1f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f200
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f208
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f210
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f218
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f220
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f228
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f230
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f238
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f240
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f248
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f250
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f258
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f260
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f268
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f270
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f278
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f280
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f288
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f290
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f298
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f2f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f300
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f308
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f310
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f318
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f320
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f328
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f330
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f338
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f340
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f348
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f350
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f358
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f360
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f368
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f370
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f378
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f380
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f388
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f390
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f398
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f3f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f400
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f408
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f410
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f418
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f420
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f428
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f430
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f438
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f440
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f448
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f450
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f458
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f460
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f468
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f470
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f478
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f480
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f488
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f490
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f498
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f4f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f500
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f508
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f510
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f518
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f520
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f528
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f530
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f538
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f540
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f548
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f550
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f558
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f560
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f568
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f570
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f578
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f580
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f588
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f590
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f598
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f5f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f600
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f608
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f610
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f618
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f620
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f628
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f630
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f638
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f640
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f648
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f650
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f658
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f660
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f668
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f670
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f678
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f680
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f688
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f690
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f698
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f6f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f700
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f708
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f710
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f718
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f720
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f728
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f730
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f738
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f740
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f748
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f750
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f758
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f760
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f768
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f770
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f778
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f780
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f788
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f790
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f798
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f7f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f800
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f808
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f810
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f818
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f820
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f828
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f830
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f838
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f840
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f848
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f850
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f858
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f860
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f868
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f870
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f878
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f880
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f888
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f890
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f898
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f8f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f900
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f908
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f910
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f918
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f920
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f928
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f930
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f938
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f940
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f948
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f950
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f958
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f960
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f968
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f970
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f978
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f980
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f988
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f990
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f998
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9a0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9a8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9b0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9b8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9c0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9c8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9d0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9d8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9e0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9e8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9f0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ f9f8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fa98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ faa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ faa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fab0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fab8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fac0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fac8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fad0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fad8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fae0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fae8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ faf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ faf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fb98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fba0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fba8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fbf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fc98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fca0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fca8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fce0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fce8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fcf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fd98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fda0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fda8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fde0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fde8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdf0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fdf8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fe98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fea0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fea8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ feb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ feb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fec0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fec8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fed0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fed8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fee0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fee8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fef0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fef8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff00
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff08
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff10
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff18
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff20
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff28
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff30
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff38
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff40
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff48
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff50
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff58
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff60
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff68
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff70
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff78
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff80
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff88
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff90
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ff98
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffa0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffa8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffb0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffb8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffc0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffc8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffd0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffd8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffe0
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ ffe8
  dcd Op____,Op____,Op____,Op____,Op____,Op____,Op____,Op____ ;@ fff0
  dcd Op____,Op____,Op____,Op____,Op____,Op____
  END
