/*
	File:			Application.cpp	

	Description:	Contains the 68K portion of the Palm ARMlet.  It sets up a form so that
					the UI works with the Palm UI, and it calls the ARMlet.

	Author:		

	Copyright:		Copyright  2003 by Tapwave, Inc.

	Disclaimer:		IMPORTANT:  This Tapwave software is provided by Tapwave, Inc. ("Tapwave").  Your 
					use is subject to and governed by terms and conditions of the Software Development 
					Kit Agreement ("SDK Agreement") between you and Tapwave.  If you have not entered 
					into Tapwaves standard SDK Agreement with Tapwave, you have no right or license 
					to use, reproduce, modify, distribute or otherwise exploit this Tapwave software.  
					You may obtain a copy of Tapwaves standard SDK Agreement by calling 650-960-1817
					or visiting Tapwave at http://www.tapwave.com/developers/. 

	Change History (most recent first):
				
*/

//#define TESTING_ONLY


#include <TapWave.h>
#include <TwRuntime.h>
#include <StringMgr.h>


#include "ljZRsc.h"

#include "App_Cst.h"

#include "App_FX.h"

static unsigned char *locked_scr=NULL;
static Int16 g_prev_form;
static int g_launch_emu;
static char **g_listchoice=NULL;


#define MAX_ROM_LISTED 512
static int g_num_entries=0;

static char g_labelbutton[0x12][10]=
{"none","Red","Green","Yellow","Blue","L","R","Func","Home",
"Up","Down","Left","Right","TL","BL","TR","BR","Analog"};

static char g_inputfld[255];

static char g_lastQuickRomChar;

static char g_helpmsg[]="\
[LJZ] v0.7\n\
--------------\n\
Compiled on 2004/08/31\n\
\n\
* multiple screen mode (zoom/smoothing)\n\
* adjustable frameskipping\n\
* SRAM support\n\
* Load/Save state\n\
* Turbo mode\n\
* sound (except Wonderswan)\n\
* zipped rom\n\
\n\
Put your ROMS in \n\
/PALM/Programs/LJZ/xxx/Roms/\n\
Where xxx is NES,SNES,...\n\
\n\
[GB/GBC]\n\
* powered by fGB32(gnuboy)\n\
* from great Rlyeh & gnuboy team\n\
\n\
[GENESIS]\n\
* powered by GenesisPlus\n\
from Charles Mac Donald\n\
* powered by Cyclone 68000 ARM cpu core\n\
from Dave\n\
\n\
[TG16]\n\
* powered by TG16Emu\n\
from Charles Mac Donald\n\
* no savestate\n\
* still in alpha state (buggy)\n\
\n\
[NEOGEO POCKET]\n\
* powered by NeoGP32 from great THUNDERZ!\n\
\n\
[NES]\n\
* using/inspired by\n\
Shatbox,Nofrendo,Nester,NNNEsterJ,\n\
NesterDC,PocketNester,Nestopia\n\
* support mappers :\n\
0,1,2,3,4,5,6,7,8,9,10,11,13,15,\n\
16,17,18,19,21,22,23,24,25,26,\n\
32,33,34,40,41,42,43,44,45,64,\n\
66,68,69,70,71,72,73,74,75,76,\n\
77,78,79,80,82,83,85,86,87,88,\n\
89,90,91,92,93,94,95,96,97,99,\n\
100,101,105,112,113,114,115,117,\n\
118,119,122,187,255\n\
\n\
[SNES]\n\
* powered by OpenSnes9xGP(Snes9X 1.39a)\n\
from Laxer3a,ThunderZ,YoyoFR\n\
* spc playback through openspc++\n\
\n\
[WONDERSWAN]\n\
* powered by Oswan 0.7 \n\
* No audio, slow emulation\n\
* still beta emu\n\
----------\n\
Greetings to\n\
Skeezix\n\
Rlyeh\n\
ThunderZ\n\
Laxer3a\n\
Dave\n\
Sine909\n\
All the donators!\n\
....\n\
\n\
\n\
\tYoyoFR\n";


static int g_midi;
static unsigned char* g_midiData;
static MemHandle g_midiDataH;
static TwSmfType* g_smfHandle;

//int g_frmloaded;
static int g_config_mode=0;
static int g_rom_system;

typedef struct {
UInt16 g_SnesFastSprite;
UInt16 g_SnesTransparency;
UInt16 g_SnesRender;
UInt16 g_SnesSound;
UInt16 g_SnesSmoothing;
UInt16 g_SnesScreenMode;
UInt16 g_SnesFSkip;
UInt16 g_SnesSoundFreq;
UInt16 g_SnesASMCPU;

UInt16 g_NesSound;
UInt16 g_NesSmoothing;
UInt16 g_NesScreenMode;
UInt16 g_NesFSkip;
UInt16 g_NesSoundFreq;

UInt16 g_GBSound;
UInt16 g_GBSmoothing;
UInt16 g_GBScreenMode;
UInt16 g_GBFSkip;
UInt16 g_GBSoundFreq;

UInt16 g_GENZSound;
UInt16 g_GENZSmoothing;
UInt16 g_GENZScreenMode;
UInt16 g_GENZFSkip;
UInt16 g_GENZSoundFreq;
UInt16 g_GENZASMCPU;


UInt16 g_NEOPOPZSound;
UInt16 g_NEOPOPZSmoothing;
UInt16 g_NEOPOPZScreenMode;
UInt16 g_NEOPOPZFSkip;
UInt16 g_NEOPOPZSoundFreq;
UInt16 g_NEOPOPZEmuAccuracy;


UInt16 g_OSWANZSound;
UInt16 g_OSWANZSmoothing;
UInt16 g_OSWANZScreenMode;
UInt16 g_OSWANZFSkip;
UInt16 g_OSWANZSoundFreq;

UInt16 g_TG16ZSound;
UInt16 g_TG16ZSmoothing;
UInt16 g_TG16ZScreenMode;
UInt16 g_TG16ZFSkip;
UInt16 g_TG16ZSoundFreq;


UInt16 g_sysfilter;//|SYS_SNES|SYS_GB|SYS_GENZ;
UInt16 g_conf_system;
unsigned char g_inputconf[7][20];
char g_lastrom[256];
UInt16 g_firsttime;
} t_ljz_pref;

static t_ljz_pref g_ljz_pref;
/**********************************/

#define UPDATE_UI_COLORS(v) \
{ \
			RGBColorType col; \
			int cr,cg,cb;\
			switch (v) \
			{\
				case SYS_NES:cr=200;cg=200;cb=200; \
				break; \
				case SYS_SNES:cr=210;cg=160;cb=230; \
				break; \
				case SYS_GB:cr=255;cg=255;cb=128; \
				break; \
				case SYS_GENZ:cr=230;cg=190;cb=220; \
				break; \
				case SYS_NEOPOPZ:cr=200;cg=220;cb=240; \
				break; \
				case SYS_OSWANZ:cr=200;cg=240;cb=220; \
				break; \
				case SYS_TG16Z:cr=255;cg=230;cb=230; \
				break; \
			}			 \
			col.r=cr;col.g=cg;col.b=cb; \
			UIColorSetTableEntry(UIFormFill,&col);			 \
			col.r>>=1;col.g>>=1;col.b>>=1; \
			UIColorSetTableEntry(UIFormFrame,&col); \
			col.r=255;col.g=245;col.b=250; \
			UIColorSetTableEntry(UIObjectFill,&col);			 \
			col.r=(cr)>>1;\
			col.g=(cg)>>1;\
			col.b=(cb)>>1;\
			UIColorSetTableEntry(UIObjectFrame,&col); \
}

#define ASK_FOR_UPDATE \
	{EventType my_event; \
	my_event.eType=frmUpdateEvent;\
	EvtAddEventToQueue(&my_event);}
	
#define DRAW_SYSTEM_ICON(S,X,Y)\
	{\
		MemHandle hRsc=NULL;\
			switch (S)\
			{\
				case SYS_NES:hRsc = DmGetResource(bitmapRsc, NESBMP);break;\
				case SYS_SNES:hRsc = DmGetResource(bitmapRsc, SNESBMP);break;\
				case SYS_GB:hRsc = DmGetResource(bitmapRsc, GBBMP);break;\
				case SYS_GENZ:hRsc = DmGetResource(bitmapRsc, GENESISBMP);break;\
				case SYS_NEOPOPZ:hRsc = DmGetResource(bitmapRsc, NGPBMP);break;\
				case SYS_OSWANZ:hRsc = DmGetResource(bitmapRsc, WSBMP);break;\
				case SYS_TG16Z:hRsc = DmGetResource(bitmapRsc, TG16BMP);break;\
			}\
			if (hRsc)\
			{	BitmapType* bitmapP = (BitmapType*) MemHandleLock(hRsc);\
				WinPaintBitmap(bitmapP, X, Y);\
				MemHandleUnlock(hRsc);\
				DmReleaseResource(hRsc);}\
   }


static void Start_Midi(UInt32 num);
static void Stop_Midi(void);

void ResetQuickRomLabel(FormType *frmP)
{
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrm0BButton) ),"_ab");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmCFButton) ),"cdef");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmGJButton) ),"ghij");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmKNButton) ),"klmn");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmORButton) ),"opqr");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmSVButton) ),"stuv");
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrmWZButton) ),"wxyz");
}

/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version is meet your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags that indicate if the application 
 *                                UI is initialized.
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err RomVersionCompatible(UInt16 launchFlags)
{
	UInt32 romVersion;

    // Check Tapwave Native Application glue.
    if (errNone == FtrGet(twFtrCreator, twFtrAPIGlue, &romVersion)) {
    
/*    char strinfo[64];
	StrPrintF(strinfo,"%X.%X.%X %X %X",
	sysGetROMVerMajor(romVersion),
	sysGetROMVerMinor(romVersion),
	sysGetROMVerFix(romVersion),
	sysGetROMVerStage(romVersion),
	sysGetROMVerBuild(romVersion));
	menu_message(strinfo);
	menu_waitBtn(1,1);
    */
        return errNone;
    }

    // Show dialog if we can.
    if ((launchFlags & sysAppLaunchFlagNewGlobals) && (launchFlags & sysAppLaunchFlagUIApp)) {
        FrmAlert(RomIncompatibleAlert);
    }

    // Palm OS 1.0 will continuously relaunch this app unless we
    // switch to another safe one.
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion <= sysMakeROMVersion(1,0,0,sysROMStageRelease,0)) {
        AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
    }
    
    

    return sysErrRomIncompatible;
}

static void Sort_Entries(char **listchoice,int num)
{
	int i,j;
    char *a,*b;

    if (num<2)
      return;
    
    for (i=0;i<num;i++)
      {
        for (j=0;j<num-1;j++)
          {
          	a=listchoice[j]+3;
          	b=listchoice[j+1]+3;
            if (StrCaselessCompare(a, b)>0)
              {
                char * temp;
                temp = listchoice[j];
                listchoice[j] = listchoice[j+1];
                listchoice[j+1] = temp;
              }
          }
      }
}


static void InitRomList(ListType *list,int _system2list)
{
	Err			err = errNone;
	char		tmpName[256];	
	int 		cur_entry=0;	
	int 		total_size=0;
	UInt16 		volRefNum;			
	UInt32 		volIterator; 
	
	g_num_entries = MAX_ROM_LISTED;//max 512 roms
	g_listchoice=(char**)MemPtrNew(g_num_entries*sizeof(char*));	

	if (!g_listchoice) 			
	{		
		LstSetListChoices(list,(char**)NULL,0);		
		menu_message("Browser : not enough memory!");
		menu_waitBtn(1,1);
		return;
	}			
	
/*************************
  First pass : NES  
/*************************/  
	if (_system2list&SYS_NES)
	{
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __NES_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".nes") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_N-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}
	}
/*************************
  Second pass : SNES  
/*************************/  
    if ((_system2list&SYS_SNES)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __SNES_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".spc") != 0)&&(StrCaselessCompare(ext, ".smc") != 0)&&(StrCaselessCompare(ext, ".sfc") != 0)&&(StrCaselessCompare(ext, ".fig") != 0)&&(StrCaselessCompare(ext, ".bin") != 0)&&(StrCaselessCompare(ext, ".1") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_S-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}		
	
/*************************
  Third pass : GB
/*************************/  
    if ((_system2list&SYS_GB)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __GB_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".gb") != 0)&&(StrCaselessCompare(ext, ".gbc") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_G-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}		
/*************************
  Fourth pass : Genesis
/*************************/  
    if ((_system2list&SYS_GENZ)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __GENZ_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".smd") != 0)&&(StrCaselessCompare(ext, ".bin") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_M-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}		
	
/*************************
  Fourth pass : Neogeo Pocket
/*************************/  
    if ((_system2list&SYS_NEOPOPZ)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __NEOPOPZ_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{

						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".ngp") != 0)&&(StrCaselessCompare(ext, ".ngc") != 0)&&(StrCaselessCompare(ext, ".npc") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_P-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}		

/*************************
  Fifth pass : wonderswan
/*************************/  
    if ((_system2list&SYS_OSWANZ)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __OSWANZ_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".ws") != 0)&&(StrCaselessCompare(ext, ".wsc") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_W-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}	

/*************************
  Sixth pass : tg16
/*************************/  
    if ((_system2list&SYS_TG16Z)&&(cur_entry<g_num_entries))
    {
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef dirRef;
			UInt32 dirIterator;					
			// open the directory first, to get the directory reference
			// volRefNum must have already been defined
			err = VFSFileOpen(volRefNum, __TG16Z_ROMS_DIR__, vfsModeRead, &dirRef);
			if(err == errNone) 
			{
				info.nameP = tmpName; // point to local buffer												
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
				while (dirIterator != vfsIteratorStop) 
				{
					// Get the next file
					err = VFSDirEntryEnumerate (dirRef, &dirIterator,&info);
					if (err == errNone) 
					{
						// Do something with the directory entry information
						// Pull the attributes from info.attributes
						// The file name is in fileName
						char ext[5];
						char *qext,*pext;
						int add_entry=1;
						qext = info.nameP;
						pext = NULL;
						while (qext = StrChr(qext, '.'))
						{
							pext = qext;
							qext++;
						}
														
						if (pext) {
							StrNCopy(ext, pext, sizeof(ext));
						}
						else ext[0]=0;							
						if ( (StrCaselessCompare(ext, ".zip") != 0)&&(StrCaselessCompare(ext, ".pce") != 0)&&(StrCaselessCompare(ext, ".gsx") != 0))
						{
							add_entry=0;
						}
						if (add_entry)
						{								
							g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
							StrCopy(g_listchoice[cur_entry],"_T-");
							g_listchoice[cur_entry][0]=volRefNum;
							StrCat(g_listchoice[cur_entry],tmpName);
							total_size+=StrLen(tmpName)+1+3;
							cur_entry++;
							if (cur_entry>=g_num_entries) break;
							info.nameP = tmpName;
							info.nameBufLen = 256;
						}
					}
					else break;
				} 
			}
			else 
			{
				// handle error, possibly by breaking out of the loop
				break;
			}
			VFSFileClose (dirRef);
		} 
		else 
		{
			// handle directory open error here
			break;
		}
		if (cur_entry>=g_num_entries) break;
	}		
	}	
		
	
	
#ifdef TESTING_ONLY	
	g_listchoice[cur_entry] = (char*)MemPtrNew(StrLen(tmpName)+1+3);
	StrCopy(g_listchoice[cur_entry],"_N-");
	g_listchoice[cur_entry][0]=0;
	StrCat(g_listchoice[cur_entry],"testyoN");
	total_size+=StrLen(tmpName)+1+3;
	cur_entry++;
	
#endif

    if (cur_entry==g_num_entries)
    {
    	char str[128];
    	StrPrintF(str,"Cannot show more than %d files",g_num_entries);
    	menu_message(str);menu_waitBtn(1,1);

    }

			
	if (cur_entry)
	{
		signed short selindex=-1;
		Sort_Entries(g_listchoice,cur_entry);
		
		for (int i=0;i<cur_entry;i++)
		if (StrCompare(g_ljz_pref.g_lastrom,g_listchoice[i])==0)
		{selindex=i;break;}
				
		LstSetListChoices(list,(char**)NULL,cur_entry);
		
		if (selindex>=0) 
		{
			LstMakeItemVisible(list,selindex);
			LstSetSelection(list,selindex);
		}
		else
		{
			LstMakeItemVisible(list,0);
			LstSetSelection(list,0);
		}
		g_num_entries=cur_entry;
	}
	else
	{		
		MemPtrFree(g_listchoice);
		g_listchoice=NULL;
		LstSetListChoices(list,(char**)NULL,0);		
		switch (_system2list)
		{
			case SYS_NES:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__NES_ROMS_DIR__);
				menu_waitBtn(1,1);
				break;
			case SYS_SNES:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__SNES_ROMS_DIR__);
				menu_waitBtn(1,1);
				break;
			case SYS_GB:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__GB_ROMS_DIR__);
				menu_waitBtn(1,1);
				break;
			case SYS_GENZ:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__GENZ_ROMS_DIR__);
				menu_waitBtn(1,1);
				break;
			case SYS_NEOPOPZ:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__NEOPOPZ_ROMS_DIR__);
				menu_waitBtn(1,1);				
				break;
			case SYS_OSWANZ:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__OSWANZ_ROMS_DIR__);
				menu_waitBtn(1,1);				
				break;
			case SYS_TG16Z:
				menu_messageXY(220-100,100,200,40,"Put ROMS in : ",0x0AC00A);		
				menu_message(__TG16Z_ROMS_DIR__);
				menu_waitBtn(1,1);				
				break;
		}
		
	}
}

/***********************************************************************
 *
 * FUNCTION:    ConfigFrmHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "ConfigForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
static void ConfigUpdateControls(int *config_mode,FormType *frmP,int only_set)
{	
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_GENZ:
		case SYS_SNES:	
		case SYS_NEOPOPZ:
		break;
		case SYS_NES:
		case SYS_GB:
		case SYS_TG16Z:
			if (*config_mode>2) *config_mode=2;
		break;
		case SYS_OSWANZ:
			if (*config_mode>1) *config_mode=1;
		break;
	}

	ListType *list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOZoomList) );
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_SNES:LstSetSelection(list,g_ljz_pref.g_SnesScreenMode);
		break;
		case SYS_NES:LstSetSelection(list,g_ljz_pref.g_NesScreenMode);
		break;
		case SYS_GB:LstSetSelection(list,g_ljz_pref.g_GBScreenMode);
		break;
		case SYS_GENZ:LstSetSelection(list,g_ljz_pref.g_GENZScreenMode);
		break;
		case SYS_NEOPOPZ:LstSetSelection(list,g_ljz_pref.g_NEOPOPZScreenMode);
		break;
		case SYS_OSWANZ:LstSetSelection(list,g_ljz_pref.g_OSWANZScreenMode);
		break;
		case SYS_TG16Z:LstSetSelection(list,g_ljz_pref.g_TG16ZScreenMode);
		break;
	}
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOZoomPopup) ),LstGetSelectionText(list,LstGetSelection(list)));
	
	list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipList) );
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_SNES:LstSetSelection(list,g_ljz_pref.g_SnesFSkip);
		break;
		case SYS_NES:LstSetSelection(list,g_ljz_pref.g_NesFSkip);
		break;
		case SYS_GB:LstSetSelection(list,g_ljz_pref.g_GBFSkip);
		break;
		case SYS_GENZ:LstSetSelection(list,g_ljz_pref.g_GENZFSkip);
		break;
		case SYS_NEOPOPZ:LstSetSelection(list,g_ljz_pref.g_NEOPOPZFSkip);
		break;
		case SYS_OSWANZ:LstSetSelection(list,g_ljz_pref.g_OSWANZFSkip);
		break;
		case SYS_TG16Z:LstSetSelection(list,g_ljz_pref.g_TG16ZFSkip);
		break;
	}
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipPopup) ),LstGetSelectionText(list,LstGetSelection(list)));
	
	list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqList) );
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_SNES:LstSetSelection(list,g_ljz_pref.g_SnesSoundFreq);
		break;
		case SYS_NES:LstSetSelection(list,g_ljz_pref.g_NesSoundFreq);
		break;
		case SYS_GB:LstSetSelection(list,g_ljz_pref.g_GBSoundFreq);
		break;
		case SYS_GENZ:LstSetSelection(list,g_ljz_pref.g_GENZSoundFreq);
		break;
		case SYS_NEOPOPZ:LstSetSelection(list,g_ljz_pref.g_NEOPOPZSoundFreq);
		break;
		case SYS_OSWANZ:LstSetSelection(list,g_ljz_pref.g_OSWANZSoundFreq);
		break;
		case SYS_TG16Z:LstSetSelection(list,g_ljz_pref.g_TG16ZSoundFreq);
		break;
	}
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqPopup) ),LstGetSelectionText(list,LstGetSelection(list)));
	
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_SNES:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_SnesSmoothing);
		
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_SnesSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_SnesSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_SnesSound==2));				
		

		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmASMCPUCORE),(g_ljz_pref.g_SnesASMCPU==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmCCPUCORE),(g_ljz_pref.g_SnesASMCPU==0));
		
		
		break;
		case SYS_NES:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_NesSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_NesSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_NesSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_NesSound==2));
		break;
		case SYS_GB:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_GBSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_GBSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_GBSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_GBSound==2));
		break;		
		case SYS_GENZ:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_GENZSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),!(g_ljz_pref.g_GENZSound&3));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_GENZSound&3)==1);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_GENZSound&3)==2);
		
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDZX80Checkbox),(g_ljz_pref.g_GENZSound&4));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFASTCheckbox),(g_ljz_pref.g_GENZSound&8));
		
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmASMCPUCORE),(g_ljz_pref.g_GENZASMCPU==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmCCPUCORE),(g_ljz_pref.g_GENZASMCPU==0));
		
		break;
		case SYS_NEOPOPZ:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_NEOPOPZSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_NEOPOPZSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_NEOPOPZSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_NEOPOPZSound==2));
		
		
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmCOMPEMUACC),(g_ljz_pref.g_NEOPOPZEmuAccuracy==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmMEDEMUACC),(g_ljz_pref.g_NEOPOPZEmuAccuracy==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmFASTEMUACC),(g_ljz_pref.g_NEOPOPZEmuAccuracy==2));
		break;
		case SYS_OSWANZ:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_OSWANZSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_OSWANZSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_OSWANZSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_OSWANZSound==2));
		break;
		case SYS_TG16Z:
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox),g_ljz_pref.g_TG16ZSmoothing);
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF),(g_ljz_pref.g_TG16ZSound==0));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM),(g_ljz_pref.g_TG16ZSound==1));
		FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON),(g_ljz_pref.g_TG16ZSound==2));
		break;
	}
	
	FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESFastSpritesCheckbox),g_ljz_pref.g_SnesFastSprite);
	FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESTranspCheckbox),g_ljz_pref.g_SnesTransparency);
	list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderList) );
	LstSetSelection(list,g_ljz_pref.g_SnesRender);
	CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderPopup) ),LstGetSelectionText(list,LstGetSelection(list)));	
	
	switch (*config_mode)
		{
			case 0://input
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),1);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton),0);			
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpVIDEOButton),0);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton),0);
				
				int i,j;
				switch (g_ljz_pref.g_conf_system)
				{	
					case SYS_SNES:j=0;break;
					case SYS_NES:j=1;break;
					case SYS_GB:j=2;break;
					case SYS_GENZ:j=3;break;
					case SYS_NEOPOPZ:j=4;break;
					case SYS_OSWANZ:j=5;break;
					case SYS_TG16Z:j=6;break;
				}	
				g_inputfld[0]=0;			
				for (i=0;i<20;i++)
				{					
					StrCat(g_inputfld,g_labelbutton[g_ljz_pref.g_inputconf[j][i]]);
					StrCat(g_inputfld,"\n");
				}
				
				FieldType *field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
				FldSetTextPtr(field,g_inputfld);
				FldRecalculateField (field,1);
				
			break;
			case 2://audio
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),0);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton),1);			
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpVIDEOButton),0);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton),0);
				
			break;
			case 1://video
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),0);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton),0);			
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpVIDEOButton),1);					
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton),0);
			break;
			case 3://other
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),0);
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton),0);			
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpVIDEOButton),0);					
				FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton),1);
			break;
		}
						
	if (only_set) return;
	
	switch (g_ljz_pref.g_conf_system)
	{
		case SYS_GENZ:
		case SYS_SNES:
		case SYS_NEOPOPZ:
			CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton) ));
			CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton) ));
		break;
		case SYS_NES:
		case SYS_GB:				
		case SYS_TG16Z:
			CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton) ));
			CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton) ));
		break;
		case SYS_OSWANZ:
			CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton) ));			
			CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton) ));
		break;
	}

	
	
	/*CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTSNESList) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNESList) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGBList) ));*/
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGBList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGENZList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNESList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTSNESList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNGPList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTWSList) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTTG16ZList) );

	
	
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmASSIGNButton) ));
	
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESFastSpritesCheckbox) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESTranspCheckbox) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderPopup) ));
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEODispsizeLabel));
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipLabel) );
	
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOZoomPopup) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipPopup) ));		
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderLabel) );
	
	
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqPopup) ));		
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDZX80Checkbox) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFASTCheckbox) ));
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDLabel) );
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqLabel) );


	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmCPUCoreLabel));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmASMCPUCORE) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmCCPUCORE) ));	
	
	FrmHideObject(frmP,FrmGetObjectIndex(frmP,ConfFrmEMUACCLabel));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmCOMPEMUACC) ));
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmMEDEMUACC) ));	
	CtlHideControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmFASTEMUACC) ));	
	
	
	
	switch (*config_mode)
		{
			case 0://input
				CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmASSIGNButton) ));				
				switch (g_ljz_pref.g_conf_system)
				{
					case SYS_SNES:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTSNESList) );break;
					case SYS_NES:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNESList) );break;
					case SYS_GB:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGBList) );break;
					case SYS_GENZ:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGENZList) );break;
					case SYS_NEOPOPZ:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNGPList) );break;
					case SYS_OSWANZ:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTWSList) );break;
					case SYS_TG16Z:FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTTG16ZList) );break;
				}				
				FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );

			break;
			case 2://audio
				
				switch (g_ljz_pref.g_conf_system)
				{					
					case SYS_GENZ:
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDZX80Checkbox) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFASTCheckbox) ));
					case SYS_SNES:
					case SYS_NES:
					case SYS_GB:
					case SYS_TG16Z:
					case SYS_NEOPOPZ:
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDON) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDSIM) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDOFF) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqPopup) ));		
						FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDLabel) );
						FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFreqLabel) );
						break;
				}
				
			break;
			case 1://video
				CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox) ));
				CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOZoomPopup) ));
				CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipPopup) ));				
				FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEODispsizeLabel));
				FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOFSkipLabel) );
				if (g_ljz_pref.g_conf_system==SYS_SNES)
				{		
					FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderLabel) );
					CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESFastSpritesCheckbox) ));
					CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESTranspCheckbox) ));
					CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESRenderPopup) ));
				}
			break;
			case 3://other
			switch (g_ljz_pref.g_conf_system)
				{					
					case SYS_SNES:
					case SYS_GENZ:
						FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmCPUCoreLabel));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmASMCPUCORE) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmCCPUCORE) ));	
					break;
					case SYS_NEOPOPZ:
						FrmShowObject(frmP,FrmGetObjectIndex(frmP,ConfFrmEMUACCLabel));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmCOMPEMUACC) ));
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmMEDEMUACC) ));	
						CtlShowControl((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmFASTEMUACC) ));	
					break;
					
				}
			break;
		}
		
	
}

static Boolean ConfigFrmHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err				err = errNone;
//	static int		frmloaded=0;
	

	switch (eventP->eType) 
	{
		case frmOpenEvent:
		{			
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),1);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpAUDIOButton),0);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpVIDEOButton),0);			
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpINPUTButton),0);
			FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmGrpOTHERButton),0);
			
			ListType *list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemListConf) );
			switch (g_ljz_pref.g_conf_system)
			{
				case SYS_SNES:LstSetSelection(list,0);
				break;
				case SYS_NES:LstSetSelection(list,1);
				break;
				case SYS_GB:LstSetSelection(list,2);
				break;
				case SYS_GENZ:LstSetSelection(list,3);
				break;
				case SYS_NEOPOPZ:LstSetSelection(list,4);
				break;
				case SYS_OSWANZ:LstSetSelection(list,5);
				break;
				case SYS_TG16Z:LstSetSelection(list,6);
				break;
			}
			CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemPopupConf) ),LstGetSelectionText(list,LstGetSelection(list)));

			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			DRAW_SYSTEM_ICON(g_ljz_pref.g_conf_system,10,120)
			if (scr) WinScreenUnlock();
			}
			ConfigUpdateControls(&g_config_mode,frmP,0);			
			
						
			ASK_FOR_UPDATE	
			handled = true;
			break;
		}
		case frmCloseEvent:							
		{	
			
			break;
		}		
		case popSelectEvent:
			switch ( eventP->data.ctlSelect.controlID ) 
			{
				case SystemPopupConf:
				{
					switch (eventP->data.popSelect.selection)
					{
						case 0:g_ljz_pref.g_conf_system=SYS_SNES;
						break;
						case 1:g_ljz_pref.g_conf_system=SYS_NES;break;
						case 2:g_ljz_pref.g_conf_system=SYS_GB;break;
						case 3:g_ljz_pref.g_conf_system=SYS_GENZ;
						break;
						case 4:g_ljz_pref.g_conf_system=SYS_NEOPOPZ;
						break;
						case 5:g_ljz_pref.g_conf_system=SYS_OSWANZ;						
						break;
						case 6:g_ljz_pref.g_conf_system=SYS_TG16Z;break;
					}
					
					ListType *list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemListConf) );
					if (list) CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemPopupConf) ),LstGetSelectionText(list,LstGetSelection(list)));
					
					UPDATE_UI_COLORS(g_ljz_pref.g_conf_system)
					
					
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					break;
				}								
				/*VIDEO*/
				case ConfFrmVIDEOZoomPopup:
				{
					int val=eventP->data.popSelect.selection;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesScreenMode=val;break;
						case SYS_NES:g_ljz_pref.g_NesScreenMode=val;break;
						case SYS_GB:g_ljz_pref.g_GBScreenMode=val;break;
						case SYS_GENZ:g_ljz_pref.g_GENZScreenMode=val;break;
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZScreenMode=val;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZScreenMode=val;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZScreenMode=val;break;
					}
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					break;
				}
				case ConfFrmVIDEOFSkipPopup:
				{
					int val=eventP->data.popSelect.selection;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesFSkip=val;break;
						case SYS_NES:g_ljz_pref.g_NesFSkip=val;
						if (val==11) {menu_message("Not implemented, same as 0");menu_waitBtn(1,1);}
						break;
						case SYS_GB:g_ljz_pref.g_GBFSkip=val;
						if (val==11) {menu_message("Not implemented, same as 0");menu_waitBtn(1,1);}
						break;
						case SYS_GENZ:g_ljz_pref.g_GENZFSkip=val;
						break;
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZFSkip=val;
						if (val==11) {menu_message("Not implemented, same as 0");menu_waitBtn(1,1);}
						break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZFSkip=val;
						if (val==11) {menu_message("Not implemented, same as 0");menu_waitBtn(1,1);}
						break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZFSkip=val;break;
					}
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					break;
				}
				case ConfFrmVIDEOSNESRenderPopup:
				{
					g_ljz_pref.g_SnesRender=eventP->data.popSelect.selection;
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					break;
				}
				/*SOUND*/
				case ConfFrmSOUNDFreqPopup:
				{
					int val=eventP->data.popSelect.selection;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesSoundFreq=val;break;
						case SYS_NES:g_ljz_pref.g_NesSoundFreq=val;break;
						case SYS_GB:g_ljz_pref.g_GBSoundFreq=val;break;
						case SYS_GENZ:g_ljz_pref.g_GENZSoundFreq=val;break;
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZSoundFreq=val;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZSoundFreq=val;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZSoundFreq=val;break;
					}
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					break;
				}
			}
			break;
		case ctlSelectEvent:
		{
			switch ( eventP->data.ctlSelect.controlID ) 
			{			
				case ConfFrmHELPButton:
					g_prev_form=ConfigForm;
					//g_frmloaded=0;
					FrmGotoForm(HelpForm);										
					handled=true;
					break;
				case ConfFrmHOMEButton:
					//g_frmloaded=0;
					FrmGotoForm(LauncherForm);
					handled=true;
					break;
				case ConfFrmGrpINPUTButton:
					g_config_mode=0;
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					handled=true;
					break;							
				case ConfFrmGrpAUDIOButton:					
					g_config_mode=2;
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					handled=true;
					break;	
				case ConfFrmGrpVIDEOButton:
					g_config_mode=1;
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					handled=true;
					break;
				case ConfFrmGrpOTHERButton:
					g_config_mode=3;
					ConfigUpdateControls(&g_config_mode,frmP,0);
					ASK_FOR_UPDATE
					handled=true;
					break;
		/*INPUT*/
				case ConfFrmASSIGNButton:
				{
					ListType *list;
					int i,j;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:j=0;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTSNESList) ); break;
						case SYS_NES:j=1;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNESList) ); break;
						case SYS_GB:j=2;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGBList) ); break;
						case SYS_GENZ:j=3;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGENZList) ); break;
						case SYS_NEOPOPZ:j=4;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNGPList) ); break;
						case SYS_OSWANZ:j=5;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTWSList) ); break;
						case SYS_TG16Z:j=6;list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTTG16ZList) ); break;
					}	
					i=LstGetSelection(list);
					menu_messageXY(0,0,120,80,"TL",0x0AC00A);
					menu_messageXY(326,0,120,80,"TR",0x0AC00A);
					menu_messageXY(0,236,120,80,"BL",0x0AC00A);
					menu_messageXY(326,236,120,80,"BR",0x0AC00A);
					menu_messageXY(223-80-40,180,80,30,"Cancel",0x80600A);
					menu_messageXY(223+80-40,180,80,30,"None",0x80600A);
					menu_message("Press a button or hit screen");
					int k=menu_waitBtn(1,0);
					if (k>=0) 
					{	
						if (k)
							for (int l=0;l<20;l++) 
								if (g_ljz_pref.g_inputconf[j][l]==k) g_ljz_pref.g_inputconf[j][l]=0;
						g_ljz_pref.g_inputconf[j][i]=k;
						ConfigUpdateControls(&g_config_mode,frmP,0);
					}
					
					FieldType *field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
					FldSetScrollPosition(field,0);				
					FldScrollField(field,LstGetTopItem(list),winDown);		
					ASK_FOR_UPDATE
				}
		/*OTHER*/
				case ConfFrmASMCPUCORE:
				{
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesASMCPU=1;break;
						case SYS_GENZ:g_ljz_pref.g_GENZASMCPU=1;break;
					}
					handled=true;
					break;
				}
				case ConfFrmCCPUCORE:
				{
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesASMCPU=0;break;
						case SYS_GENZ:g_ljz_pref.g_GENZASMCPU=0;break;
					}
					handled=true;
					break;
				}
				case ConfFrmCOMPEMUACC:
				{
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZEmuAccuracy=0;break;
					}
					handled=true;
					break;
				}
				case ConfFrmMEDEMUACC:
				{
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZEmuAccuracy=1;break;
					}
					handled=true;
					break;
				}
				case ConfFrmFASTEMUACC:
				{
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZEmuAccuracy=2;break;
					}
					handled=true;
					break;
				}
		/*VIDEO*/															
				case ConfFrmVIDEOSmoothingCheckbox:
				{
					int val=FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSmoothingCheckbox));
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesSmoothing=val;break;
						case SYS_NES:g_ljz_pref.g_NesSmoothing=val;break;
						case SYS_GB:g_ljz_pref.g_GBSmoothing=val;break;
						case SYS_GENZ:g_ljz_pref.g_GENZSmoothing=val;break;
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZSmoothing=val;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZSmoothing=val;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZSmoothing=val;break;
					}
					handled = true;
					break;
				}
				case ConfFrmVIDEOSNESTranspCheckbox:
					g_ljz_pref.g_SnesTransparency=FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESTranspCheckbox));
					handled = true;
					break;
				case ConfFrmVIDEOSNESFastSpritesCheckbox:
					g_ljz_pref.g_SnesFastSprite=FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmVIDEOSNESFastSpritesCheckbox));
					handled = true;
					break;
			/*SOUND*/
				case ConfFrmSOUNDOFF:
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesSound=0;break;
						case SYS_NES:g_ljz_pref.g_NesSound=0;break;
						case SYS_GB:g_ljz_pref.g_GBSound=0;break;						
						case SYS_GENZ:g_ljz_pref.g_GENZSound=0;
						ConfigUpdateControls(&g_config_mode,frmP,0);
						ASK_FOR_UPDATE
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZSound=0;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZSound=0;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZSound=0;break;
					}
					handled = true;
					break;
				case ConfFrmSOUNDSIM:
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesSound=1;break;
						case SYS_NES:g_ljz_pref.g_NesSound=1;break;
						case SYS_GB:g_ljz_pref.g_GBSound=1;break;
						case SYS_GENZ:g_ljz_pref.g_GENZSound&=~3;g_ljz_pref.g_GENZSound|=1|4;
						ConfigUpdateControls(&g_config_mode,frmP,0);
						ASK_FOR_UPDATE
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZSound=1;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZSound=1;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZSound=1;break;
					}
					handled = true;
					break;

				case ConfFrmSOUNDON:
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_SNES:g_ljz_pref.g_SnesSound=2;break;
						case SYS_NES:g_ljz_pref.g_NesSound=2;break;
						case SYS_GB:g_ljz_pref.g_GBSound=2;break;
						case SYS_GENZ:g_ljz_pref.g_GENZSound&=~3;g_ljz_pref.g_GENZSound|=2;break;
						case SYS_NEOPOPZ:g_ljz_pref.g_NEOPOPZSound=2;break;
						case SYS_OSWANZ:g_ljz_pref.g_OSWANZSound=2;break;
						case SYS_TG16Z:g_ljz_pref.g_TG16ZSound=2;break;
					}
					handled = true;
					break;
				case ConfFrmSOUNDZX80Checkbox:
					handled = true;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_GENZ:
							if (!g_ljz_pref.g_GENZSound) g_ljz_pref.g_GENZSound|=1;//if sound off,the put in sim
							g_ljz_pref.g_GENZSound&=~4;g_ljz_pref.g_GENZSound|=(FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDZX80Checkbox))?4:0);
							ConfigUpdateControls(&g_config_mode,frmP,0);
							ASK_FOR_UPDATE
							break;
					}
					break;
				case ConfFrmSOUNDFASTCheckbox:
					handled = true;
					switch (g_ljz_pref.g_conf_system)
					{
						case SYS_GENZ:
							//force sound on
							g_ljz_pref.g_GENZSound&=~(8|3);g_ljz_pref.g_GENZSound|=2|(FrmGetControlValue (frmP,FrmGetObjectIndex(frmP,ConfFrmSOUNDFASTCheckbox))?8:0);
							ConfigUpdateControls(&g_config_mode,frmP,0);
							ASK_FOR_UPDATE
							break;
					}
					break;
					
					
			}	
			break;			
		}
		case penUpEvent:
		{
			if (g_config_mode==0)
			{			
				FieldType *field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTButtonField) );
				FldSetScrollPosition(field,0);
				ListType *list;
				switch (g_ljz_pref.g_conf_system)
				{
					case SYS_SNES:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTSNESList) ); break;
					case SYS_NES:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNESList) ); break;
					case SYS_GB:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGBList) ); break;;
					case SYS_GENZ:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTGENZList) ); break;
					case SYS_NEOPOPZ:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTNGPList) ); break;
					case SYS_OSWANZ:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTWSList) ); break;
					case SYS_TG16Z:list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,ConfFrmINPUTTG16ZList) ); break;;
				}
				FldScrollField(field,LstGetTopItem(list),winDown);		
				ASK_FOR_UPDATE
			}
			break;
		}
		
		case winEnterEvent:	
		case winDisplayChangedEvent:	
			SysSetOrientation(sysOrientationLandscape);
			PINSetInputAreaState(pinInputAreaClosed);
		case frmUpdateEvent:		
			/**/						
			//ConfigUpdateControls(&g_config_mode,frmP,0);
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			DRAW_SYSTEM_ICON(g_ljz_pref.g_conf_system,10,120)
			if (scr) WinScreenUnlock();
			}			
			handled=true;
			break;
		
	}
    
	return handled;
}



void	drawList	(Int16	i,RectangleType	*bounds,Char		**items)
	{
	if (i>=g_num_entries) return;
	
	Char *p=g_listchoice[i]+3;
	Int16 j=0;
	Int16 wid;
	Boolean notruncated;
	IndexedColorType selectColor = UIColorGetTableEntryIndex(UIObjectSelectedForeground);
	IndexedColorType myColor=0;
	IndexedColorType itemColor = WinSetForeColor(myColor);
	WinSetForeColor(itemColor);
	

	
	while (p[j++]);
	RGBColorType fg,bg;	
	WinPushDrawState();	
    WinSetDrawMode(winPaint);
    
	j--;    
	
	switch (g_listchoice[i][0])
	{
		case 3: //SLOT 2
			fg.r=100;fg.g=50;fg.b=200;
		break;
		case 2: //SLOT 1
			fg.r=120;fg.g=140;fg.b=40;
		break;
		case 1:
		default:  //SLOT 0 (internal)
			fg.r=50;fg.g=150;fg.b=70;
		break;
	}
	
	bg.r=255;bg.g=255;bg.b=255;       
	
    if (((p[j-1]=='P')||(p[j-1]=='p'))&&((p[j-2]=='I')||(p[j-2]=='i'))&&((p[j-3]=='Z')||(p[j-3]=='z')))
    {
	    //FntSetFont(boldFont);
	}
	else
	{
	    UInt16 cr=(UInt16)(fg.r)*4/3;if (cr>255) cr=255;
	    UInt16 cg=(UInt16)(fg.g)*4/3;if (cg>255) cg=255;
	    UInt16 cb=(UInt16)(fg.b)*4/3;if (cb>255) cb=255;
	    fg.r=cr;fg.b=cb;fg.b=cb;
//		fg.r<<=1;fg.g<<=1;fg.b<<=1;
	}   
	
	if (itemColor == selectColor)
		{fg.r^=255;fg.g^=255;fg.b^=255;
		bg.r^=255;bg.g^=255;bg.b^=255;}
	        	
    WinSetTextColorRGB(&fg,NULL);    
//    WinSetBackColorRGB(&bg,NULL);


	wid=155;
	FntCharsInWidth(p,&wid,&j,&notruncated);
	if (!notruncated)
	{
		p=(Char*)MemPtrNew(j);
		MemMove(p,g_listchoice[i]+3,j);
		p[j]=0;
		p[j-1]='.';p[j-2]='.';p[j-3]='.';
	}
	
	
	WinSetCoordinateSystem(kCoordinatesNative);
	
	WinPaintChars(p,j,bounds->topLeft.x*2,
			bounds->topLeft.y*2);
			
	WinSetDrawMode(winOverlay);
	
	if (itemColor == selectColor)
	{
	fg.r<<=1;fg.g<<=1;fg.b<<=1;
    WinSetTextColorRGB(&fg,NULL);
	WinPaintChars(p,j,bounds->topLeft.x*2-1,
			bounds->topLeft.y*2-1);
	}		
	if (!notruncated) MemPtrFree(p);
	WinPopDrawState();	
	}

/***********************************************************************
 *
 * FUNCTION:    LauncherFrmHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "LauncherForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
static Boolean LauncherFrmHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err				err = errNone;


	switch (eventP->eType) 
	{
		case frmOpenEvent:
		{														
			UPDATE_UI_COLORS(g_ljz_pref.g_sysfilter)
			
			ResetQuickRomLabel(frmP);
			g_lastQuickRomChar='0';
			CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,LauncherFrm0BButton) ),"#ab");
		
			// Setup the form state
			WinSetConstraintsSize(FrmGetWindowHandle(frmP), 160, 160, 240, 160, 240, 240);
			FrmSetDIAPolicyAttr( frmP, frmDIAPolicyCustom );			
			// Check to make sure that the device we're running on supports
			// version 1.1 of the Pin Input Manager before making our calls.
			err = FtrGet( pinCreator, pinFtrAPIVersion, &pinMgrVersion );
			if (!err && pinMgrVersion == pinAPIVersion1_1)
			{
				flags = FtrGet( sysFtrCreator, sysFtrNumInputAreaFlags, &flags );								
				if (flags & grfFtrInputAreaFlagCollapsible)
					PINSetInputAreaState(pinInputAreaUser);
				if (flags & grfFtrInputAreaFlagCollapsible)
					StatShow();
				if (flags & grfFtrInputAreaFlagCollapsible)
					PINSetInputTriggerState(pinInputTriggerEnabled);								
				SysSetOrientationTriggerState(sysOrientationTriggerEnabled);									
				SysSetOrientation(/*sysOrientationUser*/sysOrientationLandscape);
			}	
			StatShow();
			PINSetInputAreaState(pinInputAreaClosed);
			
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			DRAW_SYSTEM_ICON(g_ljz_pref.g_sysfilter,10,120)
			if (scr) WinScreenUnlock();
			}			
						
			if (!g_ljz_pref.g_firsttime)
			{
				g_ljz_pref.g_firsttime=1;				
				menu_message(WELCOME_MESSAGE);
				menu_waitBtn(1,1);			
			}
			
			LstNewList((void**)&frmP,FORM_LISTROM_ID,60,25,161,120,FntGetFont(),11,0);
			ListType *list=(ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );			
			LstSetDrawFunction (list, (ListDrawDataFuncPtr) drawList);
			
			
			InitRomList(list,g_ljz_pref.g_sysfilter);			
			
						
			list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemList) );
			switch (g_ljz_pref.g_sysfilter)
			{
				/*case (SYS_NES|SYS_SNES|SYS_GB|SYS_GENZ):LstSetSelection(list,0);
				break;*/
				case SYS_NES:LstSetSelection(list,0);g_ljz_pref.g_conf_system=SYS_NES;
				break;
				case SYS_SNES:LstSetSelection(list,1);g_ljz_pref.g_conf_system=SYS_SNES;
				break;
				case SYS_GB:LstSetSelection(list,2);g_ljz_pref.g_conf_system=SYS_GB;
				break;
				case SYS_GENZ:LstSetSelection(list,3);g_ljz_pref.g_conf_system=SYS_GENZ;
				break;
				case SYS_NEOPOPZ:LstSetSelection(list,4);g_ljz_pref.g_conf_system=SYS_NEOPOPZ;
				break;
				case SYS_OSWANZ:LstSetSelection(list,5);g_ljz_pref.g_conf_system=SYS_OSWANZ;
				break;
				case SYS_TG16Z:LstSetSelection(list,6);g_ljz_pref.g_conf_system=SYS_TG16Z;
				break;
			}
			CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemPopup) ),LstGetSelectionText(list,LstGetSelection(list)));
			
			
						
			ASK_FOR_UPDATE			
			handled = true;
			break;
		}
		case frmCloseEvent:							
		{
			ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );	
			if (list) 
			{
				LstEraseList(list);
				LstSetListChoices(list,(char**)NULL,0);										
				if (g_listchoice) 
				{
					for (int i=0;i<g_num_entries;i++)
						MemPtrFree(g_listchoice[i]);
					MemPtrFree(g_listchoice);
					g_listchoice=NULL;
				}
			}													
			break;
		}
		case keyDownEvent:
			switch (eventP->data.keyDown.chr)
			{
				case vchrNavDown:
				{
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=LstGetSelection(list);
					if (currentItem<LstGetNumberOfItems(list)) LstSetSelection(list,currentItem+1);
					handled = true;					
				}
				break;
				case vchrNavUp:
				{
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=LstGetSelection(list);
					if (currentItem>0) LstSetSelection(list,currentItem-1);
					handled = true;					
				}
				break;
				case vchrNavSelect:
				{
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					if (LstGetSelection(list)<g_num_entries)
					{
						char *filename=g_listchoice[LstGetSelection(list)];
						char *ftrptr;
						if (filename)
						{
							FtrPtrNew(APPLI_ID, FTR_ROM_FILENAME_ID , 256, (void **)&ftrptr);
							DmStrCopy(ftrptr,0,filename);					
							g_rom_system=0;
							if (filename[1]=='N') g_rom_system=SYS_NES;
							if (filename[1]=='S') g_rom_system=SYS_SNES;
							if (filename[1]=='G') g_rom_system=SYS_GB;
							if (filename[1]=='M') g_rom_system=SYS_GENZ;
							if (filename[1]=='P') g_rom_system=SYS_NEOPOPZ;
							if (filename[1]=='W') g_rom_system=SYS_OSWANZ;
							if (filename[1]=='T') g_rom_system=SYS_TG16Z;
//							g_launch_emu=1;							
							
							WinPushDrawState();					
					RGBColorType col;
					col.r=0;col.g=0;col.b=0;
					UIColorSetTableEntry(UIFormFill,&col);
					UIColorSetTableEntry(UIObjectFill,&col);			
					col.r=0;col.g=0;col.b=0;
					UIColorSetTableEntry(UIFormFrame,&col);
					UIColorSetTableEntry(UIObjectFrame,&col);			
#ifndef TESTING_ONLY					
  					locked_scr=DoNiceTransition(1);
#endif  										
  					//FrmCloseAllForms();  					
  					//FrmEraseForm(FrmGetActiveForm());
  					FrmGotoForm(BlankForm);
  					
  					WinPopDrawState(); 
						}
					}
					handled = true;					
				}
				break;
			}
			break;

		case popSelectEvent:
			switch ( eventP->data.ctlSelect.controlID ) 
			{
				case SystemPopup:
				{
					int oldsys=g_ljz_pref.g_sysfilter;
					switch (eventP->data.popSelect.selection)
					{
						//case 0:g_ljz_pref.g_sysfilter=SYS_NES|SYS_SNES|SYS_GB|SYS_GENZ;break;
						case 0:g_ljz_pref.g_sysfilter=SYS_NES;g_ljz_pref.g_conf_system=SYS_NES;break;
						case 1:g_ljz_pref.g_sysfilter=SYS_SNES;g_ljz_pref.g_conf_system=SYS_SNES;break;
						case 2:g_ljz_pref.g_sysfilter=SYS_GB;g_ljz_pref.g_conf_system=SYS_GB;break;
						case 3:g_ljz_pref.g_sysfilter=SYS_GENZ;g_ljz_pref.g_conf_system=SYS_GENZ;break;
						case 4:g_ljz_pref.g_sysfilter=SYS_NEOPOPZ;g_ljz_pref.g_conf_system=SYS_NEOPOPZ;break;
						case 5:g_ljz_pref.g_sysfilter=SYS_OSWANZ;g_ljz_pref.g_conf_system=SYS_OSWANZ;break;
						case 6:g_ljz_pref.g_sysfilter=SYS_TG16Z;g_ljz_pref.g_conf_system=SYS_TG16Z;break;
					}
					
					if (oldsys!=g_ljz_pref.g_sysfilter)
					{
						Stop_Midi();
						int i;
						switch (g_ljz_pref.g_sysfilter)
						{
							case SYS_NES:i=0;break;
							case SYS_SNES:i=1;break;
							case SYS_GB:i=2;break;
							case SYS_GENZ:i=3;break;
							case SYS_NEOPOPZ:i=4;break;
							case SYS_OSWANZ:i=5;break;
							case SYS_TG16Z:i=6;break;
						}
						Start_Midi(i);
					}
										
					UPDATE_UI_COLORS(g_ljz_pref.g_sysfilter)

						
					ListType *list=(ListType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemList) );
					if (list) CtlSetLabel((ControlType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SystemPopup) ),LstGetSelectionText(list,LstGetSelection(list)));
					
					list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					if (list) 
					{
						LstEraseList(list);
						LstSetListChoices(list,(char**)NULL,0);								
						if (g_listchoice) 
						{
							for (int i=0;i<g_num_entries;i++)
								MemPtrFree(g_listchoice[i]);
							MemPtrFree(g_listchoice);
							g_listchoice=NULL;
						}
						InitRomList(list,g_ljz_pref.g_sysfilter);																			
					}
					
					ASK_FOR_UPDATE
					handled = true;													
					break;
				}
			}
			break;
		case lstSelectEvent:
		{
			switch ( eventP->data.lstSelect.listID)
			{
				case FORM_LISTROM_ID:
				
				if (eventP->data.lstSelect.selection<g_num_entries)
					StrCopy(g_ljz_pref.g_lastrom,g_listchoice[eventP->data.lstSelect.selection]);
				break;
			}			
		}
		case ctlSelectEvent:
		{
			switch ( eventP->data.ctlSelect.controlID ) 
			{
				case LauncherFrmHELPButton:
					g_prev_form=LauncherForm;
					//g_frmloaded=0;
					FrmGotoForm(HelpForm);					
					handled=true;
					break;
				case LauncherFrmPLSButton:
					menu_message("Not implemented yet...");
					menu_waitBtn(1,1);
					handled=true;
					ASK_FOR_UPDATE					
					break;
				case LauncherFrmCONFButton:
					//g_frmloaded=0;
					FrmGotoForm(ConfigForm);
					handled=true;
					break;
				case LauncherFrmLAUNCHButton:
				{
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					if (LstGetSelection(list)<g_num_entries)
					{
						char *filename=g_listchoice[LstGetSelection(list)];
						char *ftrptr;
						if (filename)
						{
							FtrPtrNew(APPLI_ID, FTR_ROM_FILENAME_ID , 256, (void **)&ftrptr);
							DmStrCopy(ftrptr,0,filename);										
					
							g_rom_system=0;
							if (filename[1]=='N') g_rom_system=SYS_NES;
							if (filename[1]=='S') g_rom_system=SYS_SNES;
							if (filename[1]=='G') g_rom_system=SYS_GB;
							if (filename[1]=='M') g_rom_system=SYS_GENZ;
							if (filename[1]=='P') g_rom_system=SYS_NEOPOPZ;
							if (filename[1]=='W') g_rom_system=SYS_OSWANZ;
							if (filename[1]=='T') g_rom_system=SYS_TG16Z;
//							g_launch_emu=1;							
				WinPushDrawState();					

					RGBColorType col;
					col.r=0;col.g=0;col.b=0;
					UIColorSetTableEntry(UIFormFill,&col);
					col.r=0;col.g=0;col.b=0;
					UIColorSetTableEntry(UIFormFrame,&col);
#ifndef TESTING_ONLY					
  					locked_scr=DoNiceTransition(1);
#endif  										
  					//FrmCloseAllForms();  					
  					//FrmEraseForm(FrmGetActiveForm());
  					FrmGotoForm(BlankForm);
  					
  					WinPopDrawState(); 
						}
					}
					handled = true;
					break;
				}				
				case LauncherFrm0BButton:
				{									
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrm0BButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='0') {g_lastQuickRomChar='a';CtlSetLabel(eventP->data.ctlSelect.pControl,"_Ab");}
					else if (g_lastQuickRomChar=='a') {g_lastQuickRomChar='b';CtlSetLabel(eventP->data.ctlSelect.pControl,"_aB");}
					else {g_lastQuickRomChar='0';CtlSetLabel(eventP->data.ctlSelect.pControl,"#ab");}
					
					
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );					
					Int16 currentItem=0;
					if (!list) break;
					if (g_lastQuickRomChar!='0')
					{						
						Char p[256];
						while (1)
						{
							if (currentItem>=LstGetNumberOfItems(list)-1) break;
							StrToLower(p,g_listchoice[currentItem]);
							if (p[3]>=g_lastQuickRomChar) break;
							currentItem++;						
						}									
					}
					LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE
					break;
				}
				case LauncherFrmCFButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmCFButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='c') {g_lastQuickRomChar='d';CtlSetLabel(eventP->data.ctlSelect.pControl,"cDef");}
					else if (g_lastQuickRomChar=='d') {g_lastQuickRomChar='e';CtlSetLabel(eventP->data.ctlSelect.pControl,"cdEf");}
					else if (g_lastQuickRomChar=='e') {g_lastQuickRomChar='f';CtlSetLabel(eventP->data.ctlSelect.pControl,"cdeF");}
					else {g_lastQuickRomChar='c';CtlSetLabel(eventP->data.ctlSelect.pControl,"Cdef");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );					
					Int16 currentItem=0;
					Char p[256];
					
					if (!list) break;
					
					while (1)
					{
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;						
					}
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE					
					break;
				}
				case LauncherFrmGJButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmGJButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='g') {g_lastQuickRomChar='h';CtlSetLabel(eventP->data.ctlSelect.pControl,"gHij");}
					else if (g_lastQuickRomChar=='h') {g_lastQuickRomChar='i';CtlSetLabel(eventP->data.ctlSelect.pControl,"ghIj");}
					else if (g_lastQuickRomChar=='i') {g_lastQuickRomChar='j';CtlSetLabel(eventP->data.ctlSelect.pControl,"ghiJ");}
					else {g_lastQuickRomChar='g';CtlSetLabel(eventP->data.ctlSelect.pControl,"Ghij");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=0;
					Char p[256];
					if (!list) break;
					while (1)
					{
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;						
					}
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE
					break;
				}
				case LauncherFrmKNButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmKNButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='k') {g_lastQuickRomChar='l';CtlSetLabel(eventP->data.ctlSelect.pControl,"kLmn");}
					else if (g_lastQuickRomChar=='l') {g_lastQuickRomChar='m';CtlSetLabel(eventP->data.ctlSelect.pControl,"klMn");}
					else if (g_lastQuickRomChar=='m') {g_lastQuickRomChar='n';CtlSetLabel(eventP->data.ctlSelect.pControl,"klmN");}
					else {g_lastQuickRomChar='k';CtlSetLabel(eventP->data.ctlSelect.pControl,"Klmn");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=0;
					Char p[256];
					if (!list) break;
					while (1)
					{
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;						
					}
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE					
					break;
				}
				case LauncherFrmORButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmORButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='o') {g_lastQuickRomChar='p';CtlSetLabel(eventP->data.ctlSelect.pControl,"oPqr");}
					else if (g_lastQuickRomChar=='p') {g_lastQuickRomChar='q';CtlSetLabel(eventP->data.ctlSelect.pControl,"opQr");}
					else if (g_lastQuickRomChar=='q') {g_lastQuickRomChar='r';CtlSetLabel(eventP->data.ctlSelect.pControl,"opqR");}
					else {g_lastQuickRomChar='o';CtlSetLabel(eventP->data.ctlSelect.pControl,"Opqr");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=0;
					Char p[256];
					if (!list) break;
					while (1)
					{
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;						
					}
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE					
					break;
				}
				case LauncherFrmSVButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmSVButton),0);
					handled = true;
					
					ResetQuickRomLabel(frmP);
					if (g_lastQuickRomChar=='s') {g_lastQuickRomChar='t';CtlSetLabel(eventP->data.ctlSelect.pControl,"sTuv");}
					else if (g_lastQuickRomChar=='t') {g_lastQuickRomChar='u';CtlSetLabel(eventP->data.ctlSelect.pControl,"stUv");}
					else if (g_lastQuickRomChar=='u') {g_lastQuickRomChar='v';CtlSetLabel(eventP->data.ctlSelect.pControl,"stuV");}
					else {g_lastQuickRomChar='s';CtlSetLabel(eventP->data.ctlSelect.pControl,"Stuv");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=0;
					Char p[256];
					if (!list) break;
					while (1)
					{
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;						
					}
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE					
					break;
				}
				case LauncherFrmWZButton:
				{					
					FrmSetControlValue (frmP,FrmGetObjectIndex(frmP,LauncherFrmWZButton),0);
					handled = true;
						
					ResetQuickRomLabel(frmP);				
					if (g_lastQuickRomChar=='w') {g_lastQuickRomChar='x';CtlSetLabel(eventP->data.ctlSelect.pControl,"wXyz");}
					else if (g_lastQuickRomChar=='x') {g_lastQuickRomChar='y';CtlSetLabel(eventP->data.ctlSelect.pControl,"wxYz");}
					else if (g_lastQuickRomChar=='y') {g_lastQuickRomChar='z';CtlSetLabel(eventP->data.ctlSelect.pControl,"wxyZ");}
					else {g_lastQuickRomChar='w';CtlSetLabel(eventP->data.ctlSelect.pControl,"Wxyz");}
					
					ListType *list= (ListType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,FORM_LISTROM_ID) );
					Int16 currentItem=0;
					Char p[256];
					if (!list) break;
					while (1)
					{											
						if (currentItem>=LstGetNumberOfItems(list)-1) break;
						StrToLower(p,g_listchoice[currentItem]);
						if (p[3]>=g_lastQuickRomChar) break;
						currentItem++;
					}					
					
					/*if (currentItem<LstGetNumberOfItems(list)-1)*/ LstSetTopItem(list,currentItem);
					ASK_FOR_UPDATE					
					break;
				}
			}	
			break;			
		}
		case winEnterEvent:	
		case winDisplayChangedEvent:	
			SysSetOrientation(sysOrientationLandscape);
			PINSetInputAreaState(pinInputAreaClosed);
		case frmUpdateEvent:
		{
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			DRAW_SYSTEM_ICON(g_ljz_pref.g_sysfilter,10,120)			
			if (scr) WinScreenUnlock();
			}			
			handled=true;
		}
			break;
		
	}
    
	return handled;
}




/***********************************************************************
 *
 * FUNCTION:    HelpFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "HelpForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
static Boolean HelpFormHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err			err = errNone;

	switch (eventP->eType) 
	{
		case frmOpenEvent:
		{	
			UInt16 scrollPos,textHeight,fieldHeight,maxValue;
				
			FieldType *field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,HelpHelpTxtField) );
			FldSetTextPtr(field,g_helpmsg);
			FldRecalculateField (field,1);
//			FldMakeFullyVisible(field);			
			ScrollBarType *sb=(ScrollBarType*)FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,SBHELP) );
			 
			
			FldGetScrollValues (field, &scrollPos,&textHeight, &fieldHeight);
			if (textHeight > fieldHeight)
				maxValue = textHeight - fieldHeight;
			else if (scrollPos)
				maxValue = scrollPos;
			else
				maxValue = 0;
			SclSetScrollBar (sb, scrollPos, 0, maxValue,fieldHeight-1);
			
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			if (scr) WinScreenUnlock();
			}			
						
			handled = true;
			break;
		}
		case sclRepeatEvent:
		{
			FieldType *field=(FieldType*) FrmGetObjectPtr ( frmP,FrmGetObjectIndex(frmP,HelpHelpTxtField) );
			int line=eventP->data.sclExit.newValue-eventP->data.sclExit.value;
			if (line>0)
			{
				FldScrollField (field, line,winDown);
			}
			else
			{
				FldScrollField (field, -line,winUp);
			}
		}
			break;
		case ctlSelectEvent:
		{
			switch ( eventP->data.ctlSelect.controlID ) 
			{
				case HelpFrmHOMEButton:
					//g_frmloaded=0;
					FrmGotoForm(g_prev_form);															
					handled=true;
					break;
			}
		}
		case winEnterEvent:	
		case winDisplayChangedEvent:	
			SysSetOrientation(sysOrientationLandscape);
			PINSetInputAreaState(pinInputAreaClosed);
		case frmUpdateEvent:
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			if (scr) WinScreenUnlock();
			}			
			handled=true;
			break;
			
		
	}
    
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    BlankFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the "BlankForm" of this 
 * 				application.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should be passed to
 *				FrmHandleEvent.
 *
 ***********************************************************************/
static Boolean BlankFormHandleEvent(EventType * eventP)
{
	Boolean 		handled = false;
	FormType* 		frmP = FrmGetActiveForm();
	UInt32 			flags, pinMgrVersion;
	Err			err = errNone;

	switch (eventP->eType) 
	{
		case frmOpenEvent:
		{	
			g_launch_emu=1;			
			{//UInt8 *scr=WinScreenLock(winLockCopy);
			RGBColorType col;
			col.r=0;col.g=0;col.b=0;
			UIColorSetTableEntry(UIFormFill,&col);
			UIColorSetTableEntry(UIFormFrame,&col);
			StatHide();
			FrmDrawForm(frmP);//g_frmloaded=1;
			if (/*scr*/locked_scr) WinScreenUnlock();
			locked_scr=NULL;
			
			}									
			handled = true;
			break;
		}
		case winEnterEvent:	
		case winDisplayChangedEvent:	
			SysSetOrientation(sysOrientationLandscape);
			PINSetInputAreaState(pinInputAreaClosed);
		case frmUpdateEvent:
			{UInt8 *scr=WinScreenLock(winLockCopy);
			FrmDrawForm(frmP);//g_frmloaded=1;
			if (scr) WinScreenUnlock();
			}			
			handled=true;
			break;
			
		
	}
    
	return handled;
}



/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 * 				handler for the form loaded.
 *
 * PARAMETERS:  eventP - A pointer to the EventType structure.
 *
 * RETURNED:    true if the event was handled and should not be passed
 *				to a higher level handler.
 *
 ***********************************************************************/
static Boolean AppHandleEvent(EventType* eventP)
{
	UInt16 formId;
	FormType * frmP= FrmGetActiveForm();
	
	if (eventP->eType == frmLoadEvent)
	{
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		/* 
		 * Set the event handler for the form.  The handler of the
		 * currently active form is called by FrmHandleEvent each
		 * time is receives an event. 
		 */
		switch (formId)
		{
			case LauncherForm:
				FrmSetEventHandler( frmP, LauncherFrmHandleEvent );
				break;
			case ConfigForm:
				FrmSetEventHandler( frmP, ConfigFrmHandleEvent );
				break;
			case HelpForm:
				FrmSetEventHandler( frmP, HelpFormHandleEvent );
				break;
			case BlankForm:
				FrmSetEventHandler( frmP, BlankFormHandleEvent );
				break;
		}
		return true;
	}		
	return false;
}


static void AppEventLoop(void)
{
	UInt16 error;
	EventType event;

	g_launch_emu=0;
	g_rom_system=0;	


	do 
	{
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, evtWaitForever);				

		if (! SysHandleEvent(&event))
		{
			if (! AppHandleEvent(&event))
			{
				if (! MenuHandleEvent(0, &event, &error))
				{
					FrmDispatchEvent(&event);
				}	
			}		
		}		
		if (g_launch_emu) 
		{ break;}
	} while (event.eType != appStopEvent);
}

static void AppStop(void)
{
	Stop_Midi();

/*SAVE PREF*/
	PrefSetAppPreferences(APPLI_ID,0x0001,(short)VERSION,(void*)NULL,(short)0,(unsigned char)1);
	PrefSetAppPreferences(APPLI_ID,0x0001,(short)VERSION,(void*)&g_ljz_pref,(short)sizeof(g_ljz_pref),(unsigned char)1);
		
	if (!g_launch_emu)
	{
		PINSetInputAreaState(pinInputAreaUser);
		SysSetOrientation(sysOrientationUser);
		FrmCloseAllForms();	
	}
			/**/							
}

static void CheckDirs(void)
{
	UInt16 volRefNum;			
	UInt32 volIterator; 
	Err err;
	
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NES/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NES/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NES/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/SNES/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/SNES/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/SNES/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GB/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GB/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GB/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GENZ/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GENZ/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/GENZ/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NGP/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NGP/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/NGP/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/WS/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/WS/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/WS/Saves/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/TG16/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/TG16/Roms/");
		    VFSDirCreate(volRefNum, "/PALM/Programs/LJZ/TG16/Saves/");
		 }
	}
}

static MemHandle readMidiData(UInt32 num/*char* Path*/) {
/*    UInt32 size, nb;    
    FileRef ref;
    Err err;
    unsigned char* buf = NULL;
    UInt16 VolRefNum;
    UInt32 volIterator; 	
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&VolRefNum, &volIterator);
		if (!err) 
		{        
		    err = VFSFileOpen(VolRefNum, Path, vfsModeRead, &ref);
		    if (!err) 
		    {
        		err = VFSFileSize(ref, &size);
		        if (!err) 
		        {
           			buf = (unsigned char*)MemPtrNew(size);
	  	            if (buf) 
	  	            {
                		err = VFSFileRead(ref, size, buf, &nb);
            		}
        		}
        		VFSFileClose(ref);
        		return buf;
        	}
        }
    }*/
    
    
	MemHandle hRsc=NULL;
	int rd=SysRandom(0);
	switch (num)
	{
		case 0:
			if (rd&1) hRsc = DmGetResource('MIDI', MIDI_NES1);
			else hRsc = DmGetResource('MIDI', MIDI_NES2);
		break;
		case 1:
			if ((rd&3)==0) hRsc = DmGetResource('MIDI', MIDI_SNES1);
			else if ((rd&3)==1) hRsc = DmGetResource('MIDI', MIDI_SNES2);
			else if ((rd&3)==2) hRsc = DmGetResource('MIDI', MIDI_SNES3);
			else hRsc = DmGetResource('MIDI', MIDI_SNES4);
		break;
		case 2:
			if ((rd&3)==0) hRsc = DmGetResource('MIDI', MIDI_GB1);
			else if ((rd&3)==1) hRsc = DmGetResource('MIDI', MIDI_GB2);
			else if ((rd&3)==2) hRsc = DmGetResource('MIDI', MIDI_GB3);
			else hRsc = DmGetResource('MIDI', MIDI_GB4);
		break;
		case 3:
			if (rd&1) hRsc = DmGetResource('MIDI', MIDI_GENESIS1);
			else hRsc = DmGetResource('MIDI', MIDI_GENESIS2);
		break;
		case 4:hRsc = DmGetResource('MIDI', MIDI_NGP);
		break;
		case 5:
			if (rd&1) hRsc = DmGetResource('MIDI', MIDI_WS1);
			else hRsc = DmGetResource('MIDI', MIDI_WS2);
		break;
	}	
    return hRsc;
}



static void MidiCallBack(UInt32 userdata)
{
	SndCallbackInfoType cbinfo;
	cbinfo.funcP=MidiCallBack;
	cbinfo.dwUserData=NULL;
	TwSmfPlay(g_smfHandle, NULL, NULL, &cbinfo);
}

static void Start_Midi(UInt32 num)
{
//return;
	TwMidiSetMasterVolume(twMidiMaxVolume>>1);
    if (num<6) 
    {	    
        UInt32 duration;                                
        
        g_midiDataH=readMidiData(num);
        if (g_midiDataH)
		{	
			SndCallbackInfoType cbinfo;
			cbinfo.funcP=MidiCallBack;
			cbinfo.dwUserData=NULL;
			g_midiData = (unsigned char*) MemHandleLock(g_midiDataH);
							            
	        Err err = TwSmfOpen(&g_smfHandle, g_midiData, &duration);
    	    if (!err) {g_midi=1;TwSmfPlay(g_smfHandle, NULL, NULL, &cbinfo);}
        	else 
	        {
    	    	g_midi=0;
        		menu_message("Cannot load midi file");
		    	menu_waitBtn(1,1);
	        }
	    }
    }
}

static void Stop_Midi(void)
{
	if (g_midi)
	{
	    TwSmfStop(g_smfHandle);
	    SysTaskDelay(10);
	    TwSmfClose(g_smfHandle);
	    //MemPtrFree(g_midiData);
	    
	    MemHandleUnlock(g_midiDataH);
		DmReleaseResource(g_midiDataH);
	    
	    g_midi=0;
	}
}

void InitPrefDefaultValues()
{
	unsigned char l_default_inputconf[7][20]=
	{ 
	//SNES
	{BTN_Yellow,BTN_Green,BTN_Blue,BTN_Red,BTN_L,
	BTN_R,BTN_Func,BTN_Home,BTN_Up,BTN_Down,
	BTN_Left,BTN_Right,BTN_TL,BTN_BL,BTN_TR,
	BTN_BR,0,0,0,0},
	//NES
	{BTN_Yellow,BTN_Green,BTN_Red,BTN_Blue,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0},
	//GB/GBC
	{BTN_Yellow,BTN_Green,BTN_Red,BTN_Blue,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0},
	//GENESIS
	{BTN_Red,BTN_Green,BTN_Yellow,BTN_Blue,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0},
	//NGP
	{BTN_Yellow,BTN_Green,BTN_Red,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0,0},
	//WS
	{BTN_Yellow,BTN_Green,BTN_Red,BTN_Blue,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0},
	//TG16
	{BTN_Yellow,BTN_Green,BTN_Red,BTN_Blue,BTN_Up,BTN_Down,BTN_Left,BTN_Right,
	BTN_L,BTN_R,BTN_TR,BTN_BR,BTN_Func,
	0,0,0,0,0}
	};
	
	MemMove(g_ljz_pref.g_inputconf,l_default_inputconf,7*20);
	
	g_ljz_pref.g_SnesFastSprite=0;
	g_ljz_pref.g_SnesTransparency=0;
	g_ljz_pref.g_SnesRender=0;
	g_ljz_pref.g_SnesSound=0;
	g_ljz_pref.g_SnesSmoothing=0;
	g_ljz_pref.g_SnesScreenMode=0;
	g_ljz_pref.g_SnesFSkip=1;
	g_ljz_pref.g_SnesSoundFreq=1;
	g_ljz_pref.g_SnesASMCPU=1;

	g_ljz_pref.g_NesSound=2;
	g_ljz_pref.g_NesSmoothing=0;
	g_ljz_pref.g_NesScreenMode=0;
	g_ljz_pref.g_NesFSkip=0;
	g_ljz_pref.g_NesSoundFreq=0;

	g_ljz_pref.g_GBSound=2;
	g_ljz_pref.g_GBSmoothing=0;
	g_ljz_pref.g_GBScreenMode=0;
	g_ljz_pref.g_GBFSkip=1;
	g_ljz_pref.g_GBSoundFreq=0;

	g_ljz_pref.g_GENZSound=0;
	g_ljz_pref.g_GENZSmoothing=0;
	g_ljz_pref.g_GENZScreenMode=0;
	g_ljz_pref.g_GENZFSkip=1;
	g_ljz_pref.g_GENZSoundFreq=1;
	g_ljz_pref.g_GENZASMCPU=1;


	g_ljz_pref.g_NEOPOPZSound=2;
	g_ljz_pref.g_NEOPOPZSmoothing=0;
	g_ljz_pref.g_NEOPOPZScreenMode=0;
	g_ljz_pref.g_NEOPOPZFSkip=1;
	g_ljz_pref.g_NEOPOPZSoundFreq=1;
	g_ljz_pref.g_NEOPOPZEmuAccuracy=0;


	g_ljz_pref.g_OSWANZSound=0;
	g_ljz_pref.g_OSWANZSmoothing=0;
	g_ljz_pref.g_OSWANZScreenMode=0;
	g_ljz_pref.g_OSWANZFSkip=1;
	g_ljz_pref.g_OSWANZSoundFreq=1;
	
	g_ljz_pref.g_TG16ZSound=0;
	g_ljz_pref.g_TG16ZSmoothing=0;
	g_ljz_pref.g_TG16ZScreenMode=0;
	g_ljz_pref.g_TG16ZFSkip=1;
	g_ljz_pref.g_TG16ZSoundFreq=0;

	g_ljz_pref.g_sysfilter=SYS_NES;//|SYS_SNES|SYS_GB|SYS_GENZ;
	g_ljz_pref.g_conf_system=g_ljz_pref.g_sysfilter;	
	g_ljz_pref.g_lastrom[0]=0;
	g_ljz_pref.g_firsttime=0;

}

void LoadModuleFromCard(char *name,UInt16 *cardNo,LocalID *dbID)
{
	UInt16 		volRefNum;			
	UInt32 		volIterator; 
	Err			err;
	
	volIterator = vfsIteratorStart;
	while (volIterator != vfsIteratorStop) 
	{
		err = VFSVolumeEnumerate(&volRefNum, &volIterator);
		if (!err) 
		{
			FileInfoType info;
			FileRef fileRef;
			
			err = VFSFileOpen(volRefNum, name, vfsModeRead, &fileRef);
			if(err == errNone) 
			{
				err=VFSImportDatabaseFromFile (volRefNum, name, cardNo, dbID);
				if (err==dmErrMemError)
				{
					menu_message("out of mem");
					menu_waitBtn(1,1);
				}
				VFSFileClose(fileRef);
				break;
			}
			
		}
	}
}

UInt32 LJZLaunchModule(UInt16 system)
{
	UInt16 cardNo=0;
	LocalID dbID;
	DmSearchStateType sst;

	
	UInt32 res;
	res=0;
	{
		switch (system)
		{
			case SYS_SNES:dbID = DmFindDatabase(cardNo,"Little John Z-SNES");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-SNES.prc",&cardNo,&dbID);
			break;
			case SYS_NES:dbID = DmFindDatabase(cardNo,"Little John Z-NES");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-NES.prc",&cardNo,&dbID);				
			break;
			case SYS_GB:dbID = DmFindDatabase(cardNo,"Little John Z-GB");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-GB.prc",&cardNo,&dbID);
			break;
			case SYS_NEOPOPZ:dbID = DmFindDatabase(cardNo,"Little John Z-NGP");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-NGP.prc",&cardNo,&dbID);
			break;
			case SYS_OSWANZ:dbID = DmFindDatabase(cardNo,"Little John Z-WS");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-WS.prc",&cardNo,&dbID);
			break;
			case SYS_GENZ:dbID = DmFindDatabase(cardNo,"Little John Z-GENESIS");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-GENESIS.prc",&cardNo,&dbID);
			break;
			case SYS_TG16Z:dbID = DmFindDatabase(cardNo,"Little John Z-TG16");
			if (!dbID) LoadModuleFromCard("/PALM/Launcher/ljZ-TG16.prc",&cardNo,&dbID);
			break;
		}
		
		if (dbID) {SysAppLaunch(cardNo,dbID,0,0,NULL,&res);}
		else 
		{
			char str[128];		
			menu_message("Module not found");
			menu_waitBtn(1,1);
		}
	}
	return res;
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	UInt32 res = errNone;
    NativeFuncType* entry;
	
	if (cmd == sysAppLaunchCmdNormalLaunch) {
		if (!RomVersionCompatible(launchFlags)) 
		{							
			/*Setup dir if necessary*/
			CheckDirs();
		
			
			unsigned short val;



			/* DEFAULT PREF */
			InitPrefDefaultValues();
			
			/* LOAD PREF IF AVAILABLE WITH GOOD VERSION NUMBER*/
			char c;
			val=1;
			if (PrefGetAppPreferences(APPLI_ID,0x0001,(void*)&c,(unsigned short*)&val,(unsigned char)1)==VERSION);
			{
				val=sizeof(g_ljz_pref);
				PrefGetAppPreferences(APPLI_ID,0x0001,(void*)&g_ljz_pref,(unsigned short*)&val,(unsigned char)1);
			}
			
			/* init random seed */
			SysRandom(TimGetTicks());
												
			for (;;)
			{
						/*Start music*/
				int i;
				switch (g_ljz_pref.g_sysfilter)
				{
					case SYS_NES:i=0;break;
					case SYS_SNES:i=1;break;
					case SYS_GB:i=2;break;
					case SYS_GENZ:i=3;break;
					case SYS_NEOPOPZ:i=4;break;
					case SYS_OSWANZ:i=5;break;					
					case SYS_TG16Z:i=6;break;
				}
				Start_Midi(i);

				//g_frmloaded=0;
				if (FrmGetActiveFormID()!=LauncherForm)
					FrmGotoForm( LauncherForm );
				else FrmUpdateForm( LauncherForm ,0);
				AppEventLoop();
				AppStop();			
				
				
				if (g_launch_emu) 
				{									
					char *ptr;
					UInt32 val;
					
					

  					  					  					  					
					RectangleType rclip;
					UInt16 err;
					WinHandle runningWnd,oldWnd;
					rclip.topLeft.x=0;
					rclip.topLeft.y=0;
					rclip.extent.x=240;
					rclip.extent.y=160;
					oldWnd=WinGetActiveWindow();
					runningWnd=WinCreateWindow( &rclip,noFrame,false,true, &err );
					if (err) runningWnd=NULL;
					  					
					if (runningWnd)
					{
						WinSetActiveWindow(runningWnd);
//#ifndef TESTING_ONLY								
						switch (g_rom_system)
						{
							case SYS_SNES:						
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[0][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}																				
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_SnesSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_SnesSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_SnesSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_SnesScreenMode);							
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_SnesFSkip);	
							FtrSet(APPLI_ID, FTR_OPT_CPUCORE_ID,g_ljz_pref.g_SnesASMCPU);								
							FtrSet(APPLI_ID, FTR_OPT_FAST_SPR_ID, g_ljz_pref.g_SnesFastSprite);
							FtrSet(APPLI_ID, FTR_OPT_SNESRENDER_ID,g_ljz_pref.g_SnesRender);
							FtrSet(APPLI_ID, FTR_OPT_SNESTRANSP_ID,g_ljz_pref.g_SnesTransparency);														
							
							res = LJZLaunchModule(SYS_SNES);
							
							/*res = TwLoadModule(0, 0, 0, 1, twLoadFlagTNA|twLoadFlagQuickRun, &entry);*/
							

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID, &val);g_ljz_pref.g_SnesSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_SnesSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_SnesSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_SnesScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_SnesFSkip=val;
							FtrGet(APPLI_ID, FTR_OPT_CPUCORE_ID,&val);g_ljz_pref.g_SnesASMCPU=val;
							FtrGet(APPLI_ID, FTR_OPT_FAST_SPR_ID, &val);g_ljz_pref.g_SnesFastSprite=val;
							FtrGet(APPLI_ID, FTR_OPT_SNESRENDER_ID,&val);g_ljz_pref.g_SnesRender=val;
							FtrGet(APPLI_ID, FTR_OPT_SNESTRANSP_ID,&val);g_ljz_pref.g_SnesTransparency=val;														
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);														
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_CPUCORE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FAST_SPR_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SNESRENDER_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SNESTRANSP_ID);
							break;
							case SYS_NES:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[1][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_NesSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_NesSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_NesSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_NesScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_NesFSkip);	
							
							//res = TwLoadModule(0, 0, 0, 2, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_NES);

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_NesSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_NesSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_NesSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_NesScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_NesFSkip=val;	
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);
							
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							break;
							case SYS_GB:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[2][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}
						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_GBSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_GBSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_GBSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_GBScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_GBFSkip);	
							
							//res = TwLoadModule(0, 0, 0, 3, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_GB);

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_GBSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_GBSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_GBSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_GBScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_GBFSkip=val;

							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							break;
							case SYS_GENZ:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[3][k]==i) break;
								DmWrite(ptr,i,(void*)&k,1);
							}
						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_GENZSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_GENZSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_GENZSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_GENZScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_GENZFSkip);	
							FtrSet(APPLI_ID, FTR_OPT_CPUCORE_ID,g_ljz_pref.g_GENZASMCPU);	
							
							//res = TwLoadModule(0, 0, 0, 4, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_GENZ);

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_GENZSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_GENZSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_GENZSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_GENZScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_GENZFSkip=val;
							FtrGet(APPLI_ID, FTR_OPT_CPUCORE_ID,&val);g_ljz_pref.g_GENZASMCPU=val;
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_CPUCORE_ID);							
							break;
						
							case SYS_NEOPOPZ:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[4][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}
						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_NEOPOPZSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_NEOPOPZSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_NEOPOPZSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_NEOPOPZScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_NEOPOPZFSkip);															
							FtrSet(APPLI_ID, FTR_OPT_EMU_ACCURACY_ID,g_ljz_pref.g_NEOPOPZEmuAccuracy);								
							
							//res = TwLoadModule(0, 0, 0, 5, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_NEOPOPZ);

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_NEOPOPZSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_NEOPOPZSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_NEOPOPZSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_NEOPOPZScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_NEOPOPZFSkip=val;
							FtrGet(APPLI_ID, FTR_OPT_EMU_ACCURACY_ID,&val);g_ljz_pref.g_NEOPOPZEmuAccuracy=val;
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_EMU_ACCURACY_ID);
							break;
										
					
							case SYS_OSWANZ:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[5][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}
						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_OSWANZSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_OSWANZSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_OSWANZSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_OSWANZScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_OSWANZFSkip);	
							
							//res = TwLoadModule(0, 0, 0, 6, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_OSWANZ);

							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_OSWANZSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_OSWANZSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_OSWANZSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_OSWANZScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_OSWANZFSkip=val;
							
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);							
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							break;
							
							case SYS_TG16Z:	
							FtrPtrNew(APPLI_ID, FTR_OPT_INPUTCONF_ID, 20, (void**)&ptr);
							for (unsigned char i=0;i<20;i++)
							{
								unsigned char k;
								for (k=0;k<20;k++)
									if (g_ljz_pref.g_inputconf[6][k]==i) break;									
								DmWrite(ptr,i,(void*)&k,1);
							}
						
							FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID, g_ljz_pref.g_TG16ZSmoothing);
							FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,g_ljz_pref.g_TG16ZSound);
							FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,g_ljz_pref.g_TG16ZSoundFreq);
							FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,g_ljz_pref.g_TG16ZScreenMode);	
							FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,g_ljz_pref.g_TG16ZFSkip);	
							
							//res = TwLoadModule(0, 0, 0, 7, twLoadFlagTNA|twLoadFlagQuickRun, &entry);
							res = LJZLaunchModule(SYS_TG16Z);

/*							FtrGet(APPLI_ID, FTR_OPT_SMOOTHING_ID,&val);g_ljz_pref.g_TG16ZSmoothing=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUND_ID,&val);g_ljz_pref.g_TG16ZSound=val;
							FtrGet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,&val);g_ljz_pref.g_TG16ZSoundFreq=val;
							FtrGet(APPLI_ID, FTR_OPT_SCREENMODE_ID,&val);g_ljz_pref.g_TG16ZScreenMode=val;
							FtrGet(APPLI_ID, FTR_OPT_FSKIP_ID,&val);g_ljz_pref.g_TG16ZFSkip=val;
*/
							FtrPtrFree(APPLI_ID, FTR_OPT_INPUTCONF_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SMOOTHING_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUND_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SOUNDFREQ_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_SCREENMODE_ID);
							FtrUnregister(APPLI_ID, FTR_OPT_FSKIP_ID);
							break;
						}					
//#else
					FtrPtrFree(APPLI_ID, FTR_ROM_FILENAME_ID );						
//#endif				
					
					if (runningWnd) WinDeleteWindow(runningWnd,false);
					else
					{menu_message("cannot free window");menu_waitBtn(1,1);}					
					WinSetActiveWindow(oldWnd);
					}				
					
					
					
				}
				else 
				{
					//FrmCloseAllForms();	
					break; //tmp bug fixes due to snes bug when snes launched 2 times
				}
				
			}								

        }

	}

	
	return res;
}
